// Copyright (c) 2022 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "mfx_common.h"
#include "mfxenctools-int.h"
#include "mfx_enctools_loader.h"
#include "mfx_loader_utils.h"

const char* const encToolsDLLName = "enctools.so";

typedef void* (MFX_CDECL* FuncMFXVideoENCODE_NewEncTools) (void*, const char*);

void* mfxEncToolsLoader::LoadEncToolsLib()
{
    void* hModule = nullptr;

    hModule = LoadDependency(encToolsDLLName);

    return hModule;
}

mfxStatus mfxEncToolsLoader::UnLoadEncToolsLib(void** phModule)
{
    if (phModule)
    {
        UnLoadModule(*phModule);
        *phModule = nullptr;
    }
    return MFX_ERR_NONE;
}

void* MFXNewEncTools(void* hModule)
{
    FuncMFXVideoENCODE_NewEncTools pfnNew = nullptr;
    void* encTools = nullptr;
    if (hModule)
    {
        pfnNew = (FuncMFXVideoENCODE_NewEncTools)GetModuleFuncAddress(hModule, "MFXVideoENCODE_NewEncTools");
    }

    if (pfnNew)
    {
        std::string path = GetSelfModuleFullPath();
        encTools = (void*)pfnNew(hModule, path.c_str());
    }

    return encTools;

}
