// Copyright (c) 2018 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

/* //////////////////////////////// "pjname.h" ////////////////////////////////
//
//  Purpose:
//
*/

#ifndef __PJNAME_H__
#define __PJNAME_H__

/*
//  The prefix of library without the quotes. The prefix is directly used
//  to generate the GetLibVersion function. It is used to generate the names in
//  the dispatcher code, in the version description and in the resource file.
*/

#define LIB_PREFIX          mfxj

/* Short name of library. It is used in the resource file only */
#define IPP_LIB_LONGNAME()  "JPEG processing"
#define IPP_LIB_SHORTNAME()    "ippJP"


#define GET_STR2(x)      #x
#define GET_STR(x)       GET_STR2(x)
#define IPP_LIB_PREFIX() GET_STR(LIB_PREFIX)

#define IPP_INC_NAME()   "ippj.h"

#endif /* __PJNAME_H__ */
/* //////////////////////// End of file "pjname.h" ///////////////////////// */
