// Copyright (c) 2003-2018 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "umc_event.h"

namespace UMC
{

Event::Event(void)
{
    vm_event_set_invalid(&m_handle);

} // Event::Event(void)

Event::~Event(void)
{
    if (vm_event_is_valid(&m_handle))
    {
        vm_event_signal(&m_handle);
        vm_event_destroy(&m_handle);
    }
} // Event::~Event(void)

Status Event::Init(int32_t iManual, int32_t iState)
{
    Status umcRes = UMC_OK;

    if (vm_event_is_valid(&m_handle))
    {
        umcRes = vm_event_signal(&m_handle);
        vm_event_destroy(&m_handle);
    }

    if (UMC_OK == umcRes)
        umcRes = vm_event_init(&m_handle, iManual, iState);

    return umcRes;

} // Status Event::Init(int32_t iManual, int32_t iState)

} // namespace UMC
