// Copyright (c) 2016-2018 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "mfx_critical_error_handler.h"

MfxCriticalErrorHandler::MfxCriticalErrorHandler(void):
    m_CriticalErrorReportedAtLeastOnce(false),
    m_CriticalErrorStatus(MFX_ERR_NONE)
{
}

mfxStatus MfxCriticalErrorHandler::ReturningCriticalStatus()
{
    assert(m_CriticalErrorStatus != MFX_ERR_NONE);
    m_CriticalErrorReportedAtLeastOnce = true;
    if (m_CriticalErrorStatus == MFX_ERR_NONE)
        m_CriticalErrorStatus = MFX_ERR_UNKNOWN;

    return m_CriticalErrorStatus;
}

bool MfxCriticalErrorHandler::NeedToReturnCriticalStatus(mfxBitstream *bs)
{
    bool RetrievingCachedFrames_and_CriticalErrorReportedAtLeastOnce = m_CriticalErrorReportedAtLeastOnce && bs == NULL;
    return CriticalErrorOccured() && !RetrievingCachedFrames_and_CriticalErrorReportedAtLeastOnce;
}

void MfxCriticalErrorHandler::SetCriticalErrorOccured(mfxStatus errorStatus)
{
    assert(errorStatus != MFX_ERR_NONE);
    m_CriticalErrorStatus = errorStatus;
}

bool MfxCriticalErrorHandler::CriticalErrorOccured()
{
    return m_CriticalErrorStatus != MFX_ERR_NONE;
}