// Copyright (c) 2003-2019 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "umc_defs.h"
#if defined (MFX_ENABLE_H264_VIDEO_DECODE)

#ifndef __UMC_H264_DEC_DEBUG_H
#define __UMC_H264_DEC_DEBUG_H

#include "umc_h264_dec_defs_dec.h"

//#define ENABLE_TRACE

namespace UMC
{

#ifdef ENABLE_TRACE

void Trace(vm_char const* format, ...);

#if defined _DEBUG
#define DEBUG_PRINT(x) Trace x
static int pppp = 0;
#else
#define DEBUG_PRINT(x)
#endif

#else //  ENABLE_TRACE
#define DEBUG_PRINT(x)
#endif //  ENABLE_TRACE

} // namespace UMC

#endif // __UMC_H264_DEC_DEBUG_H
#endif // MFX_ENABLE_H264_VIDEO_DECODE
