// Copyright (c) 2018 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

/*M*
//
//  Purpose: Common header file for IPPs arithmetic functions
//
*M*/

#if !defined(__PS_ANARITH_H__)

#define __PS_ANARITH_H__

#if !defined(__OWNS_H__)

#include "owns.h"

#endif /* __OWNS_H__ */

#if (_IPP64 >= _IPP64_I7)

#define IF_I7(EXP) if(EXP) {
#define ENDIF_I7 }
#define ELSE_I7 } else {

#else

#define IF_I7(EXP)
#define ENDIF_I7
#define ELSE_I7

#endif

extern void mfxownsMulC_16s_I         (Ipp16s val, Ipp16s* pSrcDst, int len);
extern void mfxownsMul_16s_I          (const Ipp16s* pSrc, Ipp16s* pSrcDst, int len);

#endif /* __PS_ANARITH_H__ */

