// Copyright (c) 2017-2018 Intel Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
#include "../include/tree.h"
#include "../include/asc_defs.h"

bool SCDetect1(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (MVDiff < 93) {
            if (diffMVdiffVal < 31) {
                if (negBalance < 9) {
                    if (gchDC < 133) {
                        if (diffTSC < 1) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (TSCindex < 1) {
                                    if (SC < 1376) {
                                        if (negBalance < 1) {
                                            return 1;
                                        }
                                        else if (negBalance >= 1) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 1376) {
                                        return 0;
                                    }
                                }
                                else if (TSCindex >= 1) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 1) {
                            if (ssDCval < 21) {
                                if (Cs < 363) {
                                    return 0;
                                }
                                else if (Cs >= 363) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 21) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 133) {
                        return 1;
                    }
                }
                else if (negBalance >= 9) {
                    if (RsDiff < 111) {
                        if (gchDC < 31) {
                            return 0;
                        }
                        else if (gchDC >= 31) {
                            if (refDCval < 213) {
                                return 0;
                            }
                            else if (refDCval >= 213) {
                                return 1;
                            }
                        }
                    }
                    else if (RsDiff >= 111) {
                        if (SC < 3109) {
                            return 0;
                        }
                        else if (SC >= 3109) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 31) {
                if (CsDiff < 23) {
                    if (ssDCval < 35) {
                        if (RsDiff < 6) {
                            if (posBalance < 45) {
                                return 0;
                            }
                            else if (posBalance >= 45) {
                                if (AFD < 10) {
                                    return 0;
                                }
                                else if (AFD >= 10) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 6) {
                            if (CsDiff < 18) {
                                return 1;
                            }
                            else if (CsDiff >= 18) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 35) {
                        return 0;
                    }
                }
                else if (CsDiff >= 23) {
                    if (posBalance < 70) {
                        return 0;
                    }
                    else if (posBalance >= 70) {
                        if (TSC < 23) {
                            if (AFD < 19) {
                                return 1;
                            }
                            else if (AFD >= 19) {
                                return 0;
                            }
                        }
                        else if (TSC >= 23) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 93) {
            if (diffAFD < 8) {
                if (diffTSC < 4) {
                    if (diffRsCsdiff < 34) {
                        if (Rs < 68) {
                            if (gchDC < 66) {
                                if (RsCsDiff < 5) {
                                    if (gchDC < 2) {
                                        return 0;
                                    }
                                    else if (gchDC >= 2) {
                                        if (MVDiff < 375) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 375) {
                                            if (posBalance < 58) {
                                                if (TSCindex < 6) {
                                                    return 0;
                                                }
                                                else if (TSCindex >= 6) {
                                                    if (refDCval < 227) {
                                                        return 0;
                                                    }
                                                    else if (refDCval >= 227) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 58) {
                                                if (gchDC < 3) {
                                                    if (Cs < 7) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 7) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 3) {
                                                    if (negBalance < 57) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 57) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 5) {
                                    if (refDCval < 30) {
                                        if (Cs < 50) {
                                            return 0;
                                        }
                                        else if (Cs >= 50) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 30) {
                                        if (diffRsCsdiff < -12) {
                                            if (TSC < 4) {
                                                return 1;
                                            }
                                            else if (TSC >= 4) {
                                                if (AFD < 31) {
                                                    return 0;
                                                }
                                                else if (AFD >= 31) {
                                                    if (gchDC < 14) {
                                                        return 0;
                                                    }
                                                    else if (gchDC >= 14) {
                                                        if (Cs < 18) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 18) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= -12) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 66) {
                                if (AFD < 26) {
                                    if (refDCval < 201) {
                                        if (Cs < 1) {
                                            return 1;
                                        }
                                        else if (Cs >= 1) {
                                            if (Rs < 29) {
                                                return 0;
                                            }
                                            else if (Rs >= 29) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 201) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 26) {
                                    return 0;
                                }
                            }
                        }
                        else if (Rs >= 68) {
                            if (gchDC < 6) {
                                return 0;
                            }
                            else if (gchDC >= 6) {
                                if (RsCsDiff < 43) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 43) {
                                    if (AFD < 62) {
                                        if (RsDiff < 6) {
                                            if (TSC < 7) {
                                                if (AFD < 17) {
                                                    return 0;
                                                }
                                                else if (AFD >= 17) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 7) {
                                                if (Rs < 449) {
                                                    if (diffRsCsdiff < 4) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= 4) {
                                                        if (negBalance < 80) {
                                                            if (AFD < 22) {
                                                                return 0;
                                                            }
                                                            else if (AFD >= 22) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 80) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 449) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 6) {
                                            if (Rs < 716) {
                                                if (Rs < 713) {
                                                    if (AFD < 40) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 40) {
                                                        if (SC < 449) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 449) {
                                                            if (MVDiff < 161) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 161) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 713) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 716) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 62) {
                                        if (Rs < 1425) {
                                            return 0;
                                        }
                                        else if (Rs >= 1425) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 34) {
                        if (refDCval < 50) {
                            if (ssDCval < 31) {
                                return 1;
                            }
                            else if (ssDCval >= 31) {
                                if (diffTSC < -5) {
                                    if (ssDCval < 36) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 36) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= -5) {
                                    if (diffRsCsdiff < 282) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 282) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 50) {
                            if (gchDC < 32) {
                                return 0;
                            }
                            else if (gchDC >= 32) {
                                if (refDCval < 72) {
                                    if (Rs < 364) {
                                        return 1;
                                    }
                                    else if (Rs >= 364) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 72) {
                                    if (RsCsDiff < 901) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 901) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 4) {
                    if (MVDiff < 227) {
                        if (TSCindex < 8) {
                            if (refDCval < 42) {
                                return 0;
                            }
                            else if (refDCval >= 42) {
                                return 1;
                            }
                        }
                        else if (TSCindex >= 8) {
                            return 0;
                        }
                    }
                    else if (MVDiff >= 227) {
                        if (refDCval < 31) {
                            if (SC < 441) {
                                if (CsDiff < 1) {
                                    return 1;
                                }
                                else if (CsDiff >= 1) {
                                    return 0;
                                }
                            }
                            else if (SC >= 441) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 31) {
                            if (AFD < 60) {
                                if (SC < 10) {
                                    if (AFD < 7) {
                                        return 0;
                                    }
                                    else if (AFD >= 7) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 10) {
                                    if (ssDCval < 58) {
                                        if (Cs < 537) {
                                            if (diffTSC < 6) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 6) {
                                                if (Rs < 396) {
                                                    return 0;
                                                }
                                                else if (Rs >= 396) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 537) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 58) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 60) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (ssDCval < 50) {
                    if (Cs < 82) {
                        if (MVDiff < 129) {
                            if (AFD < 13) {
                                return 1;
                            }
                            else if (AFD >= 13) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 129) {
                            if (posBalance < 23) {
                                if (negBalance < 100) {
                                    if (diffAFD < 13) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 13) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 100) {
                                    if (RsDiff < 5) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 5) {
                                        if (diffMVdiffVal < 344) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 344) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 23) {
                                if (Rs < 82) {
                                    return 1;
                                }
                                else if (Rs >= 82) {
                                    if (SC < 142) {
                                        return 0;
                                    }
                                    else if (SC >= 142) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 82) {
                        if (diffRsCsdiff < 2) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 2) {
                            if (AFD < 20) {
                                if (diffTSC < 4) {
                                    if (SC < 729) {
                                        return 0;
                                    }
                                    else if (SC >= 729) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 4) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 20) {
                                if (gchDC < 12) {
                                    if (Cs < 593) {
                                        if (CsDiff < 5) {
                                            if (ssDCval < 34) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 34) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 593) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 12) {
                                    if (negBalance < 96) {
                                        return 1;
                                    }
                                    else if (negBalance >= 96) {
                                        if (RsCsDiff < 43) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 43) {
                                            if (diffAFD < 10) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 10) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 50) {
                    if (gchDC < 40) {
                        if (SC < 297) {
                            if (CsDiff < 2) {
                                return 0;
                            }
                            else if (CsDiff >= 2) {
                                if (diffMVdiffVal < 142) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 142) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 297) {
                            if (MVDiff < 621) {
                                if (diffAFD < 18) {
                                    return 0;
                                }
                                else if (diffAFD >= 18) {
                                    if (Rs < 484) {
                                        if (RsDiff < 6) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 484) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 621) {
                                if (AFD < 28) {
                                    return 1;
                                }
                                else if (AFD >= 28) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 40) {
                        if (TSCindex < 9) {
                            return 1;
                        }
                        else if (TSCindex >= 9) {
                            if (diffRsCsdiff < 16) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 16) {
                                if (TSC < 19) {
                                    return 1;
                                }
                                else if (TSC >= 19) {
                                    if (MVDiff < 277) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 277) {
                                        if (RsCsDiff < 275) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 275) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (Cs < 2676) {
            if (diffMVdiffVal < 71) {
                if (diffTSC < 13) {
                    if (MVDiff < 56) {
                        if (SC < 243) {
                            if (SC < 79) {
                                return 0;
                            }
                            else if (SC >= 79) {
                                return 1;
                            }
                        }
                        else if (SC >= 243) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                if (RsDiff < 26) {
                                    return 0;
                                }
                                else if (RsDiff >= 26) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 56) {
                        if (RsCsDiff < 74) {
                            if (diffMVdiffVal < -982) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= -982) {
                                if (gchDC < 6) {
                                    if (ssDCval < 90) {
                                        if (refDCval < 37) {
                                            return 0;
                                        }
                                        else if (refDCval >= 37) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 90) {
                                        if (AFD < 54) {
                                            return 0;
                                        }
                                        else if (AFD >= 54) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 6) {
                                    if (diffMVdiffVal < 55) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 55) {
                                        if (diffRsCsdiff < 15) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 15) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 74) {
                            if (ssDCval < 76) {
                                if (diffAFD < 15) {
                                    if (gchDC < 2) {
                                        return 1;
                                    }
                                    else if (gchDC >= 2) {
                                        if (diffMVdiffVal < -25) {
                                            if (diffMVdiffVal < -46) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= -46) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -25) {
                                            if (refDCval < 34) {
                                                return 1;
                                            }
                                            else if (refDCval >= 34) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 15) {
                                    if (AFD < 35) {
                                        if (diffMVdiffVal < 51) {
                                            if (diffRsCsdiff < 161) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 161) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 51) {
                                            if (RsCsDiff < 212) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 212) {
                                                if (SC < 1860) {
                                                    return 0;
                                                }
                                                else if (SC >= 1860) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 35) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 76) {
                                if (RsDiff < 12) {
                                    return 0;
                                }
                                else if (RsDiff >= 12) {
                                    if (diffAFD < 26) {
                                        if (gchDC < 22) {
                                            if (gchDC < 18) {
                                                if (ssDCval < 134) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 134) {
                                                    if (SC < 2103) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2103) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 18) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 26) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 13) {
                    if (diffRsCsdiff < 233) {
                        if (TSC < 57) {
                            if (diffRsCsdiff < 67) {
                                if (diffTSC < 31) {
                                    if (Cs < 369) {
                                        if (refDCval < 24) {
                                            return 1;
                                        }
                                        else if (refDCval >= 24) {
                                            if (AFD < 43) {
                                                return 0;
                                            }
                                            else if (AFD >= 43) {
                                                if (AFD < 44) {
                                                    return 1;
                                                }
                                                else if (AFD >= 44) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 369) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 31) {
                                    if (MVDiff < 141) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 141) {
                                        if (Rs < 162) {
                                            return 1;
                                        }
                                        else if (Rs >= 162) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 67) {
                                if (Cs < 1173) {
                                    if (diffMVdiffVal < 28) {
                                        if (Rs < 2120) {
                                            if (RsCsDiff < 151) {
                                                if (RsDiff < 8) {
                                                    if (TSC < 20) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 20) {
                                                        if (CsDiff < 8) {
                                                            return 0;
                                                        }
                                                        else if (CsDiff >= 8) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (refDCval < 29) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 29) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 151) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 2120) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 28) {
                                        if (Rs < 3549) {
                                            return 0;
                                        }
                                        else if (Rs >= 3549) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 1173) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 57) {
                            return 1;
                        }
                    }
                    else if (diffRsCsdiff >= 233) {
                        if (gchDC < 46) {
                            if (RsDiff < 21) {
                                if (Scindex < 5) {
                                    if (MVDiff < 89) {
                                        if (TSC < 14) {
                                            return 0;
                                        }
                                        else if (TSC >= 14) {
                                            if (MVDiff < 64) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 64) {
                                                if (RsDiff < 20) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 20) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 89) {
                                        return 1;
                                    }
                                }
                                else if (Scindex >= 5) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 21) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 46) {
                            if (AFD < 54) {
                                if (MVDiff < 41) {
                                    return 1;
                                }
                                else if (MVDiff >= 41) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 54) {
                                if (MVDiff < 178) {
                                    return 1;
                                }
                                else if (MVDiff >= 178) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 71) {
                if (negBalance < 28) {
                    if (diffRsCsdiff < 157) {
                        if (Rs < 1437) {
                            if (diffAFD < 34) {
                                if (refDCval < 32) {
                                    if (diffMVdiffVal < 91) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 91) {
                                        if (diffMVdiffVal < 196) {
                                            if (diffTSC < 8) {
                                                return 1;
                                            }
                                            else if (diffTSC >= 8) {
                                                if (SC < 140) {
                                                    return 0;
                                                }
                                                else if (SC >= 140) {
                                                    if (AFD < 25) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 25) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 196) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 32) {
                                    if (negBalance < 22) {
                                        if (TSC < 22) {
                                            if (MVDiff < 783) {
                                                if (SC < 172) {
                                                    if (gchDC < 53) {
                                                        if (posBalance < 112) {
                                                            return 0;
                                                        }
                                                        else if (posBalance >= 112) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 53) {
                                                        if (Rs < 1) {
                                                            if (CsDiff < 2) {
                                                                return 0;
                                                            }
                                                            else if (CsDiff >= 2) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 172) {
                                                    if (diffTSC < 11) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 11) {
                                                        if (Rs < 1184) {
                                                            if (gchDC < 39) {
                                                                if (SC < 403) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 403) {
                                                                    if (Cs < 199) {
                                                                        if (Rs < 439) {
                                                                            return 1;
                                                                        }
                                                                        else if (Rs >= 439) {
                                                                            return 0;
                                                                        }
                                                                    }
                                                                    else if (Cs >= 199) {
                                                                        return 0;
                                                                    }
                                                                }
                                                            }
                                                            else if (gchDC >= 39) {
                                                                if (RsDiff < 3) {
                                                                    return 0;
                                                                }
                                                                else if (RsDiff >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (Rs >= 1184) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 783) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 22) {
                                        if (diffRsCsdiff < 23) {
                                            if (diffRsCsdiff < 14) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 14) {
                                                if (Rs < 343) {
                                                    return 1;
                                                }
                                                else if (Rs >= 343) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 23) {
                                            if (Rs < 593) {
                                                return 1;
                                            }
                                            else if (Rs >= 593) {
                                                if (TSC < 17) {
                                                    return 0;
                                                }
                                                else if (TSC >= 17) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 34) {
                                if (TSC < 32) {
                                    return 1;
                                }
                                else if (TSC >= 32) {
                                    if (ssDCval < 114) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 114) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 1437) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 157) {
                        if (TSC < 20) {
                            return 1;
                        }
                        else if (TSC >= 20) {
                            if (Rs < 2234) {
                                if (posBalance < 125) {
                                    if (MVDiff < 521) {
                                        if (negBalance < 12) {
                                            if (gchDC < 61) {
                                                return 0;
                                            }
                                            else if (gchDC >= 61) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 521) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 125) {
                                    if (gchDC < 121) {
                                        return 0;
                                    }
                                    else if (gchDC >= 121) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 2234) {
                                if (RsDiff < 29) {
                                    if (MVDiff < 107) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 107) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 29) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 28) {
                    if (ssDCval < 149) {
                        if (diffTSC < 12) {
                            if (ssDCval < 73) {
                                if (gchDC < 8) {
                                    if (CsDiff < 7) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 7) {
                                        if (diffMVdiffVal < 74) {
                                            if (SC < 1269) {
                                                return 0;
                                            }
                                            else if (SC >= 1269) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 74) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 8) {
                                    if (diffAFD < 15) {
                                        if (ssDCval < 41) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 41) {
                                            if (posBalance < 83) {
                                                return 0;
                                            }
                                            else if (posBalance >= 83) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 15) {
                                        if (diffRsCsdiff < -8) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= -8) {
                                            if (diffMVdiffVal < 78) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 78) {
                                                if (ssDCval < 45) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 45) {
                                                    if (Rs < 2209) {
                                                        if (negBalance < 41) {
                                                            if (diffAFD < 27) {
                                                                return 1;
                                                            }
                                                            else if (diffAFD >= 27) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (negBalance >= 41) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 2209) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 73) {
                                if (TSC < 15) {
                                    if (CsDiff < 1) {
                                        if (Rs < 249) {
                                            return 1;
                                        }
                                        else if (Rs >= 249) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 15) {
                                    if (diffRsCsdiff < 43) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 43) {
                                        if (Scindex < 4) {
                                            if (Rs < 1129) {
                                                return 1;
                                            }
                                            else if (Rs >= 1129) {
                                                return 0;
                                            }
                                        }
                                        else if (Scindex >= 4) {
                                            if (posBalance < 77) {
                                                if (SC < 3679) {
                                                    if (MVDiff < 267) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 267) {
                                                        if (CsDiff < 22) {
                                                            if (MVDiff < 484) {
                                                                return 0;
                                                            }
                                                            else if (MVDiff >= 484) {
                                                                if (posBalance < 42) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 42) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (CsDiff >= 22) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (SC >= 3679) {
                                                    if (diffMVdiffVal < 243) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 243) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 77) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 12) {
                            if (RsCsDiff < 245) {
                                if (gchDC < 20) {
                                    if (AFD < 83) {
                                        if (diffMVdiffVal < 81) {
                                            if (SC < 1253) {
                                                return 1;
                                            }
                                            else if (SC >= 1253) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 81) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 83) {
                                        if (Rs < 387) {
                                            return 0;
                                        }
                                        else if (Rs >= 387) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 20) {
                                    if (negBalance < 81) {
                                        if (posBalance < 89) {
                                            if (diffMVdiffVal < 103) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 103) {
                                                if (RsDiff < 8) {
                                                    if (diffRsCsdiff < 68) {
                                                        if (diffTSC < 20) {
                                                            if (Cs < 417) {
                                                                if (CsDiff < 3) {
                                                                    return 0;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Cs >= 417) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (diffTSC >= 20) {
                                                            if (Cs < 526) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 526) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 68) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 8) {
                                                    if (RsDiff < 14) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 14) {
                                                        if (Cs < 679) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 679) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 89) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 81) {
                                        if (MVDiff < 1062) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 1062) {
                                            if (TSC < 24) {
                                                return 1;
                                            }
                                            else if (TSC >= 24) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 245) {
                                if (diffTSC < 16) {
                                    if (diffAFD < 26) {
                                        if (refDCval < 86) {
                                            return 1;
                                        }
                                        else if (refDCval >= 86) {
                                            if (MVDiff < 193) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 193) {
                                                if (gchDC < 37) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 37) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 26) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 16) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 149) {
                        if (diffRsCsdiff < 17) {
                            if (MVDiff < 177) {
                                return 1;
                            }
                            else if (MVDiff >= 177) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 17) {
                            if (diffAFD < 21) {
                                return 0;
                            }
                            else if (diffAFD >= 21) {
                                if (TSC < 23) {
                                    if (gchDC < 3) {
                                        if (CsDiff < 7) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 7) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 3) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 23) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (Cs >= 2676) {
            if (diffMVdiffVal < 63) {
                if (diffTSC < 13) {
                    return 0;
                }
                else if (diffTSC >= 13) {
                    if (gchDC < 37) {
                        if (ssDCval < 133) {
                            return 0;
                        }
                        else if (ssDCval >= 133) {
                            if (diffRsCsdiff < 4855) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= 4855) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 37) {
                        if (RsDiff < 41) {
                            return 1;
                        }
                        else if (RsDiff >= 41) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 63) {
                if (TSC < 20) {
                    if (Rs < 4782) {
                        if (MVDiff < 143) {
                            if (Cs < 4645) {
                                return 0;
                            }
                            else if (Cs >= 4645) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 143) {
                            return 1;
                        }
                    }
                    else if (Rs >= 4782) {
                        return 0;
                    }
                }
                else if (TSC >= 20) {
                    if (MVDiff < 75) {
                        if (refDCval < 129) {
                            return 1;
                        }
                        else if (refDCval >= 129) {
                            return 0;
                        }
                    }
                    else if (MVDiff >= 75) {
                        if (gchDC < 29) {
                            return 1;
                        }
                        else if (gchDC >= 29) {
                            if (CsDiff < 46) {
                                if (CsDiff < 31) {
                                    return 1;
                                }
                                else if (CsDiff >= 31) {
                                    if (Rs < 6212) {
                                        return 0;
                                    }
                                    else if (Rs >= 6212) {
                                        if (gchDC < 38) {
                                            return 0;
                                        }
                                        else if (gchDC >= 38) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 46) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect2(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (MVDiff < 67) {
        if (diffTSC < 13) {
            if (gchDC < 7) {
                if (refDCval < 21) {
                    if (diffRsCsdiff < 77) {
                        if (diffMVdiffVal < -1) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= -1) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (ssDCval < 18) {
                                    if (MVDiff < 9) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 9) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 18) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 77) {
                        if (CsDiff < 11) {
                            if (Cs < 1352) {
                                return 1;
                            }
                            else if (Cs >= 1352) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 11) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 21) {
                    if (diffTSC < 8) {
                        if (diffRsCsdiff < 260) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 260) {
                            if (refDCval < 25) {
                                return 1;
                            }
                            else if (refDCval >= 25) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 8) {
                        if (MVDiff < 59) {
                            if (diffMVdiffVal < 14) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 14) {
                                if (ssDCval < 43) {
                                    return 1;
                                }
                                else if (ssDCval >= 43) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 59) {
                            if (Cs < 1683) {
                                if (RsCsDiff < 191) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 191) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 1683) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 7) {
                if (Scindex < 1) {
                    if (gchDC < 41) {
                        if (ssDCval < 175) {
                            return 0;
                        }
                        else if (ssDCval >= 175) {
                            if (MVDiff < 41) {
                                return 1;
                            }
                            else if (MVDiff >= 41) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 41) {
                        return 1;
                    }
                }
                else if (Scindex >= 1) {
                    if (diffAFD < -17) {
                        if (Cs < 1865) {
                            return 0;
                        }
                        else if (Cs >= 1865) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= -17) {
                        if (diffTSC < 5) {
                            return 0;
                        }
                        else if (diffTSC >= 5) {
                            if (MVDiff < 63) {
                                if (Cs < 79) {
                                    return 1;
                                }
                                else if (Cs >= 79) {
                                    if (diffMVdiffVal < 54) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 54) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 63) {
                                if (diffRsCsdiff < 393) {
                                    if (CsDiff < 12) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 12) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 393) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 13) {
            if (Cs < 3011) {
                if (SC < 2704) {
                    if (diffAFD < 29) {
                        if (RsDiff < 13) {
                            return 0;
                        }
                        else if (RsDiff >= 13) {
                            if (AFD < 29) {
                                return 1;
                            }
                            else if (AFD >= 29) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 29) {
                        return 1;
                    }
                }
                else if (SC >= 2704) {
                    if (diffRsCsdiff < 271) {
                        return 0;
                    }
                    else if (diffRsCsdiff >= 271) {
                        if (gchDC < 1) {
                            return 0;
                        }
                        else if (gchDC >= 1) {
                            if (RsCsDiff < 539) {
                                if (posBalance < 61) {
                                    return 1;
                                }
                                else if (posBalance >= 61) {
                                    if (Cs < 1801) {
                                        return 1;
                                    }
                                    else if (Cs >= 1801) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 539) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (Cs >= 3011) {
                if (MVDiff < 46) {
                    return 0;
                }
                else if (MVDiff >= 46) {
                    return 1;
                }
            }
        }
    }
    else if (MVDiff >= 67) {
        if (diffAFD < 13) {
            if (gchDC < 22) {
                if (diffAFD < 9) {
                    if (MVDiff < 149) {
                        if (diffAFD < 3) {
                            return 0;
                        }
                        else if (diffAFD >= 3) {
                            if (diffRsCsdiff < 398) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 398) {
                                if (Scindex < 5) {
                                    return 1;
                                }
                                else if (Scindex >= 5) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 149) {
                        if (diffTSC < 1) {
                            if (refDCval < 56) {
                                if (SC < 7282) {
                                    if (RsDiff < 2) {
                                        if (refDCval < 24) {
                                            if (posBalance < 38) {
                                                return 0;
                                            }
                                            else if (posBalance >= 38) {
                                                if (diffAFD < -19) {
                                                    if (ssDCval < 26) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 26) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= -19) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 24) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 2) {
                                        if (ssDCval < 24) {
                                            if (negBalance < 4) {
                                                return 1;
                                            }
                                            else if (negBalance >= 4) {
                                                if (TSC < 3) {
                                                    return 0;
                                                }
                                                else if (TSC >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 24) {
                                            if (RsCsDiff < 19) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 19) {
                                                if (TSC < 13) {
                                                    if (SC < 855) {
                                                        if (SC < 841) {
                                                            if (AFD < 31) {
                                                                if (Rs < 94) {
                                                                    if (posBalance < 42) {
                                                                        return 0;
                                                                    }
                                                                    else if (posBalance >= 42) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (Rs >= 94) {
                                                                    return 0;
                                                                }
                                                            }
                                                            else if (AFD >= 31) {
                                                                if (Rs < 213) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 213) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 841) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 855) {
                                                        return 0;
                                                    }
                                                }
                                                else if (TSC >= 13) {
                                                    if (Rs < 945) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 945) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 7282) {
                                    if (Rs < 3078) {
                                        return 1;
                                    }
                                    else if (Rs >= 3078) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 56) {
                                if (AFD < 34) {
                                    return 0;
                                }
                                else if (AFD >= 34) {
                                    if (SC < 1068) {
                                        if (RsDiff < 10) {
                                            if (CsDiff < 6) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (RsDiff >= 10) {
                                            if (Rs < 865) {
                                                return 0;
                                            }
                                            else if (Rs >= 865) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 1068) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 1) {
                            if (AFD < 15) {
                                if (ssDCval < 30) {
                                    if (CsDiff < 1) {
                                        if (diffTSC < 4) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 4) {
                                            if (Rs < 94) {
                                                return 1;
                                            }
                                            else if (Rs >= 94) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 30) {
                                    if (TSC < 12) {
                                        if (AFD < 14) {
                                            if (diffAFD < 7) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 7) {
                                                if (negBalance < 61) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 61) {
                                                    if (RsDiff < 1) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 14) {
                                            if (Cs < 79) {
                                                return 0;
                                            }
                                            else if (Cs >= 79) {
                                                if (diffRsCsdiff < 11) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 11) {
                                                    if (Cs < 367) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 367) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 12) {
                                        if (SC < 336) {
                                            return 0;
                                        }
                                        else if (SC >= 336) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 15) {
                                if (refDCval < 48) {
                                    if (ssDCval < 48) {
                                        if (diffRsCsdiff < 244) {
                                            if (negBalance < 84) {
                                                if (Scindex < 3) {
                                                    if (SC < 197) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 197) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Scindex >= 3) {
                                                    if (refDCval < 32) {
                                                        if (gchDC < 11) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 11) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (refDCval >= 32) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 84) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 244) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 48) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 48) {
                                    if (diffRsCsdiff < 94) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 94) {
                                        if (SC < 1492) {
                                            return 1;
                                        }
                                        else if (SC >= 1492) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 9) {
                    if (Cs < 210) {
                        if (refDCval < 51) {
                            if (CsDiff < 3) {
                                if (ssDCval < 45) {
                                    if (Rs < 148) {
                                        if (refDCval < 39) {
                                            return 1;
                                        }
                                        else if (refDCval >= 39) {
                                            if (AFD < 18) {
                                                return 1;
                                            }
                                            else if (AFD >= 18) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 148) {
                                        if (RsDiff < 2) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 2) {
                                            if (RsCsDiff < 25) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 25) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 45) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 3) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 51) {
                            if (gchDC < 15) {
                                if (Rs < 94) {
                                    if (RsCsDiff < 11) {
                                        if (Rs < 47) {
                                            return 0;
                                        }
                                        else if (Rs >= 47) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 11) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 94) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 15) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 210) {
                        if (ssDCval < 46) {
                            if (diffMVdiffVal < 92) {
                                if (AFD < 18) {
                                    if (Cs < 783) {
                                        return 0;
                                    }
                                    else if (Cs >= 783) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 18) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 92) {
                                if (Rs < 307) {
                                    if (SC < 393) {
                                        return 1;
                                    }
                                    else if (SC >= 393) {
                                        if (Cs < 309) {
                                            return 0;
                                        }
                                        else if (Cs >= 309) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 307) {
                                    if (MVDiff < 364) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 364) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 46) {
                            if (diffRsCsdiff < 56) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 56) {
                                if (Rs < 916) {
                                    if (negBalance < 36) {
                                        return 0;
                                    }
                                    else if (negBalance >= 36) {
                                        if (MVDiff < 265) {
                                            if (ssDCval < 59) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 59) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 265) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 916) {
                                    if (RsCsDiff < 1401) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 1401) {
                                        if (TSC < 26) {
                                            return 0;
                                        }
                                        else if (TSC >= 26) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 22) {
                if (SC < 1) {
                    if (TSC < 6) {
                        return 1;
                    }
                    else if (TSC >= 6) {
                        if (RsDiff < 5) {
                            if (ssDCval < 89) {
                                return 1;
                            }
                            else if (ssDCval >= 89) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 5) {
                            return 1;
                        }
                    }
                }
                else if (SC >= 1) {
                    if (RsCsDiff < 749) {
                        if (SC < 2185) {
                            if (SC < 2170) {
                                if (ssDCval < 91) {
                                    if (diffRsCsdiff < 38) {
                                        if (gchDC < 59) {
                                            if (posBalance < 28) {
                                                if (diffTSC < 9) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 9) {
                                                    if (Rs < 574) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 574) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 28) {
                                                if (RsDiff < 5) {
                                                    if (refDCval < 38) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 38) {
                                                        if (negBalance < 90) {
                                                            return 0;
                                                        }
                                                        else if (negBalance >= 90) {
                                                            if (ssDCval < 47) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 47) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 5) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 59) {
                                            if (diffAFD < 2) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 2) {
                                                if (diffRsCsdiff < 8) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= 8) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 38) {
                                        if (Rs < 302) {
                                            if (CsDiff < 16) {
                                                if (gchDC < 31) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 31) {
                                                    return 1;
                                                }
                                            }
                                            else if (CsDiff >= 16) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 302) {
                                            if (refDCval < 39) {
                                                return 1;
                                            }
                                            else if (refDCval >= 39) {
                                                if (gchDC < 85) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 85) {
                                                    if (Cs < 768) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 768) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 91) {
                                    if (diffTSC < 25) {
                                        if (AFD < 40) {
                                            return 0;
                                        }
                                        else if (AFD >= 40) {
                                            if (refDCval < 78) {
                                                if (Cs < 136) {
                                                    return 1;
                                                }
                                                else if (Cs >= 136) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 78) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 25) {
                                        if (Rs < 508) {
                                            return 0;
                                        }
                                        else if (Rs >= 508) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 2170) {
                                return 1;
                            }
                        }
                        else if (SC >= 2185) {
                            return 0;
                        }
                    }
                    else if (RsCsDiff >= 749) {
                        if (AFD < 39) {
                            return 0;
                        }
                        else if (AFD >= 39) {
                            if (diffAFD < -2) {
                                return 1;
                            }
                            else if (diffAFD >= -2) {
                                if (AFD < 59) {
                                    if (diffRsCsdiff < 928) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 928) {
                                        if (SC < 573) {
                                            return 0;
                                        }
                                        else if (SC >= 573) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 59) {
                                    if (diffTSC < 3) {
                                        if (negBalance < 45) {
                                            return 1;
                                        }
                                        else if (negBalance >= 45) {
                                            if (diffMVdiffVal < -13) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -13) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 3) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 13) {
            if (CsDiff < 7) {
                if (negBalance < 22) {
                    if (MVDiff < 862) {
                        if (SC < 788) {
                            if (diffMVdiffVal < 144) {
                                if (posBalance < 128) {
                                    if (diffRsCsdiff < 18) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 18) {
                                        if (diffMVdiffVal < 98) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 98) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 128) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 144) {
                                if (AFD < 24) {
                                    return 1;
                                }
                                else if (AFD >= 24) {
                                    if (RsCsDiff < 29) {
                                        if (diffTSC < 16) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 16) {
                                            if (RsDiff < 2) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 2) {
                                                if (Rs < 301) {
                                                    return 1;
                                                }
                                                else if (Rs >= 301) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 29) {
                                        if (ssDCval < 96) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 96) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 788) {
                            if (diffRsCsdiff < 57) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 57) {
                                if (diffAFD < 29) {
                                    return 0;
                                }
                                else if (diffAFD >= 29) {
                                    if (AFD < 42) {
                                        return 1;
                                    }
                                    else if (AFD >= 42) {
                                        if (Rs < 1905) {
                                            return 0;
                                        }
                                        else if (Rs >= 1905) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 862) {
                        if (ssDCval < 79) {
                            return 0;
                        }
                        else if (ssDCval >= 79) {
                            return 1;
                        }
                    }
                }
                else if (negBalance >= 22) {
                    if (diffMVdiffVal < 83) {
                        if (posBalance < 56) {
                            if (gchDC < 30) {
                                if (RsDiff < 10) {
                                    if (negBalance < 62) {
                                        if (SC < 247) {
                                            if (Cs < 17) {
                                                return 0;
                                            }
                                            else if (Cs >= 17) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 247) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 62) {
                                        if (RsDiff < 7) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 7) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 10) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 30) {
                                if (diffTSC < 7) {
                                    return 1;
                                }
                                else if (diffTSC >= 7) {
                                    if (Cs < 570) {
                                        return 0;
                                    }
                                    else if (Cs >= 570) {
                                        if (TSC < 27) {
                                            return 1;
                                        }
                                        else if (TSC >= 27) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 56) {
                            if (diffTSC < 22) {
                                if (AFD < 30) {
                                    if (SC < 1449) {
                                        if (posBalance < 60) {
                                            if (Cs < 273) {
                                                return 1;
                                            }
                                            else if (Cs >= 273) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 60) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 1449) {
                                        if (ssDCval < 109) {
                                            if (Cs < 805) {
                                                return 1;
                                            }
                                            else if (Cs >= 805) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 109) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (AFD >= 30) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 22) {
                                if (SC < 1603) {
                                    return 0;
                                }
                                else if (SC >= 1603) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 83) {
                        if (MVDiff < 417) {
                            if (Cs < 636) {
                                if (ssDCval < 105) {
                                    if (diffRsCsdiff < -1) {
                                        if (AFD < 36) {
                                            return 0;
                                        }
                                        else if (AFD >= 36) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -1) {
                                        if (diffMVdiffVal < 130) {
                                            if (gchDC < 7) {
                                                return 1;
                                            }
                                            else if (gchDC >= 7) {
                                                if (gchDC < 9) {
                                                    if (posBalance < 76) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 76) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 9) {
                                                    if (ssDCval < 84) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 84) {
                                                        if (RsCsDiff < 47) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 47) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 130) {
                                            if (posBalance < 2) {
                                                if (Cs < 11) {
                                                    return 0;
                                                }
                                                else if (Cs >= 11) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 2) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 105) {
                                    if (refDCval < 107) {
                                        return 1;
                                    }
                                    else if (refDCval >= 107) {
                                        if (diffAFD < 24) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 24) {
                                            if (refDCval < 135) {
                                                return 0;
                                            }
                                            else if (refDCval >= 135) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 636) {
                                if (diffAFD < 21) {
                                    return 0;
                                }
                                else if (diffAFD >= 21) {
                                    if (TSC < 25) {
                                        return 1;
                                    }
                                    else if (TSC >= 25) {
                                        if (SC < 3466) {
                                            return 0;
                                        }
                                        else if (SC >= 3466) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 417) {
                            if (Cs < 492) {
                                if (diffAFD < 18) {
                                    if (refDCval < 53) {
                                        return 1;
                                    }
                                    else if (refDCval >= 53) {
                                        if (SC < 3) {
                                            return 1;
                                        }
                                        else if (SC >= 3) {
                                            if (diffMVdiffVal < 1381) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 1381) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 18) {
                                    if (diffRsCsdiff < 9) {
                                        if (TSC < 19) {
                                            return 1;
                                        }
                                        else if (TSC >= 19) {
                                            if (Cs < 115) {
                                                return 0;
                                            }
                                            else if (Cs >= 115) {
                                                if (Cs < 216) {
                                                    return 1;
                                                }
                                                else if (Cs >= 216) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 9) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 492) {
                                if (MVDiff < 1057) {
                                    return 0;
                                }
                                else if (MVDiff >= 1057) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (CsDiff >= 7) {
                if (gchDC < 16) {
                    if (diffAFD < 22) {
                        if (ssDCval < 78) {
                            if (diffTSC < 6) {
                                if (AFD < 24) {
                                    return 1;
                                }
                                else if (AFD >= 24) {
                                    if (Cs < 512) {
                                        return 1;
                                    }
                                    else if (Cs >= 512) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 6) {
                                if (Rs < 643) {
                                    return 1;
                                }
                                else if (Rs >= 643) {
                                    if (RsDiff < 10) {
                                        if (ssDCval < 46) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 46) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 10) {
                                        if (refDCval < 82) {
                                            return 1;
                                        }
                                        else if (refDCval >= 82) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 78) {
                            if (TSC < 21) {
                                if (RsCsDiff < 106) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 106) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 21) {
                                if (refDCval < 87) {
                                    return 1;
                                }
                                else if (refDCval >= 87) {
                                    if (diffTSC < 11) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 11) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 22) {
                        if (diffMVdiffVal < 73) {
                            if (refDCval < 131) {
                                if (RsDiff < 48) {
                                    if (gchDC < 8) {
                                        if (ssDCval < 52) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 52) {
                                            if (diffTSC < 9) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 9) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 8) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 48) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 131) {
                                if (diffTSC < 13) {
                                    return 0;
                                }
                                else if (diffTSC >= 13) {
                                    if (Rs < 7533) {
                                        return 1;
                                    }
                                    else if (Rs >= 7533) {
                                        if (Cs < 7014) {
                                            return 0;
                                        }
                                        else if (Cs >= 7014) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 73) {
                            if (gchDC < 3) {
                                if (AFD < 54) {
                                    if (TSC < 16) {
                                        if (ssDCval < 87) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 87) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 16) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 54) {
                                    if (Rs < 1589) {
                                        return 0;
                                    }
                                    else if (Rs >= 1589) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 3) {
                                return 1;
                            }
                        }
                    }
                }
                else if (gchDC >= 16) {
                    if (diffRsCsdiff < 65) {
                        if (TSC < 65) {
                            if (CsDiff < 10) {
                                if (Cs < 439) {
                                    if (MVDiff < 202) {
                                        if (MVDiff < 169) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 169) {
                                            if (Rs < 371) {
                                                return 1;
                                            }
                                            else if (Rs >= 371) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 202) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 439) {
                                    if (diffTSC < 20) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 20) {
                                        if (gchDC < 34) {
                                            return 1;
                                        }
                                        else if (gchDC >= 34) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 10) {
                                return 0;
                            }
                        }
                        else if (TSC >= 65) {
                            return 1;
                        }
                    }
                    else if (diffRsCsdiff >= 65) {
                        if (diffMVdiffVal < 80) {
                            if (RsCsDiff < 633) {
                                if (AFD < 54) {
                                    if (MVDiff < 87) {
                                        if (diffMVdiffVal < 66) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 66) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 87) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 54) {
                                    if (RsCsDiff < 269) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 269) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 633) {
                                if (diffTSC < 16) {
                                    if (posBalance < 44) {
                                        if (CsDiff < 28) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 28) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 44) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 16) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 80) {
                            if (SC < 3327) {
                                if (ssDCval < 229) {
                                    if (Cs < 1033) {
                                        if (diffMVdiffVal < 98) {
                                            if (diffAFD < 26) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 26) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 98) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 1033) {
                                        if (diffAFD < 34) {
                                            if (TSC < 30) {
                                                if (MVDiff < 116) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 116) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 30) {
                                                if (refDCval < 42) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 42) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 34) {
                                            if (RsDiff < 7) {
                                                if (Cs < 1356) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1356) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 7) {
                                                if (AFD < 46) {
                                                    if (diffTSC < 26) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= 26) {
                                                        return 0;
                                                    }
                                                }
                                                else if (AFD >= 46) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 229) {
                                    return 0;
                                }
                            }
                            else if (SC >= 3327) {
                                if (diffTSC < 18) {
                                    if (posBalance < 104) {
                                        if (diffRsCsdiff < 439) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 439) {
                                            if (posBalance < 60) {
                                                return 1;
                                            }
                                            else if (posBalance >= 60) {
                                                if (gchDC < 44) {
                                                    if (MVDiff < 156) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 156) {
                                                        if (SC < 7554) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 7554) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 44) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 104) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 18) {
                                    if (RsDiff < 12) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 12) {
                                        if (posBalance < 125) {
                                            if (SC < 4217) {
                                                if (diffRsCsdiff < 460) {
                                                    if (RsCsDiff < 455) {
                                                        if (MVDiff < 288) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 288) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 455) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 460) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 4217) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 125) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect3(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (AFD < 29) {
        if (diffMVdiffVal < 66) {
            if (diffAFD < 9) {
                if (refDCval < 24) {
                    if (RsCsDiff < 231) {
                        if (diffRsCsdiff < 4) {
                            if (Rs < 1) {
                                if (ssDCval < 130) {
                                    return 0;
                                }
                                else if (ssDCval >= 130) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 4) {
                            if (negBalance < 4) {
                                if (CsDiff < 4) {
                                    if (refDCval < 17) {
                                        return 1;
                                    }
                                    else if (refDCval >= 17) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 4) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 231) {
                        return 1;
                    }
                }
                else if (refDCval >= 24) {
                    if (AFD < 22) {
                        if (refDCval < 173) {
                            return 0;
                        }
                        else if (refDCval >= 173) {
                            if (RsDiff < 36) {
                                if (gchDC < 30) {
                                    return 0;
                                }
                                else if (gchDC >= 30) {
                                    if (RsDiff < 1) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 1) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsDiff >= 36) {
                                if (RsDiff < 40) {
                                    return 1;
                                }
                                else if (RsDiff >= 40) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (AFD >= 22) {
                        if (refDCval < 48) {
                            if (RsDiff < 5) {
                                if (diffRsCsdiff < 31) {
                                    if (diffTSC < 5) {
                                        if (SC < 1063) {
                                            return 0;
                                        }
                                        else if (SC >= 1063) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 5) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 31) {
                                    return 1;
                                }
                            }
                            else if (RsDiff >= 5) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 48) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffAFD >= 9) {
                if (diffTSC < 13) {
                    if (refDCval < 57) {
                        if (SC < 101) {
                            if (ssDCval < 41) {
                                if (MVDiff < 73) {
                                    return 0;
                                }
                                else if (MVDiff >= 73) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 41) {
                                return 0;
                            }
                        }
                        else if (SC >= 101) {
                            if (diffMVdiffVal < 51) {
                                if (Rs < 5133) {
                                    if (diffMVdiffVal < -105) {
                                        if (Cs < 60) {
                                            return 1;
                                        }
                                        else if (Cs >= 60) {
                                            if (negBalance < 54) {
                                                return 0;
                                            }
                                            else if (negBalance >= 54) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -105) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 5133) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 51) {
                                if (CsDiff < 9) {
                                    if (gchDC < 13) {
                                        if (MVDiff < 233) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 233) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 13) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 9) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 57) {
                        if (CsDiff < 2) {
                            if (Rs < 2455) {
                                return 0;
                            }
                            else if (Rs >= 2455) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 2) {
                            return 0;
                        }
                    }
                }
                else if (diffTSC >= 13) {
                    if (RsCsDiff < 230) {
                        if (refDCval < 25) {
                            if (MVDiff < 361) {
                                return 1;
                            }
                            else if (MVDiff >= 361) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 25) {
                            if (AFD < 19) {
                                if (SC < 2080) {
                                    return 0;
                                }
                                else if (SC >= 2080) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 19) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 230) {
                        if (gchDC < 6) {
                            if (CsDiff < 12) {
                                return 1;
                            }
                            else if (CsDiff >= 12) {
                                return 0;
                            }
                        }
                        else if (gchDC >= 6) {
                            if (AFD < 25) {
                                return 1;
                            }
                            else if (AFD >= 25) {
                                if (TSC < 24) {
                                    return 0;
                                }
                                else if (TSC >= 24) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 66) {
            if (diffTSC < 6) {
                if (refDCval < 68) {
                    if (diffAFD < 8) {
                        if (CsDiff < 22) {
                            if (diffTSC < -1) {
                                return 0;
                            }
                            else if (diffTSC >= -1) {
                                if (ssDCval < 38) {
                                    if (AFD < 15) {
                                        if (diffTSC < 3) {
                                            if (Rs < 430) {
                                                return 0;
                                            }
                                            else if (Rs >= 430) {
                                                if (Rs < 483) {
                                                    return 1;
                                                }
                                                else if (Rs >= 483) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 3) {
                                            if (Cs < 31) {
                                                return 0;
                                            }
                                            else if (Cs >= 31) {
                                                if (MVDiff < 163) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 163) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 15) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 38) {
                                    if (refDCval < 35) {
                                        if (posBalance < 123) {
                                            return 0;
                                        }
                                        else if (posBalance >= 123) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 35) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 22) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (SC < 383) {
                            if (ssDCval < 31) {
                                return 1;
                            }
                            else if (ssDCval >= 31) {
                                if (diffMVdiffVal < 125) {
                                    if (CsDiff < 10) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 10) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 125) {
                                    if (Cs < 43) {
                                        if (Cs < 22) {
                                            return 1;
                                        }
                                        else if (Cs >= 22) {
                                            if (diffAFD < 11) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 11) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 43) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (SC >= 383) {
                            if (ssDCval < 40) {
                                if (MVDiff < 170) {
                                    return 1;
                                }
                                else if (MVDiff >= 170) {
                                    if (AFD < 22) {
                                        return 0;
                                    }
                                    else if (AFD >= 22) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 40) {
                                if (diffTSC < 5) {
                                    return 0;
                                }
                                else if (diffTSC >= 5) {
                                    if (TSC < 9) {
                                        return 1;
                                    }
                                    else if (TSC >= 9) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (refDCval >= 68) {
                    if (diffMVdiffVal < 667) {
                        if (posBalance < 128) {
                            if (diffMVdiffVal < 100) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 100) {
                                if (gchDC < 100) {
                                    if (TSC < 9) {
                                        return 0;
                                    }
                                    else if (TSC >= 9) {
                                        if (MVDiff < 159) {
                                            if (Cs < 535) {
                                                return 1;
                                            }
                                            else if (Cs >= 535) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 159) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 100) {
                                    if (SC < 13) {
                                        return 1;
                                    }
                                    else if (SC >= 13) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 128) {
                            return 1;
                        }
                    }
                    else if (diffMVdiffVal >= 667) {
                        if (TSC < 13) {
                            if (ssDCval < 34) {
                                if (AFD < 25) {
                                    return 1;
                                }
                                else if (AFD >= 25) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 34) {
                                if (RsCsDiff < 315) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 315) {
                                    return 1;
                                }
                            }
                        }
                        else if (TSC >= 13) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffTSC >= 6) {
                if (posBalance < 86) {
                    if (diffAFD < 12) {
                        if (gchDC < 10) {
                            if (SC < 499) {
                                if (SC < 71) {
                                    return 0;
                                }
                                else if (SC >= 71) {
                                    return 1;
                                }
                            }
                            else if (SC >= 499) {
                                if (diffMVdiffVal < 158) {
                                    if (Rs < 801) {
                                        return 1;
                                    }
                                    else if (Rs >= 801) {
                                        if (SC < 4114) {
                                            return 0;
                                        }
                                        else if (SC >= 4114) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 158) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 10) {
                            return 0;
                        }
                    }
                    else if (diffAFD >= 12) {
                        if (SC < 3578) {
                            if (negBalance < 34) {
                                if (CsDiff < 8) {
                                    if (Cs < 171) {
                                        return 1;
                                    }
                                    else if (Cs >= 171) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 8) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 34) {
                                if (SC < 524) {
                                    return 1;
                                }
                                else if (SC >= 524) {
                                    if (RsDiff < 8) {
                                        if (gchDC < 12) {
                                            if (TSC < 9) {
                                                if (RsCsDiff < 59) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 59) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 9) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 12) {
                                            if (gchDC < 34) {
                                                if (gchDC < 17) {
                                                    if (refDCval < 53) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 53) {
                                                        if (RsCsDiff < 99) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 99) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 17) {
                                                    if (RsDiff < 5) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 34) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 8) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (SC >= 3578) {
                            if (Rs < 4691) {
                                if (refDCval < 172) {
                                    return 0;
                                }
                                else if (refDCval >= 172) {
                                    if (AFD < 24) {
                                        return 0;
                                    }
                                    else if (AFD >= 24) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 4691) {
                                return 1;
                            }
                        }
                    }
                }
                else if (posBalance >= 86) {
                    if (negBalance < 25) {
                        if (Scindex < 3) {
                            if (RsCsDiff < 19) {
                                if (Cs < 9) {
                                    return 0;
                                }
                                else if (Cs >= 9) {
                                    if (TSC < 12) {
                                        if (SC < 258) {
                                            if (diffRsCsdiff < 1) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 1) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 258) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 12) {
                                        if (SC < 314) {
                                            return 0;
                                        }
                                        else if (SC >= 314) {
                                            if (TSC < 19) {
                                                return 1;
                                            }
                                            else if (TSC >= 19) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 19) {
                                return 1;
                            }
                        }
                        else if (Scindex >= 3) {
                            if (Rs < 1323) {
                                if (diffMVdiffVal < 657) {
                                    if (TSC < 11) {
                                        if (AFD < 16) {
                                            return 0;
                                        }
                                        else if (AFD >= 16) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 11) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 657) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1323) {
                                if (Rs < 1459) {
                                    return 1;
                                }
                                else if (Rs >= 1459) {
                                    if (RsDiff < 23) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 23) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 25) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (AFD >= 29) {
        if (diffAFD < 23) {
            if (TSC < 16) {
                if (MVDiff < 125) {
                    return 0;
                }
                else if (MVDiff >= 125) {
                    if (ssDCval < 61) {
                        if (diffTSC < -2) {
                            if (CsDiff < 3) {
                                if (CsDiff < 2) {
                                    return 0;
                                }
                                else if (CsDiff >= 2) {
                                    if (diffTSC < -5) {
                                        return 0;
                                    }
                                    else if (diffTSC >= -5) {
                                        return 1;
                                    }
                                }
                            }
                            else if (CsDiff >= 3) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= -2) {
                            if (RsCsDiff < 4) {
                                return 1;
                            }
                            else if (RsCsDiff >= 4) {
                                if (CsDiff < 6) {
                                    if (TSC < 13) {
                                        if (diffRsCsdiff < 6) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 6) {
                                            if (RsCsDiff < 33) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 33) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 13) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 6) {
                                    if (Cs < 435) {
                                        return 1;
                                    }
                                    else if (Cs >= 435) {
                                        if (MVDiff < 278) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 278) {
                                            if (diffRsCsdiff < -27) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= -27) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 61) {
                        if (gchDC < 119) {
                            return 0;
                        }
                        else if (gchDC >= 119) {
                            return 1;
                        }
                    }
                }
            }
            else if (TSC >= 16) {
                if (diffMVdiffVal < 92) {
                    if (refDCval < 26) {
                        if (Cs < 298) {
                            return 0;
                        }
                        else if (Cs >= 298) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 26) {
                        if (MVDiff < 46) {
                            return 0;
                        }
                        else if (MVDiff >= 46) {
                            if (AFD < 58) {
                                if (diffAFD < 19) {
                                    if (diffRsCsdiff < 96) {
                                        if (TSC < 20) {
                                            return 0;
                                        }
                                        else if (TSC >= 20) {
                                            if (Scindex < 1) {
                                                if (Rs < 21) {
                                                    return 1;
                                                }
                                                else if (Rs >= 21) {
                                                    return 0;
                                                }
                                            }
                                            else if (Scindex >= 1) {
                                                if (diffMVdiffVal < -38) {
                                                    if (diffMVdiffVal < -40) {
                                                        if (ssDCval < 59) {
                                                            if (AFD < 45) {
                                                                return 0;
                                                            }
                                                            else if (AFD >= 45) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 59) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -40) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= -38) {
                                                    if (Rs < 1738) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 1738) {
                                                        if (Rs < 1777) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1777) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 96) {
                                        if (RsDiff < 13) {
                                            if (diffTSC < 9) {
                                                if (Cs < 278) {
                                                    return 1;
                                                }
                                                else if (Cs >= 278) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 9) {
                                                if (diffTSC < 15) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 15) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 13) {
                                            if (refDCval < 79) {
                                                if (gchDC < 9) {
                                                    if (MVDiff < 148) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 148) {
                                                        return 0;
                                                    }
                                                }
                                                else if (gchDC >= 9) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 79) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 19) {
                                    if (refDCval < 61) {
                                        return 0;
                                    }
                                    else if (refDCval >= 61) {
                                        if (diffMVdiffVal < 56) {
                                            if (Rs < 622) {
                                                if (Rs < 193) {
                                                    return 0;
                                                }
                                                else if (Rs >= 193) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 622) {
                                                if (RsCsDiff < 16225) {
                                                    if (diffMVdiffVal < -57) {
                                                        if (AFD < 43) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 43) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= -57) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsCsDiff >= 16225) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 56) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 58) {
                                if (Scindex < 4) {
                                    return 0;
                                }
                                else if (Scindex >= 4) {
                                    if (RsDiff < 24) {
                                        if (diffMVdiffVal < -5) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= -5) {
                                            if (Cs < 1466) {
                                                return 0;
                                            }
                                            else if (Cs >= 1466) {
                                                if (SC < 3716) {
                                                    return 1;
                                                }
                                                else if (SC >= 3716) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 24) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 92) {
                    if (diffTSC < 1) {
                        if (ssDCval < 39) {
                            if (Cs < 302) {
                                return 0;
                            }
                            else if (Cs >= 302) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 39) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= 1) {
                        if (diffRsCsdiff < 85) {
                            if (refDCval < 127) {
                                if (ssDCval < 90) {
                                    if (negBalance < 70) {
                                        if (TSC < 20) {
                                            if (diffRsCsdiff < 36) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 36) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 70) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 90) {
                                    if (Cs < 450) {
                                        if (ssDCval < 108) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 108) {
                                            if (AFD < 48) {
                                                return 1;
                                            }
                                            else if (AFD >= 48) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Cs >= 450) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 127) {
                                if (Cs < 53) {
                                    return 1;
                                }
                                else if (Cs >= 53) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 85) {
                            if (Rs < 1456) {
                                if (gchDC < 31) {
                                    if (posBalance < 89) {
                                        if (diffTSC < 7) {
                                            if (RsCsDiff < 1287) {
                                                if (gchDC < 3) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 3) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsCsDiff >= 1287) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 89) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 31) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1456) {
                                if (gchDC < 21) {
                                    if (diffTSC < 9) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 9) {
                                        if (gchDC < 19) {
                                            return 1;
                                        }
                                        else if (gchDC >= 19) {
                                            if (AFD < 42) {
                                                return 1;
                                            }
                                            else if (AFD >= 42) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 21) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 23) {
            if (posBalance < 81) {
                if (diffRsCsdiff < 1) {
                    if (Cs < 204) {
                        return 1;
                    }
                    else if (Cs >= 204) {
                        if (AFD < 40) {
                            return 1;
                        }
                        else if (AFD >= 40) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 1) {
                    if (ssDCval < 82) {
                        if (TSC < 14) {
                            if (diffMVdiffVal < 132) {
                                if (AFD < 30) {
                                    if (ssDCval < 64) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 64) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 30) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 132) {
                                return 1;
                            }
                        }
                        else if (TSC >= 14) {
                            if (TSC < 23) {
                                if (diffMVdiffVal < 77) {
                                    if (RsCsDiff < 291) {
                                        if (AFD < 40) {
                                            return 1;
                                        }
                                        else if (AFD >= 40) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 291) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 77) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 23) {
                                if (negBalance < 39) {
                                    if (MVDiff < 244) {
                                        if (diffAFD < 33) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 33) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 244) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 39) {
                                    if (negBalance < 68) {
                                        if (ssDCval < 55) {
                                            if (RsCsDiff < 670) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 670) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 55) {
                                            if (AFD < 64) {
                                                return 1;
                                            }
                                            else if (AFD >= 64) {
                                                if (Rs < 1449) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1449) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 68) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 82) {
                        if (TSC < 18) {
                            if (diffRsCsdiff < 101) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= 101) {
                                if (gchDC < 8) {
                                    if (TSC < 16) {
                                        return 0;
                                    }
                                    else if (TSC >= 16) {
                                        if (negBalance < 40) {
                                            return 0;
                                        }
                                        else if (negBalance >= 40) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 8) {
                                    if (TSC < 13) {
                                        return 0;
                                    }
                                    else if (TSC >= 13) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 18) {
                            if (diffAFD < 29) {
                                if (AFD < 32) {
                                    return 1;
                                }
                                else if (AFD >= 32) {
                                    if (diffRsCsdiff < 186) {
                                        if (TSC < 51) {
                                            if (Cs < 737) {
                                                if (MVDiff < 262) {
                                                    if (Cs < 553) {
                                                        if (Rs < 136) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 136) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (Cs >= 553) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 262) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 737) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 51) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 186) {
                                        if (SC < 6949) {
                                            return 1;
                                        }
                                        else if (SC >= 6949) {
                                            if (TSC < 27) {
                                                return 0;
                                            }
                                            else if (TSC >= 27) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 29) {
                                if (RsDiff < 6) {
                                    if (refDCval < 67) {
                                        return 0;
                                    }
                                    else if (refDCval >= 67) {
                                        if (TSC < 46) {
                                            return 1;
                                        }
                                        else if (TSC >= 46) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 6) {
                                    if (AFD < 45) {
                                        if (Scindex < 7) {
                                            if (SC < 2767) {
                                                if (posBalance < 36) {
                                                    if (posBalance < 33) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 33) {
                                                        return 0;
                                                    }
                                                }
                                                else if (posBalance >= 36) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2767) {
                                                return 1;
                                            }
                                        }
                                        else if (Scindex >= 7) {
                                            if (CsDiff < 33) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 33) {
                                                if (TSC < 22) {
                                                    return 0;
                                                }
                                                else if (TSC >= 22) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 45) {
                                        if (gchDC < 6) {
                                            if (negBalance < 54) {
                                                if (diffRsCsdiff < 1516) {
                                                    if (CsDiff < 16) {
                                                        if (SC < 1792) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 1792) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (CsDiff >= 16) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 1516) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 54) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (posBalance >= 81) {
                if (CsDiff < 16) {
                    if (diffMVdiffVal < 106) {
                        if (RsCsDiff < 831) {
                            if (MVDiff < 40) {
                                return 1;
                            }
                            else if (MVDiff >= 40) {
                                if (MVDiff < 645) {
                                    if (diffAFD < 30) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 30) {
                                        if (AFD < 44) {
                                            if (diffMVdiffVal < 90) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 90) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 44) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 645) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsCsDiff >= 831) {
                            return 1;
                        }
                    }
                    else if (diffMVdiffVal >= 106) {
                        if (AFD < 51) {
                            if (diffAFD < 32) {
                                if (negBalance < 6) {
                                    return 0;
                                }
                                else if (negBalance >= 6) {
                                    if (diffRsCsdiff < 88) {
                                        if (SC < 797) {
                                            return 1;
                                        }
                                        else if (SC >= 797) {
                                            if (negBalance < 10) {
                                                return 1;
                                            }
                                            else if (negBalance >= 10) {
                                                if (TSC < 19) {
                                                    return 1;
                                                }
                                                else if (TSC >= 19) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 88) {
                                        if (Rs < 2087) {
                                            return 1;
                                        }
                                        else if (Rs >= 2087) {
                                            if (Cs < 652) {
                                                return 0;
                                            }
                                            else if (Cs >= 652) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 32) {
                                if (negBalance < 8) {
                                    if (RsDiff < 11) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 11) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 8) {
                                    if (RsDiff < 5) {
                                        if (Cs < 600) {
                                            return 1;
                                        }
                                        else if (Cs >= 600) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 51) {
                            if (RsDiff < 7) {
                                if (MVDiff < 262) {
                                    if (Cs < 38) {
                                        return 0;
                                    }
                                    else if (Cs >= 38) {
                                        if (gchDC < 54) {
                                            return 0;
                                        }
                                        else if (gchDC >= 54) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 262) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 7) {
                                if (gchDC < 121) {
                                    if (Rs < 1599) {
                                        return 1;
                                    }
                                    else if (Rs >= 1599) {
                                        if (refDCval < 89) {
                                            return 1;
                                        }
                                        else if (refDCval >= 89) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 121) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (CsDiff >= 16) {
                    if (negBalance < 6) {
                        if (ssDCval < 123) {
                            return 1;
                        }
                        else if (ssDCval >= 123) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 6) {
                        if (diffTSC < 15) {
                            if (diffTSC < 14) {
                                return 1;
                            }
                            else if (diffTSC >= 14) {
                                if (negBalance < 30) {
                                    return 0;
                                }
                                else if (negBalance >= 30) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffTSC >= 15) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect4(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (TSC < 16) {
        if (Rs < 569) {
            if (diffAFD < 9) {
                if (MVDiff < 127) {
                    if (RsCsDiff < 510) {
                        if (posBalance < 2) {
                            if (Rs < 566) {
                                if (SC < 76) {
                                    if (SC < 73) {
                                        if (negBalance < 106) {
                                            return 0;
                                        }
                                        else if (negBalance >= 106) {
                                            if (refDCval < 211) {
                                                return 0;
                                            }
                                            else if (refDCval >= 211) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 73) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 76) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 566) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 2) {
                            if (diffAFD < 8) {
                                return 0;
                            }
                            else if (diffAFD >= 8) {
                                if (diffMVdiffVal < 85) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 85) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 510) {
                        if (RsCsDiff < 1568) {
                            return 1;
                        }
                        else if (RsCsDiff >= 1568) {
                            return 0;
                        }
                    }
                }
                else if (MVDiff >= 127) {
                    if (negBalance < 66) {
                        if (CsDiff < 3) {
                            if (diffMVdiffVal < 224) {
                                if (MVDiff < 1768) {
                                    return 0;
                                }
                                else if (MVDiff >= 1768) {
                                    if (gchDC < 101) {
                                        return 0;
                                    }
                                    else if (gchDC >= 101) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 224) {
                                if (MVDiff < 393) {
                                    if (Rs < 69) {
                                        if (AFD < 7) {
                                            return 0;
                                        }
                                        else if (AFD >= 7) {
                                            if (AFD < 9) {
                                                return 1;
                                            }
                                            else if (AFD >= 9) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 69) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 393) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 3) {
                            if (Rs < 1) {
                                return 1;
                            }
                            else if (Rs >= 1) {
                                if (refDCval < 28) {
                                    if (TSC < 8) {
                                        return 1;
                                    }
                                    else if (TSC >= 8) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 28) {
                                    if (SC < 539) {
                                        if (diffRsCsdiff < 8) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 8) {
                                            if (Cs < 243) {
                                                return 0;
                                            }
                                            else if (Cs >= 243) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 539) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 66) {
                        if (ssDCval < 25) {
                            if (refDCval < 27) {
                                if (negBalance < 71) {
                                    return 0;
                                }
                                else if (negBalance >= 71) {
                                    if (MVDiff < 2641) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 2641) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 27) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 25) {
                            if (posBalance < 42) {
                                if (diffRsCsdiff < 36) {
                                    if (refDCval < 227) {
                                        if (diffRsCsdiff < -241) {
                                            if (diffRsCsdiff < -282) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= -282) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -241) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 227) {
                                        if (ssDCval < 226) {
                                            if (TSC < 7) {
                                                return 1;
                                            }
                                            else if (TSC >= 7) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 226) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 36) {
                                    if (Rs < 296) {
                                        return 1;
                                    }
                                    else if (Rs >= 296) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 42) {
                                if (refDCval < 51) {
                                    if (diffAFD < 2) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 2) {
                                        if (gchDC < 2) {
                                            return 0;
                                        }
                                        else if (gchDC >= 2) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 51) {
                                    if (posBalance < 59) {
                                        return 0;
                                    }
                                    else if (posBalance >= 59) {
                                        if (TSC < 14) {
                                            return 1;
                                        }
                                        else if (TSC >= 14) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 9) {
                if (diffMVdiffVal < 64) {
                    if (refDCval < 65) {
                        if (Rs < 99) {
                            if (negBalance < 43) {
                                if (RsDiff < 13) {
                                    return 0;
                                }
                                else if (RsDiff >= 13) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 43) {
                                if (RsDiff < 2) {
                                    return 1;
                                }
                                else if (RsDiff >= 2) {
                                    if (Rs < 86) {
                                        return 0;
                                    }
                                    else if (Rs >= 86) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 99) {
                            if (CsDiff < 13) {
                                if (negBalance < 54) {
                                    return 0;
                                }
                                else if (negBalance >= 54) {
                                    if (diffAFD < 11) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 11) {
                                        if (SC < 1066) {
                                            return 1;
                                        }
                                        else if (SC >= 1066) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 13) {
                                return 1;
                            }
                        }
                    }
                    else if (refDCval >= 65) {
                        if (posBalance < 7) {
                            if (posBalance < 6) {
                                return 0;
                            }
                            else if (posBalance >= 6) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 7) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 64) {
                    if (diffTSC < 6) {
                        if (diffRsCsdiff < 53) {
                            if (SC < 302) {
                                if (refDCval < 78) {
                                    if (diffMVdiffVal < 234) {
                                        if (diffRsCsdiff < 14) {
                                            if (gchDC < 9) {
                                                if (SC < 58) {
                                                    if (TSC < 6) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 6) {
                                                        return 0;
                                                    }
                                                }
                                                else if (SC >= 58) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 9) {
                                                if (gchDC < 20) {
                                                    if (diffTSC < 4) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 4) {
                                                        if (SC < 125) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 125) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 20) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 14) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 234) {
                                        if (gchDC < 5) {
                                            if (Rs < 16) {
                                                return 0;
                                            }
                                            else if (Rs >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 5) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 78) {
                                    return 0;
                                }
                            }
                            else if (SC >= 302) {
                                if (refDCval < 45) {
                                    if (diffAFD < 10) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 10) {
                                        if (RsDiff < 3) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 45) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 53) {
                            if (diffTSC < 1) {
                                if (Rs < 149) {
                                    return 1;
                                }
                                else if (Rs >= 149) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 1) {
                                return 1;
                            }
                        }
                    }
                    else if (diffTSC >= 6) {
                        if (ssDCval < 127) {
                            if (diffAFD < 13) {
                                if (SC < 274) {
                                    if (diffTSC < 8) {
                                        return 1;
                                    }
                                    else if (diffTSC >= 8) {
                                        if (Cs < 87) {
                                            return 1;
                                        }
                                        else if (Cs >= 87) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 274) {
                                    if (ssDCval < 47) {
                                        if (AFD < 11) {
                                            return 0;
                                        }
                                        else if (AFD >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 47) {
                                        if (diffRsCsdiff < 31) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 31) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 13) {
                                if (AFD < 31) {
                                    if (CsDiff < 1) {
                                        if (SC < 185) {
                                            return 1;
                                        }
                                        else if (SC >= 185) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 31) {
                                    if (RsCsDiff < 46) {
                                        if (Rs < 278) {
                                            if (Cs < 55) {
                                                if (Rs < 33) {
                                                    return 1;
                                                }
                                                else if (Rs >= 33) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 55) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 278) {
                                            if (ssDCval < 61) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 61) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 46) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 127) {
                            if (RsCsDiff < 43) {
                                if (CsDiff < 4) {
                                    return 0;
                                }
                                else if (CsDiff >= 4) {
                                    if (TSC < 12) {
                                        return 1;
                                    }
                                    else if (TSC >= 12) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 43) {
                                if (diffAFD < 32) {
                                    if (CsDiff < 5) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 32) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (Rs >= 569) {
            if (diffAFD < 11) {
                if (refDCval < 21) {
                    if (diffAFD < 2) {
                        if (diffMVdiffVal < -1) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= -1) {
                            if (MVDiff < 6) {
                                if (AFD < 1) {
                                    if (RsCsDiff < 4) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 4) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 1) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 6) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 2) {
                        return 1;
                    }
                }
                else if (refDCval >= 21) {
                    if (gchDC < 3) {
                        return 0;
                    }
                    else if (gchDC >= 3) {
                        if (ssDCval < 43) {
                            if (RsCsDiff < 410) {
                                if (diffTSC < 4) {
                                    return 0;
                                }
                                else if (diffTSC >= 4) {
                                    if (diffMVdiffVal < 92) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 92) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 410) {
                                if (diffMVdiffVal < 12) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 12) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 43) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffAFD >= 11) {
                if (MVDiff < 56) {
                    if (Cs < 340) {
                        if (Rs < 1979) {
                            return 0;
                        }
                        else if (Rs >= 1979) {
                            if (AFD < 13) {
                                return 0;
                            }
                            else if (AFD >= 13) {
                                return 1;
                            }
                        }
                    }
                    else if (Cs >= 340) {
                        if (diffAFD < 17) {
                            return 0;
                        }
                        else if (diffAFD >= 17) {
                            if (SC < 3770) {
                                if (SC < 3706) {
                                    if (refDCval < 68) {
                                        if (SC < 2270) {
                                            return 0;
                                        }
                                        else if (SC >= 2270) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 68) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 3706) {
                                    return 1;
                                }
                            }
                            else if (SC >= 3770) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 56) {
                    if (diffMVdiffVal < 54) {
                        if (ssDCval < 41) {
                            if (AFD < 27) {
                                return 1;
                            }
                            else if (AFD >= 27) {
                                return 0;
                            }
                        }
                        else if (ssDCval >= 41) {
                            if (RsDiff < 22) {
                                return 0;
                            }
                            else if (RsDiff >= 22) {
                                if (Rs < 1833) {
                                    return 1;
                                }
                                else if (Rs >= 1833) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 54) {
                        if (ssDCval < 75) {
                            if (diffAFD < 17) {
                                if (diffRsCsdiff < 84) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 84) {
                                    if (Rs < 2030) {
                                        return 1;
                                    }
                                    else if (Rs >= 2030) {
                                        if (AFD < 18) {
                                            return 1;
                                        }
                                        else if (AFD >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 17) {
                                if (diffRsCsdiff < 321) {
                                    if (Rs < 2062) {
                                        if (MVDiff < 178) {
                                            if (gchDC < 16) {
                                                if (diffTSC < 9) {
                                                    if (RsDiff < 12) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffTSC >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 16) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 178) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2062) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 321) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 75) {
                            if (diffAFD < 21) {
                                if (Rs < 1277) {
                                    if (diffRsCsdiff < 84) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 84) {
                                        if (CsDiff < 10) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 10) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Rs >= 1277) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 21) {
                                if (gchDC < 3) {
                                    return 0;
                                }
                                else if (gchDC >= 3) {
                                    if (Cs < 1995) {
                                        return 1;
                                    }
                                    else if (Cs >= 1995) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (TSC >= 16) {
        if (diffTSC < 10) {
            if (Scindex < 5) {
                if (diffAFD < 13) {
                    if (CsDiff < 20) {
                        if (CsDiff < 4) {
                            return 0;
                        }
                        else if (CsDiff >= 4) {
                            if (AFD < 47) {
                                if (CsDiff < 10) {
                                    if (Rs < 449) {
                                        if (SC < 968) {
                                            return 0;
                                        }
                                        else if (SC >= 968) {
                                            if (Rs < 444) {
                                                return 0;
                                            }
                                            else if (Rs >= 444) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 449) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 10) {
                                    if (gchDC < 13) {
                                        if (MVDiff < 122) {
                                            if (RsDiff < 81) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 81) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 122) {
                                            if (MVDiff < 140) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 140) {
                                                if (negBalance < 54) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 54) {
                                                    if (RsCsDiff < 547) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 547) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 13) {
                                        if (refDCval < 47) {
                                            return 1;
                                        }
                                        else if (refDCval >= 47) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 47) {
                                if (posBalance < 51) {
                                    if (diffRsCsdiff < 40) {
                                        if (Cs < 349) {
                                            return 1;
                                        }
                                        else if (Cs >= 349) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 40) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 51) {
                                    if (RsCsDiff < 200) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 200) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 20) {
                        if (SC < 3703) {
                            if (diffMVdiffVal < 5) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= 5) {
                                if (diffMVdiffVal < 98) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 98) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 3703) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 13) {
                    if (RsCsDiff < 147) {
                        if (negBalance < 89) {
                            if (diffRsCsdiff < 82) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 82) {
                                if (MVDiff < 1455) {
                                    return 0;
                                }
                                else if (MVDiff >= 1455) {
                                    return 1;
                                }
                            }
                        }
                        else if (negBalance >= 89) {
                            if (posBalance < 29) {
                                if (Cs < 53) {
                                    return 1;
                                }
                                else if (Cs >= 53) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 29) {
                                return 1;
                            }
                        }
                    }
                    else if (RsCsDiff >= 147) {
                        if (MVDiff < 526) {
                            if (diffMVdiffVal < 193) {
                                if (posBalance < 76) {
                                    if (MVDiff < 116) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 116) {
                                        if (TSC < 20) {
                                            if (Cs < 969) {
                                                return 1;
                                            }
                                            else if (Cs >= 969) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            if (Cs < 1425) {
                                                return 0;
                                            }
                                            else if (Cs >= 1425) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 76) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 193) {
                                if (diffMVdiffVal < 277) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 277) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 526) {
                            if (Rs < 1749) {
                                return 1;
                            }
                            else if (Rs >= 1749) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (Scindex >= 5) {
                if (diffMVdiffVal < 36) {
                    return 0;
                }
                else if (diffMVdiffVal >= 36) {
                    if (CsDiff < 23) {
                        return 0;
                    }
                    else if (CsDiff >= 23) {
                        if (Cs < 2238) {
                            return 1;
                        }
                        else if (Cs >= 2238) {
                            if (refDCval < 160) {
                                if (diffMVdiffVal < 205) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 205) {
                                    if (Cs < 4026) {
                                        return 1;
                                    }
                                    else if (Cs >= 4026) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 160) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 10) {
            if (negBalance < 25) {
                if (CsDiff < 13) {
                    if (diffMVdiffVal < 111) {
                        if (gchDC < 127) {
                            if (SC < 113) {
                                return 1;
                            }
                            else if (SC >= 113) {
                                if (RsCsDiff < 127) {
                                    if (MVDiff < 533) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 533) {
                                        if (RsDiff < 3) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 127) {
                                    if (Cs < 683) {
                                        return 1;
                                    }
                                    else if (Cs >= 683) {
                                        if (refDCval < 26) {
                                            if (AFD < 33) {
                                                return 1;
                                            }
                                            else if (AFD >= 33) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 26) {
                                            if (Rs < 1155) {
                                                if (Cs < 995) {
                                                    return 1;
                                                }
                                                else if (Cs >= 995) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 1155) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 127) {
                            return 1;
                        }
                    }
                    else if (diffMVdiffVal >= 111) {
                        if (negBalance < 10) {
                            if (diffAFD < 56) {
                                if (ssDCval < 127) {
                                    if (SC < 122) {
                                        return 1;
                                    }
                                    else if (SC >= 122) {
                                        if (diffAFD < 17) {
                                            if (AFD < 22) {
                                                return 1;
                                            }
                                            else if (AFD >= 22) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 17) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 127) {
                                    if (Cs < 1006) {
                                        return 1;
                                    }
                                    else if (Cs >= 1006) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 56) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 10) {
                            if (refDCval < 58) {
                                if (posBalance < 84) {
                                    return 1;
                                }
                                else if (posBalance >= 84) {
                                    if (SC < 290) {
                                        return 0;
                                    }
                                    else if (SC >= 290) {
                                        if (diffRsCsdiff < 187) {
                                            if (Rs < 904) {
                                                if (Cs < 576) {
                                                    if (ssDCval < 68) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 68) {
                                                        if (diffTSC < 22) {
                                                            return 1;
                                                        }
                                                        else if (diffTSC >= 22) {
                                                            if (Cs < 272) {
                                                                return 0;
                                                            }
                                                            else if (Cs >= 272) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 576) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 904) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 187) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 58) {
                                if (Rs < 1354) {
                                    return 1;
                                }
                                else if (Rs >= 1354) {
                                    if (Cs < 1131) {
                                        return 1;
                                    }
                                    else if (Cs >= 1131) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (CsDiff >= 13) {
                    if (AFD < 44) {
                        if (refDCval < 27) {
                            return 1;
                        }
                        else if (refDCval >= 27) {
                            if (Cs < 1316) {
                                return 1;
                            }
                            else if (Cs >= 1316) {
                                if (negBalance < 14) {
                                    return 0;
                                }
                                else if (negBalance >= 14) {
                                    if (diffTSC < 18) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 18) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 44) {
                        if (MVDiff < 359) {
                            if (SC < 4852) {
                                return 1;
                            }
                            else if (SC >= 4852) {
                                if (Cs < 5815) {
                                    return 0;
                                }
                                else if (Cs >= 5815) {
                                    return 1;
                                }
                            }
                        }
                        else if (MVDiff >= 359) {
                            if (CsDiff < 18) {
                                if (AFD < 63) {
                                    return 0;
                                }
                                else if (AFD >= 63) {
                                    return 1;
                                }
                            }
                            else if (CsDiff >= 18) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 25) {
                if (diffRsCsdiff < 12) {
                    if (diffMVdiffVal < 96) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 96) {
                        if (diffAFD < 11) {
                            if (MVDiff < 1095) {
                                return 0;
                            }
                            else if (MVDiff >= 1095) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 11) {
                            if (AFD < 48) {
                                if (Cs < 345) {
                                    if (refDCval < 39) {
                                        return 0;
                                    }
                                    else if (refDCval >= 39) {
                                        if (Rs < 22) {
                                            return 0;
                                        }
                                        else if (Rs >= 22) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 345) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 48) {
                                if (RsDiff < 5) {
                                    return 0;
                                }
                                else if (RsDiff >= 5) {
                                    if (TSC < 33) {
                                        if (AFD < 53) {
                                            return 0;
                                        }
                                        else if (AFD >= 53) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 33) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 12) {
                    if (diffMVdiffVal < 62) {
                        if (Cs < 448) {
                            return 1;
                        }
                        else if (Cs >= 448) {
                            if (diffAFD < 23) {
                                if (refDCval < 59) {
                                    if (Rs < 5774) {
                                        return 0;
                                    }
                                    else if (Rs >= 5774) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 59) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 23) {
                                if (TSC < 24) {
                                    if (negBalance < 63) {
                                        if (diffMVdiffVal < 50) {
                                            if (diffAFD < 35) {
                                                if (gchDC < 5) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 5) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffAFD >= 35) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 50) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 63) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 24) {
                                    if (Cs < 2677) {
                                        if (SC < 2388) {
                                            if (SC < 1637) {
                                                return 1;
                                            }
                                            else if (SC >= 1637) {
                                                return 0;
                                            }
                                        }
                                        else if (SC >= 2388) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 2677) {
                                        if (posBalance < 50) {
                                            return 0;
                                        }
                                        else if (posBalance >= 50) {
                                            if (Rs < 9944) {
                                                return 1;
                                            }
                                            else if (Rs >= 9944) {
                                                if (Rs < 11454) {
                                                    return 0;
                                                }
                                                else if (Rs >= 11454) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 62) {
                        if (diffRsCsdiff < 238) {
                            if (SC < 3032) {
                                if (diffMVdiffVal < 82) {
                                    if (diffMVdiffVal < 76) {
                                        if (diffAFD < 35) {
                                            if (Cs < 624) {
                                                return 1;
                                            }
                                            else if (Cs >= 624) {
                                                if (AFD < 55) {
                                                    return 1;
                                                }
                                                else if (AFD >= 55) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 35) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 76) {
                                        if (MVDiff < 93) {
                                            if (Rs < 1285) {
                                                return 1;
                                            }
                                            else if (Rs >= 1285) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 93) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 82) {
                                    if (ssDCval < 92) {
                                        if (negBalance < 74) {
                                            if (MVDiff < 634) {
                                                if (RsDiff < 14) {
                                                    if (SC < 324) {
                                                        if (CsDiff < 5) {
                                                            return 0;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 324) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 14) {
                                                    if (Cs < 679) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 679) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 634) {
                                                if (Rs < 762) {
                                                    if (refDCval < 62) {
                                                        if (Rs < 689) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 689) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (refDCval >= 62) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Rs >= 762) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 74) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 92) {
                                        if (refDCval < 83) {
                                            if (TSC < 20) {
                                                return 0;
                                            }
                                            else if (TSC >= 20) {
                                                if (diffAFD < 32) {
                                                    if (Rs < 380) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 380) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 32) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 83) {
                                            if (RsCsDiff < 98) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 98) {
                                                if (refDCval < 151) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 151) {
                                                    if (CsDiff < 6) {
                                                        return 1;
                                                    }
                                                    else if (CsDiff >= 6) {
                                                        if (SC < 437) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 437) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 3032) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 238) {
                            if (Rs < 2300) {
                                if (MVDiff < 358) {
                                    return 1;
                                }
                                else if (MVDiff >= 358) {
                                    if (negBalance < 28) {
                                        if (AFD < 50) {
                                            return 0;
                                        }
                                        else if (AFD >= 50) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 28) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 2300) {
                                if (diffRsCsdiff < 833) {
                                    if (diffAFD < 33) {
                                        if (diffMVdiffVal < 97) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 97) {
                                            if (gchDC < 17) {
                                                return 1;
                                            }
                                            else if (gchDC >= 17) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 33) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 833) {
                                    if (negBalance < 121) {
                                        if (diffAFD < 15) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 15) {
                                            if (RsDiff < 22) {
                                                if (AFD < 66) {
                                                    return 1;
                                                }
                                                else if (AFD >= 66) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 22) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 121) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect5(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 12) {
        if (diffTSC < 4) {
            if (diffMVdiffVal < 53) {
                if (posBalance < 9) {
                    if (RsDiff < 11) {
                        if (CsDiff < 2) {
                            if (negBalance < 123) {
                                return 0;
                            }
                            else if (negBalance >= 123) {
                                if (AFD < 1) {
                                    if (Rs < 17) {
                                        return 0;
                                    }
                                    else if (Rs >= 17) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 1) {
                                    if (diffTSC < -2) {
                                        if (MVDiff < 570) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 570) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= -2) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 2) {
                            if (ssDCval < 18) {
                                if (MVDiff < 6) {
                                    return 1;
                                }
                                else if (MVDiff >= 6) {
                                    if (TSCindex < 1) {
                                        if (Cs < 588) {
                                            return 1;
                                        }
                                        else if (Cs >= 588) {
                                            return 0;
                                        }
                                    }
                                    else if (TSCindex >= 1) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 18) {
                                return 0;
                            }
                        }
                    }
                    else if (RsDiff >= 11) {
                        if (refDCval < 85) {
                            if (CsDiff < 12) {
                                return 0;
                            }
                            else if (CsDiff >= 12) {
                                if (Rs < 234) {
                                    if (TSC < 1) {
                                        return 0;
                                    }
                                    else if (TSC >= 1) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 234) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 85) {
                            return 0;
                        }
                    }
                }
                else if (posBalance >= 9) {
                    if (AFD < 43) {
                        if (RsCsDiff < 2455) {
                            if (diffAFD < -7) {
                                if (Cs < 27) {
                                    if (RsDiff < 4) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 4) {
                                        if (Rs < 74) {
                                            return 1;
                                        }
                                        else if (Rs >= 74) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Cs >= 27) {
                                    if (negBalance < 46) {
                                        return 0;
                                    }
                                    else if (negBalance >= 46) {
                                        if (gchDC < 15) {
                                            return 0;
                                        }
                                        else if (gchDC >= 15) {
                                            if (posBalance < 62) {
                                                return 0;
                                            }
                                            else if (posBalance >= 62) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= -7) {
                                if (diffTSC < -4) {
                                    if (SC < 307) {
                                        if (diffRsCsdiff < 11) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 307) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= -4) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsCsDiff >= 2455) {
                            if (Scindex < 5) {
                                if (Rs < 1541) {
                                    return 0;
                                }
                                else if (Rs >= 1541) {
                                    return 1;
                                }
                            }
                            else if (Scindex >= 5) {
                                return 0;
                            }
                        }
                    }
                    else if (AFD >= 43) {
                        if (refDCval < 117) {
                            if (refDCval < 116) {
                                if (CsDiff < 19) {
                                    return 0;
                                }
                                else if (CsDiff >= 19) {
                                    if (Scindex < 5) {
                                        if (Cs < 1905) {
                                            return 1;
                                        }
                                        else if (Cs >= 1905) {
                                            return 0;
                                        }
                                    }
                                    else if (Scindex >= 5) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 116) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 117) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 53) {
                if (gchDC < 124) {
                    if (diffAFD < 7) {
                        if (diffRsCsdiff < 12) {
                            if (CsDiff < 6) {
                                if (refDCval < 74) {
                                    if (ssDCval < 76) {
                                        if (posBalance < 80) {
                                            return 0;
                                        }
                                        else if (posBalance >= 80) {
                                            if (gchDC < 3) {
                                                if (diffMVdiffVal < 275) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 275) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 3) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 76) {
                                        if (Rs < 901) {
                                            return 0;
                                        }
                                        else if (Rs >= 901) {
                                            if (AFD < 31) {
                                                return 0;
                                            }
                                            else if (AFD >= 31) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 74) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 6) {
                                if (ssDCval < 33) {
                                    return 1;
                                }
                                else if (ssDCval >= 33) {
                                    if (Scindex < 3) {
                                        if (TSC < 18) {
                                            return 0;
                                        }
                                        else if (TSC >= 18) {
                                            return 1;
                                        }
                                    }
                                    else if (Scindex >= 3) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 12) {
                            if (SC < 980) {
                                if (refDCval < 34) {
                                    return 1;
                                }
                                else if (refDCval >= 34) {
                                    if (Cs < 280) {
                                        return 0;
                                    }
                                    else if (Cs >= 280) {
                                        if (Cs < 290) {
                                            return 1;
                                        }
                                        else if (Cs >= 290) {
                                            if (MVDiff < 733) {
                                                if (AFD < 38) {
                                                    return 0;
                                                }
                                                else if (AFD >= 38) {
                                                    if (RsDiff < 5) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 733) {
                                                if (AFD < 25) {
                                                    return 1;
                                                }
                                                else if (AFD >= 25) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 980) {
                                if (CsDiff < 26) {
                                    return 0;
                                }
                                else if (CsDiff >= 26) {
                                    if (RsDiff < 23) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 23) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (posBalance < 50) {
                            if (AFD < 16) {
                                if (TSCindex < 7) {
                                    if (AFD < 9) {
                                        if (MVDiff < 201) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 201) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 9) {
                                        return 1;
                                    }
                                }
                                else if (TSCindex >= 7) {
                                    if (negBalance < 70) {
                                        return 1;
                                    }
                                    else if (negBalance >= 70) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 16) {
                                if (RsDiff < 6) {
                                    return 0;
                                }
                                else if (RsDiff >= 6) {
                                    if (Cs < 460) {
                                        return 1;
                                    }
                                    else if (Cs >= 460) {
                                        if (AFD < 41) {
                                            if (diffAFD < 11) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 11) {
                                                if (TSC < 10) {
                                                    return 0;
                                                }
                                                else if (TSC >= 10) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 41) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 50) {
                            if (Cs < 1085) {
                                return 0;
                            }
                            else if (Cs >= 1085) {
                                if (Cs < 1120) {
                                    return 1;
                                }
                                else if (Cs >= 1120) {
                                    if (TSC < 6) {
                                        return 1;
                                    }
                                    else if (TSC >= 6) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (gchDC >= 124) {
                    if (AFD < 27) {
                        if (Rs < 7) {
                            return 1;
                        }
                        else if (Rs >= 7) {
                            if (AFD < 17) {
                                return 1;
                            }
                            else if (AFD >= 17) {
                                return 0;
                            }
                        }
                    }
                    else if (AFD >= 27) {
                        return 0;
                    }
                }
            }
        }
        else if (diffTSC >= 4) {
            if (diffMVdiffVal < 66) {
                if (refDCval < 34) {
                    if (CsDiff < 9) {
                        if (diffRsCsdiff < -15) {
                            if (Cs < 385) {
                                if (CsDiff < 1) {
                                    return 0;
                                }
                                else if (CsDiff >= 1) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 385) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= -15) {
                            if (SC < 5136) {
                                return 0;
                            }
                            else if (SC >= 5136) {
                                return 1;
                            }
                        }
                    }
                    else if (CsDiff >= 9) {
                        return 1;
                    }
                }
                else if (refDCval >= 34) {
                    if (MVDiff < 63) {
                        return 0;
                    }
                    else if (MVDiff >= 63) {
                        if (TSC < 17) {
                            if (diffAFD < 11) {
                                return 0;
                            }
                            else if (diffAFD >= 11) {
                                if (refDCval < 46) {
                                    if (refDCval < 41) {
                                        return 0;
                                    }
                                    else if (refDCval >= 41) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 46) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 17) {
                            if (ssDCval < 87) {
                                if (refDCval < 72) {
                                    if (MVDiff < 218) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 218) {
                                        if (RsDiff < 11) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 11) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 72) {
                                    if (SC < 1397) {
                                        return 0;
                                    }
                                    else if (SC >= 1397) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 87) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 66) {
                if (ssDCval < 50) {
                    if (negBalance < 25) {
                        if (SC < 2673) {
                            return 0;
                        }
                        else if (SC >= 2673) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 25) {
                        if (MVDiff < 135) {
                            return 1;
                        }
                        else if (MVDiff >= 135) {
                            if (RsDiff < 2) {
                                if (negBalance < 103) {
                                    return 1;
                                }
                                else if (negBalance >= 103) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 2) {
                                if (Cs < 588) {
                                    if (MVDiff < 142) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 142) {
                                        if (refDCval < 52) {
                                            if (Rs < 75) {
                                                return 0;
                                            }
                                            else if (Rs >= 75) {
                                                if (SC < 944) {
                                                    return 1;
                                                }
                                                else if (SC >= 944) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 52) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Cs >= 588) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 50) {
                    if (SC < 2) {
                        if (RsCsDiff < 31) {
                            if (AFD < 15) {
                                return 1;
                            }
                            else if (AFD >= 15) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 31) {
                            return 1;
                        }
                    }
                    else if (SC >= 2) {
                        if (diffAFD < 8) {
                            if (RsCsDiff < 775) {
                                return 0;
                            }
                            else if (RsCsDiff >= 775) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 8) {
                            if (RsDiff < 7) {
                                if (posBalance < 115) {
                                    return 0;
                                }
                                else if (posBalance >= 115) {
                                    if (SC < 861) {
                                        return 1;
                                    }
                                    else if (SC >= 861) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsDiff >= 7) {
                                if (Cs < 383) {
                                    return 1;
                                }
                                else if (Cs >= 383) {
                                    if (diffRsCsdiff < 255) {
                                        if (MVDiff < 229) {
                                            if (diffMVdiffVal < 88) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 88) {
                                                if (RsCsDiff < 90) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 90) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 229) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 255) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 12) {
        if (diffMVdiffVal < 65) {
            if (MVDiff < 45) {
                if (refDCval < 32) {
                    return 1;
                }
                else if (refDCval >= 32) {
                    if (RsDiff < 68) {
                        if (Cs < 542) {
                            if (diffTSC < 13) {
                                return 0;
                            }
                            else if (diffTSC >= 13) {
                                return 1;
                            }
                        }
                        else if (Cs >= 542) {
                            return 0;
                        }
                    }
                    else if (RsDiff >= 68) {
                        if (TSC < 22) {
                            return 0;
                        }
                        else if (TSC >= 22) {
                            return 1;
                        }
                    }
                }
            }
            else if (MVDiff >= 45) {
                if (diffAFD < 23) {
                    if (diffRsCsdiff < 5087) {
                        if (diffAFD < 13) {
                            return 0;
                        }
                        else if (diffAFD >= 13) {
                            if (MVDiff < 706) {
                                if (ssDCval < 33) {
                                    if (SC < 54) {
                                        return 0;
                                    }
                                    else if (SC >= 54) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 33) {
                                    if (diffMVdiffVal < -71) {
                                        if (negBalance < 78) {
                                            if (ssDCval < 100) {
                                                if (Rs < 1466) {
                                                    if (gchDC < 8) {
                                                        if (ssDCval < 74) {
                                                            if (MVDiff < 308) {
                                                                return 0;
                                                            }
                                                            else if (MVDiff >= 308) {
                                                                if (Rs < 204) {
                                                                    return 0;
                                                                }
                                                                else if (Rs >= 204) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (ssDCval >= 74) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 8) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Rs >= 1466) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 100) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 78) {
                                            if (Rs < 1390) {
                                                if (RsCsDiff < 20) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 20) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1390) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= -71) {
                                        if (diffTSC < 7) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 7) {
                                            if (TSC < 19) {
                                                if (refDCval < 22) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 22) {
                                                    if (diffAFD < 14) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= 14) {
                                                        if (AFD < 33) {
                                                            if (negBalance < 18) {
                                                                return 0;
                                                            }
                                                            else if (negBalance >= 18) {
                                                                if (diffMVdiffVal < 36) {
                                                                    return 0;
                                                                }
                                                                else if (diffMVdiffVal >= 36) {
                                                                    if (RsCsDiff < 137) {
                                                                        if (Cs < 222) {
                                                                            return 1;
                                                                        }
                                                                        else if (Cs >= 222) {
                                                                            return 0;
                                                                        }
                                                                    }
                                                                    else if (RsCsDiff >= 137) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 33) {
                                                            if (Cs < 777) {
                                                                if (RsCsDiff < 124) {
                                                                    if (TSC < 17) {
                                                                        return 0;
                                                                    }
                                                                    else if (TSC >= 17) {
                                                                        return 1;
                                                                    }
                                                                }
                                                                else if (RsCsDiff >= 124) {
                                                                    return 0;
                                                                }
                                                            }
                                                            else if (Cs >= 777) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (TSC >= 19) {
                                                if (Rs < 7129) {
                                                    return 0;
                                                }
                                                else if (Rs >= 7129) {
                                                    if (Rs < 8092) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 8092) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 706) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 5087) {
                        return 1;
                    }
                }
                else if (diffAFD >= 23) {
                    if (RsCsDiff < 313) {
                        if (gchDC < 20) {
                            if (refDCval < 64) {
                                return 0;
                            }
                            else if (refDCval >= 64) {
                                if (TSC < 12) {
                                    return 0;
                                }
                                else if (TSC >= 12) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 20) {
                            return 0;
                        }
                    }
                    else if (RsCsDiff >= 313) {
                        if (negBalance < 31) {
                            if (TSC < 29) {
                                return 0;
                            }
                            else if (TSC >= 29) {
                                if (posBalance < 110) {
                                    return 1;
                                }
                                else if (posBalance >= 110) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 31) {
                            if (Scindex < 6) {
                                if (TSC < 12) {
                                    return 0;
                                }
                                else if (TSC >= 12) {
                                    if (AFD < 45) {
                                        if (MVDiff < 76) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 76) {
                                            if (AFD < 30) {
                                                return 1;
                                            }
                                            else if (AFD >= 30) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 45) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Scindex >= 6) {
                                if (negBalance < 56) {
                                    if (Rs < 8691) {
                                        if (SC < 7680) {
                                            return 1;
                                        }
                                        else if (SC >= 7680) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 8691) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 56) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffAFD < 23) {
                if (negBalance < 34) {
                    if (refDCval < 33) {
                        if (TSC < 13) {
                            return 1;
                        }
                        else if (TSC >= 13) {
                            if (RsCsDiff < 103) {
                                if (diffMVdiffVal < 385) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 385) {
                                    return 1;
                                }
                            }
                            else if (RsCsDiff >= 103) {
                                return 1;
                            }
                        }
                    }
                    else if (refDCval >= 33) {
                        if (AFD < 20) {
                            if (Cs < 128) {
                                return 1;
                            }
                            else if (Cs >= 128) {
                                if (TSC < 12) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (AFD < 18) {
                                            return 0;
                                        }
                                        else if (AFD >= 18) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 12) {
                                    return 0;
                                }
                            }
                        }
                        else if (AFD >= 20) {
                            if (diffRsCsdiff < 295) {
                                if (diffRsCsdiff < 25) {
                                    if (AFD < 22) {
                                        if (Cs < 60) {
                                            return 1;
                                        }
                                        else if (Cs >= 60) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 22) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 25) {
                                    if (SC < 437) {
                                        return 1;
                                    }
                                    else if (SC >= 437) {
                                        if (Cs < 810) {
                                            if (RsCsDiff < 81) {
                                                if (Cs < 532) {
                                                    if (AFD < 22) {
                                                        if (refDCval < 69) {
                                                            return 0;
                                                        }
                                                        else if (refDCval >= 69) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 22) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Cs >= 532) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 81) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 810) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 295) {
                                if (RsDiff < 19) {
                                    return 1;
                                }
                                else if (RsDiff >= 19) {
                                    if (SC < 7507) {
                                        if (MVDiff < 785) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 785) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 7507) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 34) {
                    if (Rs < 635) {
                        if (diffTSC < -2) {
                            return 0;
                        }
                        else if (diffTSC >= -2) {
                            if (MVDiff < 119) {
                                return 1;
                            }
                            else if (MVDiff >= 119) {
                                if (MVDiff < 124) {
                                    return 0;
                                }
                                else if (MVDiff >= 124) {
                                    if (Rs < 141) {
                                        if (refDCval < 59) {
                                            return 1;
                                        }
                                        else if (refDCval >= 59) {
                                            if (CsDiff < 6) {
                                                if (diffMVdiffVal < 119) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 119) {
                                                    if (CsDiff < 5) {
                                                        if (AFD < 32) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 32) {
                                                            if (ssDCval < 31) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 31) {
                                                                if (diffTSC < 9) {
                                                                    return 0;
                                                                }
                                                                else if (diffTSC >= 9) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 141) {
                                        if (AFD < 45) {
                                            if (CsDiff < 3) {
                                                if (AFD < 19) {
                                                    return 1;
                                                }
                                                else if (AFD >= 19) {
                                                    if (ssDCval < 48) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 48) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (gchDC < 12) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 12) {
                                                    if (diffAFD < 14) {
                                                        if (gchDC < 26) {
                                                            return 0;
                                                        }
                                                        else if (gchDC >= 26) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffAFD >= 14) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 45) {
                                            if (Cs < 727) {
                                                if (ssDCval < 62) {
                                                    if (TSC < 19) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 19) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 62) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 727) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 635) {
                        if (diffTSC < 6) {
                            if (MVDiff < 1256) {
                                return 0;
                            }
                            else if (MVDiff >= 1256) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 6) {
                            if (diffRsCsdiff < 80) {
                                if (RsCsDiff < 80) {
                                    if (SC < 1448) {
                                        return 0;
                                    }
                                    else if (SC >= 1448) {
                                        if (refDCval < 116) {
                                            return 0;
                                        }
                                        else if (refDCval >= 116) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 80) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 80) {
                                if (Rs < 1860) {
                                    if (diffRsCsdiff < 126) {
                                        if (negBalance < 45) {
                                            return 1;
                                        }
                                        else if (negBalance >= 45) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 126) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 1860) {
                                    if (refDCval < 135) {
                                        if (diffTSC < 10) {
                                            if (diffAFD < 17) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 17) {
                                                if (RsCsDiff < 515) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 515) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 10) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 135) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 23) {
                if (negBalance < 27) {
                    if (CsDiff < 1) {
                        return 0;
                    }
                    else if (CsDiff >= 1) {
                        if (refDCval < 54) {
                            if (TSC < 23) {
                                if (diffTSC < 20) {
                                    if (diffTSC < 14) {
                                        if (TSC < 13) {
                                            return 1;
                                        }
                                        else if (TSC >= 13) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 14) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 20) {
                                    if (MVDiff < 694) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 694) {
                                        return 0;
                                    }
                                }
                            }
                            else if (TSC >= 23) {
                                if (RsDiff < 4) {
                                    return 0;
                                }
                                else if (RsDiff >= 4) {
                                    if (diffMVdiffVal < 154) {
                                        if (posBalance < 86) {
                                            return 1;
                                        }
                                        else if (posBalance >= 86) {
                                            if (AFD < 49) {
                                                if (AFD < 29) {
                                                    if (Cs < 1742) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 1742) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 29) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 49) {
                                                if (AFD < 61) {
                                                    return 1;
                                                }
                                                else if (AFD >= 61) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 154) {
                                        if (gchDC < 76) {
                                            if (negBalance < 12) {
                                                return 0;
                                            }
                                            else if (negBalance >= 12) {
                                                if (ssDCval < 116) {
                                                    if (gchDC < 42) {
                                                        if (Rs < 1077) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1077) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (gchDC >= 42) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 116) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 76) {
                                            if (Rs < 3011) {
                                                return 1;
                                            }
                                            else if (Rs >= 3011) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 54) {
                            if (diffMVdiffVal < 132) {
                                if (MVDiff < 103) {
                                    return 1;
                                }
                                else if (MVDiff >= 103) {
                                    if (RsDiff < 10) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 10) {
                                        if (SC < 4408) {
                                            return 1;
                                        }
                                        else if (SC >= 4408) {
                                            if (RsCsDiff < 989) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 989) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 132) {
                                if (RsDiff < 3) {
                                    if (Rs < 151) {
                                        return 1;
                                    }
                                    else if (Rs >= 151) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 3) {
                                    if (diffAFD < 24) {
                                        if (RsDiff < 32) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 32) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 24) {
                                        if (TSC < 34) {
                                            return 1;
                                        }
                                        else if (TSC >= 34) {
                                            if (RsCsDiff < 183) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 183) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 27) {
                    if (diffRsCsdiff < 254) {
                        if (AFD < 54) {
                            if (RsDiff < 15) {
                                if (diffMVdiffVal < 80) {
                                    if (posBalance < 50) {
                                        return 1;
                                    }
                                    else if (posBalance >= 50) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 80) {
                                    if (MVDiff < 595) {
                                        if (diffMVdiffVal < 129) {
                                            if (MVDiff < 158) {
                                                if (AFD < 39) {
                                                    return 1;
                                                }
                                                else if (AFD >= 39) {
                                                    if (diffTSC < 11) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 11) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 158) {
                                                if (AFD < 51) {
                                                    if (CsDiff < 7) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 51) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 129) {
                                            if (AFD < 41) {
                                                return 1;
                                            }
                                            else if (AFD >= 41) {
                                                if (ssDCval < 36) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 36) {
                                                    if (Cs < 1469) {
                                                        if (TSC < 38) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 38) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (Cs >= 1469) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 595) {
                                        if (diffAFD < 25) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 25) {
                                            if (TSC < 21) {
                                                return 1;
                                            }
                                            else if (TSC >= 21) {
                                                if (refDCval < 79) {
                                                    if (diffMVdiffVal < 663) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 663) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 79) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 15) {
                                if (MVDiff < 229) {
                                    return 0;
                                }
                                else if (MVDiff >= 229) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 54) {
                            if (MVDiff < 331) {
                                if (diffAFD < 39) {
                                    if (MVDiff < 149) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 149) {
                                        if (Rs < 38) {
                                            return 1;
                                        }
                                        else if (Rs >= 38) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 39) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 331) {
                                if (diffAFD < 43) {
                                    if (Cs < 883) {
                                        if (diffRsCsdiff < 54) {
                                            if (Rs < 1055) {
                                                return 0;
                                            }
                                            else if (Rs >= 1055) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 54) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 883) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 43) {
                                    if (TSC < 59) {
                                        if (diffMVdiffVal < 886) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 886) {
                                            if (AFD < 56) {
                                                return 0;
                                            }
                                            else if (AFD >= 56) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 59) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 254) {
                        if (diffTSC < 16) {
                            if (gchDC < 22) {
                                if (TSC < 10) {
                                    if (Rs < 2761) {
                                        return 0;
                                    }
                                    else if (Rs >= 2761) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 10) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 22) {
                                if (diffAFD < 27) {
                                    if (TSC < 23) {
                                        if (Rs < 2076) {
                                            return 1;
                                        }
                                        else if (Rs >= 2076) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 23) {
                                        if (SC < 4161) {
                                            return 0;
                                        }
                                        else if (SC >= 4161) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 27) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffTSC >= 16) {
                            if (RsCsDiff < 4065) {
                                if (SC < 4994) {
                                    return 1;
                                }
                                else if (SC >= 4994) {
                                    if (RsCsDiff < 1233) {
                                        if (AFD < 63) {
                                            return 1;
                                        }
                                        else if (AFD >= 63) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 1233) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 4065) {
                                if (negBalance < 114) {
                                    if (negBalance < 43) {
                                        if (gchDC < 7) {
                                            return 0;
                                        }
                                        else if (gchDC >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 114) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect6(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 66) {
        if (diffTSC < 12) {
            if (MVDiff < 42) {
                if (diffAFD < 4) {
                    if (Scindex < 1) {
                        if (negBalance < 106) {
                            if (diffRsCsdiff < -2) {
                                if (MVDiff < 4) {
                                    if (gchDC < 104) {
                                        return 0;
                                    }
                                    else if (gchDC >= 104) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 4) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -2) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 106) {
                            if (refDCval < 226) {
                                return 0;
                            }
                            else if (refDCval >= 226) {
                                return 1;
                            }
                        }
                    }
                    else if (Scindex >= 1) {
                        if (TSC < 2) {
                            if (CsDiff < 4) {
                                if (diffRsCsdiff < 7) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 7) {
                                    if (refDCval < 20) {
                                        return 1;
                                    }
                                    else if (refDCval >= 20) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 4) {
                                if (SC < 2341) {
                                    if (Rs < 613) {
                                        return 0;
                                    }
                                    else if (Rs >= 613) {
                                        if (MVDiff < 2) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 2) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 2341) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 2) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 4) {
                    if (refDCval < 22) {
                        if (Cs < 742) {
                            return 0;
                        }
                        else if (Cs >= 742) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 22) {
                        return 0;
                    }
                }
            }
            else if (MVDiff >= 42) {
                if (diffAFD < 10) {
                    if (CsDiff < 16) {
                        if (ssDCval < 38) {
                            if (TSCindex < 8) {
                                if (RsCsDiff < 5) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 5) {
                                    if (AFD < 1) {
                                        if (ssDCval < 17) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 17) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 1) {
                                        if (Scindex < 1) {
                                            if (Cs < 44) {
                                                return 0;
                                            }
                                            else if (Cs >= 44) {
                                                return 1;
                                            }
                                        }
                                        else if (Scindex >= 1) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (TSCindex >= 8) {
                                if (CsDiff < 5) {
                                    if (MVDiff < 143) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 143) {
                                        if (negBalance < 66) {
                                            return 1;
                                        }
                                        else if (negBalance >= 66) {
                                            if (posBalance < 36) {
                                                return 0;
                                            }
                                            else if (posBalance >= 36) {
                                                if (RsCsDiff < 39) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 39) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 5) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 38) {
                            if (diffTSC < 6) {
                                if (refDCval < 230) {
                                    if (TSC < 17) {
                                        return 0;
                                    }
                                    else if (TSC >= 17) {
                                        if (ssDCval < 46) {
                                            if (RsCsDiff < 296) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 296) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 46) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 230) {
                                    if (gchDC < 5) {
                                        return 0;
                                    }
                                    else if (gchDC >= 5) {
                                        if (negBalance < 126) {
                                            return 1;
                                        }
                                        else if (negBalance >= 126) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 6) {
                                if (Rs < 601) {
                                    return 0;
                                }
                                else if (Rs >= 601) {
                                    if (diffMVdiffVal < -26) {
                                        if (diffMVdiffVal < -40) {
                                            if (MVDiff < 458) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 458) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -40) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -26) {
                                        if (gchDC < 2) {
                                            if (AFD < 24) {
                                                return 0;
                                            }
                                            else if (AFD >= 24) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 16) {
                        if (AFD < 10) {
                            if (posBalance < 6) {
                                if (SC < 604) {
                                    return 0;
                                }
                                else if (SC >= 604) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 6) {
                                if (diffAFD < 6) {
                                    if (diffMVdiffVal < 22) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 22) {
                                        if (Cs < 2019) {
                                            return 1;
                                        }
                                        else if (Cs >= 2019) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 6) {
                                    if (Rs < 3753) {
                                        return 0;
                                    }
                                    else if (Rs >= 3753) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 10) {
                            if (AFD < 57) {
                                if (MVDiff < 499) {
                                    return 0;
                                }
                                else if (MVDiff >= 499) {
                                    if (SC < 6135) {
                                        return 1;
                                    }
                                    else if (SC >= 6135) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 57) {
                                if (SC < 3705) {
                                    return 1;
                                }
                                else if (SC >= 3705) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 10) {
                    if (ssDCval < 52) {
                        if (RsCsDiff < 236) {
                            if (negBalance < 41) {
                                return 0;
                            }
                            else if (negBalance >= 41) {
                                if (negBalance < 42) {
                                    return 1;
                                }
                                else if (negBalance >= 42) {
                                    if (MVDiff < 155) {
                                        if (diffMVdiffVal < -6) {
                                            if (AFD < 16) {
                                                return 0;
                                            }
                                            else if (AFD >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -6) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 155) {
                                        if (Cs < 177) {
                                            return 1;
                                        }
                                        else if (Cs >= 177) {
                                            if (Cs < 482) {
                                                return 0;
                                            }
                                            else if (Cs >= 482) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 236) {
                            if (diffAFD < 19) {
                                return 1;
                            }
                            else if (diffAFD >= 19) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 52) {
                        if (diffAFD < 21) {
                            if (AFD < 35) {
                                return 0;
                            }
                            else if (AFD >= 35) {
                                if (diffRsCsdiff < 1075) {
                                    if (Rs < 41) {
                                        if (SC < 13) {
                                            return 0;
                                        }
                                        else if (SC >= 13) {
                                            if (Cs < 23) {
                                                return 1;
                                            }
                                            else if (Cs >= 23) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 41) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 1075) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 21) {
                            if (diffTSC < 8) {
                                return 0;
                            }
                            else if (diffTSC >= 8) {
                                if (Scindex < 5) {
                                    if (posBalance < 57) {
                                        return 1;
                                    }
                                    else if (posBalance >= 57) {
                                        if (diffMVdiffVal < 50) {
                                            if (MVDiff < 306) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 306) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 50) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Scindex >= 5) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 12) {
            if (diffAFD < 25) {
                if (diffTSC < 13) {
                    if (MVDiff < 370) {
                        return 0;
                    }
                    else if (MVDiff >= 370) {
                        if (RsCsDiff < 125) {
                            return 0;
                        }
                        else if (RsCsDiff >= 125) {
                            return 1;
                        }
                    }
                }
                else if (diffTSC >= 13) {
                    if (RsCsDiff < 618) {
                        if (Cs < 1132) {
                            if (RsCsDiff < 115) {
                                if (MVDiff < 140) {
                                    if (refDCval < 24) {
                                        return 1;
                                    }
                                    else if (refDCval >= 24) {
                                        if (negBalance < 27) {
                                            return 0;
                                        }
                                        else if (negBalance >= 27) {
                                            if (Rs < 783) {
                                                return 1;
                                            }
                                            else if (Rs >= 783) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 140) {
                                    if (MVDiff < 544) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 544) {
                                        if (RsCsDiff < 6) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 115) {
                                if (RsDiff < 13) {
                                    if (RsCsDiff < 149) {
                                        if (TSC < 25) {
                                            return 0;
                                        }
                                        else if (TSC >= 25) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 149) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 13) {
                                    if (CsDiff < 4) {
                                        if (TSC < 18) {
                                            return 1;
                                        }
                                        else if (TSC >= 18) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 4) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 1132) {
                            if (RsCsDiff < 381) {
                                return 0;
                            }
                            else if (RsCsDiff >= 381) {
                                if (TSC < 42) {
                                    return 0;
                                }
                                else if (TSC >= 42) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 618) {
                        if (gchDC < 3) {
                            return 0;
                        }
                        else if (gchDC >= 3) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 25) {
                if (RsCsDiff < 389) {
                    if (MVDiff < 22) {
                        return 1;
                    }
                    else if (MVDiff >= 22) {
                        if (gchDC < 20) {
                            if (TSC < 16) {
                                return 0;
                            }
                            else if (TSC >= 16) {
                                if (refDCval < 56) {
                                    if (SC < 381) {
                                        return 0;
                                    }
                                    else if (SC >= 381) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 56) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 20) {
                            if (SC < 2725) {
                                return 0;
                            }
                            else if (SC >= 2725) {
                                if (MVDiff < 89) {
                                    return 0;
                                }
                                else if (MVDiff >= 89) {
                                    if (RsCsDiff < 258) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 258) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (RsCsDiff >= 389) {
                    if (ssDCval < 111) {
                        if (refDCval < 82) {
                            return 1;
                        }
                        else if (refDCval >= 82) {
                            if (SC < 5837) {
                                return 1;
                            }
                            else if (SC >= 5837) {
                                if (Rs < 4505) {
                                    return 0;
                                }
                                else if (Rs >= 4505) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 111) {
                        if (AFD < 36) {
                            return 0;
                        }
                        else if (AFD >= 36) {
                            if (Rs < 1752) {
                                if (RsCsDiff < 502) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 502) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 1752) {
                                if (negBalance < 28) {
                                    if (CsDiff < 13) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 13) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 28) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 66) {
        if (diffAFD < 12) {
            if (diffAFD < 7) {
                if (diffTSC < 4) {
                    if (gchDC < 108) {
                        if (CsDiff < 5) {
                            if (ssDCval < 24) {
                                if (gchDC < 42) {
                                    return 0;
                                }
                                else if (gchDC >= 42) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 24) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 5) {
                            if (SC < 1444) {
                                if (diffRsCsdiff < 65) {
                                    if (gchDC < 13) {
                                        if (TSC < 14) {
                                            if (refDCval < 39) {
                                                if (Rs < 508) {
                                                    if (diffAFD < -3) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= -3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 508) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 39) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 14) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 13) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 65) {
                                    if (refDCval < 111) {
                                        return 1;
                                    }
                                    else if (refDCval >= 111) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 1444) {
                                if (Cs < 6689) {
                                    return 0;
                                }
                                else if (Cs >= 6689) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 108) {
                        if (Cs < 9) {
                            if (refDCval < 99) {
                                return 0;
                            }
                            else if (refDCval >= 99) {
                                return 1;
                            }
                        }
                        else if (Cs >= 9) {
                            return 0;
                        }
                    }
                }
                else if (diffTSC >= 4) {
                    if (CsDiff < 19) {
                        if (SC < 4) {
                            if (TSC < 14) {
                                return 0;
                            }
                            else if (TSC >= 14) {
                                return 1;
                            }
                        }
                        else if (SC >= 4) {
                            if (diffMVdiffVal < 208) {
                                if (negBalance < 122) {
                                    return 0;
                                }
                                else if (negBalance >= 122) {
                                    if (Cs < 471) {
                                        return 0;
                                    }
                                    else if (Cs >= 471) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 208) {
                                if (diffTSC < 5) {
                                    if (Cs < 23) {
                                        return 0;
                                    }
                                    else if (Cs >= 23) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 5) {
                                    if (Cs < 612) {
                                        return 0;
                                    }
                                    else if (Cs >= 612) {
                                        if (Rs < 1548) {
                                            return 1;
                                        }
                                        else if (Rs >= 1548) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 19) {
                        if (MVDiff < 198) {
                            return 0;
                        }
                        else if (MVDiff >= 198) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 7) {
                if (ssDCval < 45) {
                    if (gchDC < 5) {
                        if (RsCsDiff < 23) {
                            return 1;
                        }
                        else if (RsCsDiff >= 23) {
                            if (Rs < 468) {
                                if (diffAFD < 10) {
                                    if (RsDiff < 4) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 4) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 10) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 468) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 5) {
                        if (RsDiff < 5) {
                            if (AFD < 13) {
                                if (Rs < 35) {
                                    return 0;
                                }
                                else if (Rs >= 35) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 13) {
                                if (diffRsCsdiff < 17) {
                                    if (TSC < 6) {
                                        if (AFD < 16) {
                                            return 1;
                                        }
                                        else if (AFD >= 16) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 6) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 17) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 5) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 45) {
                    if (AFD < 13) {
                        if (SC < 958) {
                            return 1;
                        }
                        else if (SC >= 958) {
                            return 0;
                        }
                    }
                    else if (AFD >= 13) {
                        if (Cs < 694) {
                            if (RsDiff < 5) {
                                if (diffMVdiffVal < 391) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 391) {
                                    if (AFD < 16) {
                                        return 1;
                                    }
                                    else if (AFD >= 16) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsDiff >= 5) {
                                if (Scindex < 3) {
                                    if (diffRsCsdiff < 31) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 31) {
                                        return 1;
                                    }
                                }
                                else if (Scindex >= 3) {
                                    if (gchDC < 8) {
                                        if (TSC < 12) {
                                            return 0;
                                        }
                                        else if (TSC >= 12) {
                                            if (diffTSC < -4) {
                                                return 0;
                                            }
                                            else if (diffTSC >= -4) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 8) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 694) {
                            if (RsDiff < 33) {
                                return 0;
                            }
                            else if (RsDiff >= 33) {
                                if (Rs < 2231) {
                                    return 1;
                                }
                                else if (Rs >= 2231) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 12) {
            if (diffMVdiffVal < 93) {
                if (MVDiff < 100) {
                    if (posBalance < 100) {
                        if (AFD < 32) {
                            if (RsCsDiff < 145) {
                                if (Scindex < 4) {
                                    if (Cs < 678) {
                                        return 1;
                                    }
                                    else if (Cs >= 678) {
                                        return 0;
                                    }
                                }
                                else if (Scindex >= 4) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 145) {
                                return 1;
                            }
                        }
                        else if (AFD >= 32) {
                            if (Scindex < 8) {
                                return 1;
                            }
                            else if (Scindex >= 8) {
                                if (ssDCval < 136) {
                                    return 1;
                                }
                                else if (ssDCval >= 136) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 100) {
                        if (MVDiff < 75) {
                            return 1;
                        }
                        else if (MVDiff >= 75) {
                            if (Cs < 661) {
                                return 1;
                            }
                            else if (Cs >= 661) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 100) {
                    if (diffRsCsdiff < 27) {
                        return 0;
                    }
                    else if (diffRsCsdiff >= 27) {
                        if (posBalance < 61) {
                            if (diffMVdiffVal < 89) {
                                if (diffAFD < 18) {
                                    return 0;
                                }
                                else if (diffAFD >= 18) {
                                    if (MVDiff < 144) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 144) {
                                        if (Cs < 404) {
                                            return 1;
                                        }
                                        else if (Cs >= 404) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 89) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 61) {
                            if (Cs < 3592) {
                                if (Rs < 411) {
                                    return 1;
                                }
                                else if (Rs >= 411) {
                                    if (Cs < 225) {
                                        return 1;
                                    }
                                    else if (Cs >= 225) {
                                        if (MVDiff < 107) {
                                            if (Cs < 1368) {
                                                return 1;
                                            }
                                            else if (Cs >= 1368) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 107) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 3592) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 93) {
                if (diffRsCsdiff < 58) {
                    if (Cs < 591) {
                        if (diffRsCsdiff < 2) {
                            if (Cs < 264) {
                                if (AFD < 46) {
                                    if (negBalance < 43) {
                                        if (TSC < 11) {
                                            if (Rs < 119) {
                                                return 1;
                                            }
                                            else if (Rs >= 119) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 11) {
                                            if (TSC < 21) {
                                                if (diffMVdiffVal < 169) {
                                                    if (MVDiff < 254) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 254) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 169) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 21) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (CsDiff < 3) {
                                            if (RsDiff < 1) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 1) {
                                                if (Cs < 33) {
                                                    return 0;
                                                }
                                                else if (Cs >= 33) {
                                                    if (Cs < 146) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 146) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 46) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 264) {
                                if (diffRsCsdiff < -152) {
                                    return 1;
                                }
                                else if (diffRsCsdiff >= -152) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 2) {
                            if (gchDC < 14) {
                                if (gchDC < 5) {
                                    return 1;
                                }
                                else if (gchDC >= 5) {
                                    if (posBalance < 71) {
                                        if (ssDCval < 44) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 44) {
                                            if (CsDiff < 3) {
                                                if (posBalance < 46) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 46) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 3) {
                                                if (ssDCval < 46) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 46) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 71) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 14) {
                                if (Rs < 151) {
                                    if (diffAFD < 15) {
                                        if (MVDiff < 183) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 183) {
                                            if (gchDC < 43) {
                                                return 0;
                                            }
                                            else if (gchDC >= 43) {
                                                if (SC < 130) {
                                                    return 0;
                                                }
                                                else if (SC >= 130) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 15) {
                                        if (gchDC < 42) {
                                            return 1;
                                        }
                                        else if (gchDC >= 42) {
                                            if (gchDC < 46) {
                                                if (CsDiff < 2) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 0;
                                                }
                                            }
                                            else if (gchDC >= 46) {
                                                if (refDCval < 54) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 54) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 151) {
                                    if (posBalance < 92) {
                                        if (diffTSC < 12) {
                                            if (gchDC < 27) {
                                                if (posBalance < 80) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 80) {
                                                    if (gchDC < 18) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 18) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 27) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= 12) {
                                            if (diffTSC < 47) {
                                                if (diffMVdiffVal < 219) {
                                                    if (AFD < 39) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 39) {
                                                        if (AFD < 43) {
                                                            return 0;
                                                        }
                                                        else if (AFD >= 43) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 219) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffTSC >= 47) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 92) {
                                        if (diffTSC < 13) {
                                            if (CsDiff < 6) {
                                                if (Cs < 271) {
                                                    if (AFD < 21) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 21) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Cs >= 271) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= 13) {
                                            if (negBalance < 10) {
                                                if (MVDiff < 823) {
                                                    if (AFD < 21) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 21) {
                                                        return 0;
                                                    }
                                                }
                                                else if (MVDiff >= 823) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 10) {
                                                if (diffAFD < 39) {
                                                    if (diffTSC < 19) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= 19) {
                                                        if (ssDCval < 116) {
                                                            return 0;
                                                        }
                                                        else if (ssDCval >= 116) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffAFD >= 39) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 591) {
                        if (Cs < 803) {
                            if (ssDCval < 66) {
                                if (diffAFD < 21) {
                                    return 0;
                                }
                                else if (diffAFD >= 21) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 66) {
                                if (SC < 1651) {
                                    if (SC < 985) {
                                        if (diffMVdiffVal < 536) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 536) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 985) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 1651) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 803) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 58) {
                    if (diffTSC < 12) {
                        if (SC < 1749) {
                            if (CsDiff < 3) {
                                return 0;
                            }
                            else if (CsDiff >= 3) {
                                if (negBalance < 8) {
                                    if (diffTSC < 8) {
                                        return 1;
                                    }
                                    else if (diffTSC >= 8) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 8) {
                                    if (diffMVdiffVal < 105) {
                                        if (MVDiff < 133) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 133) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 105) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (SC >= 1749) {
                            if (ssDCval < 58) {
                                if (SC < 3505) {
                                    return 1;
                                }
                                else if (SC >= 3505) {
                                    if (TSC < 13) {
                                        return 0;
                                    }
                                    else if (TSC >= 13) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 58) {
                                if (posBalance < 69) {
                                    if (diffTSC < 6) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 6) {
                                        if (CsDiff < 10) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 10) {
                                            if (RsDiff < 34) {
                                                if (MVDiff < 227) {
                                                    if (ssDCval < 75) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 75) {
                                                        return 0;
                                                    }
                                                }
                                                else if (MVDiff >= 227) {
                                                    if (posBalance < 39) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 39) {
                                                        if (CsDiff < 14) {
                                                            return 0;
                                                        }
                                                        else if (CsDiff >= 14) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 34) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 69) {
                                    if (SC < 7378) {
                                        return 0;
                                    }
                                    else if (SC >= 7378) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 12) {
                        if (gchDC < 21) {
                            if (MVDiff < 1206) {
                                return 1;
                            }
                            else if (MVDiff >= 1206) {
                                if (Scindex < 4) {
                                    if (Rs < 513) {
                                        return 0;
                                    }
                                    else if (Rs >= 513) {
                                        return 1;
                                    }
                                }
                                else if (Scindex >= 4) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 21) {
                            if (negBalance < 3) {
                                if (Rs < 261) {
                                    return 1;
                                }
                                else if (Rs >= 261) {
                                    if (diffTSC < 41) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 41) {
                                        if (AFD < 56) {
                                            return 0;
                                        }
                                        else if (AFD >= 56) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 3) {
                                if (Rs < 724) {
                                    return 1;
                                }
                                else if (Rs >= 724) {
                                    if (diffTSC < 17) {
                                        if (Rs < 2290) {
                                            if (TSC < 29) {
                                                return 1;
                                            }
                                            else if (TSC >= 29) {
                                                if (Cs < 2405) {
                                                    return 0;
                                                }
                                                else if (Cs >= 2405) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 2290) {
                                            if (AFD < 34) {
                                                return 1;
                                            }
                                            else if (AFD >= 34) {
                                                if (refDCval < 116) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 116) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 17) {
                                        if (diffRsCsdiff < 181) {
                                            if (diffMVdiffVal < 450) {
                                                if (MVDiff < 415) {
                                                    if (CsDiff < 7) {
                                                        if (diffAFD < 35) {
                                                            if (negBalance < 89) {
                                                                return 0;
                                                            }
                                                            else if (negBalance >= 89) {
                                                                if (AFD < 40) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 40) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (diffAFD >= 35) {
                                                            if (CsDiff < 5) {
                                                                return 0;
                                                            }
                                                            else if (CsDiff >= 5) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 7) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 415) {
                                                    if (SC < 1589) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 1589) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 450) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 181) {
                                            if (negBalance < 21) {
                                                if (diffMVdiffVal < 117) {
                                                    if (Rs < 1867) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 1867) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 117) {
                                                    if (posBalance < 107) {
                                                        if (MVDiff < 302) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 302) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (posBalance >= 107) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 21) {
                                                if (TSC < 40) {
                                                    return 1;
                                                }
                                                else if (TSC >= 40) {
                                                    if (diffAFD < 38) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= 38) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect7(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (TSC < 16) {
        if (diffMVdiffVal < 65) {
            if (gchDC < 6) {
                if (ssDCval < 21) {
                    if (RsDiff < 15) {
                        if (Cs < 1155) {
                            if (Cs < 1151) {
                                if (RsDiff < 2) {
                                    return 0;
                                }
                                else if (RsDiff >= 2) {
                                    if (TSCindex < 2) {
                                        if (Cs < 652) {
                                            return 1;
                                        }
                                        else if (Cs >= 652) {
                                            if (MVDiff < 5) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 5) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSCindex >= 2) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Cs >= 1151) {
                                return 1;
                            }
                        }
                        else if (Cs >= 1155) {
                            return 0;
                        }
                    }
                    else if (RsDiff >= 15) {
                        return 1;
                    }
                }
                else if (ssDCval >= 21) {
                    if (diffAFD < 7) {
                        if (diffRsCsdiff < 1477) {
                            if (gchDC < 4) {
                                return 0;
                            }
                            else if (gchDC >= 4) {
                                if (refDCval < 43) {
                                    if (ssDCval < 25) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 25) {
                                        if (diffTSC < -6) {
                                            if (RsDiff < 4) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 4) {
                                                if (Cs < 133) {
                                                    return 1;
                                                }
                                                else if (Cs >= 133) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= -6) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 43) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 1477) {
                            if (diffRsCsdiff < 1531) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= 1531) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (MVDiff < 70) {
                            if (TSC < 14) {
                                return 0;
                            }
                            else if (TSC >= 14) {
                                if (RsCsDiff < 189) {
                                    if (SC < 3434) {
                                        return 1;
                                    }
                                    else if (SC >= 3434) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 189) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 70) {
                            if (ssDCval < 47) {
                                if (SC < 623) {
                                    if (Rs < 8) {
                                        return 0;
                                    }
                                    else if (Rs >= 8) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 623) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 47) {
                                if (negBalance < 22) {
                                    return 1;
                                }
                                else if (negBalance >= 22) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 6) {
                if (diffRsCsdiff < 247) {
                    if (SC < 125) {
                        if (refDCval < 24) {
                            if (diffAFD < 3) {
                                return 1;
                            }
                            else if (diffAFD >= 3) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 24) {
                            if (SC < 122) {
                                if (diffRsCsdiff < -97) {
                                    if (diffAFD < -11) {
                                        return 0;
                                    }
                                    else if (diffAFD >= -11) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= -97) {
                                    if (RsCsDiff < 2) {
                                        if (diffRsCsdiff < -22) {
                                            if (AFD < 6) {
                                                return 1;
                                            }
                                            else if (AFD >= 6) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -22) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 2) {
                                        if (TSCindex < 8) {
                                            return 0;
                                        }
                                        else if (TSCindex >= 8) {
                                            if (diffMVdiffVal < -17) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= -17) {
                                                if (Rs < 87) {
                                                    return 1;
                                                }
                                                else if (Rs >= 87) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 122) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 125) {
                        if (MVDiff < 175) {
                            if (RsDiff < 10) {
                                return 0;
                            }
                            else if (RsDiff >= 10) {
                                if (diffMVdiffVal < 54) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 54) {
                                    if (SC < 2716) {
                                        if (RsCsDiff < 192) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 192) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 2716) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 175) {
                            if (gchDC < 51) {
                                if (diffAFD < 15) {
                                    return 0;
                                }
                                else if (diffAFD >= 15) {
                                    if (diffAFD < 16) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 16) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 51) {
                                if (gchDC < 55) {
                                    return 1;
                                }
                                else if (gchDC >= 55) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 247) {
                    if (SC < 3613) {
                        if (diffTSC < 7) {
                            if (refDCval < 78) {
                                if (CsDiff < 10) {
                                    return 0;
                                }
                                else if (CsDiff >= 10) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 78) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 7) {
                            if (posBalance < 61) {
                                return 1;
                            }
                            else if (posBalance >= 61) {
                                if (Rs < 366) {
                                    return 1;
                                }
                                else if (Rs >= 366) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (SC >= 3613) {
                        if (ssDCval < 46) {
                            return 1;
                        }
                        else if (ssDCval >= 46) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffRsCsdiff < 1) {
                if (Rs < 94) {
                    if (diffAFD < 7) {
                        if (diffMVdiffVal < 397) {
                            if (diffTSC < -21) {
                                if (refDCval < 94) {
                                    return 1;
                                }
                                else if (refDCval >= 94) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= -21) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 397) {
                            if (RsCsDiff < 15) {
                                if (negBalance < 127) {
                                    return 0;
                                }
                                else if (negBalance >= 127) {
                                    if (MVDiff < 611) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 611) {
                                        if (MVDiff < 1653) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 1653) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 15) {
                                return 1;
                            }
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (ssDCval < 94) {
                            if (ssDCval < 21) {
                                return 0;
                            }
                            else if (ssDCval >= 21) {
                                if (diffAFD < 8) {
                                    return 0;
                                }
                                else if (diffAFD >= 8) {
                                    if (refDCval < 38) {
                                        return 1;
                                    }
                                    else if (refDCval >= 38) {
                                        if (diffMVdiffVal < 119) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 119) {
                                            if (gchDC < 34) {
                                                if (refDCval < 40) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 40) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 34) {
                                                if (Rs < 2) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 94) {
                            if (diffMVdiffVal < 233) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 233) {
                                if (AFD < 18) {
                                    return 1;
                                }
                                else if (AFD >= 18) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (Rs >= 94) {
                    if (gchDC < 41) {
                        return 0;
                    }
                    else if (gchDC >= 41) {
                        if (gchDC < 45) {
                            return 1;
                        }
                        else if (gchDC >= 45) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 1) {
                if (Cs < 1209) {
                    if (diffTSC < 3) {
                        if (ssDCval < 51) {
                            if (diffRsCsdiff < 87) {
                                if (diffTSC < -2) {
                                    return 0;
                                }
                                else if (diffTSC >= -2) {
                                    if (diffAFD < 1) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 1) {
                                        if (refDCval < 29) {
                                            return 1;
                                        }
                                        else if (refDCval >= 29) {
                                            if (Rs < 130) {
                                                if (CsDiff < 2) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 2) {
                                                    if (Rs < 10) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 10) {
                                                        if (gchDC < 96) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 96) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 130) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 87) {
                                if (RsDiff < 45) {
                                    return 1;
                                }
                                else if (RsDiff >= 45) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 51) {
                            if (AFD < 15) {
                                if (SC < 621) {
                                    if (RsCsDiff < 11) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 11) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 621) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 15) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 3) {
                        if (diffAFD < 13) {
                            if (ssDCval < 50) {
                                if (Rs < 278) {
                                    if (gchDC < 10) {
                                        if (Rs < 230) {
                                            return 1;
                                        }
                                        else if (Rs >= 230) {
                                            if (Rs < 233) {
                                                return 0;
                                            }
                                            else if (Rs >= 233) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 10) {
                                        if (diffMVdiffVal < 224) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 224) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 278) {
                                    if (Cs < 170) {
                                        return 1;
                                    }
                                    else if (Cs >= 170) {
                                        if (MVDiff < 135) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 135) {
                                            if (diffMVdiffVal < 139) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 139) {
                                                if (SC < 661) {
                                                    return 0;
                                                }
                                                else if (SC >= 661) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 50) {
                                if (Rs < 99) {
                                    if (CsDiff < 4) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 4) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 99) {
                                    if (CsDiff < 7) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 7) {
                                        if (Rs < 803) {
                                            if (AFD < 31) {
                                                return 1;
                                            }
                                            else if (AFD >= 31) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 803) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 13) {
                            if (posBalance < 112) {
                                if (diffTSC < 5) {
                                    if (refDCval < 68) {
                                        return 1;
                                    }
                                    else if (refDCval >= 68) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 5) {
                                    if (diffMVdiffVal < 180) {
                                        if (negBalance < 43) {
                                            if (diffAFD < 30) {
                                                if (Cs < 383) {
                                                    return 1;
                                                }
                                                else if (Cs >= 383) {
                                                    if (diffRsCsdiff < 134) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= 134) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 30) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 43) {
                                            if (diffTSC < 15) {
                                                if (diffRsCsdiff < 86) {
                                                    if (RsDiff < 3) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        if (diffAFD < 14) {
                                                            return 0;
                                                        }
                                                        else if (diffAFD >= 14) {
                                                            if (diffMVdiffVal < 87) {
                                                                if (SC < 898) {
                                                                    return 0;
                                                                }
                                                                else if (SC >= 898) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 87) {
                                                                if (SC < 1459) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1459) {
                                                                    if (Cs < 661) {
                                                                        return 0;
                                                                    }
                                                                    else if (Cs >= 661) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 86) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffTSC >= 15) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 180) {
                                        if (gchDC < 74) {
                                            return 1;
                                        }
                                        else if (gchDC >= 74) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 112) {
                                if (ssDCval < 121) {
                                    if (SC < 1348) {
                                        return 1;
                                    }
                                    else if (SC >= 1348) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 121) {
                                    if (CsDiff < 6) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 6) {
                                        if (refDCval < 155) {
                                            if (CsDiff < 9) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 9) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 155) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 1209) {
                    if (diffRsCsdiff < 249) {
                        return 0;
                    }
                    else if (diffRsCsdiff >= 249) {
                        if (Cs < 1483) {
                            if (RsDiff < 22) {
                                return 1;
                            }
                            else if (RsDiff >= 22) {
                                if (AFD < 23) {
                                    return 1;
                                }
                                else if (AFD >= 23) {
                                    return 0;
                                }
                            }
                        }
                        else if (Cs >= 1483) {
                            if (diffAFD < 21) {
                                if (AFD < 19) {
                                    return 1;
                                }
                                else if (AFD >= 19) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 21) {
                                if (gchDC < 3) {
                                    return 0;
                                }
                                else if (gchDC >= 3) {
                                    if (gchDC < 21) {
                                        return 1;
                                    }
                                    else if (gchDC >= 21) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (TSC >= 16) {
        if (diffMVdiffVal < 65) {
            if (diffAFD < 19) {
                if (refDCval < 61) {
                    if (refDCval < 25) {
                        if (SC < 1212) {
                            if (TSC < 28) {
                                return 1;
                            }
                            else if (TSC >= 28) {
                                return 0;
                            }
                        }
                        else if (SC >= 1212) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 25) {
                        if (negBalance < 42) {
                            if (refDCval < 54) {
                                if (RsDiff < 4) {
                                    if (ssDCval < 57) {
                                        if (Cs < 579) {
                                            return 0;
                                        }
                                        else if (Cs >= 579) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 57) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 4) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 54) {
                                if (SC < 2255) {
                                    if (TSC < 20) {
                                        return 1;
                                    }
                                    else if (TSC >= 20) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 2255) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 42) {
                            if (gchDC < 6) {
                                if (refDCval < 47) {
                                    return 1;
                                }
                                else if (refDCval >= 47) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 6) {
                                return 1;
                            }
                        }
                    }
                }
                else if (refDCval >= 61) {
                    if (Cs < 2381) {
                        if (AFD < 59) {
                            if (diffMVdiffVal < 43) {
                                if (diffTSC < -2) {
                                    if (CsDiff < 8) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 8) {
                                        if (AFD < 49) {
                                            if (RsDiff < 92) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 92) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 49) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= -2) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 43) {
                                if (posBalance < 48) {
                                    return 0;
                                }
                                else if (posBalance >= 48) {
                                    if (Cs < 2322) {
                                        if (CsDiff < 6) {
                                            if (gchDC < 17) {
                                                if (RsCsDiff < 590) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 590) {
                                                    return 0;
                                                }
                                            }
                                            else if (gchDC >= 17) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (MVDiff < 66) {
                                                if (AFD < 32) {
                                                    return 0;
                                                }
                                                else if (AFD >= 32) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 66) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Cs >= 2322) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 59) {
                            if (ssDCval < 108) {
                                return 1;
                            }
                            else if (ssDCval >= 108) {
                                if (RsCsDiff < 470) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 470) {
                                    if (SC < 3672) {
                                        return 1;
                                    }
                                    else if (SC >= 3672) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 2381) {
                        return 0;
                    }
                }
            }
            else if (diffAFD >= 19) {
                if (diffAFD < 30) {
                    if (RsCsDiff < 5423) {
                        if (RsCsDiff < 80) {
                            if (refDCval < 29) {
                                return 1;
                            }
                            else if (refDCval >= 29) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 80) {
                            if (RsCsDiff < 1213) {
                                if (posBalance < 103) {
                                    if (Rs < 630) {
                                        return 1;
                                    }
                                    else if (Rs >= 630) {
                                        if (Rs < 1387) {
                                            if (diffMVdiffVal < -401) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -401) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 1387) {
                                            if (diffAFD < 22) {
                                                if (negBalance < 10) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 10) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 22) {
                                                if (Cs < 1464) {
                                                    if (SC < 2958) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2958) {
                                                        if (posBalance < 71) {
                                                            return 1;
                                                        }
                                                        else if (posBalance >= 71) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 1464) {
                                                    if (CsDiff < 15) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 15) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 103) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 1213) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 5423) {
                        return 1;
                    }
                }
                else if (diffAFD >= 30) {
                    if (refDCval < 136) {
                        if (AFD < 39) {
                            return 1;
                        }
                        else if (AFD >= 39) {
                            if (RsCsDiff < 209) {
                                if (Rs < 206) {
                                    if (SC < 253) {
                                        return 0;
                                    }
                                    else if (SC >= 253) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 206) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 209) {
                                if (MVDiff < 68) {
                                    return 1;
                                }
                                else if (MVDiff >= 68) {
                                    if (MVDiff < 113) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 113) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 136) {
                        if (posBalance < 6) {
                            return 1;
                        }
                        else if (posBalance >= 6) {
                            if (diffAFD < 58) {
                                return 0;
                            }
                            else if (diffAFD >= 58) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffRsCsdiff < 66) {
                if (diffMVdiffVal < 96) {
                    if (ssDCval < 184) {
                        return 0;
                    }
                    else if (ssDCval >= 184) {
                        if (ssDCval < 185) {
                            return 1;
                        }
                        else if (ssDCval >= 185) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 96) {
                    if (Cs < 691) {
                        if (negBalance < 14) {
                            if (gchDC < 77) {
                                if (AFD < 19) {
                                    return 1;
                                }
                                else if (AFD >= 19) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 77) {
                                if (MVDiff < 433) {
                                    return 1;
                                }
                                else if (MVDiff >= 433) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 14) {
                            if (diffAFD < 24) {
                                if (diffRsCsdiff < 36) {
                                    if (diffTSC < 5) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 5) {
                                        if (TSC < 17) {
                                            return 1;
                                        }
                                        else if (TSC >= 17) {
                                            if (gchDC < 7) {
                                                return 1;
                                            }
                                            else if (gchDC >= 7) {
                                                if (negBalance < 53) {
                                                    if (diffMVdiffVal < 138) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 138) {
                                                        if (Cs < 404) {
                                                            if (SC < 298) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 298) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (Cs >= 404) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 53) {
                                                    if (SC < 481) {
                                                        if (CsDiff < 2) {
                                                            return 0;
                                                        }
                                                        else if (CsDiff >= 2) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 481) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 36) {
                                    if (MVDiff < 298) {
                                        if (Rs < 938) {
                                            return 1;
                                        }
                                        else if (Rs >= 938) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 298) {
                                        if (diffMVdiffVal < 192) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 192) {
                                            if (AFD < 50) {
                                                return 1;
                                            }
                                            else if (AFD >= 50) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 24) {
                                if (AFD < 82) {
                                    if (SC < 216) {
                                        if (CsDiff < 2) {
                                            if (MVDiff < 324) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 324) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 2) {
                                            if (posBalance < 22) {
                                                return 0;
                                            }
                                            else if (posBalance >= 22) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 216) {
                                        if (negBalance < 50) {
                                            if (posBalance < 99) {
                                                if (negBalance < 28) {
                                                    if (AFD < 28) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 28) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 28) {
                                                    if (posBalance < 71) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 71) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (posBalance >= 99) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 50) {
                                            if (Rs < 73) {
                                                return 0;
                                            }
                                            else if (Rs >= 73) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 82) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (Cs >= 691) {
                        if (gchDC < 24) {
                            if (AFD < 53) {
                                if (diffAFD < 19) {
                                    return 0;
                                }
                                else if (diffAFD >= 19) {
                                    if (negBalance < 37) {
                                        return 0;
                                    }
                                    else if (negBalance >= 37) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 53) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 24) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 66) {
                if (diffAFD < 22) {
                    if (ssDCval < 63) {
                        if (MVDiff < 146) {
                            if (RsDiff < 15) {
                                return 0;
                            }
                            else if (RsDiff >= 15) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 146) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 63) {
                        if (diffRsCsdiff < 174) {
                            if (MVDiff < 740) {
                                return 0;
                            }
                            else if (MVDiff >= 740) {
                                if (RsCsDiff < 296) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 296) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 174) {
                            if (diffTSC < 10) {
                                if (CsDiff < 15) {
                                    return 0;
                                }
                                else if (CsDiff >= 15) {
                                    if (Rs < 1541) {
                                        return 1;
                                    }
                                    else if (Rs >= 1541) {
                                        if (Rs < 8557) {
                                            if (TSC < 22) {
                                                return 0;
                                            }
                                            else if (TSC >= 22) {
                                                if (Cs < 2574) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2574) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (Rs >= 8557) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 10) {
                                if (SC < 3908) {
                                    return 1;
                                }
                                else if (SC >= 3908) {
                                    if (diffRsCsdiff < 1276) {
                                        if (TSC < 27) {
                                            return 0;
                                        }
                                        else if (TSC >= 27) {
                                            if (AFD < 51) {
                                                return 1;
                                            }
                                            else if (AFD >= 51) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 1276) {
                                        if (Cs < 4743) {
                                            return 1;
                                        }
                                        else if (Cs >= 4743) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 22) {
                    if (diffAFD < 38) {
                        if (negBalance < 27) {
                            if (AFD < 54) {
                                if (diffMVdiffVal < 151) {
                                    if (diffTSC < 20) {
                                        if (Scindex < 5) {
                                            if (negBalance < 16) {
                                                return 0;
                                            }
                                            else if (negBalance >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (Scindex >= 5) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 20) {
                                        if (SC < 1727) {
                                            return 1;
                                        }
                                        else if (SC >= 1727) {
                                            if (diffRsCsdiff < 1180) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 1180) {
                                                if (SC < 10013) {
                                                    return 1;
                                                }
                                                else if (SC >= 10013) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 151) {
                                    if (diffAFD < 30) {
                                        if (diffRsCsdiff < 414) {
                                            if (Rs < 874) {
                                                return 1;
                                            }
                                            else if (Rs >= 874) {
                                                if (gchDC < 48) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 48) {
                                                    if (RsCsDiff < 195) {
                                                        return 0;
                                                    }
                                                    else if (RsCsDiff >= 195) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 414) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 30) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 54) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 27) {
                            if (MVDiff < 144) {
                                return 1;
                            }
                            else if (MVDiff >= 144) {
                                if (diffMVdiffVal < 98) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 98) {
                                    if (SC < 2204) {
                                        return 1;
                                    }
                                    else if (SC >= 2204) {
                                        if (gchDC < 20) {
                                            return 1;
                                        }
                                        else if (gchDC >= 20) {
                                            if (RsCsDiff < 1327) {
                                                if (negBalance < 73) {
                                                    if (AFD < 41) {
                                                        if (posBalance < 90) {
                                                            if (SC < 4121) {
                                                                return 0;
                                                            }
                                                            else if (SC >= 4121) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 90) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 41) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 73) {
                                                    if (Cs < 2197) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2197) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 1327) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 38) {
                        if (posBalance < 125) {
                            if (diffRsCsdiff < 194) {
                                if (refDCval < 47) {
                                    if (TSC < 33) {
                                        return 1;
                                    }
                                    else if (TSC >= 33) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 47) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 194) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 125) {
                            return 0;
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect8(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (diffMVdiffVal < 56) {
            if (TSC < 24) {
                if (diffRsCsdiff < 248) {
                    if (Cs < 56) {
                        if (TSC < 5) {
                            if (posBalance < 128) {
                                if (refDCval < 228) {
                                    if (negBalance < 1) {
                                        if (AFD < 1) {
                                            return 0;
                                        }
                                        else if (AFD >= 1) {
                                            if (MVDiff < 1179) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 1179) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 1) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 228) {
                                    if (SC < 38) {
                                        return 0;
                                    }
                                    else if (SC >= 38) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 128) {
                                return 1;
                            }
                        }
                        else if (TSC >= 5) {
                            if (MVDiff < 232) {
                                return 0;
                            }
                            else if (MVDiff >= 232) {
                                if (diffMVdiffVal < -16) {
                                    if (MVDiff < 747) {
                                        if (diffRsCsdiff < 25) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 25) {
                                            if (Rs < 126) {
                                                return 1;
                                            }
                                            else if (Rs >= 126) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 747) {
                                        if (AFD < 18) {
                                            return 1;
                                        }
                                        else if (AFD >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -16) {
                                    if (refDCval < 227) {
                                        if (diffRsCsdiff < -96) {
                                            if (RsCsDiff < 15) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 15) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -96) {
                                            if (diffMVdiffVal < -9) {
                                                if (AFD < 12) {
                                                    return 1;
                                                }
                                                else if (AFD >= 12) {
                                                    if (TSC < 12) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 12) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= -9) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 227) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 56) {
                        if (negBalance < 67) {
                            if (refDCval < 18) {
                                if (diffRsCsdiff < 7) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 7) {
                                    if (SC < 882) {
                                        return 0;
                                    }
                                    else if (SC >= 882) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 18) {
                                if (diffMVdiffVal < -9) {
                                    if (TSC < 11) {
                                        return 0;
                                    }
                                    else if (TSC >= 11) {
                                        if (SC < 1136) {
                                            if (ssDCval < 57) {
                                                if (CsDiff < 4) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 4) {
                                                    if (MVDiff < 110) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 110) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 57) {
                                                return 0;
                                            }
                                        }
                                        else if (SC >= 1136) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -9) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 67) {
                            if (ssDCval < 50) {
                                if (diffRsCsdiff < 12) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 12) {
                                    if (diffTSC < 5) {
                                        if (AFD < 41) {
                                            if (diffMVdiffVal < -273) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -273) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 41) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 5) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 50) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 248) {
                    if (negBalance < 9) {
                        if (Cs < 2604) {
                            if (ssDCval < 21) {
                                return 1;
                            }
                            else if (ssDCval >= 21) {
                                if (SC < 4038) {
                                    if (TSC < 3) {
                                        return 0;
                                    }
                                    else if (TSC >= 3) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 4038) {
                                    return 0;
                                }
                            }
                        }
                        else if (Cs >= 2604) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 9) {
                        if (RsDiff < 14) {
                            if (Cs < 1771) {
                                if (ssDCval < 63) {
                                    if (AFD < 29) {
                                        return 1;
                                    }
                                    else if (AFD >= 29) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 63) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 1771) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 14) {
                            return 0;
                        }
                    }
                }
            }
            else if (TSC >= 24) {
                if (ssDCval < 111) {
                    if (RsCsDiff < 387) {
                        return 0;
                    }
                    else if (RsCsDiff >= 387) {
                        if (Cs < 2454) {
                            if (diffAFD < 11) {
                                if (CsDiff < 10) {
                                    return 0;
                                }
                                else if (CsDiff >= 10) {
                                    if (SC < 3716) {
                                        return 1;
                                    }
                                    else if (SC >= 3716) {
                                        if (SC < 5712) {
                                            return 0;
                                        }
                                        else if (SC >= 5712) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 11) {
                                return 0;
                            }
                        }
                        else if (Cs >= 2454) {
                            return 0;
                        }
                    }
                }
                else if (ssDCval >= 111) {
                    return 0;
                }
            }
        }
        else if (diffMVdiffVal >= 56) {
            if (diffTSC < 4) {
                if (negBalance < 59) {
                    if (gchDC < 103) {
                        if (RsCsDiff < 12) {
                            if (ssDCval < 24) {
                                if (negBalance < 46) {
                                    return 0;
                                }
                                else if (negBalance >= 46) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 24) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 12) {
                            if (diffAFD < 18) {
                                if (Cs < 99) {
                                    if (SC < 145) {
                                        return 1;
                                    }
                                    else if (SC >= 145) {
                                        if (Cs < 91) {
                                            return 0;
                                        }
                                        else if (Cs >= 91) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 99) {
                                    if (TSC < 1) {
                                        if (SC < 920) {
                                            return 0;
                                        }
                                        else if (SC >= 920) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 1) {
                                        if (RsCsDiff < 15) {
                                            if (posBalance < 87) {
                                                return 0;
                                            }
                                            else if (posBalance >= 87) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 15) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 18) {
                                if (refDCval < 64) {
                                    return 1;
                                }
                                else if (refDCval >= 64) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 103) {
                        if (CsDiff < 2) {
                            return 0;
                        }
                        else if (CsDiff >= 2) {
                            return 1;
                        }
                    }
                }
                else if (negBalance >= 59) {
                    if (diffAFD < 8) {
                        if (diffAFD < -36) {
                            if (SC < 431) {
                                return 1;
                            }
                            else if (SC >= 431) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= -36) {
                            if (negBalance < 83) {
                                if (gchDC < 5) {
                                    return 0;
                                }
                                else if (gchDC >= 5) {
                                    if (ssDCval < 31) {
                                        if (RsDiff < 3) {
                                            if (diffRsCsdiff < -21) {
                                                return 1;
                                            }
                                            else if (diffRsCsdiff >= -21) {
                                                return 0;
                                            }
                                        }
                                        else if (RsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 31) {
                                        if (CsDiff < 5) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 5) {
                                            if (posBalance < 48) {
                                                return 0;
                                            }
                                            else if (posBalance >= 48) {
                                                if (ssDCval < 48) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 48) {
                                                    if (ssDCval < 124) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 124) {
                                                        if (AFD < 56) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 56) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 83) {
                                if (ssDCval < 24) {
                                    if (CsDiff < 1) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 24) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (posBalance < 2) {
                            return 1;
                        }
                        else if (posBalance >= 2) {
                            if (SC < 2894) {
                                if (diffRsCsdiff < 85) {
                                    if (AFD < 16) {
                                        return 1;
                                    }
                                    else if (AFD >= 16) {
                                        if (Rs < 146) {
                                            if (diffMVdiffVal < 236) {
                                                if (CsDiff < 4) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 4) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 236) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 146) {
                                            if (MVDiff < 1442) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 1442) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 85) {
                                    if (AFD < 69) {
                                        if (CsDiff < 20) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 20) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 69) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 2894) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 4) {
                if (Rs < 102) {
                    if (diffAFD < 9) {
                        if (Rs < 6) {
                            return 1;
                        }
                        else if (Rs >= 6) {
                            if (ssDCval < 28) {
                                return 1;
                            }
                            else if (ssDCval >= 28) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 9) {
                        if (refDCval < 64) {
                            if (diffMVdiffVal < 135) {
                                if (AFD < 15) {
                                    return 1;
                                }
                                else if (AFD >= 15) {
                                    if (diffTSC < 6) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 6) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 135) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 64) {
                            if (RsDiff < 7) {
                                if (AFD < 23) {
                                    if (Rs < 42) {
                                        return 0;
                                    }
                                    else if (Rs >= 42) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 23) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 7) {
                                return 1;
                            }
                        }
                    }
                }
                else if (Rs >= 102) {
                    if (TSC < 34) {
                        if (diffRsCsdiff < 8) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 8) {
                            if (SC < 1410) {
                                if (diffRsCsdiff < 50) {
                                    if (ssDCval < 43) {
                                        if (MVDiff < 97) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 97) {
                                            if (RsDiff < 5) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 5) {
                                                if (Cs < 164) {
                                                    return 1;
                                                }
                                                else if (Cs >= 164) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 43) {
                                        if (ssDCval < 120) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 120) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 50) {
                                    if (Rs < 792) {
                                        if (TSC < 8) {
                                            if (AFD < 24) {
                                                return 1;
                                            }
                                            else if (AFD >= 24) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 8) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 792) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 1410) {
                                if (ssDCval < 40) {
                                    return 1;
                                }
                                else if (ssDCval >= 40) {
                                    if (MVDiff < 534) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 534) {
                                        if (refDCval < 52) {
                                            return 0;
                                        }
                                        else if (refDCval >= 52) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 34) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (MVDiff < 56) {
            if (Cs < 2138) {
                if (TSC < 27) {
                    if (diffRsCsdiff < 565) {
                        if (refDCval < 29) {
                            if (SC < 79) {
                                return 0;
                            }
                            else if (SC >= 79) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 29) {
                            if (ssDCval < 70) {
                                if (ssDCval < 63) {
                                    return 0;
                                }
                                else if (ssDCval >= 63) {
                                    if (diffAFD < 16) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 16) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 70) {
                                return 0;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 565) {
                        if (negBalance < 35) {
                            if (diffTSC < 8) {
                                return 0;
                            }
                            else if (diffTSC >= 8) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 35) {
                            return 1;
                        }
                    }
                }
                else if (TSC >= 27) {
                    return 1;
                }
            }
            else if (Cs >= 2138) {
                if (TSC < 33) {
                    return 0;
                }
                else if (TSC >= 33) {
                    if (Rs < 4830) {
                        return 1;
                    }
                    else if (Rs >= 4830) {
                        return 0;
                    }
                }
            }
        }
        else if (MVDiff >= 56) {
            if (negBalance < 30) {
                if (diffMVdiffVal < 99) {
                    if (RsDiff < 10) {
                        if (Scindex < 3) {
                            if (negBalance < 1) {
                                return 1;
                            }
                            else if (negBalance >= 1) {
                                if (Cs < 104) {
                                    if (diffMVdiffVal < 89) {
                                        if (negBalance < 12) {
                                            return 0;
                                        }
                                        else if (negBalance >= 12) {
                                            if (Cs < 99) {
                                                return 0;
                                            }
                                            else if (Cs >= 99) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 89) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 104) {
                                    if (SC < 488) {
                                        return 0;
                                    }
                                    else if (SC >= 488) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Scindex >= 3) {
                            if (refDCval < 39) {
                                if (diffRsCsdiff < 40) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 40) {
                                    if (refDCval < 37) {
                                        if (SC < 1054) {
                                            return 1;
                                        }
                                        else if (SC >= 1054) {
                                            if (diffTSC < 26) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 26) {
                                                if (Cs < 1213) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1213) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 37) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 39) {
                                return 0;
                            }
                        }
                    }
                    else if (RsDiff >= 10) {
                        if (SC < 2375) {
                            if (refDCval < 97) {
                                if (MVDiff < 68) {
                                    return 0;
                                }
                                else if (MVDiff >= 68) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 97) {
                                if (TSC < 26) {
                                    return 0;
                                }
                                else if (TSC >= 26) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 2375) {
                            if (RsDiff < 22) {
                                if (diffAFD < 19) {
                                    return 0;
                                }
                                else if (diffAFD >= 19) {
                                    if (MVDiff < 141) {
                                        if (diffRsCsdiff < 222) {
                                            if (Cs < 742) {
                                                return 1;
                                            }
                                            else if (Cs >= 742) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 222) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 141) {
                                        if (Rs < 2152) {
                                            if (gchDC < 92) {
                                                if (Rs < 1771) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1771) {
                                                    if (SC < 3012) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3012) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 92) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 2152) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 22) {
                                if (diffTSC < 14) {
                                    if (AFD < 19) {
                                        return 1;
                                    }
                                    else if (AFD >= 19) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 14) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 99) {
                    if (diffRsCsdiff < 66) {
                        if (Rs < 377) {
                            if (RsCsDiff < 53) {
                                if (RsCsDiff < 31) {
                                    if (SC < 685) {
                                        if (MVDiff < 353) {
                                            if (TSC < 20) {
                                                if (Cs < 94) {
                                                    return 1;
                                                }
                                                else if (Cs >= 94) {
                                                    if (Cs < 126) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 126) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 20) {
                                                if (MVDiff < 232) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 232) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 353) {
                                            if (AFD < 28) {
                                                if (RsCsDiff < 3) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 3) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 28) {
                                                if (negBalance < 9) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 9) {
                                                    if (SC < 111) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 111) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 685) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 31) {
                                    if (gchDC < 30) {
                                        if (Rs < 206) {
                                            return 1;
                                        }
                                        else if (Rs >= 206) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 30) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 53) {
                                if (MVDiff < 1028) {
                                    return 0;
                                }
                                else if (MVDiff >= 1028) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 377) {
                            if (gchDC < 71) {
                                if (RsDiff < 5) {
                                    return 0;
                                }
                                else if (RsDiff >= 5) {
                                    if (ssDCval < 111) {
                                        if (AFD < 31) {
                                            if (diffTSC < 16) {
                                                if (diffAFD < 20) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 20) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffTSC >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 31) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 111) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 71) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 66) {
                        if (diffRsCsdiff < 740) {
                            if (Cs < 990) {
                                if (Rs < 2373) {
                                    if (diffTSC < 10) {
                                        if (MVDiff < 715) {
                                            if (TSC < 22) {
                                                return 1;
                                            }
                                            else if (TSC >= 22) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 715) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 10) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 2373) {
                                    if (Cs < 394) {
                                        return 1;
                                    }
                                    else if (Cs >= 394) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Cs >= 990) {
                                if (diffMVdiffVal < 377) {
                                    if (gchDC < 30) {
                                        return 1;
                                    }
                                    else if (gchDC >= 30) {
                                        if (posBalance < 104) {
                                            return 0;
                                        }
                                        else if (posBalance >= 104) {
                                            if (Rs < 836) {
                                                return 0;
                                            }
                                            else if (Rs >= 836) {
                                                if (diffRsCsdiff < 433) {
                                                    if (MVDiff < 139) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 139) {
                                                        if (RsCsDiff < 138) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 138) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 433) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 377) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 740) {
                            return 1;
                        }
                    }
                }
            }
            else if (negBalance >= 30) {
                if (diffTSC < 16) {
                    if (diffAFD < 12) {
                        if (diffMVdiffVal < 74) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 74) {
                            if (posBalance < 16) {
                                return 0;
                            }
                            else if (posBalance >= 16) {
                                if (ssDCval < 55) {
                                    if (AFD < 11) {
                                        return 0;
                                    }
                                    else if (AFD >= 11) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 55) {
                                    if (TSC < 22) {
                                        return 0;
                                    }
                                    else if (TSC >= 22) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            if (Rs < 1366) {
                                                return 1;
                                            }
                                            else if (Rs >= 1366) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 12) {
                        if (diffMVdiffVal < 63) {
                            if (diffTSC < 11) {
                                if (AFD < 21) {
                                    if (TSC < 10) {
                                        return 0;
                                    }
                                    else if (TSC >= 10) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 21) {
                                    if (Cs < 699) {
                                        if (refDCval < 93) {
                                            if (CsDiff < 4) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (Rs < 475) {
                                                    return 0;
                                                }
                                                else if (Rs >= 475) {
                                                    if (negBalance < 62) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 62) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (refDCval >= 93) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 699) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (diffRsCsdiff < 18) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 18) {
                                    if (Cs < 505) {
                                        if (RsDiff < 6) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 505) {
                                        if (MVDiff < 69) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 69) {
                                            if (refDCval < 63) {
                                                return 0;
                                            }
                                            else if (refDCval >= 63) {
                                                if (diffRsCsdiff < 493) {
                                                    if (TSC < 28) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 28) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 493) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 63) {
                            if (TSC < 25) {
                                if (ssDCval < 93) {
                                    if (negBalance < 43) {
                                        if (RsDiff < 4) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 4) {
                                            if (SC < 732) {
                                                if (Scindex < 3) {
                                                    return 1;
                                                }
                                                else if (Scindex >= 3) {
                                                    return 0;
                                                }
                                            }
                                            else if (SC >= 732) {
                                                if (CsDiff < 6) {
                                                    if (SC < 1013) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 1013) {
                                                        return 0;
                                                    }
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (diffMVdiffVal < 83) {
                                                        if (refDCval < 63) {
                                                            return 1;
                                                        }
                                                        else if (refDCval >= 63) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 83) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (gchDC < 7) {
                                            return 1;
                                        }
                                        else if (gchDC >= 7) {
                                            if (Cs < 677) {
                                                if (gchDC < 79) {
                                                    if (diffRsCsdiff < 1) {
                                                        if (diffTSC < 8) {
                                                            return 0;
                                                        }
                                                        else if (diffTSC >= 8) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffRsCsdiff >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 79) {
                                                    if (SC < 213) {
                                                        if (Cs < 45) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 45) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (SC >= 213) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 677) {
                                                if (AFD < 34) {
                                                    if (ssDCval < 36) {
                                                        if (ssDCval < 33) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 33) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (ssDCval >= 36) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 34) {
                                                    if (Cs < 811) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 811) {
                                                        if (CsDiff < 20) {
                                                            if (RsDiff < 35) {
                                                                return 0;
                                                            }
                                                            else if (RsDiff >= 35) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (CsDiff >= 20) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 93) {
                                    if (diffRsCsdiff < 2) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 2) {
                                        if (diffAFD < 25) {
                                            if (diffAFD < 19) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 19) {
                                                if (Scindex < 4) {
                                                    return 1;
                                                }
                                                else if (Scindex >= 4) {
                                                    if (diffAFD < 20) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 20) {
                                                        if (diffRsCsdiff < 2026) {
                                                            return 0;
                                                        }
                                                        else if (diffRsCsdiff >= 2026) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 25) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 25) {
                                if (gchDC < 7) {
                                    return 1;
                                }
                                else if (gchDC >= 7) {
                                    if (CsDiff < 15) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 15) {
                                        if (CsDiff < 31) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 31) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 16) {
                    if (diffRsCsdiff < 63) {
                        if (diffMVdiffVal < 95) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 95) {
                            if (diffTSC < 25) {
                                if (diffRsCsdiff < 1) {
                                    if (gchDC < 29) {
                                        return 1;
                                    }
                                    else if (gchDC >= 29) {
                                        if (Rs < 2319) {
                                            return 0;
                                        }
                                        else if (Rs >= 2319) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 1) {
                                    if (diffRsCsdiff < 62) {
                                        if (Cs < 754) {
                                            if (refDCval < 56) {
                                                if (AFD < 43) {
                                                    return 1;
                                                }
                                                else if (AFD >= 43) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 56) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 754) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 62) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 25) {
                                if (refDCval < 183) {
                                    return 1;
                                }
                                else if (refDCval >= 183) {
                                    if (gchDC < 47) {
                                        return 1;
                                    }
                                    else if (gchDC >= 47) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 63) {
                        if (posBalance < 36) {
                            if (posBalance < 35) {
                                if (SC < 1745) {
                                    return 1;
                                }
                                else if (SC >= 1745) {
                                    if (diffMVdiffVal < 120) {
                                        if (MVDiff < 170) {
                                            if (RsDiff < 9) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 9) {
                                                if (AFD < 59) {
                                                    return 1;
                                                }
                                                else if (AFD >= 59) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 170) {
                                            if (refDCval < 195) {
                                                return 0;
                                            }
                                            else if (refDCval >= 195) {
                                                if (Rs < 1676) {
                                                    return 0;
                                                }
                                                else if (Rs >= 1676) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 120) {
                                        if (gchDC < 82) {
                                            return 1;
                                        }
                                        else if (gchDC >= 82) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 35) {
                                if (TSC < 22) {
                                    return 0;
                                }
                                else if (TSC >= 22) {
                                    return 1;
                                }
                            }
                        }
                        else if (posBalance >= 36) {
                            if (MVDiff < 95) {
                                if (RsCsDiff < 245) {
                                    if (refDCval < 68) {
                                        return 1;
                                    }
                                    else if (refDCval >= 68) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 245) {
                                    if (Cs < 6363) {
                                        return 1;
                                    }
                                    else if (Cs >= 6363) {
                                        if (TSC < 32) {
                                            return 0;
                                        }
                                        else if (TSC >= 32) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 95) {
                                if (diffAFD < 25) {
                                    if (TSC < 44) {
                                        if (diffAFD < 24) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 24) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 44) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 25) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect9(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 14) {
        if (diffMVdiffVal < 54) {
            if (refDCval < 24) {
                if (diffRsCsdiff < 115) {
                    if (gchDC < 32) {
                        if (diffRsCsdiff < 5) {
                            if (diffAFD < -1) {
                                if (diffMVdiffVal < 2) {
                                    if (TSCindex < 2) {
                                        return 0;
                                    }
                                    else if (TSCindex >= 2) {
                                        if (diffTSC < -2) {
                                            if (diffTSC < -3) {
                                                return 0;
                                            }
                                            else if (diffTSC >= -3) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= -2) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 2) {
                                    if (MVDiff < 10) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 10) {
                                        if (ssDCval < 30) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 30) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= -1) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (negBalance < 4) {
                                if (diffRsCsdiff < 7) {
                                    if (TSC < 1) {
                                        return 0;
                                    }
                                    else if (TSC >= 1) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 7) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 4) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 32) {
                        return 1;
                    }
                }
                else if (diffRsCsdiff >= 115) {
                    return 1;
                }
            }
            else if (refDCval >= 24) {
                if (gchDC < 5) {
                    if (refDCval < 44) {
                        if (diffAFD < 10) {
                            if (TSCindex < 9) {
                                return 0;
                            }
                            else if (TSCindex >= 9) {
                                if (MVDiff < 311) {
                                    return 0;
                                }
                                else if (MVDiff >= 311) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 10) {
                            if (diffTSC < 5) {
                                return 0;
                            }
                            else if (diffTSC >= 5) {
                                if (Cs < 2320) {
                                    return 1;
                                }
                                else if (Cs >= 2320) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 44) {
                        if (ssDCval < 174) {
                            return 0;
                        }
                        else if (ssDCval >= 174) {
                            if (refDCval < 174) {
                                if (diffRsCsdiff < 814) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 814) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 174) {
                                return 0;
                            }
                        }
                    }
                }
                else if (gchDC >= 5) {
                    if (diffRsCsdiff < 255) {
                        if (MVDiff < 224) {
                            if (SC < 81) {
                                if (Cs < 61) {
                                    if (refDCval < 227) {
                                        return 0;
                                    }
                                    else if (refDCval >= 227) {
                                        if (Rs < 1) {
                                            return 1;
                                        }
                                        else if (Rs >= 1) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Cs >= 61) {
                                    if (Rs < 8) {
                                        return 0;
                                    }
                                    else if (Rs >= 8) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 81) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 224) {
                            if (diffTSC < 4) {
                                if (gchDC < 7) {
                                    if (Rs < 1) {
                                        return 1;
                                    }
                                    else if (Rs >= 1) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            if (RsDiff < 8) {
                                                if (ssDCval < 65) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 65) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 8) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 7) {
                                    if (AFD < 62) {
                                        return 0;
                                    }
                                    else if (AFD >= 62) {
                                        if (SC < 3672) {
                                            return 1;
                                        }
                                        else if (SC >= 3672) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 4) {
                                if (refDCval < 114) {
                                    if (negBalance < 48) {
                                        if (MVDiff < 241) {
                                            if (MVDiff < 233) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 233) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 241) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 48) {
                                        if (SC < 647) {
                                            return 0;
                                        }
                                        else if (SC >= 647) {
                                            if (diffRsCsdiff < 62) {
                                                if (TSC < 14) {
                                                    if (Rs < 540) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 540) {
                                                        return 0;
                                                    }
                                                }
                                                else if (TSC >= 14) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 62) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 114) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 255) {
                        if (ssDCval < 47) {
                            if (Rs < 2156) {
                                if (diffAFD < 5) {
                                    if (RsCsDiff < 6622) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 6622) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 5) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 2156) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 47) {
                            if (Cs < 177) {
                                return 1;
                            }
                            else if (Cs >= 177) {
                                if (AFD < 58) {
                                    return 0;
                                }
                                else if (AFD >= 58) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 54) {
            if (diffAFD < 7) {
                if (ssDCval < 38) {
                    if (Cs < 63) {
                        if (MVDiff < 1653) {
                            if (RsDiff < 4) {
                                return 0;
                            }
                            else if (RsDiff >= 4) {
                                if (TSC < 4) {
                                    return 1;
                                }
                                else if (TSC >= 4) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 1653) {
                            if (diffMVdiffVal < 1131) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 1131) {
                                return 1;
                            }
                        }
                    }
                    else if (Cs >= 63) {
                        if (TSC < 5) {
                            if (Rs < 40) {
                                return 1;
                            }
                            else if (Rs >= 40) {
                                if (Cs < 574) {
                                    return 0;
                                }
                                else if (Cs >= 574) {
                                    if (Cs < 992) {
                                        return 1;
                                    }
                                    else if (Cs >= 992) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 5) {
                            if (diffMVdiffVal < 74) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 74) {
                                if (CsDiff < 1) {
                                    return 0;
                                }
                                else if (CsDiff >= 1) {
                                    if (diffTSC < -6) {
                                        return 1;
                                    }
                                    else if (diffTSC >= -6) {
                                        if (refDCval < 34) {
                                            if (posBalance < 72) {
                                                return 1;
                                            }
                                            else if (posBalance >= 72) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 34) {
                                            if (MVDiff < 1099) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 1099) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 38) {
                    if (TSC < 36) {
                        if (AFD < 34) {
                            if (SC < 1) {
                                if (diffTSC < -7) {
                                    return 1;
                                }
                                else if (diffTSC >= -7) {
                                    if (MVDiff < 272) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 272) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 1) {
                                return 0;
                            }
                        }
                        else if (AFD >= 34) {
                            if (SC < 1104) {
                                if (diffRsCsdiff < 18) {
                                    if (refDCval < 76) {
                                        if (RsCsDiff < 77) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 77) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 76) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 18) {
                                    if (diffTSC < -6) {
                                        return 0;
                                    }
                                    else if (diffTSC >= -6) {
                                        if (gchDC < 9) {
                                            return 0;
                                        }
                                        else if (gchDC >= 9) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 1104) {
                                return 0;
                            }
                        }
                    }
                    else if (TSC >= 36) {
                        if (SC < 2017) {
                            return 0;
                        }
                        else if (SC >= 2017) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 7) {
                if (SC < 825) {
                    if (ssDCval < 50) {
                        if (gchDC < 10) {
                            if (SC < 740) {
                                if (diffRsCsdiff < 5) {
                                    if (MVDiff < 200) {
                                        if (diffTSC < 5) {
                                            if (AFD < 12) {
                                                if (posBalance < 65) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 65) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 12) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 200) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 5) {
                                    return 1;
                                }
                            }
                            else if (SC >= 740) {
                                if (ssDCval < 44) {
                                    return 0;
                                }
                                else if (ssDCval >= 44) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 10) {
                            if (diffMVdiffVal < 214) {
                                if (gchDC < 20) {
                                    if (Rs < 458) {
                                        return 0;
                                    }
                                    else if (Rs >= 458) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 20) {
                                    if (CsDiff < 9) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 214) {
                                if (gchDC < 23) {
                                    return 1;
                                }
                                else if (gchDC >= 23) {
                                    if (diffAFD < 10) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 10) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 50) {
                        if (ssDCval < 224) {
                            if (CsDiff < 12) {
                                if (refDCval < 65) {
                                    if (negBalance < 34) {
                                        return 0;
                                    }
                                    else if (negBalance >= 34) {
                                        if (gchDC < 8) {
                                            return 0;
                                        }
                                        else if (gchDC >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 65) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 12) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 224) {
                            if (AFD < 20) {
                                return 1;
                            }
                            else if (AFD >= 20) {
                                return 0;
                            }
                        }
                    }
                }
                else if (SC >= 825) {
                    if (SC < 2557) {
                        if (diffRsCsdiff < 45) {
                            if (diffRsCsdiff < -82) {
                                if (refDCval < 81) {
                                    return 1;
                                }
                                else if (refDCval >= 81) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -82) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 45) {
                            if (TSC < 27) {
                                if (negBalance < 40) {
                                    if (negBalance < 24) {
                                        return 0;
                                    }
                                    else if (negBalance >= 24) {
                                        if (diffAFD < 12) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 12) {
                                            if (Rs < 1391) {
                                                return 1;
                                            }
                                            else if (Rs >= 1391) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 40) {
                                    if (diffMVdiffVal < 128) {
                                        if (ssDCval < 71) {
                                            if (RsDiff < 10) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 10) {
                                                if (SC < 1198) {
                                                    return 0;
                                                }
                                                else if (SC >= 1198) {
                                                    if (RsCsDiff < 521) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 521) {
                                                        if (Cs < 1428) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1428) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 71) {
                                            if (gchDC < 34) {
                                                return 0;
                                            }
                                            else if (gchDC >= 34) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 128) {
                                        return 1;
                                    }
                                }
                            }
                            else if (TSC >= 27) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 2557) {
                        return 0;
                    }
                }
            }
        }
    }
    else if (diffAFD >= 14) {
        if (diffMVdiffVal < 65) {
            if (diffAFD < 25) {
                if (MVDiff < 153) {
                    if (gchDC < 4) {
                        if (ssDCval < 36) {
                            if (SC < 1436) {
                                return 0;
                            }
                            else if (SC >= 1436) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 36) {
                            if (diffMVdiffVal < 59) {
                                if (SC < 2658) {
                                    if (TSC < 13) {
                                        return 0;
                                    }
                                    else if (TSC >= 13) {
                                        if (RsCsDiff < 158) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 158) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 2658) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 59) {
                                if (Cs < 557) {
                                    return 0;
                                }
                                else if (Cs >= 557) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 4) {
                        if (Rs < 7070) {
                            if (ssDCval < 53) {
                                if (Rs < 1146) {
                                    if (diffRsCsdiff < 169) {
                                        if (diffTSC < 17) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 17) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 169) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 1146) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 53) {
                                if (diffMVdiffVal < -506) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= -506) {
                                    if (Cs < 4218) {
                                        return 0;
                                    }
                                    else if (Cs >= 4218) {
                                        if (Cs < 4368) {
                                            return 1;
                                        }
                                        else if (Cs >= 4368) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 7070) {
                            return 1;
                        }
                    }
                }
                else if (MVDiff >= 153) {
                    if (gchDC < 8) {
                        if (AFD < 42) {
                            return 1;
                        }
                        else if (AFD >= 42) {
                            return 0;
                        }
                    }
                    else if (gchDC >= 8) {
                        if (gchDC < 56) {
                            if (RsCsDiff < 345) {
                                if (ssDCval < 44) {
                                    return 1;
                                }
                                else if (ssDCval >= 44) {
                                    if (diffAFD < 23) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 23) {
                                        if (RsCsDiff < 73) {
                                            if (Rs < 425) {
                                                return 0;
                                            }
                                            else if (Rs >= 425) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 73) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 345) {
                                if (RsCsDiff < 549) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 549) {
                                    if (refDCval < 132) {
                                        return 0;
                                    }
                                    else if (refDCval >= 132) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 56) {
                            if (AFD < 32) {
                                return 0;
                            }
                            else if (AFD >= 32) {
                                if (CsDiff < 8) {
                                    if (MVDiff < 720) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 720) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 8) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 25) {
                if (diffAFD < 47) {
                    if (Rs < 1682) {
                        if (ssDCval < 76) {
                            if (Rs < 127) {
                                return 0;
                            }
                            else if (Rs >= 127) {
                                if (SC < 948) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 948) {
                                    if (Cs < 486) {
                                        return 1;
                                    }
                                    else if (Cs >= 486) {
                                        if (CsDiff < 18) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 18) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 76) {
                            if (AFD < 57) {
                                if (posBalance < 124) {
                                    return 0;
                                }
                                else if (posBalance >= 124) {
                                    if (MVDiff < 120) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 120) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 57) {
                                if (diffMVdiffVal < -2) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= -2) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (Rs >= 1682) {
                        if (MVDiff < 152) {
                            if (Cs < 3562) {
                                if (TSC < 18) {
                                    return 0;
                                }
                                else if (TSC >= 18) {
                                    if (RsCsDiff < 809) {
                                        if (Rs < 2145) {
                                            return 1;
                                        }
                                        else if (Rs >= 2145) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 809) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 3562) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 152) {
                            return 1;
                        }
                    }
                }
                else if (diffAFD >= 47) {
                    return 1;
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (RsDiff < 14) {
                if (negBalance < 30) {
                    if (diffMVdiffVal < 142) {
                        if (ssDCval < 69) {
                            if (CsDiff < 1) {
                                return 0;
                            }
                            else if (CsDiff >= 1) {
                                if (MVDiff < 117) {
                                    if (Rs < 322) {
                                        return 0;
                                    }
                                    else if (Rs >= 322) {
                                        if (Cs < 1093) {
                                            return 1;
                                        }
                                        else if (Cs >= 1093) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 117) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 69) {
                            if (diffMVdiffVal < 99) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 99) {
                                if (TSC < 30) {
                                    if (AFD < 45) {
                                        if (gchDC < 73) {
                                            if (negBalance < 16) {
                                                return 0;
                                            }
                                            else if (negBalance >= 16) {
                                                if (RsDiff < 10) {
                                                    if (ssDCval < 89) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 89) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsDiff >= 10) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 73) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 45) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 30) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 142) {
                        if (TSC < 11) {
                            return 1;
                        }
                        else if (TSC >= 11) {
                            if (negBalance < 10) {
                                if (diffMVdiffVal < 415) {
                                    if (CsDiff < 4) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 4) {
                                        if (TSC < 18) {
                                            if (refDCval < 107) {
                                                return 1;
                                            }
                                            else if (refDCval >= 107) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 18) {
                                            if (SC < 2876) {
                                                if (MVDiff < 190) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 190) {
                                                    return 0;
                                                }
                                            }
                                            else if (SC >= 2876) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 415) {
                                    if (RsCsDiff < 14) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 14) {
                                        if (refDCval < 125) {
                                            return 1;
                                        }
                                        else if (refDCval >= 125) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 10) {
                                if (diffTSC < 9) {
                                    return 0;
                                }
                                else if (diffTSC >= 9) {
                                    if (MVDiff < 1439) {
                                        if (TSC < 20) {
                                            return 1;
                                        }
                                        else if (TSC >= 20) {
                                            if (ssDCval < 103) {
                                                if (CsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 7) {
                                                    if (Rs < 1199) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1199) {
                                                        if (Cs < 1280) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 1280) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 103) {
                                                if (RsDiff < 13) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 13) {
                                                    if (posBalance < 103) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 103) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1439) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 30) {
                    if (diffTSC < 7) {
                        if (AFD < 43) {
                            if (Rs < 541) {
                                if (diffMVdiffVal < 93) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 93) {
                                    if (refDCval < 111) {
                                        return 1;
                                    }
                                    else if (refDCval >= 111) {
                                        if (CsDiff < 6) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 541) {
                                return 0;
                            }
                        }
                        else if (AFD >= 43) {
                            if (diffAFD < 26) {
                                return 0;
                            }
                            else if (diffAFD >= 26) {
                                if (CsDiff < 3) {
                                    return 0;
                                }
                                else if (CsDiff >= 3) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 7) {
                        if (TSC < 29) {
                            if (diffRsCsdiff < 1) {
                                if (negBalance < 51) {
                                    return 0;
                                }
                                else if (negBalance >= 51) {
                                    if (CsDiff < 12) {
                                        if (gchDC < 34) {
                                            return 1;
                                        }
                                        else if (gchDC >= 34) {
                                            if (Cs < 124) {
                                                return 0;
                                            }
                                            else if (Cs >= 124) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 12) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 1) {
                                if (diffAFD < 25) {
                                    if (CsDiff < 5) {
                                        if (Cs < 43) {
                                            if (diffMVdiffVal < 328) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= 328) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 43) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 5) {
                                        if (diffRsCsdiff < 98) {
                                            if (SC < 627) {
                                                return 1;
                                            }
                                            else if (SC >= 627) {
                                                if (diffMVdiffVal < 85) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 85) {
                                                    if (AFD < 33) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 33) {
                                                        if (diffMVdiffVal < 146) {
                                                            return 0;
                                                        }
                                                        else if (diffMVdiffVal >= 146) {
                                                            if (ssDCval < 81) {
                                                                return 1;
                                                            }
                                                            else if (ssDCval >= 81) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 98) {
                                            if (TSC < 24) {
                                                return 1;
                                            }
                                            else if (TSC >= 24) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 25) {
                                    if (diffMVdiffVal < 80) {
                                        if (AFD < 30) {
                                            return 1;
                                        }
                                        else if (AFD >= 30) {
                                            if (posBalance < 34) {
                                                return 1;
                                            }
                                            else if (posBalance >= 34) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 80) {
                                        if (ssDCval < 46) {
                                            if (diffRsCsdiff < 46) {
                                                if (gchDC < 7) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 7) {
                                                    if (negBalance < 71) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 71) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 46) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 46) {
                                            if (CsDiff < 10) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 10) {
                                                if (diffRsCsdiff < 91) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 91) {
                                                    if (diffAFD < 52) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 52) {
                                                        if (SC < 1118) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 1118) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 29) {
                            if (diffRsCsdiff < 64) {
                                if (SC < 897) {
                                    if (diffTSC < 27) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 27) {
                                        if (posBalance < 17) {
                                            return 0;
                                        }
                                        else if (posBalance >= 17) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 897) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 64) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (RsDiff >= 14) {
                if (negBalance < 31) {
                    if (SC < 3198) {
                        if (CsDiff < 3) {
                            return 0;
                        }
                        else if (CsDiff >= 3) {
                            return 1;
                        }
                    }
                    else if (SC >= 3198) {
                        if (RsDiff < 30) {
                            if (CsDiff < 19) {
                                if (RsCsDiff < 373) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 373) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 19) {
                                if (SC < 4537) {
                                    if (AFD < 32) {
                                        return 1;
                                    }
                                    else if (AFD >= 32) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 4537) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 30) {
                            if (negBalance < 2) {
                                return 0;
                            }
                            else if (negBalance >= 2) {
                                return 1;
                            }
                        }
                    }
                }
                else if (negBalance >= 31) {
                    if (diffRsCsdiff < 218) {
                        if (SC < 2115) {
                            if (TSC < 26) {
                                return 0;
                            }
                            else if (TSC >= 26) {
                                if (refDCval < 149) {
                                    return 1;
                                }
                                else if (refDCval >= 149) {
                                    return 0;
                                }
                            }
                        }
                        else if (SC >= 2115) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 218) {
                        if (TSC < 16) {
                            if (refDCval < 71) {
                                return 1;
                            }
                            else if (refDCval >= 71) {
                                if (MVDiff < 272) {
                                    if (diffTSC < 10) {
                                        if (posBalance < 75) {
                                            return 0;
                                        }
                                        else if (posBalance >= 75) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 10) {
                                        if (negBalance < 72) {
                                            return 0;
                                        }
                                        else if (negBalance >= 72) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 272) {
                                    return 1;
                                }
                            }
                        }
                        else if (TSC >= 16) {
                            if (AFD < 76) {
                                if (diffMVdiffVal < 113) {
                                    if (MVDiff < 128) {
                                        if (SC < 16399) {
                                            return 1;
                                        }
                                        else if (SC >= 16399) {
                                            if (TSC < 32) {
                                                return 0;
                                            }
                                            else if (TSC >= 32) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 128) {
                                        if (diffAFD < 25) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 25) {
                                            if (CsDiff < 28) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 28) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 113) {
                                    if (Rs < 2296) {
                                        return 1;
                                    }
                                    else if (Rs >= 2296) {
                                        if (Rs < 2341) {
                                            if (MVDiff < 213) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 213) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 2341) {
                                            if (diffMVdiffVal < 126) {
                                                if (MVDiff < 149) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 149) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 126) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 76) {
                                if (Rs < 4638) {
                                    return 1;
                                }
                                else if (Rs >= 4638) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectA(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 53) {
        if (diffTSC < 12) {
            if (MVDiff < 54) {
                if (ssDCval < 20) {
                    if (diffRsCsdiff < 5) {
                        if (RsDiff < 2) {
                            return 0;
                        }
                        else if (RsDiff >= 2) {
                            if (RsCsDiff < 9) {
                                return 0;
                            }
                            else if (RsCsDiff >= 9) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 5) {
                        if (RsDiff < 3) {
                            if (posBalance < 2) {
                                if (refDCval < 17) {
                                    return 1;
                                }
                                else if (refDCval >= 17) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 2) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 3) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 20) {
                    if (SC < 158) {
                        if (diffRsCsdiff < 5) {
                            if (posBalance < 124) {
                                if (diffMVdiffVal < -443) {
                                    if (negBalance < 87) {
                                        return 0;
                                    }
                                    else if (negBalance >= 87) {
                                        if (Rs < 17) {
                                            return 0;
                                        }
                                        else if (Rs >= 17) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -443) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 124) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (CsDiff < 30) {
                                if (CsDiff < 4) {
                                    return 0;
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                            else if (CsDiff >= 30) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 158) {
                        if (diffRsCsdiff < 260) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 260) {
                            if (RsCsDiff < 261) {
                                return 1;
                            }
                            else if (RsCsDiff >= 261) {
                                if (refDCval < 44) {
                                    if (RsDiff < 25) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 25) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 44) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 54) {
                if (TSCindex < 9) {
                    if (CsDiff < 17) {
                        if (diffRsCsdiff < 1) {
                            if (Rs < 5) {
                                if (diffAFD < -21) {
                                    if (TSCindex < 2) {
                                        return 0;
                                    }
                                    else if (TSCindex >= 2) {
                                        if (Rs < 4) {
                                            return 0;
                                        }
                                        else if (Rs >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= -21) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 5) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 1) {
                            if (ssDCval < 36) {
                                if (MVDiff < 152) {
                                    return 0;
                                }
                                else if (MVDiff >= 152) {
                                    if (SC < 329) {
                                        return 1;
                                    }
                                    else if (SC >= 329) {
                                        if (Rs < 288) {
                                            return 0;
                                        }
                                        else if (Rs >= 288) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 36) {
                                return 0;
                            }
                        }
                    }
                    else if (CsDiff >= 17) {
                        if (diffAFD < 4) {
                            return 0;
                        }
                        else if (diffAFD >= 4) {
                            if (RsDiff < 14) {
                                return 0;
                            }
                            else if (RsDiff >= 14) {
                                if (SC < 6059) {
                                    return 1;
                                }
                                else if (SC >= 6059) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (TSCindex >= 9) {
                    if (MVDiff < 155) {
                        if (RsCsDiff < 17299) {
                            if (ssDCval < 87) {
                                if (diffRsCsdiff < 1181) {
                                    if (SC < 80) {
                                        if (Cs < 44) {
                                            return 0;
                                        }
                                        else if (Cs >= 44) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 80) {
                                        if (diffMVdiffVal < 36) {
                                            if (diffAFD < 11) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 11) {
                                                if (Cs < 163) {
                                                    if (Rs < 230) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 230) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 163) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 36) {
                                            if (ssDCval < 35) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 35) {
                                                if (diffTSC < 11) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 11) {
                                                    if (Rs < 914) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 914) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 1181) {
                                    if (TSC < 20) {
                                        return 0;
                                    }
                                    else if (TSC >= 20) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 87) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 17299) {
                            return 1;
                        }
                    }
                    else if (MVDiff >= 155) {
                        if (TSC < 30) {
                            if (SC < 1290) {
                                if (ssDCval < 66) {
                                    if (posBalance < 29) {
                                        return 0;
                                    }
                                    else if (posBalance >= 29) {
                                        if (Cs < 582) {
                                            if (negBalance < 49) {
                                                return 0;
                                            }
                                            else if (negBalance >= 49) {
                                                if (SC < 227) {
                                                    if (CsDiff < 2) {
                                                        if (AFD < 23) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 23) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (CsDiff >= 2) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 227) {
                                                    if (posBalance < 67) {
                                                        if (Rs < 120) {
                                                            if (TSC < 13) {
                                                                return 0;
                                                            }
                                                            else if (TSC >= 13) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 120) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (posBalance >= 67) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 582) {
                                            if (RsCsDiff < 38) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 38) {
                                                if (CsDiff < 11) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 11) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 66) {
                                    return 0;
                                }
                            }
                            else if (SC >= 1290) {
                                if (diffAFD < 27) {
                                    if (diffMVdiffVal < -121) {
                                        if (ssDCval < 46) {
                                            if (CsDiff < 8) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 46) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -121) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 27) {
                                    return 1;
                                }
                            }
                        }
                        else if (TSC >= 30) {
                            if (Cs < 1905) {
                                if (MVDiff < 366) {
                                    if (CsDiff < 12) {
                                        if (Cs < 819) {
                                            return 1;
                                        }
                                        else if (Cs >= 819) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 12) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 366) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 1905) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 12) {
            if (RsCsDiff < 245) {
                if (Rs < 291) {
                    if (RsDiff < 5) {
                        if (diffRsCsdiff < -12) {
                            if (ssDCval < 70) {
                                return 0;
                            }
                            else if (ssDCval >= 70) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= -12) {
                            if (RsDiff < 3) {
                                return 0;
                            }
                            else if (RsDiff >= 3) {
                                return 1;
                            }
                        }
                    }
                    else if (RsDiff >= 5) {
                        return 1;
                    }
                }
                else if (Rs >= 291) {
                    if (diffMVdiffVal < -70) {
                        if (Rs < 602) {
                            return 0;
                        }
                        else if (Rs >= 602) {
                            if (negBalance < 36) {
                                if (MVDiff < 1058) {
                                    return 0;
                                }
                                else if (MVDiff >= 1058) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 36) {
                                return 1;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= -70) {
                        if (diffTSC < 43) {
                            if (ssDCval < 60) {
                                if (CsDiff < 5) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 5) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 60) {
                                if (RsDiff < 12) {
                                    return 0;
                                }
                                else if (RsDiff >= 12) {
                                    if (Rs < 1806) {
                                        if (RsCsDiff < 206) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 206) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1806) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 43) {
                            return 1;
                        }
                    }
                }
            }
            else if (RsCsDiff >= 245) {
                if (Cs < 3054) {
                    if (negBalance < 5) {
                        return 1;
                    }
                    else if (negBalance >= 5) {
                        if (RsDiff < 27) {
                            if (posBalance < 91) {
                                if (diffTSC < 22) {
                                    if (diffRsCsdiff < 564) {
                                        if (diffMVdiffVal < 46) {
                                            if (SC < 1467) {
                                                return 1;
                                            }
                                            else if (SC >= 1467) {
                                                if (posBalance < 76) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 76) {
                                                    if (Cs < 1534) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1534) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 46) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 564) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 22) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 91) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 27) {
                            if (gchDC < 61) {
                                return 1;
                            }
                            else if (gchDC >= 61) {
                                return 0;
                            }
                        }
                    }
                }
                else if (Cs >= 3054) {
                    if (gchDC < 7) {
                        return 0;
                    }
                    else if (gchDC >= 7) {
                        if (MVDiff < 45) {
                            if (Rs < 3654) {
                                return 1;
                            }
                            else if (Rs >= 3654) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 45) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 53) {
        if (diffRsCsdiff < 1) {
            if (diffMVdiffVal < 219) {
                if (Rs < 145) {
                    if (CsDiff < 5) {
                        if (TSC < 4) {
                            return 0;
                        }
                        else if (TSC >= 4) {
                            if (TSC < 22) {
                                if (ssDCval < 24) {
                                    return 1;
                                }
                                else if (ssDCval >= 24) {
                                    if (diffAFD < 9) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 9) {
                                        if (ssDCval < 38) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 38) {
                                            if (posBalance < 117) {
                                                return 0;
                                            }
                                            else if (posBalance >= 117) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 22) {
                                return 1;
                            }
                        }
                    }
                    else if (CsDiff >= 5) {
                        if (TSC < 15) {
                            if (diffTSC < 4) {
                                return 0;
                            }
                            else if (diffTSC >= 4) {
                                return 1;
                            }
                        }
                        else if (TSC >= 15) {
                            if (AFD < 52) {
                                if (Cs < 6) {
                                    return 0;
                                }
                                else if (Cs >= 6) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 52) {
                                return 0;
                            }
                        }
                    }
                }
                else if (Rs >= 145) {
                    if (RsCsDiff < 821) {
                        if (AFD < 34) {
                            return 0;
                        }
                        else if (AFD >= 34) {
                            if (AFD < 35) {
                                if (Cs < 161) {
                                    return 1;
                                }
                                else if (Cs >= 161) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 35) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 821) {
                        if (diffMVdiffVal < 93) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 93) {
                            if (SC < 4164) {
                                return 1;
                            }
                            else if (SC >= 4164) {
                                if (CsDiff < 53) {
                                    return 0;
                                }
                                else if (CsDiff >= 53) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 219) {
                if (diffRsCsdiff < -2) {
                    if (MVDiff < 1805) {
                        if (CsDiff < 5) {
                            if (posBalance < 1) {
                                if (Rs < 38) {
                                    return 0;
                                }
                                else if (Rs >= 38) {
                                    if (Cs < 324) {
                                        return 1;
                                    }
                                    else if (Cs >= 324) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 1) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 5) {
                            if (TSC < 17) {
                                if (Cs < 100) {
                                    return 1;
                                }
                                else if (Cs >= 100) {
                                    if (diffTSC < 10) {
                                        if (diffTSC < -3) {
                                            if (AFD < 31) {
                                                return 0;
                                            }
                                            else if (AFD >= 31) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= -3) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 10) {
                                        return 1;
                                    }
                                }
                            }
                            else if (TSC >= 17) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 1805) {
                        return 1;
                    }
                }
                else if (diffRsCsdiff >= -2) {
                    if (MVDiff < 328) {
                        if (Cs < 300) {
                            return 1;
                        }
                        else if (Cs >= 300) {
                            return 0;
                        }
                    }
                    else if (MVDiff >= 328) {
                        if (diffTSC < 4) {
                            if (gchDC < 83) {
                                if (diffMVdiffVal < 797) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 797) {
                                    if (posBalance < 56) {
                                        if (negBalance < 119) {
                                            return 1;
                                        }
                                        else if (negBalance >= 119) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 56) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 83) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 4) {
                            if (Rs < 132) {
                                if (TSC < 11) {
                                    if (ssDCval < 120) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 120) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 11) {
                                    if (diffTSC < 8) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 8) {
                                        if (Rs < 79) {
                                            return 1;
                                        }
                                        else if (Rs >= 79) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 132) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (diffRsCsdiff >= 1) {
            if (negBalance < 33) {
                if (diffTSC < 10) {
                    if (SC < 239) {
                        if (diffAFD < 10) {
                            return 0;
                        }
                        else if (diffAFD >= 10) {
                            if (refDCval < 103) {
                                if (diffAFD < 12) {
                                    if (Cs < 103) {
                                        return 1;
                                    }
                                    else if (Cs >= 103) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 12) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 103) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 239) {
                        if (diffRsCsdiff < 3281) {
                            if (posBalance < 10) {
                                return 1;
                            }
                            else if (posBalance >= 10) {
                                if (diffMVdiffVal < 1317) {
                                    if (ssDCval < 83) {
                                        if (Rs < 304) {
                                            if (Rs < 252) {
                                                return 0;
                                            }
                                            else if (Rs >= 252) {
                                                if (AFD < 14) {
                                                    return 0;
                                                }
                                                else if (AFD >= 14) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 304) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 83) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 1317) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 3281) {
                            return 1;
                        }
                    }
                }
                else if (diffTSC >= 10) {
                    if (diffAFD < 28) {
                        if (ssDCval < 58) {
                            if (Rs < 343) {
                                return 1;
                            }
                            else if (Rs >= 343) {
                                if (refDCval < 33) {
                                    return 1;
                                }
                                else if (refDCval >= 33) {
                                    if (TSC < 12) {
                                        return 1;
                                    }
                                    else if (TSC >= 12) {
                                        if (MVDiff < 812) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 812) {
                                            if (SC < 1202) {
                                                return 1;
                                            }
                                            else if (SC >= 1202) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 58) {
                            if (negBalance < 17) {
                                if (diffMVdiffVal < 505) {
                                    if (diffMVdiffVal < 92) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 92) {
                                        if (diffRsCsdiff < 30) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 30) {
                                            if (refDCval < 41) {
                                                return 1;
                                            }
                                            else if (refDCval >= 41) {
                                                if (TSC < 13) {
                                                    return 1;
                                                }
                                                else if (TSC >= 13) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 505) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 17) {
                                if (Cs < 199) {
                                    return 1;
                                }
                                else if (Cs >= 199) {
                                    if (Cs < 3354) {
                                        if (diffMVdiffVal < 93) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 93) {
                                            if (ssDCval < 92) {
                                                if (SC < 3234) {
                                                    return 1;
                                                }
                                                else if (SC >= 3234) {
                                                    return 0;
                                                }
                                            }
                                            else if (ssDCval >= 92) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Cs >= 3354) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 28) {
                        if (diffRsCsdiff < 124) {
                            if (TSC < 26) {
                                if (posBalance < 119) {
                                    if (diffTSC < 12) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 12) {
                                        if (Cs < 620) {
                                            return 1;
                                        }
                                        else if (Cs >= 620) {
                                            if (Cs < 789) {
                                                return 0;
                                            }
                                            else if (Cs >= 789) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 119) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 26) {
                                if (ssDCval < 107) {
                                    return 0;
                                }
                                else if (ssDCval >= 107) {
                                    if (SC < 2082) {
                                        if (diffRsCsdiff < 26) {
                                            if (SC < 665) {
                                                return 1;
                                            }
                                            else if (SC >= 665) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 26) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 2082) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 124) {
                            if (diffAFD < 40) {
                                if (gchDC < 58) {
                                    if (diffTSC < 32) {
                                        if (negBalance < 19) {
                                            return 1;
                                        }
                                        else if (negBalance >= 19) {
                                            if (AFD < 42) {
                                                if (ssDCval < 75) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 75) {
                                                    if (Cs < 863) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 863) {
                                                        if (SC < 4864) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 4864) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (AFD >= 42) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 32) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 58) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 40) {
                                if (ssDCval < 86) {
                                    if (RsDiff < 11) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 11) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 86) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 33) {
                if (posBalance < 29) {
                    if (gchDC < 27) {
                        if (diffRsCsdiff < 102) {
                            if (diffAFD < 14) {
                                if (RsDiff < 2) {
                                    if (diffAFD < 12) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 12) {
                                        if (AFD < 17) {
                                            return 1;
                                        }
                                        else if (AFD >= 17) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 2) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 14) {
                                if (Rs < 897) {
                                    return 1;
                                }
                                else if (Rs >= 897) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 102) {
                            if (diffRsCsdiff < 165) {
                                if (Scindex < 4) {
                                    return 1;
                                }
                                else if (Scindex >= 4) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 165) {
                                if (ssDCval < 42) {
                                    return 1;
                                }
                                else if (ssDCval >= 42) {
                                    if (diffMVdiffVal < 1728) {
                                        if (posBalance < 26) {
                                            return 0;
                                        }
                                        else if (posBalance >= 26) {
                                            if (SC < 3799) {
                                                return 0;
                                            }
                                            else if (SC >= 3799) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 1728) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 27) {
                        if (TSC < 25) {
                            if (diffMVdiffVal < 97) {
                                if (ssDCval < 68) {
                                    if (MVDiff < 106) {
                                        if (SC < 1205) {
                                            return 1;
                                        }
                                        else if (SC >= 1205) {
                                            if (CsDiff < 16) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 16) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 106) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 68) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 97) {
                                if (diffAFD < 17) {
                                    if (TSC < 22) {
                                        if (MVDiff < 362) {
                                            if (AFD < 12) {
                                                return 0;
                                            }
                                            else if (AFD >= 12) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 362) {
                                            if (AFD < 27) {
                                                if (AFD < 23) {
                                                    return 0;
                                                }
                                                else if (AFD >= 23) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 27) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 22) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 17) {
                                    if (Rs < 4411) {
                                        if (negBalance < 69) {
                                            if (RsDiff < 13) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 69) {
                                            if (Rs < 7) {
                                                if (Rs < 4) {
                                                    return 1;
                                                }
                                                else if (Rs >= 4) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 7) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 4411) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 25) {
                            if (posBalance < 12) {
                                if (diffMVdiffVal < 312) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 312) {
                                    if (SC < 533) {
                                        return 0;
                                    }
                                    else if (SC >= 533) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 12) {
                                if (diffTSC < 16) {
                                    if (ssDCval < 121) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 121) {
                                        if (TSC < 27) {
                                            return 1;
                                        }
                                        else if (TSC >= 27) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 16) {
                                    if (negBalance < 70) {
                                        return 0;
                                    }
                                    else if (negBalance >= 70) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (posBalance >= 29) {
                    if (diffTSC < 6) {
                        if (Scindex < 3) {
                            if (refDCval < 62) {
                                if (diffAFD < 7) {
                                    if (refDCval < 34) {
                                        return 1;
                                    }
                                    else if (refDCval >= 34) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 7) {
                                    if (ssDCval < 52) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 52) {
                                        if (RsDiff < 3) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 62) {
                                if (diffRsCsdiff < 18) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 18) {
                                    if (SC < 341) {
                                        return 1;
                                    }
                                    else if (SC >= 341) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Scindex >= 3) {
                            if (TSC < 34) {
                                if (diffAFD < 7) {
                                    if (ssDCval < 38) {
                                        if (negBalance < 74) {
                                            return 1;
                                        }
                                        else if (negBalance >= 74) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 38) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 7) {
                                    if (negBalance < 66) {
                                        if (SC < 1438) {
                                            if (RsCsDiff < 83) {
                                                if (Cs < 171) {
                                                    return 1;
                                                }
                                                else if (Cs >= 171) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsCsDiff >= 83) {
                                                if (RsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 7) {
                                                    if (Cs < 422) {
                                                        if (Rs < 950) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 950) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 422) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 1438) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 66) {
                                        if (Cs < 498) {
                                            if (posBalance < 46) {
                                                return 1;
                                            }
                                            else if (posBalance >= 46) {
                                                if (diffAFD < 15) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 15) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 498) {
                                            if (TSC < 20) {
                                                return 0;
                                            }
                                            else if (TSC >= 20) {
                                                if (RsDiff < 28) {
                                                    if (SC < 2822) {
                                                        if (Rs < 1123) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 1123) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 2822) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsDiff >= 28) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 34) {
                                return 1;
                            }
                        }
                    }
                    else if (diffTSC >= 6) {
                        if (diffAFD < 25) {
                            if (SC < 594) {
                                return 1;
                            }
                            else if (SC >= 594) {
                                if (AFD < 31) {
                                    if (Cs < 375) {
                                        if (diffMVdiffVal < 84) {
                                            if (MVDiff < 292) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 292) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 84) {
                                            if (Cs < 371) {
                                                return 1;
                                            }
                                            else if (Cs >= 371) {
                                                if (Rs < 450) {
                                                    return 1;
                                                }
                                                else if (Rs >= 450) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 375) {
                                        if (diffAFD < 13) {
                                            if (ssDCval < 70) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 70) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 31) {
                                    if (diffTSC < 9) {
                                        if (refDCval < 90) {
                                            if (TSC < 18) {
                                                return 0;
                                            }
                                            else if (TSC >= 18) {
                                                if (RsDiff < 28) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 28) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 90) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 9) {
                                        if (negBalance < 39) {
                                            if (AFD < 40) {
                                                if (Cs < 506) {
                                                    return 0;
                                                }
                                                else if (Cs >= 506) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 40) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 39) {
                                            if (diffRsCsdiff < 180) {
                                                if (ssDCval < 84) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 84) {
                                                    if (Cs < 678) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 678) {
                                                        if (ssDCval < 131) {
                                                            return 0;
                                                        }
                                                        else if (ssDCval >= 131) {
                                                            if (Rs < 813) {
                                                                return 0;
                                                            }
                                                            else if (Rs >= 813) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 180) {
                                                if (diffMVdiffVal < 92) {
                                                    if (diffAFD < 16) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= 16) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 92) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 25) {
                            if (diffMVdiffVal < 55) {
                                if (Rs < 1339) {
                                    return 1;
                                }
                                else if (Rs >= 1339) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 55) {
                                if (diffTSC < 12) {
                                    if (negBalance < 64) {
                                        if (SC < 2363) {
                                            if (RsCsDiff < 47) {
                                                if (posBalance < 64) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 64) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 2363) {
                                            if (diffMVdiffVal < 95) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= 95) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 64) {
                                        if (posBalance < 30) {
                                            return 0;
                                        }
                                        else if (posBalance >= 30) {
                                            if (AFD < 54) {
                                                return 1;
                                            }
                                            else if (AFD >= 54) {
                                                if (CsDiff < 9) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 9) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    if (diffAFD < 38) {
                                        if (gchDC < 21) {
                                            if (AFD < 41) {
                                                return 1;
                                            }
                                            else if (AFD >= 41) {
                                                if (diffMVdiffVal < 77) {
                                                    if (ssDCval < 56) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 56) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 77) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 21) {
                                            if (posBalance < 52) {
                                                return 1;
                                            }
                                            else if (posBalance >= 52) {
                                                if (ssDCval < 113) {
                                                    if (ssDCval < 86) {
                                                        if (posBalance < 76) {
                                                            if (RsDiff < 19) {
                                                                return 0;
                                                            }
                                                            else if (RsDiff >= 19) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (posBalance >= 76) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 86) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 113) {
                                                    if (SC < 1058) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 1058) {
                                                        if (TSC < 33) {
                                                            return 0;
                                                        }
                                                        else if (TSC >= 33) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 38) {
                                        if (ssDCval < 48) {
                                            if (CsDiff < 4) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 4) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 48) {
                                            if (Rs < 9291) {
                                                return 1;
                                            }
                                            else if (Rs >= 9291) {
                                                if (TSC < 32) {
                                                    return 0;
                                                }
                                                else if (TSC >= 32) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectB(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 56) {
        if (AFD < 27) {
            if (diffAFD < 9) {
                if (diffAFD < 3) {
                    if (SC < 87) {
                        if (gchDC < 4) {
                            if (diffTSC < -2) {
                                if (gchDC < 1) {
                                    return 0;
                                }
                                else if (gchDC >= 1) {
                                    if (ssDCval < 25) {
                                        if (diffTSC < -3) {
                                            return 0;
                                        }
                                        else if (diffTSC >= -3) {
                                            if (Cs < 2) {
                                                return 0;
                                            }
                                            else if (Cs >= 2) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 25) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= -2) {
                                return 0;
                            }
                        }
                        else if (gchDC >= 4) {
                            if (RsCsDiff < 35) {
                                if (refDCval < 226) {
                                    if (MVDiff < 4) {
                                        if (MVDiff < 3) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 3) {
                                            if (SC < 5) {
                                                return 1;
                                            }
                                            else if (SC >= 5) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 4) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 226) {
                                    return 1;
                                }
                            }
                            else if (RsCsDiff >= 35) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 87) {
                        if (diffRsCsdiff < 5) {
                            if (diffTSC < 6) {
                                return 0;
                            }
                            else if (diffTSC >= 6) {
                                if (Cs < 68) {
                                    return 1;
                                }
                                else if (Cs >= 68) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (AFD < 2) {
                                if (refDCval < 20) {
                                    return 1;
                                }
                                else if (refDCval >= 20) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 2) {
                                if (negBalance < 9) {
                                    if (CsDiff < 33) {
                                        if (Rs < 525) {
                                            if (Rs < 414) {
                                                return 0;
                                            }
                                            else if (Rs >= 414) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 525) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 33) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 9) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 3) {
                    if (posBalance < 9) {
                        if (ssDCval < 31) {
                            if (RsDiff < 6) {
                                return 0;
                            }
                            else if (RsDiff >= 6) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 31) {
                            return 0;
                        }
                    }
                    else if (posBalance >= 9) {
                        if (negBalance < 17) {
                            if (RsDiff < 44) {
                                if (MVDiff < 379) {
                                    return 0;
                                }
                                else if (MVDiff >= 379) {
                                    if (Rs < 117) {
                                        return 0;
                                    }
                                    else if (Rs >= 117) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsDiff >= 44) {
                                if (Cs < 2386) {
                                    return 0;
                                }
                                else if (Cs >= 2386) {
                                    return 1;
                                }
                            }
                        }
                        else if (negBalance >= 17) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffAFD >= 9) {
                if (refDCval < 43) {
                    if (SC < 162) {
                        if (CsDiff < 1) {
                            return 0;
                        }
                        else if (CsDiff >= 1) {
                            return 1;
                        }
                    }
                    else if (SC >= 162) {
                        if (CsDiff < 15) {
                            if (diffRsCsdiff < -459) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -459) {
                                if (ssDCval < 44) {
                                    if (TSCindex < 9) {
                                        return 0;
                                    }
                                    else if (TSCindex >= 9) {
                                        if (negBalance < 1) {
                                            return 1;
                                        }
                                        else if (negBalance >= 1) {
                                            if (RsCsDiff < 110) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 110) {
                                                if (RsDiff < 11) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 11) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 44) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 15) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 43) {
                    if (diffMVdiffVal < 48) {
                        if (diffTSC < 20) {
                            if (gchDC < 7) {
                                return 0;
                            }
                            else if (gchDC >= 7) {
                                if (Cs < 192) {
                                    if (Rs < 2389) {
                                        if (diffTSC < 3) {
                                            if (ssDCval < 56) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 56) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 3) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 2389) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 192) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 20) {
                            return 1;
                        }
                    }
                    else if (diffMVdiffVal >= 48) {
                        if (SC < 6361) {
                            if (diffRsCsdiff < 148) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 148) {
                                return 1;
                            }
                        }
                        else if (SC >= 6361) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (AFD >= 27) {
            if (diffRsCsdiff < 56) {
                if (AFD < 58) {
                    if (diffTSC < 20) {
                        if (diffAFD < -7) {
                            if (RsDiff < 103) {
                                if (refDCval < 56) {
                                    if (posBalance < 39) {
                                        return 0;
                                    }
                                    else if (posBalance >= 39) {
                                        if (Rs < 794) {
                                            return 1;
                                        }
                                        else if (Rs >= 794) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 56) {
                                    if (AFD < 48) {
                                        return 0;
                                    }
                                    else if (AFD >= 48) {
                                        if (gchDC < 37) {
                                            return 0;
                                        }
                                        else if (gchDC >= 37) {
                                            if (RsDiff < 11) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 11) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 103) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= -7) {
                            if (Rs < 115) {
                                if (Rs < 114) {
                                    return 0;
                                }
                                else if (Rs >= 114) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 115) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 20) {
                        if (ssDCval < 66) {
                            if (AFD < 40) {
                                return 1;
                            }
                            else if (AFD >= 40) {
                                return 0;
                            }
                        }
                        else if (ssDCval >= 66) {
                            if (CsDiff < 3) {
                                if (Cs < 224) {
                                    return 0;
                                }
                                else if (Cs >= 224) {
                                    return 1;
                                }
                            }
                            else if (CsDiff >= 3) {
                                return 0;
                            }
                        }
                    }
                }
                else if (AFD >= 58) {
                    if (CsDiff < 12) {
                        if (diffRsCsdiff < 21) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 21) {
                            if (TSC < 24) {
                                return 0;
                            }
                            else if (TSC >= 24) {
                                return 1;
                            }
                        }
                    }
                    else if (CsDiff >= 12) {
                        if (ssDCval < 105) {
                            return 1;
                        }
                        else if (ssDCval >= 105) {
                            if (Rs < 1778) {
                                return 1;
                            }
                            else if (Rs >= 1778) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 56) {
                if (refDCval < 69) {
                    if (TSC < 30) {
                        if (SC < 1428) {
                            if (RsDiff < 6) {
                                return 0;
                            }
                            else if (RsDiff >= 6) {
                                if (RsCsDiff < 158) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 158) {
                                    return 0;
                                }
                            }
                        }
                        else if (SC >= 1428) {
                            if (diffMVdiffVal < 49) {
                                if (diffRsCsdiff < 19722) {
                                    if (diffRsCsdiff < 257) {
                                        if (SC < 2249) {
                                            if (posBalance < 52) {
                                                return 0;
                                            }
                                            else if (posBalance >= 52) {
                                                if (Rs < 962) {
                                                    if (MVDiff < 508) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 508) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 962) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (SC >= 2249) {
                                            if (diffAFD < -5) {
                                                if (RsDiff < 23) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 23) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= -5) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 257) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 19722) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 49) {
                                if (Rs < 1453) {
                                    return 0;
                                }
                                else if (Rs >= 1453) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (TSC >= 30) {
                        if (diffMVdiffVal < 35) {
                            if (refDCval < 16) {
                                return 0;
                            }
                            else if (refDCval >= 16) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 35) {
                            return 0;
                        }
                    }
                }
                else if (refDCval >= 69) {
                    if (gchDC < 7) {
                        if (MVDiff < 120) {
                            return 0;
                        }
                        else if (MVDiff >= 120) {
                            if (diffTSC < 8) {
                                return 0;
                            }
                            else if (diffTSC >= 8) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 7) {
                        if (AFD < 49) {
                            if (diffTSC < 11) {
                                if (diffRsCsdiff < 924) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 924) {
                                    if (SC < 6789) {
                                        if (CsDiff < 28) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 28) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 6789) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (refDCval < 86) {
                                    if (Cs < 3622) {
                                        return 1;
                                    }
                                    else if (Cs >= 3622) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 86) {
                                    if (negBalance < 106) {
                                        return 0;
                                    }
                                    else if (negBalance >= 106) {
                                        if (Rs < 970) {
                                            return 1;
                                        }
                                        else if (Rs >= 970) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 49) {
                            if (SC < 2266) {
                                return 0;
                            }
                            else if (SC >= 2266) {
                                if (RsDiff < 30) {
                                    if (AFD < 54) {
                                        return 0;
                                    }
                                    else if (AFD >= 54) {
                                        if (RsDiff < 23) {
                                            if (diffRsCsdiff < 144) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 144) {
                                                return 1;
                                            }
                                        }
                                        else if (RsDiff >= 23) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 30) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 56) {
        if (CsDiff < 7) {
            if (diffMVdiffVal < 129) {
                if (diffAFD < 8) {
                    if (diffTSC < 1) {
                        return 0;
                    }
                    else if (diffTSC >= 1) {
                        if (diffRsCsdiff < 12) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 12) {
                            if (RsCsDiff < 15) {
                                return 1;
                            }
                            else if (RsCsDiff >= 15) {
                                if (MVDiff < 267) {
                                    return 0;
                                }
                                else if (MVDiff >= 267) {
                                    if (MVDiff < 300) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 300) {
                                        if (gchDC < 5) {
                                            return 1;
                                        }
                                        else if (gchDC >= 5) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 8) {
                    if (MVDiff < 126) {
                        if (ssDCval < 54) {
                            if (Cs < 202) {
                                return 1;
                            }
                            else if (Cs >= 202) {
                                if (posBalance < 52) {
                                    if (Rs < 203) {
                                        return 0;
                                    }
                                    else if (Rs >= 203) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 52) {
                                    if (diffRsCsdiff < 46) {
                                        if (Rs < 416) {
                                            return 0;
                                        }
                                        else if (Rs >= 416) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 46) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 54) {
                            if (diffAFD < 25) {
                                if (RsDiff < 10) {
                                    if (posBalance < 53) {
                                        if (AFD < 22) {
                                            return 0;
                                        }
                                        else if (AFD >= 22) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 53) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 10) {
                                    if (TSC < 18) {
                                        return 1;
                                    }
                                    else if (TSC >= 18) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 25) {
                                if (RsCsDiff < 41) {
                                    if (MVDiff < 80) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 80) {
                                        return 1;
                                    }
                                }
                                else if (RsCsDiff >= 41) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 126) {
                        if (gchDC < 9) {
                            if (diffAFD < 10) {
                                return 0;
                            }
                            else if (diffAFD >= 10) {
                                if (diffMVdiffVal < 96) {
                                    if (CsDiff < 5) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 96) {
                                    if (diffAFD < 30) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 30) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 9) {
                            if (diffRsCsdiff < 25) {
                                if (diffTSC < -1) {
                                    if (CsDiff < 4) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 4) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    if (posBalance < 13) {
                                        if (diffMVdiffVal < 65) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 65) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 13) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 25) {
                                if (CsDiff < 5) {
                                    if (negBalance < 29) {
                                        return 0;
                                    }
                                    else if (negBalance >= 29) {
                                        if (TSC < 30) {
                                            return 1;
                                        }
                                        else if (TSC >= 30) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (CsDiff >= 5) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 129) {
                if (diffAFD < 12) {
                    if (Scindex < 2) {
                        if (diffAFD < 7) {
                            if (diffTSC < -38) {
                                return 1;
                            }
                            else if (diffTSC >= -38) {
                                if (diffAFD < -1) {
                                    if (RsCsDiff < 6) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 6) {
                                        if (gchDC < 125) {
                                            return 0;
                                        }
                                        else if (gchDC >= 125) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= -1) {
                                    if (RsCsDiff < 47) {
                                        if (SC < 22) {
                                            if (SC < 19) {
                                                if (negBalance < 126) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 126) {
                                                    if (AFD < 8) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 8) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (SC >= 19) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 47) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 7) {
                            if (AFD < 15) {
                                if (AFD < 8) {
                                    return 0;
                                }
                                else if (AFD >= 8) {
                                    if (SC < 102) {
                                        if (Cs < 22) {
                                            return 1;
                                        }
                                        else if (Cs >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 102) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 15) {
                                if (Cs < 29) {
                                    return 0;
                                }
                                else if (Cs >= 29) {
                                    if (posBalance < 41) {
                                        if (RsCsDiff < 38) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 38) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 41) {
                                        if (TSC < 13) {
                                            return 1;
                                        }
                                        else if (TSC >= 13) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Scindex >= 2) {
                        if (ssDCval < 50) {
                            if (AFD < 13) {
                                return 0;
                            }
                            else if (AFD >= 13) {
                                if (MVDiff < 190) {
                                    return 1;
                                }
                                else if (MVDiff >= 190) {
                                    if (diffTSC < 4) {
                                        if (diffRsCsdiff < 47) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 47) {
                                            if (RsCsDiff < 89) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 89) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 4) {
                                        if (MVDiff < 570) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 570) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 50) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 12) {
                    if (negBalance < 43) {
                        if (Cs < 689) {
                            if (diffTSC < 9) {
                                if (refDCval < 84) {
                                    if (TSC < 12) {
                                        if (Rs < 540) {
                                            return 1;
                                        }
                                        else if (Rs >= 540) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 12) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 84) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 9) {
                                if (ssDCval < 47) {
                                    return 1;
                                }
                                else if (ssDCval >= 47) {
                                    if (negBalance < 21) {
                                        if (posBalance < 104) {
                                            if (SC < 1737) {
                                                return 0;
                                            }
                                            else if (SC >= 1737) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 104) {
                                            if (diffRsCsdiff < 3) {
                                                if (negBalance < 6) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 6) {
                                                    if (RsDiff < 1) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 3) {
                                                if (ssDCval < 105) {
                                                    if (MVDiff < 583) {
                                                        if (negBalance < 14) {
                                                            if (Cs < 247) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 247) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (negBalance >= 14) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 583) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 105) {
                                                    if (gchDC < 163) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 163) {
                                                        if (Rs < 1) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 1) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 21) {
                                        if (ssDCval < 100) {
                                            if (Rs < 615) {
                                                return 1;
                                            }
                                            else if (Rs >= 615) {
                                                if (negBalance < 31) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 31) {
                                                    if (MVDiff < 477) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 477) {
                                                        if (Cs < 615) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 615) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 100) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 689) {
                            if (Rs < 1362) {
                                return 0;
                            }
                            else if (Rs >= 1362) {
                                if (Cs < 825) {
                                    return 1;
                                }
                                else if (Cs >= 825) {
                                    if (RsDiff < 9) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 9) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 43) {
                        if (MVDiff < 474) {
                            if (diffMVdiffVal < 211) {
                                if (diffTSC < 9) {
                                    if (refDCval < 66) {
                                        if (SC < 1047) {
                                            return 1;
                                        }
                                        else if (SC >= 1047) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 66) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 9) {
                                    if (MVDiff < 283) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 283) {
                                        if (Rs < 468) {
                                            return 1;
                                        }
                                        else if (Rs >= 468) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 211) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 474) {
                            if (refDCval < 64) {
                                return 1;
                            }
                            else if (refDCval >= 64) {
                                if (diffMVdiffVal < 683) {
                                    if (Rs < 82) {
                                        return 0;
                                    }
                                    else if (Rs >= 82) {
                                        if (AFD < 39) {
                                            return 1;
                                        }
                                        else if (AFD >= 39) {
                                            if (ssDCval < 95) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 95) {
                                                if (ssDCval < 100) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 100) {
                                                    if (ssDCval < 113) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 113) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 683) {
                                    if (negBalance < 118) {
                                        if (CsDiff < 5) {
                                            if (diffTSC < -1) {
                                                return 0;
                                            }
                                            else if (diffTSC >= -1) {
                                                if (TSC < 24) {
                                                    if (AFD < 28) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 28) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 24) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 118) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (CsDiff >= 7) {
            if (diffAFD < 15) {
                if (SC < 2038) {
                    if (diffAFD < 7) {
                        if (refDCval < 46) {
                            if (SC < 1285) {
                                return 1;
                            }
                            else if (SC >= 1285) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 46) {
                            if (AFD < 27) {
                                return 0;
                            }
                            else if (AFD >= 27) {
                                if (AFD < 28) {
                                    return 1;
                                }
                                else if (AFD >= 28) {
                                    if (diffMVdiffVal < 168) {
                                        if (diffRsCsdiff < -82) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= -82) {
                                            if (ssDCval < 42) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 42) {
                                                if (AFD < 46) {
                                                    return 0;
                                                }
                                                else if (AFD >= 46) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 168) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (gchDC < 2) {
                            return 1;
                        }
                        else if (gchDC >= 2) {
                            if (negBalance < 44) {
                                if (diffTSC < 8) {
                                    return 0;
                                }
                                else if (diffTSC >= 8) {
                                    if (RsCsDiff < 95) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 95) {
                                        return 1;
                                    }
                                }
                            }
                            else if (negBalance >= 44) {
                                if (gchDC < 47) {
                                    if (negBalance < 72) {
                                        return 1;
                                    }
                                    else if (negBalance >= 72) {
                                        if (refDCval < 118) {
                                            if (MVDiff < 216) {
                                                if (TSC < 12) {
                                                    return 1;
                                                }
                                                else if (TSC >= 12) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 216) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 118) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 47) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (SC >= 2038) {
                    if (ssDCval < 39) {
                        if (TSC < 13) {
                            return 1;
                        }
                        else if (TSC >= 13) {
                            return 0;
                        }
                    }
                    else if (ssDCval >= 39) {
                        if (Rs < 2023) {
                            if (TSC < 22) {
                                return 0;
                            }
                            else if (TSC >= 22) {
                                if (negBalance < 77) {
                                    return 0;
                                }
                                else if (negBalance >= 77) {
                                    if (diffRsCsdiff < 118) {
                                        if (SC < 3117) {
                                            return 0;
                                        }
                                        else if (SC >= 3117) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 118) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 2023) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffAFD >= 15) {
                if (diffAFD < 30) {
                    if (ssDCval < 69) {
                        if (AFD < 34) {
                            if (AFD < 22) {
                                if (diffTSC < 14) {
                                    return 1;
                                }
                                else if (diffTSC >= 14) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 22) {
                                return 1;
                            }
                        }
                        else if (AFD >= 34) {
                            if (Rs < 606) {
                                return 1;
                            }
                            else if (Rs >= 606) {
                                if (Rs < 773) {
                                    if (CsDiff < 10) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 10) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 773) {
                                    if (TSC < 15) {
                                        return 0;
                                    }
                                    else if (TSC >= 15) {
                                        if (refDCval < 92) {
                                            return 1;
                                        }
                                        else if (refDCval >= 92) {
                                            if (SC < 4527) {
                                                return 1;
                                            }
                                            else if (SC >= 4527) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 69) {
                        if (negBalance < 24) {
                            if (diffRsCsdiff < 208) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 208) {
                                if (MVDiff < 144) {
                                    if (refDCval < 42) {
                                        if (Rs < 1058) {
                                            return 1;
                                        }
                                        else if (Rs >= 1058) {
                                            if (Rs < 2960) {
                                                return 0;
                                            }
                                            else if (Rs >= 2960) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 42) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 144) {
                                    if (RsDiff < 18) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 18) {
                                        if (SC < 6511) {
                                            return 0;
                                        }
                                        else if (SC >= 6511) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 24) {
                            if (diffTSC < 10) {
                                if (diffAFD < 21) {
                                    if (ssDCval < 73) {
                                        if (SC < 2204) {
                                            return 1;
                                        }
                                        else if (SC >= 2204) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 73) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 21) {
                                    if (diffMVdiffVal < 107) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 107) {
                                        if (refDCval < 101) {
                                            return 1;
                                        }
                                        else if (refDCval >= 101) {
                                            if (AFD < 34) {
                                                return 1;
                                            }
                                            else if (AFD >= 34) {
                                                if (diffAFD < 23) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 23) {
                                                    if (diffTSC < 8) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= 8) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 10) {
                                if (gchDC < 57) {
                                    if (posBalance < 40) {
                                        if (posBalance < 33) {
                                            return 1;
                                        }
                                        else if (posBalance >= 33) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 40) {
                                        if (CsDiff < 14) {
                                            if (negBalance < 37) {
                                                if (diffTSC < 15) {
                                                    if (ssDCval < 87) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 87) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffTSC >= 15) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 37) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 14) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 57) {
                                    if (posBalance < 2) {
                                        return 1;
                                    }
                                    else if (posBalance >= 2) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 30) {
                    if (refDCval < 46) {
                        if (gchDC < 36) {
                            return 1;
                        }
                        else if (gchDC >= 36) {
                            if (RsCsDiff < 1262) {
                                if (refDCval < 44) {
                                    if (posBalance < 123) {
                                        if (AFD < 71) {
                                            if (MVDiff < 159) {
                                                if (Cs < 1432) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1432) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 159) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 71) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 123) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 44) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 1262) {
                                return 1;
                            }
                        }
                    }
                    else if (refDCval >= 46) {
                        if (CsDiff < 13) {
                            if (diffRsCsdiff < -1) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= -1) {
                                if (posBalance < 81) {
                                    return 1;
                                }
                                else if (posBalance >= 81) {
                                    if (Rs < 785) {
                                        return 1;
                                    }
                                    else if (Rs >= 785) {
                                        if (CsDiff < 12) {
                                            if (refDCval < 74) {
                                                return 1;
                                            }
                                            else if (refDCval >= 74) {
                                                if (diffTSC < 21) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 21) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 12) {
                                            if (RsCsDiff < 401) {
                                                if (Rs < 813) {
                                                    return 0;
                                                }
                                                else if (Rs >= 813) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 401) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 13) {
                            if (gchDC < 84) {
                                if (negBalance < 43) {
                                    if (negBalance < 42) {
                                        if (AFD < 61) {
                                            return 1;
                                        }
                                        else if (AFD >= 61) {
                                            if (diffAFD < 46) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 46) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 42) {
                                        if (AFD < 53) {
                                            return 1;
                                        }
                                        else if (AFD >= 53) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (negBalance >= 43) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 84) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectC(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 58) {
        if (diffTSC < 8) {
            if (MVDiff < 140) {
                if (TSC < 24) {
                    if (posBalance < 14) {
                        if (RsCsDiff < 252) {
                            if (ssDCval < 18) {
                                if (diffRsCsdiff < 7) {
                                    if (negBalance < 1) {
                                        return 0;
                                    }
                                    else if (negBalance >= 1) {
                                        if (Cs < 646) {
                                            if (Cs < 629) {
                                                return 0;
                                            }
                                            else if (Cs >= 629) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 646) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 7) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 18) {
                                if (refDCval < 229) {
                                    return 0;
                                }
                                else if (refDCval >= 229) {
                                    if (gchDC < 5) {
                                        return 0;
                                    }
                                    else if (gchDC >= 5) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 252) {
                            if (refDCval < 74) {
                                if (RsDiff < 30) {
                                    return 1;
                                }
                                else if (RsDiff >= 30) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 74) {
                                if (RsCsDiff < 2394) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 2394) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 14) {
                        if (posBalance < 128) {
                            if (RsDiff < 45) {
                                return 0;
                            }
                            else if (RsDiff >= 45) {
                                if (TSC < 6) {
                                    return 1;
                                }
                                else if (TSC >= 6) {
                                    return 0;
                                }
                            }
                        }
                        else if (posBalance >= 128) {
                            return 1;
                        }
                    }
                }
                else if (TSC >= 24) {
                    if (refDCval < 99) {
                        if (posBalance < 86) {
                            return 1;
                        }
                        else if (posBalance >= 86) {
                            return 0;
                        }
                    }
                    else if (refDCval >= 99) {
                        return 0;
                    }
                }
            }
            else if (MVDiff >= 140) {
                if (diffRsCsdiff < -4) {
                    if (gchDC < 5) {
                        return 0;
                    }
                    else if (gchDC >= 5) {
                        if (diffTSC < -6) {
                            return 0;
                        }
                        else if (diffTSC >= -6) {
                            if (ssDCval < 24) {
                                return 1;
                            }
                            else if (ssDCval >= 24) {
                                if (refDCval < 29) {
                                    return 1;
                                }
                                else if (refDCval >= 29) {
                                    if (TSC < 14) {
                                        return 0;
                                    }
                                    else if (TSC >= 14) {
                                        if (Rs < 116) {
                                            return 1;
                                        }
                                        else if (Rs >= 116) {
                                            if (diffTSC < -5) {
                                                if (SC < 2729) {
                                                    return 1;
                                                }
                                                else if (SC >= 2729) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= -5) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= -4) {
                    if (AFD < 62) {
                        if (negBalance < 66) {
                            if (ssDCval < 62) {
                                if (Cs < 481) {
                                    if (diffAFD < -21) {
                                        if (gchDC < 1) {
                                            return 0;
                                        }
                                        else if (gchDC >= 1) {
                                            if (SC < 11) {
                                                return 1;
                                            }
                                            else if (SC >= 11) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= -21) {
                                        if (negBalance < 64) {
                                            return 0;
                                        }
                                        else if (negBalance >= 64) {
                                            if (refDCval < 33) {
                                                return 1;
                                            }
                                            else if (refDCval >= 33) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 481) {
                                    if (MVDiff < 333) {
                                        if (diffRsCsdiff < 244) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 244) {
                                            if (Cs < 963) {
                                                return 0;
                                            }
                                            else if (Cs >= 963) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 333) {
                                        if (SC < 1184) {
                                            return 1;
                                        }
                                        else if (SC >= 1184) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 62) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 66) {
                            if (ssDCval < 53) {
                                if (RsCsDiff < 2) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 2) {
                                    if (diffTSC < 2) {
                                        if (negBalance < 71) {
                                            if (gchDC < 6) {
                                                return 0;
                                            }
                                            else if (gchDC >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 71) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 2) {
                                        if (diffRsCsdiff < 1) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 1) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 53) {
                                if (ssDCval < 78) {
                                    if (negBalance < 119) {
                                        if (ssDCval < 74) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 74) {
                                            if (RsDiff < 58) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 58) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 119) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 78) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (AFD >= 62) {
                        if (AFD < 67) {
                            if (Rs < 1778) {
                                return 1;
                            }
                            else if (Rs >= 1778) {
                                if (TSC < 34) {
                                    return 0;
                                }
                                else if (TSC >= 34) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 67) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 8) {
            if (diffTSC < 13) {
                if (refDCval < 79) {
                    if (diffAFD < 15) {
                        if (diffMVdiffVal < 52) {
                            if (CsDiff < 37) {
                                if (RsDiff < 23) {
                                    if (refDCval < 32) {
                                        if (RsCsDiff < 68) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 68) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 32) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 23) {
                                    if (Cs < 1718) {
                                        return 1;
                                    }
                                    else if (Cs >= 1718) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 37) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 52) {
                            if (Cs < 1528) {
                                return 1;
                            }
                            else if (Cs >= 1528) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 15) {
                        if (gchDC < 2) {
                            return 0;
                        }
                        else if (gchDC >= 2) {
                            if (Rs < 574) {
                                if (diffTSC < 10) {
                                    if (RsDiff < 4) {
                                        if (SC < 813) {
                                            return 1;
                                        }
                                        else if (SC >= 813) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 4) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 10) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 574) {
                                if (AFD < 28) {
                                    if (RsCsDiff < 259) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 259) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 28) {
                                    if (Rs < 2553) {
                                        return 1;
                                    }
                                    else if (Rs >= 2553) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (refDCval >= 79) {
                    if (diffMVdiffVal < -28) {
                        if (TSC < 20) {
                            if (RsCsDiff < 25) {
                                return 0;
                            }
                            else if (RsCsDiff >= 25) {
                                if (ssDCval < 111) {
                                    return 1;
                                }
                                else if (ssDCval >= 111) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 20) {
                            return 0;
                        }
                    }
                    else if (diffMVdiffVal >= -28) {
                        if (SC < 1999) {
                            if (SC < 1909) {
                                return 0;
                            }
                            else if (SC >= 1909) {
                                return 1;
                            }
                        }
                        else if (SC >= 1999) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffTSC >= 13) {
                if (diffTSC < 28) {
                    if (RsCsDiff < 583) {
                        if (negBalance < 30) {
                            if (diffAFD < 5) {
                                if (Rs < 1156) {
                                    return 1;
                                }
                                else if (Rs >= 1156) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 5) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 30) {
                            if (refDCval < 103) {
                                if (ssDCval < 54) {
                                    return 0;
                                }
                                else if (ssDCval >= 54) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        if (RsCsDiff < 485) {
                                            if (MVDiff < 47) {
                                                if (diffMVdiffVal < 24) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 24) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 485) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 103) {
                                if (ssDCval < 55) {
                                    return 1;
                                }
                                else if (ssDCval >= 55) {
                                    if (AFD < 58) {
                                        return 0;
                                    }
                                    else if (AFD >= 58) {
                                        if (RsDiff < 13) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 583) {
                        if (CsDiff < 25) {
                            if (diffRsCsdiff < 344) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 344) {
                                if (ssDCval < 105) {
                                    return 1;
                                }
                                else if (ssDCval >= 105) {
                                    if (ssDCval < 134) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 134) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 25) {
                            if (gchDC < 11) {
                                return 0;
                            }
                            else if (gchDC >= 11) {
                                if (gchDC < 53) {
                                    return 1;
                                }
                                else if (gchDC >= 53) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 28) {
                    if (MVDiff < 36) {
                        return 1;
                    }
                    else if (MVDiff >= 36) {
                        if (TSC < 57) {
                            if (diffMVdiffVal < -97) {
                                if (Cs < 683) {
                                    return 0;
                                }
                                else if (Cs >= 683) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= -97) {
                                if (RsDiff < 186) {
                                    return 0;
                                }
                                else if (RsDiff >= 186) {
                                    return 1;
                                }
                            }
                        }
                        else if (TSC >= 57) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 58) {
        if (CsDiff < 8) {
            if (TSCindex < 8) {
                if (ssDCval < 41) {
                    if (diffTSC < 2) {
                        if (gchDC < 20) {
                            if (TSCindex < 5) {
                                return 0;
                            }
                            else if (TSCindex >= 5) {
                                if (negBalance < 73) {
                                    if (gchDC < 16) {
                                        if (SC < 29) {
                                            if (SC < 11) {
                                                return 0;
                                            }
                                            else if (SC >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 29) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 16) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 73) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 20) {
                            return 1;
                        }
                    }
                    else if (diffTSC >= 2) {
                        if (diffAFD < 8) {
                            if (Cs < 36) {
                                return 0;
                            }
                            else if (Cs >= 36) {
                                if (CsDiff < 2) {
                                    return 1;
                                }
                                else if (CsDiff >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 8) {
                            if (Cs < 162) {
                                if (TSC < 5) {
                                    if (RsDiff < 1) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 1) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 5) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 162) {
                                if (gchDC < 2) {
                                    return 0;
                                }
                                else if (gchDC >= 2) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 41) {
                    if (refDCval < 30) {
                        return 1;
                    }
                    else if (refDCval >= 30) {
                        if (MVDiff < 3438) {
                            return 0;
                        }
                        else if (MVDiff >= 3438) {
                            return 1;
                        }
                    }
                }
            }
            else if (TSCindex >= 8) {
                if (diffAFD < 10) {
                    if (ssDCval < 40) {
                        if (ssDCval < 24) {
                            return 1;
                        }
                        else if (ssDCval >= 24) {
                            if (RsDiff < 7) {
                                if (diffAFD < 8) {
                                    if (MVDiff < 1169) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 1169) {
                                        if (CsDiff < 1) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 1) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 8) {
                                    return 1;
                                }
                            }
                            else if (RsDiff >= 7) {
                                if (ssDCval < 32) {
                                    return 0;
                                }
                                else if (ssDCval >= 32) {
                                    if (SC < 1554) {
                                        return 1;
                                    }
                                    else if (SC >= 1554) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 40) {
                        if (negBalance < 1) {
                            if (ssDCval < 228) {
                                return 0;
                            }
                            else if (ssDCval >= 228) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 1) {
                            if (diffTSC < 5) {
                                if (posBalance < 95) {
                                    return 0;
                                }
                                else if (posBalance >= 95) {
                                    if (MVDiff < 171) {
                                        if (Rs < 334) {
                                            return 1;
                                        }
                                        else if (Rs >= 334) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 171) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 5) {
                                if (gchDC < 7) {
                                    if (refDCval < 76) {
                                        return 1;
                                    }
                                    else if (refDCval >= 76) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 7) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 10) {
                    if (negBalance < 34) {
                        if (diffAFD < 23) {
                            if (AFD < 23) {
                                if (SC < 819) {
                                    if (Cs < 83) {
                                        return 1;
                                    }
                                    else if (Cs >= 83) {
                                        if (gchDC < 27) {
                                            if (diffMVdiffVal < 106) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 106) {
                                                if (AFD < 20) {
                                                    if (CsDiff < 3) {
                                                        if (TSC < 8) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 8) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 20) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 27) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 819) {
                                    if (Cs < 545) {
                                        if (ssDCval < 43) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 43) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 545) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 23) {
                                if (gchDC < 28) {
                                    if (diffMVdiffVal < 831) {
                                        if (diffTSC < 13) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 13) {
                                            if (ssDCval < 73) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 73) {
                                                if (negBalance < 22) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 22) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 831) {
                                        if (gchDC < 20) {
                                            return 1;
                                        }
                                        else if (gchDC >= 20) {
                                            if (TSC < 20) {
                                                return 1;
                                            }
                                            else if (TSC >= 20) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 28) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 23) {
                            if (CsDiff < 4) {
                                if (TSC < 32) {
                                    if (diffAFD < 33) {
                                        if (gchDC < 21) {
                                            return 1;
                                        }
                                        else if (gchDC >= 21) {
                                            if (diffRsCsdiff < 16) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 16) {
                                                if (Rs < 454) {
                                                    if (refDCval < 111) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 111) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Rs >= 454) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 33) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 32) {
                                    if (MVDiff < 959) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 959) {
                                        if (RsCsDiff < 11) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 11) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 4) {
                                if (RsDiff < 14) {
                                    if (SC < 1015) {
                                        return 1;
                                    }
                                    else if (SC >= 1015) {
                                        if (RsDiff < 5) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 5) {
                                            if (Cs < 1017) {
                                                if (diffMVdiffVal < 87) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 87) {
                                                    if (Rs < 1495) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1495) {
                                                        if (AFD < 59) {
                                                            if (MVDiff < 174) {
                                                                return 1;
                                                            }
                                                            else if (MVDiff >= 174) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (AFD >= 59) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1017) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 14) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 34) {
                        if (diffAFD < 24) {
                            if (ssDCval < 84) {
                                if (refDCval < 49) {
                                    if (CsDiff < 5) {
                                        if (AFD < 11) {
                                            return 0;
                                        }
                                        else if (AFD >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 5) {
                                        if (Rs < 413) {
                                            if (ssDCval < 47) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 413) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 49) {
                                    if (diffAFD < 15) {
                                        if (Rs < 29) {
                                            return 1;
                                        }
                                        else if (Rs >= 29) {
                                            if (Rs < 347) {
                                                if (RsCsDiff < 9) {
                                                    if (RsDiff < 1) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        if (Cs < 46) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 46) {
                                                            if (Rs < 113) {
                                                                return 0;
                                                            }
                                                            else if (Rs >= 113) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (RsCsDiff >= 9) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 347) {
                                                if (diffAFD < 11) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 11) {
                                                    if (refDCval < 60) {
                                                        return 0;
                                                    }
                                                    else if (refDCval >= 60) {
                                                        if (Rs < 1363) {
                                                            if (diffTSC < 7) {
                                                                if (RsCsDiff < 17) {
                                                                    return 0;
                                                                }
                                                                else if (RsCsDiff >= 17) {
                                                                    if (TSC < 14) {
                                                                        return 1;
                                                                    }
                                                                    else if (TSC >= 14) {
                                                                        return 0;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffTSC >= 7) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Rs >= 1363) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 15) {
                                        if (diffRsCsdiff < 3) {
                                            if (diffMVdiffVal < 433) {
                                                if (RsCsDiff < 15) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 15) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 433) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 3) {
                                            if (refDCval < 66) {
                                                return 1;
                                            }
                                            else if (refDCval >= 66) {
                                                if (diffTSC < 5) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 5) {
                                                    if (Cs < 26) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 26) {
                                                        if (MVDiff < 592) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 592) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 84) {
                                if (diffTSC < 9) {
                                    return 0;
                                }
                                else if (diffTSC >= 9) {
                                    if (TSC < 19) {
                                        if (diffAFD < 12) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 19) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 24) {
                            if (negBalance < 49) {
                                if (diffMVdiffVal < 90) {
                                    if (Cs < 328) {
                                        return 0;
                                    }
                                    else if (Cs >= 328) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 90) {
                                    if (negBalance < 45) {
                                        return 1;
                                    }
                                    else if (negBalance >= 45) {
                                        if (diffRsCsdiff < -10) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= -10) {
                                            if (SC < 2436) {
                                                return 1;
                                            }
                                            else if (SC >= 2436) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 49) {
                                if (TSC < 21) {
                                    return 1;
                                }
                                else if (TSC >= 21) {
                                    if (RsCsDiff < 22) {
                                        if (ssDCval < 87) {
                                            if (ssDCval < 44) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 44) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 87) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 22) {
                                        if (ssDCval < 62) {
                                            if (Rs < 1147) {
                                                return 1;
                                            }
                                            else if (Rs >= 1147) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 62) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (CsDiff >= 8) {
            if (diffRsCsdiff < 74) {
                if (diffMVdiffVal < 100) {
                    return 0;
                }
                else if (diffMVdiffVal >= 100) {
                    if (SC < 1080) {
                        if (negBalance < 117) {
                            if (Cs < 723) {
                                if (MVDiff < 300) {
                                    if (Rs < 409) {
                                        if (CsDiff < 15) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 15) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 409) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 300) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 723) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 117) {
                            return 1;
                        }
                    }
                    else if (SC >= 1080) {
                        if (diffTSC < 4) {
                            return 0;
                        }
                        else if (diffTSC >= 4) {
                            if (CsDiff < 19) {
                                return 0;
                            }
                            else if (CsDiff >= 19) {
                                if (ssDCval < 116) {
                                    return 1;
                                }
                                else if (ssDCval >= 116) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 74) {
                if (diffAFD < 22) {
                    if (SC < 1997) {
                        if (posBalance < 89) {
                            if (ssDCval < 72) {
                                if (CsDiff < 10) {
                                    if (refDCval < 41) {
                                        if (SC < 1644) {
                                            return 0;
                                        }
                                        else if (SC >= 1644) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 41) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 10) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 72) {
                                if (RsDiff < 19) {
                                    return 0;
                                }
                                else if (RsDiff >= 19) {
                                    if (Cs < 273) {
                                        return 0;
                                    }
                                    else if (Cs >= 273) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 89) {
                            if (TSC < 18) {
                                if (SC < 771) {
                                    return 1;
                                }
                                else if (SC >= 771) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 18) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 1997) {
                        if (ssDCval < 36) {
                            return 1;
                        }
                        else if (ssDCval >= 36) {
                            if (diffAFD < 13) {
                                if (gchDC < 27) {
                                    return 0;
                                }
                                else if (gchDC >= 27) {
                                    if (RsCsDiff < 817) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 817) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffAFD >= 13) {
                                if (gchDC < 36) {
                                    if (ssDCval < 107) {
                                        if (diffTSC < 7) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 7) {
                                            if (gchDC < 11) {
                                                return 1;
                                            }
                                            else if (gchDC >= 11) {
                                                if (RsDiff < 30) {
                                                    if (Rs < 1053) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1053) {
                                                        if (ssDCval < 46) {
                                                            return 1;
                                                        }
                                                        else if (ssDCval >= 46) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 30) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 107) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 36) {
                                    if (Rs < 2215) {
                                        return 1;
                                    }
                                    else if (Rs >= 2215) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 22) {
                    if (negBalance < 27) {
                        if (Cs < 1211) {
                            if (TSC < 12) {
                                return 0;
                            }
                            else if (TSC >= 12) {
                                if (negBalance < 5) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 5) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 1211) {
                            if (CsDiff < 19) {
                                if (posBalance < 105) {
                                    return 0;
                                }
                                else if (posBalance >= 105) {
                                    if (RsDiff < 9) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 9) {
                                        if (negBalance < 5) {
                                            if (gchDC < 52) {
                                                return 1;
                                            }
                                            else if (gchDC >= 52) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 5) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 19) {
                                if (negBalance < 26) {
                                    return 1;
                                }
                                else if (negBalance >= 26) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 27) {
                        if (Cs < 790) {
                            return 1;
                        }
                        else if (Cs >= 790) {
                            if (diffTSC < 16) {
                                if (gchDC < 13) {
                                    if (diffMVdiffVal < 66) {
                                        if (Rs < 1116) {
                                            return 0;
                                        }
                                        else if (Rs >= 1116) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 66) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 13) {
                                    if (Rs < 775) {
                                        if (RsDiff < 4) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 775) {
                                        if (SC < 3280) {
                                            return 1;
                                        }
                                        else if (SC >= 3280) {
                                            if (CsDiff < 17) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 17) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 16) {
                                if (TSC < 40) {
                                    if (ssDCval < 46) {
                                        if (refDCval < 127) {
                                            return 1;
                                        }
                                        else if (refDCval >= 127) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 46) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 40) {
                                    if (diffAFD < 39) {
                                        if (Cs < 5764) {
                                            return 0;
                                        }
                                        else if (Cs >= 5764) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 39) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectD(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 60) {
        if (gchDC < 6) {
            if (diffRsCsdiff < 314) {
                if (MVDiff < 120) {
                    if (negBalance < 2) {
                        if (refDCval < 18) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (CsDiff < 4) {
                                    if (Rs < 653) {
                                        return 1;
                                    }
                                    else if (Rs >= 653) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 18) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 2) {
                        if (diffAFD < 17) {
                            return 0;
                        }
                        else if (diffAFD >= 17) {
                            if (AFD < 19) {
                                if (TSC < 13) {
                                    return 0;
                                }
                                else if (TSC >= 13) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 19) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 120) {
                    if (diffTSC < 6) {
                        if (gchDC < 4) {
                            if (diffRsCsdiff < 17) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 17) {
                                if (RsCsDiff < 25) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 25) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 4) {
                            if (Rs < 61) {
                                if (diffRsCsdiff < 2) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 2) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 61) {
                                if (diffMVdiffVal < 54) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 54) {
                                    if (AFD < 27) {
                                        return 0;
                                    }
                                    else if (AFD >= 27) {
                                        if (RsCsDiff < 123) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 123) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 6) {
                        if (diffAFD < 13) {
                            if (diffAFD < 2) {
                                if (RsDiff < 6) {
                                    return 1;
                                }
                                else if (RsDiff >= 6) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 2) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 13) {
                            if (MVDiff < 186) {
                                return 1;
                            }
                            else if (MVDiff >= 186) {
                                if (TSC < 13) {
                                    return 1;
                                }
                                else if (TSC >= 13) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 314) {
                if (ssDCval < 37) {
                    return 1;
                }
                else if (ssDCval >= 37) {
                    if (CsDiff < 41) {
                        if (diffTSC < 14) {
                            if (SC < 3349) {
                                if (RsDiff < 20) {
                                    return 0;
                                }
                                else if (RsDiff >= 20) {
                                    return 1;
                                }
                            }
                            else if (SC >= 3349) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 14) {
                            if (Cs < 2020) {
                                return 1;
                            }
                            else if (Cs >= 2020) {
                                return 0;
                            }
                        }
                    }
                    else if (CsDiff >= 41) {
                        if (TSC < 6) {
                            return 1;
                        }
                        else if (TSC >= 6) {
                            if (SC < 24936) {
                                return 0;
                            }
                            else if (SC >= 24936) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (gchDC >= 6) {
            if (diffRsCsdiff < 199) {
                if (diffAFD < 14) {
                    if (MVDiff < 231) {
                        if (posBalance < 1) {
                            if (ssDCval < 217) {
                                return 0;
                            }
                            else if (ssDCval >= 217) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 1) {
                            if (RsCsDiff < 15690) {
                                return 0;
                            }
                            else if (RsCsDiff >= 15690) {
                                if (Rs < 1541) {
                                    return 0;
                                }
                                else if (Rs >= 1541) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 231) {
                        if (TSC < 27) {
                            if (AFD < 6) {
                                if (AFD < 2) {
                                    return 0;
                                }
                                else if (AFD >= 2) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 6) {
                                if (diffRsCsdiff < 12) {
                                    if (Cs < 56) {
                                        if (AFD < 32) {
                                            return 0;
                                        }
                                        else if (AFD >= 32) {
                                            if (Cs < 18) {
                                                return 0;
                                            }
                                            else if (Cs >= 18) {
                                                if (ssDCval < 104) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 104) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 56) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 12) {
                                    if (posBalance < 21) {
                                        if (Scindex < 4) {
                                            return 0;
                                        }
                                        else if (Scindex >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 21) {
                                        if (ssDCval < 40) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 40) {
                                            if (TSC < 13) {
                                                if (diffRsCsdiff < 26) {
                                                    if (Rs < 413) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 413) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 26) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 27) {
                            if (MVDiff < 232) {
                                return 1;
                            }
                            else if (MVDiff >= 232) {
                                if (diffMVdiffVal < -215) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= -215) {
                                    if (AFD < 62) {
                                        if (RsCsDiff < 12) {
                                            if (gchDC < 51) {
                                                return 0;
                                            }
                                            else if (gchDC >= 51) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 12) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 62) {
                                        if (posBalance < 63) {
                                            return 1;
                                        }
                                        else if (posBalance >= 63) {
                                            if (Rs < 1778) {
                                                return 1;
                                            }
                                            else if (Rs >= 1778) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 14) {
                    if (posBalance < 95) {
                        if (Rs < 1526) {
                            if (SC < 133) {
                                return 1;
                            }
                            else if (SC >= 133) {
                                if (RsDiff < 12) {
                                    if (refDCval < 228) {
                                        if (diffMVdiffVal < 56) {
                                            if (refDCval < 73) {
                                                if (RsCsDiff < 16) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 16) {
                                                    if (gchDC < 7) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 7) {
                                                        if (diffTSC < 16) {
                                                            return 0;
                                                        }
                                                        else if (diffTSC >= 16) {
                                                            if (SC < 918) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 918) {
                                                                if (AFD < 56) {
                                                                    return 0;
                                                                }
                                                                else if (AFD >= 56) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 73) {
                                                if (CsDiff < 2) {
                                                    if (ssDCval < 100) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 100) {
                                                        return 0;
                                                    }
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 56) {
                                            if (Cs < 179) {
                                                return 0;
                                            }
                                            else if (Cs >= 179) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 228) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 12) {
                                    if (diffAFD < 22) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 22) {
                                        if (ssDCval < 99) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 99) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 1526) {
                            return 0;
                        }
                    }
                    else if (posBalance >= 95) {
                        if (MVDiff < 1179) {
                            return 0;
                        }
                        else if (MVDiff >= 1179) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 199) {
                if (diffTSC < 12) {
                    if (TSC < 25) {
                        if (Cs < 77) {
                            return 1;
                        }
                        else if (Cs >= 77) {
                            if (ssDCval < 44) {
                                if (TSC < 10) {
                                    return 1;
                                }
                                else if (TSC >= 10) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 44) {
                                if (SC < 2258) {
                                    if (MVDiff < 207) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 207) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 2258) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (TSC >= 25) {
                        if (diffTSC < -2) {
                            return 1;
                        }
                        else if (diffTSC >= -2) {
                            if (AFD < 36) {
                                return 1;
                            }
                            else if (AFD >= 36) {
                                if (RsCsDiff < 443) {
                                    if (AFD < 48) {
                                        return 0;
                                    }
                                    else if (AFD >= 48) {
                                        return 1;
                                    }
                                }
                                else if (RsCsDiff >= 443) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 12) {
                    if (diffTSC < 43) {
                        if (Cs < 576) {
                            if (ssDCval < 61) {
                                if (AFD < 13) {
                                    return 0;
                                }
                                else if (AFD >= 13) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 61) {
                                return 1;
                            }
                        }
                        else if (Cs >= 576) {
                            if (RsCsDiff < 5301) {
                                if (MVDiff < 262) {
                                    if (ssDCval < 104) {
                                        if (MVDiff < 120) {
                                            if (diffTSC < 13) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 120) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 104) {
                                        if (refDCval < 185) {
                                            if (CsDiff < 10) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 10) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 185) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 262) {
                                    if (diffMVdiffVal < 35) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= 35) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 5301) {
                                return 1;
                            }
                        }
                    }
                    else if (diffTSC >= 43) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 60) {
        if (TSC < 16) {
            if (ssDCval < 87) {
                if (TSCindex < 6) {
                    if (diffMVdiffVal < 126) {
                        if (negBalance < 60) {
                            return 0;
                        }
                        else if (negBalance >= 60) {
                            if (diffMVdiffVal < 99) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 99) {
                                if (refDCval < 29) {
                                    return 1;
                                }
                                else if (refDCval >= 29) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 126) {
                        if (diffTSC < 1) {
                            if (RsDiff < 9) {
                                if (MVDiff < 1621) {
                                    return 0;
                                }
                                else if (MVDiff >= 1621) {
                                    return 1;
                                }
                            }
                            else if (RsDiff >= 9) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 1) {
                            if (Cs < 131) {
                                if (RsDiff < 1) {
                                    if (gchDC < 1) {
                                        return 0;
                                    }
                                    else if (gchDC >= 1) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 1) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 131) {
                                return 0;
                            }
                        }
                    }
                }
                else if (TSCindex >= 6) {
                    if (diffRsCsdiff < 1) {
                        if (ssDCval < 40) {
                            if (diffAFD < 6) {
                                if (diffTSC < -32) {
                                    return 1;
                                }
                                else if (diffTSC >= -32) {
                                    if (Rs < 379) {
                                        return 0;
                                    }
                                    else if (Rs >= 379) {
                                        if (negBalance < 80) {
                                            return 1;
                                        }
                                        else if (negBalance >= 80) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 6) {
                                if (RsCsDiff < 1) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 1) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (MVDiff < 241) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 241) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 40) {
                            if (diffTSC < 5) {
                                if (Rs < 85) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (gchDC < 14) {
                                            return 1;
                                        }
                                        else if (gchDC >= 14) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Rs >= 85) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 5) {
                                if (Cs < 172) {
                                    if (diffMVdiffVal < 128) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 128) {
                                        if (AFD < 25) {
                                            if (RsDiff < 1) {
                                                if (SC < 154) {
                                                    return 1;
                                                }
                                                else if (SC >= 154) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 1) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 25) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 172) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 1) {
                        if (diffTSC < 4) {
                            if (Cs < 128) {
                                if (ssDCval < 37) {
                                    return 1;
                                }
                                else if (ssDCval >= 37) {
                                    if (Rs < 135) {
                                        if (diffMVdiffVal < 209) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 209) {
                                            if (Cs < 118) {
                                                return 1;
                                            }
                                            else if (Cs >= 118) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 135) {
                                        if (TSC < 7) {
                                            if (SC < 440) {
                                                return 0;
                                            }
                                            else if (SC >= 440) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 7) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 128) {
                                if (refDCval < 39) {
                                    if (Rs < 1160) {
                                        if (ssDCval < 29) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 29) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 1160) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 39) {
                                    if (diffAFD < 11) {
                                        if (refDCval < 69) {
                                            return 0;
                                        }
                                        else if (refDCval >= 69) {
                                            if (AFD < 21) {
                                                if (Rs < 262) {
                                                    if (Cs < 191) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 191) {
                                                        if (ssDCval < 72) {
                                                            return 0;
                                                        }
                                                        else if (ssDCval >= 72) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 262) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 21) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 11) {
                                        if (Cs < 245) {
                                            return 1;
                                        }
                                        else if (Cs >= 245) {
                                            if (diffRsCsdiff < 269) {
                                                if (ssDCval < 40) {
                                                    if (TSC < 10) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 10) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 40) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 269) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 4) {
                            if (negBalance < 25) {
                                if (refDCval < 28) {
                                    return 1;
                                }
                                else if (refDCval >= 28) {
                                    if (Rs < 128) {
                                        if (diffRsCsdiff < 4) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 128) {
                                        if (CsDiff < 13) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 25) {
                                if (SC < 354) {
                                    if (posBalance < 11) {
                                        if (gchDC < 22) {
                                            return 0;
                                        }
                                        else if (gchDC >= 22) {
                                            if (AFD < 28) {
                                                return 1;
                                            }
                                            else if (AFD >= 28) {
                                                if (CsDiff < 6) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 11) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 354) {
                                    if (diffRsCsdiff < 97) {
                                        if (RsCsDiff < 103) {
                                            if (gchDC < 11) {
                                                if (diffTSC < 8) {
                                                    if (MVDiff < 95) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 95) {
                                                        if (refDCval < 57) {
                                                            if (TSC < 8) {
                                                                if (negBalance < 55) {
                                                                    return 1;
                                                                }
                                                                else if (negBalance >= 55) {
                                                                    if (RsDiff < 2) {
                                                                        return 1;
                                                                    }
                                                                    else if (RsDiff >= 2) {
                                                                        return 0;
                                                                    }
                                                                }
                                                            }
                                                            else if (TSC >= 8) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (refDCval >= 57) {
                                                            if (SC < 1058) {
                                                                return 0;
                                                            }
                                                            else if (SC >= 1058) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (diffTSC >= 8) {
                                                    if (diffMVdiffVal < 64) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 64) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 11) {
                                                if (diffAFD < 14) {
                                                    if (diffRsCsdiff < 24) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= 24) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= 14) {
                                                    if (negBalance < 28) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 28) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 103) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 97) {
                                        if (negBalance < 74) {
                                            return 1;
                                        }
                                        else if (negBalance >= 74) {
                                            if (diffAFD < 12) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 12) {
                                                if (SC < 1789) {
                                                    return 1;
                                                }
                                                else if (SC >= 1789) {
                                                    if (CsDiff < 13) {
                                                        if (Cs < 1056) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1056) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (CsDiff >= 13) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 87) {
                if (refDCval < 81) {
                    if (Cs < 390) {
                        if (RsDiff < 3) {
                            if (diffTSC < 9) {
                                return 0;
                            }
                            else if (diffTSC >= 9) {
                                if (refDCval < 45) {
                                    return 0;
                                }
                                else if (refDCval >= 45) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 3) {
                            return 1;
                        }
                    }
                    else if (Cs >= 390) {
                        return 0;
                    }
                }
                else if (refDCval >= 81) {
                    if (diffAFD < 21) {
                        return 0;
                    }
                    else if (diffAFD >= 21) {
                        if (diffTSC < 12) {
                            if (AFD < 36) {
                                if (gchDC < 6) {
                                    return 0;
                                }
                                else if (gchDC >= 6) {
                                    if (RsDiff < 12) {
                                        if (Cs < 138) {
                                            return 0;
                                        }
                                        else if (Cs >= 138) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 12) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 36) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 12) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (TSC >= 16) {
            if (diffAFD < 25) {
                if (negBalance < 41) {
                    if (diffAFD < 19) {
                        if (diffTSC < 13) {
                            if (Cs < 6) {
                                return 1;
                            }
                            else if (Cs >= 6) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 13) {
                            if (negBalance < 17) {
                                if (ssDCval < 50) {
                                    return 1;
                                }
                                else if (ssDCval >= 50) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 17) {
                                return 1;
                            }
                        }
                    }
                    else if (diffAFD >= 19) {
                        if (ssDCval < 65) {
                            if (refDCval < 38) {
                                if (ssDCval < 56) {
                                    return 1;
                                }
                                else if (ssDCval >= 56) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 38) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 65) {
                            if (SC < 1962) {
                                return 0;
                            }
                            else if (SC >= 1962) {
                                if (Cs < 1223) {
                                    if (diffTSC < 13) {
                                        if (Cs < 541) {
                                            return 0;
                                        }
                                        else if (Cs >= 541) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 13) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 1223) {
                                    if (ssDCval < 127) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 127) {
                                        if (ssDCval < 173) {
                                            if (diffAFD < 23) {
                                                return 1;
                                            }
                                            else if (diffAFD >= 23) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 173) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 41) {
                    if (refDCval < 81) {
                        if (RsCsDiff < 107) {
                            if (posBalance < 65) {
                                if (TSC < 17) {
                                    return 1;
                                }
                                else if (TSC >= 17) {
                                    if (gchDC < 7) {
                                        return 1;
                                    }
                                    else if (gchDC >= 7) {
                                        if (Cs < 712) {
                                            return 0;
                                        }
                                        else if (Cs >= 712) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 65) {
                                if (Rs < 385) {
                                    if (AFD < 39) {
                                        return 1;
                                    }
                                    else if (AFD >= 39) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 385) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsCsDiff >= 107) {
                            if (posBalance < 8) {
                                return 0;
                            }
                            else if (posBalance >= 8) {
                                if (Cs < 120) {
                                    if (SC < 656) {
                                        return 1;
                                    }
                                    else if (SC >= 656) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 120) {
                                    if (CsDiff < 7) {
                                        if (Cs < 330) {
                                            return 1;
                                        }
                                        else if (Cs >= 330) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        if (Cs < 1345) {
                                            return 1;
                                        }
                                        else if (Cs >= 1345) {
                                            if (TSC < 17) {
                                                return 0;
                                            }
                                            else if (TSC >= 17) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 81) {
                        if (gchDC < 86) {
                            if (CsDiff < 4) {
                                return 0;
                            }
                            else if (CsDiff >= 4) {
                                if (diffRsCsdiff < 206) {
                                    if (ssDCval < 97) {
                                        if (diffAFD < 18) {
                                            if (RsDiff < 3) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 3) {
                                                if (TSC < 34) {
                                                    if (diffTSC < 6) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 6) {
                                                        if (Cs < 429) {
                                                            if (Rs < 48) {
                                                                return 0;
                                                            }
                                                            else if (Rs >= 48) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (Cs >= 429) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 34) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 18) {
                                            if (refDCval < 151) {
                                                return 1;
                                            }
                                            else if (refDCval >= 151) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 97) {
                                        if (diffAFD < -4) {
                                            if (ssDCval < 119) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 119) {
                                                return 1;
                                            }
                                        }
                                        else if (diffAFD >= -4) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 206) {
                                    if (diffAFD < 17) {
                                        if (TSC < 21) {
                                            return 0;
                                        }
                                        else if (TSC >= 21) {
                                            if (Rs < 2277) {
                                                return 1;
                                            }
                                            else if (Rs >= 2277) {
                                                if (RsCsDiff < 7652) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 7652) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 17) {
                                        if (RsCsDiff < 2212) {
                                            if (ssDCval < 88) {
                                                if (ssDCval < 59) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 59) {
                                                    if (gchDC < 24) {
                                                        return 1;
                                                    }
                                                    else if (gchDC >= 24) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 88) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 2212) {
                                            if (RsCsDiff < 3961) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 3961) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 86) {
                            if (diffMVdiffVal < 129) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 129) {
                                if (diffRsCsdiff < 2) {
                                    if (SC < 454) {
                                        return 1;
                                    }
                                    else if (SC >= 454) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 2) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 25) {
                if (diffRsCsdiff < 243) {
                    if (CsDiff < 3) {
                        if (negBalance < 28) {
                            if (TSC < 22) {
                                if (diffAFD < 31) {
                                    return 0;
                                }
                                else if (diffAFD >= 31) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 22) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 28) {
                            if (diffAFD < 39) {
                                return 1;
                            }
                            else if (diffAFD >= 39) {
                                if (Cs < 91) {
                                    return 0;
                                }
                                else if (Cs >= 91) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 3) {
                        if (Cs < 778) {
                            if (diffMVdiffVal < 101) {
                                if (refDCval < 87) {
                                    if (diffRsCsdiff < 75) {
                                        if (AFD < 41) {
                                            return 1;
                                        }
                                        else if (AFD >= 41) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 75) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 87) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 101) {
                                if (posBalance < 69) {
                                    if (TSC < 42) {
                                        return 1;
                                    }
                                    else if (TSC >= 42) {
                                        if (AFD < 61) {
                                            return 0;
                                        }
                                        else if (AFD >= 61) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 69) {
                                    if (TSC < 32) {
                                        if (refDCval < 152) {
                                            return 1;
                                        }
                                        else if (refDCval >= 152) {
                                            if (SC < 664) {
                                                return 1;
                                            }
                                            else if (SC >= 664) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 32) {
                                        if (RsDiff < 6) {
                                            if (ssDCval < 108) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 108) {
                                                return 1;
                                            }
                                        }
                                        else if (RsDiff >= 6) {
                                            if (TSC < 37) {
                                                return 0;
                                            }
                                            else if (TSC >= 37) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 778) {
                            if (SC < 2617) {
                                if (Cs < 793) {
                                    if (refDCval < 79) {
                                        return 0;
                                    }
                                    else if (refDCval >= 79) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 793) {
                                    if (CsDiff < 7) {
                                        if (diffRsCsdiff < 62) {
                                            if (AFD < 47) {
                                                return 0;
                                            }
                                            else if (AFD >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 62) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        if (CsDiff < 8) {
                                            if (diffMVdiffVal < 171) {
                                                if (Cs < 808) {
                                                    return 1;
                                                }
                                                else if (Cs >= 808) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 171) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 8) {
                                            if (negBalance < 25) {
                                                if (diffMVdiffVal < 243) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 243) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 25) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 2617) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 243) {
                    if (Rs < 1675) {
                        return 1;
                    }
                    else if (Rs >= 1675) {
                        if (negBalance < 27) {
                            if (Rs < 2872) {
                                if (diffRsCsdiff < 1044) {
                                    if (SC < 2992) {
                                        return 1;
                                    }
                                    else if (SC >= 2992) {
                                        if (ssDCval < 122) {
                                            if (TSC < 26) {
                                                return 0;
                                            }
                                            else if (TSC >= 26) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 122) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 1044) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 2872) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 27) {
                            if (Cs < 6847) {
                                if (diffRsCsdiff < 441) {
                                    if (RsDiff < 22) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 22) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 441) {
                                    if (TSC < 18) {
                                        if (negBalance < 57) {
                                            return 1;
                                        }
                                        else if (negBalance >= 57) {
                                            if (ssDCval < 70) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 70) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 18) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 6847) {
                                if (MVDiff < 75) {
                                    if (CsDiff < 57) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 57) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 75) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectE(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffRsCsdiff < 72) {
        if (MVDiff < 98) {
            if (diffAFD < 9) {
                if (gchDC < 31) {
                    if (posBalance < 2) {
                        if (RsCsDiff < 7) {
                            return 0;
                        }
                        else if (RsCsDiff >= 7) {
                            if (negBalance < 2) {
                                if (Rs < 1030) {
                                    if (SC < 1435) {
                                        return 1;
                                    }
                                    else if (SC >= 1435) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 1030) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 2) {
                                return 0;
                            }
                        }
                    }
                    else if (posBalance >= 2) {
                        return 0;
                    }
                }
                else if (gchDC >= 31) {
                    if (AFD < 3) {
                        if (ssDCval < 122) {
                            return 0;
                        }
                        else if (ssDCval >= 122) {
                            return 1;
                        }
                    }
                    else if (AFD >= 3) {
                        return 0;
                    }
                }
            }
            else if (diffAFD >= 9) {
                if (Cs < 311) {
                    if (diffMVdiffVal < 64) {
                        if (refDCval < 24) {
                            return 1;
                        }
                        else if (refDCval >= 24) {
                            return 0;
                        }
                    }
                    else if (diffMVdiffVal >= 64) {
                        if (MVDiff < 90) {
                            if (gchDC < 7) {
                                return 1;
                            }
                            else if (gchDC >= 7) {
                                if (AFD < 33) {
                                    return 0;
                                }
                                else if (AFD >= 33) {
                                    return 1;
                                }
                            }
                        }
                        else if (MVDiff >= 90) {
                            return 0;
                        }
                    }
                }
                else if (Cs >= 311) {
                    if (MVDiff < 91) {
                        return 0;
                    }
                    else if (MVDiff >= 91) {
                        if (diffRsCsdiff < 47) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 47) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 98) {
            if (gchDC < 3) {
                if (diffTSC < 5) {
                    if (diffAFD < 7) {
                        if (TSCindex < 3) {
                            if (refDCval < 24) {
                                if (Rs < 3) {
                                    return 0;
                                }
                                else if (Rs >= 3) {
                                    if (CsDiff < 3) {
                                        if (posBalance < 45) {
                                            return 0;
                                        }
                                        else if (posBalance >= 45) {
                                            if (Rs < 79) {
                                                return 1;
                                            }
                                            else if (Rs >= 79) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 24) {
                                return 0;
                            }
                        }
                        else if (TSCindex >= 3) {
                            return 0;
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (AFD < 20) {
                            if (refDCval < 45) {
                                if (negBalance < 48) {
                                    if (diffMVdiffVal < 197) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 197) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 48) {
                                    if (posBalance < 49) {
                                        return 0;
                                    }
                                    else if (posBalance >= 49) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 45) {
                                return 0;
                            }
                        }
                        else if (AFD >= 20) {
                            return 0;
                        }
                    }
                }
                else if (diffTSC >= 5) {
                    if (Cs < 793) {
                        if (RsDiff < 9) {
                            if (diffAFD < 8) {
                                return 0;
                            }
                            else if (diffAFD >= 8) {
                                return 1;
                            }
                        }
                        else if (RsDiff >= 9) {
                            return 0;
                        }
                    }
                    else if (Cs >= 793) {
                        return 0;
                    }
                }
            }
            else if (gchDC >= 3) {
                if (diffMVdiffVal < 97) {
                    if (gchDC < 57) {
                        if (AFD < 62) {
                            if (Cs < 586) {
                                if (diffTSC < 4) {
                                    if (refDCval < 74) {
                                        if (MVDiff < 175) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 175) {
                                            if (diffAFD < -8) {
                                                return 0;
                                            }
                                            else if (diffAFD >= -8) {
                                                if (refDCval < 71) {
                                                    if (negBalance < 53) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 53) {
                                                        if (ssDCval < 37) {
                                                            if (diffRsCsdiff < 12) {
                                                                if (posBalance < 33) {
                                                                    return 0;
                                                                }
                                                                else if (posBalance >= 33) {
                                                                    if (RsDiff < 1) {
                                                                        return 0;
                                                                    }
                                                                    else if (RsDiff >= 1) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (diffRsCsdiff >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 37) {
                                                            if (RsCsDiff < 71) {
                                                                return 0;
                                                            }
                                                            else if (RsCsDiff >= 71) {
                                                                if (posBalance < 60) {
                                                                    return 0;
                                                                }
                                                                else if (posBalance >= 60) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (refDCval >= 71) {
                                                    if (TSC < 12) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 12) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 74) {
                                        if (refDCval < 229) {
                                            if (diffAFD < 11) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 11) {
                                                if (Cs < 62) {
                                                    return 1;
                                                }
                                                else if (Cs >= 62) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 229) {
                                            if (diffTSC < -3) {
                                                return 0;
                                            }
                                            else if (diffTSC >= -3) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 4) {
                                    if (AFD < 24) {
                                        if (gchDC < 6) {
                                            if (diffMVdiffVal < -3) {
                                                if (MVDiff < 132) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 132) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -3) {
                                                if (CsDiff < 5) {
                                                    if (TSC < 10) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (RsCsDiff < 9) {
                                                            return 1;
                                                        }
                                                        else if (RsCsDiff >= 9) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 6) {
                                            if (diffRsCsdiff < -84) {
                                                return 1;
                                            }
                                            else if (diffRsCsdiff >= -84) {
                                                if (ssDCval < 40) {
                                                    if (TSC < 12) {
                                                        if (diffTSC < 9) {
                                                            return 0;
                                                        }
                                                        else if (diffTSC >= 9) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 40) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 24) {
                                        if (AFD < 43) {
                                            return 0;
                                        }
                                        else if (AFD >= 43) {
                                            if (AFD < 45) {
                                                if (Cs < 260) {
                                                    return 0;
                                                }
                                                else if (Cs >= 260) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 45) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 586) {
                                return 0;
                            }
                        }
                        else if (AFD >= 62) {
                            if (Cs < 1174) {
                                return 0;
                            }
                            else if (Cs >= 1174) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 57) {
                        if (SC < 3124) {
                            if (MVDiff < 879) {
                                if (refDCval < 118) {
                                    if (ssDCval < 90) {
                                        if (negBalance < 101) {
                                            if (gchDC < 73) {
                                                return 1;
                                            }
                                            else if (gchDC >= 73) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 101) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 90) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 118) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 879) {
                                return 1;
                            }
                        }
                        else if (SC >= 3124) {
                            return 1;
                        }
                    }
                }
                else if (diffMVdiffVal >= 97) {
                    if (diffRsCsdiff < 1) {
                        if (diffAFD < 7) {
                            if (gchDC < 65) {
                                if (ssDCval < 31) {
                                    if (diffAFD < -4) {
                                        return 0;
                                    }
                                    else if (diffAFD >= -4) {
                                        if (ssDCval < 24) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 24) {
                                            if (Rs < 427) {
                                                return 1;
                                            }
                                            else if (Rs >= 427) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 31) {
                                    if (RsCsDiff < 821) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 821) {
                                        if (TSC < 30) {
                                            return 0;
                                        }
                                        else if (TSC >= 30) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 65) {
                                if (TSC < 9) {
                                    if (Scindex < 3) {
                                        return 1;
                                    }
                                    else if (Scindex >= 3) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 9) {
                                    if (ssDCval < 24) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 24) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 7) {
                            if (Cs < 185) {
                                if (ssDCval < 84) {
                                    if (SC < 11) {
                                        return 0;
                                    }
                                    else if (SC >= 11) {
                                        if (AFD < 19) {
                                            return 1;
                                        }
                                        else if (AFD >= 19) {
                                            if (negBalance < 90) {
                                                if (negBalance < 12) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 12) {
                                                    return 0;
                                                }
                                            }
                                            else if (negBalance >= 90) {
                                                if (CsDiff < 6) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 6) {
                                                    if (Rs < 72) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 72) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 84) {
                                    if (AFD < 14) {
                                        return 1;
                                    }
                                    else if (AFD >= 14) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Cs >= 185) {
                                if (RsCsDiff < 173) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 173) {
                                    if (RsCsDiff < 222) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 222) {
                                        if (CsDiff < 53) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 53) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 1) {
                        if (diffAFD < 13) {
                            if (MVDiff < 175) {
                                if (SC < 881) {
                                    if (Rs < 83) {
                                        if (posBalance < 65) {
                                            return 0;
                                        }
                                        else if (posBalance >= 65) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 83) {
                                        if (CsDiff < 2) {
                                            if (gchDC < 8) {
                                                return 1;
                                            }
                                            else if (gchDC >= 8) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 2) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 881) {
                                    if (MVDiff < 163) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 163) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 175) {
                                if (ssDCval < 50) {
                                    if (refDCval < 31) {
                                        return 1;
                                    }
                                    else if (refDCval >= 31) {
                                        if (diffAFD < 7) {
                                            if (Cs < 334) {
                                                return 0;
                                            }
                                            else if (Cs >= 334) {
                                                if (gchDC < 10) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 10) {
                                                    if (RsDiff < 5) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 5) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 7) {
                                            if (MVDiff < 221) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 221) {
                                                if (Rs < 18) {
                                                    return 0;
                                                }
                                                else if (Rs >= 18) {
                                                    if (MVDiff < 504) {
                                                        if (negBalance < 35) {
                                                            if (Rs < 52) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 52) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (negBalance >= 35) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 504) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 50) {
                                    if (Rs < 5) {
                                        return 1;
                                    }
                                    else if (Rs >= 5) {
                                        if (diffTSC < 5) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 5) {
                                            if (diffAFD < 6) {
                                                if (RsCsDiff < 37) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 37) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffAFD >= 6) {
                                                if (Cs < 216) {
                                                    if (MVDiff < 213) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 213) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Cs >= 216) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 13) {
                            if (gchDC < 22) {
                                if (SC < 593) {
                                    if (refDCval < 54) {
                                        return 1;
                                    }
                                    else if (refDCval >= 54) {
                                        if (refDCval < 55) {
                                            return 0;
                                        }
                                        else if (refDCval >= 55) {
                                            if (diffAFD < 18) {
                                                if (Rs < 75) {
                                                    return 0;
                                                }
                                                else if (Rs >= 75) {
                                                    if (Rs < 197) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 197) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 18) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 593) {
                                    if (gchDC < 9) {
                                        return 1;
                                    }
                                    else if (gchDC >= 9) {
                                        if (diffTSC < 7) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 7) {
                                            if (negBalance < 41) {
                                                if (RsDiff < 7) {
                                                    if (Cs < 352) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 352) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    return 0;
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 22) {
                                if (Cs < 621) {
                                    if (refDCval < 47) {
                                        if (RsDiff < 4) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 4) {
                                            if (Rs < 524) {
                                                if (CsDiff < 6) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 6) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 524) {
                                                if (posBalance < 117) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 117) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 47) {
                                        if (posBalance < 116) {
                                            if (diffRsCsdiff < 49) {
                                                if (SC < 813) {
                                                    if (MVDiff < 258) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 258) {
                                                        if (posBalance < 3) {
                                                            if (diffTSC < 6) {
                                                                return 1;
                                                            }
                                                            else if (diffTSC >= 6) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (posBalance >= 3) {
                                                            if (negBalance < 12) {
                                                                return 0;
                                                            }
                                                            else if (negBalance >= 12) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (SC >= 813) {
                                                    if (diffAFD < 28) {
                                                        if (gchDC < 30) {
                                                            return 0;
                                                        }
                                                        else if (gchDC >= 30) {
                                                            if (AFD < 56) {
                                                                if (posBalance < 92) {
                                                                    return 1;
                                                                }
                                                                else if (posBalance >= 92) {
                                                                    return 0;
                                                                }
                                                            }
                                                            else if (AFD >= 56) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                    else if (diffAFD >= 28) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 49) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 116) {
                                            if (Rs < 92) {
                                                if (RsCsDiff < 17) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 17) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 92) {
                                                if (ssDCval < 157) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 157) {
                                                    if (RsCsDiff < 52) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 52) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 621) {
                                    if (posBalance < 95) {
                                        return 0;
                                    }
                                    else if (posBalance >= 95) {
                                        if (posBalance < 114) {
                                            if (Cs < 683) {
                                                return 1;
                                            }
                                            else if (Cs >= 683) {
                                                if (CsDiff < 5) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 114) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffRsCsdiff >= 72) {
        if (MVDiff < 66) {
            if (diffTSC < 13) {
                if (refDCval < 24) {
                    if (CsDiff < 12) {
                        return 0;
                    }
                    else if (CsDiff >= 12) {
                        return 1;
                    }
                }
                else if (refDCval >= 24) {
                    if (Cs < 199) {
                        if (RsCsDiff < 511) {
                            return 0;
                        }
                        else if (RsCsDiff >= 511) {
                            return 1;
                        }
                    }
                    else if (Cs >= 199) {
                        if (MVDiff < 55) {
                            return 0;
                        }
                        else if (MVDiff >= 55) {
                            if (diffTSC < 7) {
                                if (TSCindex < 5) {
                                    if (Rs < 2121) {
                                        return 1;
                                    }
                                    else if (Rs >= 2121) {
                                        return 0;
                                    }
                                }
                                else if (TSCindex >= 5) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 7) {
                                if (AFD < 20) {
                                    if (Rs < 2356) {
                                        return 1;
                                    }
                                    else if (Rs >= 2356) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 20) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 13) {
                if (Cs < 2361) {
                    if (diffMVdiffVal < 6) {
                        if (posBalance < 50) {
                            return 1;
                        }
                        else if (posBalance >= 50) {
                            if (MVDiff < 14) {
                                return 1;
                            }
                            else if (MVDiff >= 14) {
                                return 0;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 6) {
                        if (posBalance < 21) {
                            return 0;
                        }
                        else if (posBalance >= 21) {
                            if (diffRsCsdiff < 221) {
                                if (Cs < 975) {
                                    return 1;
                                }
                                else if (Cs >= 975) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 221) {
                                return 1;
                            }
                        }
                    }
                }
                else if (Cs >= 2361) {
                    if (TSC < 23) {
                        if (negBalance < 55) {
                            return 0;
                        }
                        else if (negBalance >= 55) {
                            if (posBalance < 40) {
                                return 1;
                            }
                            else if (posBalance >= 40) {
                                return 0;
                            }
                        }
                    }
                    else if (TSC >= 23) {
                        if (ssDCval < 118) {
                            return 1;
                        }
                        else if (ssDCval >= 118) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 66) {
            if (TSC < 17) {
                if (diffAFD < 18) {
                    if (ssDCval < 43) {
                        if (CsDiff < 12) {
                            if (negBalance < 41) {
                                return 0;
                            }
                            else if (negBalance >= 41) {
                                if (RsDiff < 7) {
                                    return 0;
                                }
                                else if (RsDiff >= 7) {
                                    if (Cs < 1032) {
                                        return 1;
                                    }
                                    else if (Cs >= 1032) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 12) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 43) {
                        if (SC < 2026) {
                            if (refDCval < 77) {
                                if (diffMVdiffVal < 61) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 61) {
                                    if (diffRsCsdiff < 103) {
                                        if (Rs < 328) {
                                            return 1;
                                        }
                                        else if (Rs >= 328) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 103) {
                                        if (TSC < 12) {
                                            if (SC < 1109) {
                                                return 1;
                                            }
                                            else if (SC >= 1109) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 12) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 77) {
                                return 0;
                            }
                        }
                        else if (SC >= 2026) {
                            if (RsCsDiff < 3703) {
                                if (Rs < 1271) {
                                    if (TSC < 15) {
                                        return 0;
                                    }
                                    else if (TSC >= 15) {
                                        if (ssDCval < 80) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 80) {
                                            if (Rs < 1135) {
                                                return 0;
                                            }
                                            else if (Rs >= 1135) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1271) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 3703) {
                                if (AFD < 12) {
                                    return 1;
                                }
                                else if (AFD >= 12) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 18) {
                    if (Cs < 818) {
                        if (SC < 940) {
                            return 1;
                        }
                        else if (SC >= 940) {
                            if (MVDiff < 110) {
                                if (Cs < 337) {
                                    return 0;
                                }
                                else if (Cs >= 337) {
                                    if (TSC < 12) {
                                        if (ssDCval < 39) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 39) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 12) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 110) {
                                return 1;
                            }
                        }
                    }
                    else if (Cs >= 818) {
                        if (diffTSC < 13) {
                            if (MVDiff < 399) {
                                if (diffMVdiffVal < 64) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 64) {
                                    if (diffMVdiffVal < 99) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= 99) {
                                        if (ssDCval < 53) {
                                            if (SC < 4409) {
                                                if (TSC < 13) {
                                                    return 0;
                                                }
                                                else if (TSC >= 13) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 4409) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 53) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 399) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 13) {
                            if (RsCsDiff < 148) {
                                return 0;
                            }
                            else if (RsCsDiff >= 148) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (TSC >= 17) {
                if (negBalance < 33) {
                    if (SC < 1410) {
                        if (posBalance < 117) {
                            return 1;
                        }
                        else if (posBalance >= 117) {
                            if (negBalance < 6) {
                                return 1;
                            }
                            else if (negBalance >= 6) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 1410) {
                        if (CsDiff < 14) {
                            if (SC < 2549) {
                                if (diffAFD < 40) {
                                    if (posBalance < 99) {
                                        if (RsCsDiff < 175) {
                                            if (TSC < 26) {
                                                return 0;
                                            }
                                            else if (TSC >= 26) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 175) {
                                            if (Rs < 1361) {
                                                if (Rs < 1053) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1053) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 1361) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 99) {
                                        if (ssDCval < 111) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 111) {
                                            if (TSC < 27) {
                                                return 0;
                                            }
                                            else if (TSC >= 27) {
                                                if (gchDC < 61) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 61) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 40) {
                                    if (MVDiff < 80) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 80) {
                                        if (Rs < 1537) {
                                            return 1;
                                        }
                                        else if (Rs >= 1537) {
                                            if (Cs < 637) {
                                                return 1;
                                            }
                                            else if (Cs >= 637) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 2549) {
                                if (diffAFD < 17) {
                                    return 0;
                                }
                                else if (diffAFD >= 17) {
                                    if (RsCsDiff < 263) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 263) {
                                        if (diffMVdiffVal < -26) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= -26) {
                                            if (TSC < 34) {
                                                if (diffAFD < 18) {
                                                    if (AFD < 30) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 30) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 18) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 34) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 14) {
                            if (Cs < 1854) {
                                return 1;
                            }
                            else if (Cs >= 1854) {
                                if (diffRsCsdiff < 659) {
                                    if (diffRsCsdiff < 533) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 533) {
                                        if (TSC < 30) {
                                            return 1;
                                        }
                                        else if (TSC >= 30) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 659) {
                                    if (diffAFD < 18) {
                                        if (TSC < 24) {
                                            return 0;
                                        }
                                        else if (TSC >= 24) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 18) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 33) {
                    if (ssDCval < 78) {
                        if (diffAFD < 23) {
                            if (negBalance < 101) {
                                if (diffAFD < 13) {
                                    if (AFD < 46) {
                                        if (RsCsDiff < 549) {
                                            if (Cs < 2069) {
                                                if (RsCsDiff < 299) {
                                                    if (ssDCval < 63) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 63) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsCsDiff >= 299) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 2069) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 549) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 46) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 13) {
                                    if (diffTSC < 14) {
                                        return 1;
                                    }
                                    else if (diffTSC >= 14) {
                                        if (Rs < 1318) {
                                            return 0;
                                        }
                                        else if (Rs >= 1318) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 101) {
                                if (RsCsDiff < 238) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 238) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 23) {
                            if (negBalance < 123) {
                                if (Cs < 787) {
                                    return 1;
                                }
                                else if (Cs >= 787) {
                                    if (diffTSC < 11) {
                                        if (RsDiff < 22) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 11) {
                                        if (SC < 1384) {
                                            return 0;
                                        }
                                        else if (SC >= 1384) {
                                            if (Cs < 871) {
                                                if (Cs < 853) {
                                                    return 1;
                                                }
                                                else if (Cs >= 853) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 871) {
                                                if (TSC < 40) {
                                                    return 1;
                                                }
                                                else if (TSC >= 40) {
                                                    if (RsCsDiff < 1893) {
                                                        return 0;
                                                    }
                                                    else if (RsCsDiff >= 1893) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 123) {
                                if (refDCval < 139) {
                                    return 0;
                                }
                                else if (refDCval >= 139) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 78) {
                        if (diffTSC < 9) {
                            if (TSC < 25) {
                                if (diffMVdiffVal < 514) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 514) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 25) {
                                if (TSC < 28) {
                                    if (AFD < 54) {
                                        if (Rs < 3498) {
                                            return 0;
                                        }
                                        else if (Rs >= 3498) {
                                            if (SC < 9206) {
                                                return 1;
                                            }
                                            else if (SC >= 9206) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 54) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 28) {
                                    if (SC < 3832) {
                                        if (AFD < 64) {
                                            return 1;
                                        }
                                        else if (AFD >= 64) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 3832) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 9) {
                            if (diffAFD < 22) {
                                if (Scindex < 4) {
                                    return 1;
                                }
                                else if (Scindex >= 4) {
                                    if (negBalance < 42) {
                                        return 0;
                                    }
                                    else if (negBalance >= 42) {
                                        if (posBalance < 40) {
                                            if (Rs < 1054) {
                                                if (MVDiff < 546) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 546) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 1054) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 40) {
                                            if (ssDCval < 117) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 117) {
                                                if (refDCval < 121) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 121) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 22) {
                                if (ssDCval < 93) {
                                    if (diffAFD < 40) {
                                        if (MVDiff < 1093) {
                                            if (diffRsCsdiff < 198) {
                                                if (diffRsCsdiff < 151) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= 151) {
                                                    if (posBalance < 56) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 56) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 198) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 1093) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 40) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 93) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectF(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (MVDiff < 69) {
        if (gchDC < 7) {
            if (diffMVdiffVal < 50) {
                if (ssDCval < 20) {
                    if (RsCsDiff < 35) {
                        if (RsDiff < 2) {
                            return 0;
                        }
                        else if (RsDiff >= 2) {
                            if (Cs < 512) {
                                return 0;
                            }
                            else if (Cs >= 512) {
                                if (Rs < 635) {
                                    return 1;
                                }
                                else if (Rs >= 635) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 35) {
                        return 1;
                    }
                }
                else if (ssDCval >= 20) {
                    if (diffTSC < 10) {
                        if (CsDiff < 33) {
                            return 0;
                        }
                        else if (CsDiff >= 33) {
                            if (SC < 4130) {
                                return 1;
                            }
                            else if (SC >= 4130) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 10) {
                        if (Cs < 1971) {
                            if (AFD < 19) {
                                return 1;
                            }
                            else if (AFD >= 19) {
                                if (SC < 15001) {
                                    return 0;
                                }
                                else if (SC >= 15001) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 1971) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 50) {
                if (diffAFD < 13) {
                    return 0;
                }
                else if (diffAFD >= 13) {
                    if (diffMVdiffVal < 55) {
                        if (Cs < 1997) {
                            return 1;
                        }
                        else if (Cs >= 1997) {
                            return 0;
                        }
                    }
                    else if (diffMVdiffVal >= 55) {
                        return 1;
                    }
                }
            }
        }
        else if (gchDC >= 7) {
            if (TSC < 24) {
                if (Rs < 7030) {
                    if (diffTSC < 9) {
                        if (Rs < 34) {
                            if (RsCsDiff < 326) {
                                if (ssDCval < 175) {
                                    return 0;
                                }
                                else if (ssDCval >= 175) {
                                    if (refDCval < 226) {
                                        return 0;
                                    }
                                    else if (refDCval >= 226) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 326) {
                                return 1;
                            }
                        }
                        else if (Rs >= 34) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= 9) {
                        if (Cs < 469) {
                            if (ssDCval < 75) {
                                if (refDCval < 41) {
                                    if (AFD < 20) {
                                        return 1;
                                    }
                                    else if (AFD >= 20) {
                                        if (Rs < 173) {
                                            return 1;
                                        }
                                        else if (Rs >= 173) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 41) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 75) {
                                if (ssDCval < 204) {
                                    return 0;
                                }
                                else if (ssDCval >= 204) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 469) {
                            if (gchDC < 15) {
                                if (diffMVdiffVal < 47) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 47) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 15) {
                                return 0;
                            }
                        }
                    }
                }
                else if (Rs >= 7030) {
                    if (diffAFD < 9) {
                        return 0;
                    }
                    else if (diffAFD >= 9) {
                        return 1;
                    }
                }
            }
            else if (TSC >= 24) {
                if (diffTSC < 20) {
                    if (MVDiff < 62) {
                        return 0;
                    }
                    else if (MVDiff >= 62) {
                        return 1;
                    }
                }
                else if (diffTSC >= 20) {
                    if (Scindex < 6) {
                        return 1;
                    }
                    else if (Scindex >= 6) {
                        if (RsDiff < 20) {
                            return 0;
                        }
                        else if (RsDiff >= 20) {
                            if (posBalance < 119) {
                                return 1;
                            }
                            else if (posBalance >= 119) {
                                if (Cs < 4577) {
                                    return 1;
                                }
                                else if (Cs >= 4577) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (MVDiff >= 69) {
        if (AFD < 30) {
            if (diffAFD < 10) {
                if (diffRsCsdiff < 5) {
                    if (gchDC < 123) {
                        if (diffTSC < 5) {
                            if (diffMVdiffVal < 283) {
                                if (ssDCval < 25) {
                                    if (TSCindex < 2) {
                                        return 0;
                                    }
                                    else if (TSCindex >= 2) {
                                        if (refDCval < 23) {
                                            return 0;
                                        }
                                        else if (refDCval >= 23) {
                                            if (negBalance < 5) {
                                                return 1;
                                            }
                                            else if (negBalance >= 5) {
                                                if (negBalance < 56) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 56) {
                                                    if (Cs < 13) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 13) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 25) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 283) {
                                if (diffMVdiffVal < 284) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 284) {
                                    if (RsDiff < 4) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 4) {
                                        if (Rs < 99) {
                                            if (ssDCval < 20) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 20) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 99) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 5) {
                            if (gchDC < 59) {
                                if (ssDCval < 53) {
                                    if (negBalance < 56) {
                                        if (diffAFD < -9) {
                                            return 1;
                                        }
                                        else if (diffAFD >= -9) {
                                            if (Cs < 50) {
                                                if (AFD < 12) {
                                                    return 1;
                                                }
                                                else if (AFD >= 12) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 50) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 56) {
                                        if (TSC < 14) {
                                            return 1;
                                        }
                                        else if (TSC >= 14) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 53) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 59) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 123) {
                        if (AFD < 16) {
                            return 1;
                        }
                        else if (AFD >= 16) {
                            if (CsDiff < 1) {
                                return 0;
                            }
                            else if (CsDiff >= 1) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 5) {
                    if (SC < 367) {
                        if (refDCval < 44) {
                            if (Rs < 174) {
                                return 1;
                            }
                            else if (Rs >= 174) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 44) {
                            if (RsDiff < 7) {
                                return 0;
                            }
                            else if (RsDiff >= 7) {
                                if (ssDCval < 131) {
                                    return 1;
                                }
                                else if (ssDCval >= 131) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (SC >= 367) {
                        if (refDCval < 49) {
                            if (posBalance < 72) {
                                if (diffTSC < 3) {
                                    if (AFD < 1) {
                                        return 1;
                                    }
                                    else if (AFD >= 1) {
                                        if (CsDiff < 15) {
                                            if (diffTSC < -8) {
                                                if (SC < 1202) {
                                                    return 1;
                                                }
                                                else if (SC >= 1202) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= -8) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 15) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 3) {
                                    if (negBalance < 63) {
                                        if (TSCindex < 8) {
                                            if (SC < 660) {
                                                return 0;
                                            }
                                            else if (SC >= 660) {
                                                return 1;
                                            }
                                        }
                                        else if (TSCindex >= 8) {
                                            if (TSC < 14) {
                                                if (refDCval < 30) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 30) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 14) {
                                                if (Cs < 578) {
                                                    return 1;
                                                }
                                                else if (Cs >= 578) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 63) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 72) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 49) {
                            if (diffMVdiffVal < 52) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 52) {
                                if (gchDC < 85) {
                                    if (diffRsCsdiff < 2663) {
                                        if (ssDCval < 83) {
                                            if (diffTSC < 7) {
                                                if (gchDC < 7) {
                                                    if (Rs < 296) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 296) {
                                                        if (negBalance < 73) {
                                                            return 0;
                                                        }
                                                        else if (negBalance >= 73) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (gchDC >= 7) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 7) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 83) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 2663) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 85) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 10) {
                if (diffMVdiffVal < 65) {
                    if (diffAFD < 22) {
                        if (refDCval < 55) {
                            if (diffTSC < 10) {
                                if (negBalance < 48) {
                                    if (Rs < 3082) {
                                        if (posBalance < 76) {
                                            if (diffTSC < 7) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 7) {
                                                if (Rs < 470) {
                                                    return 0;
                                                }
                                                else if (Rs >= 470) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 76) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 3082) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 48) {
                                    if (Rs < 558) {
                                        if (negBalance < 56) {
                                            if (RsDiff < 2) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 2) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 56) {
                                            if (diffAFD < 13) {
                                                if (Cs < 216) {
                                                    return 1;
                                                }
                                                else if (Cs >= 216) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 13) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 558) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 10) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 55) {
                            if (negBalance < 105) {
                                return 0;
                            }
                            else if (negBalance >= 105) {
                                if (posBalance < 16) {
                                    return 0;
                                }
                                else if (posBalance >= 16) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 22) {
                        if (Rs < 524) {
                            return 0;
                        }
                        else if (Rs >= 524) {
                            if (MVDiff < 182) {
                                return 1;
                            }
                            else if (MVDiff >= 182) {
                                if (Cs < 758) {
                                    return 0;
                                }
                                else if (Cs >= 758) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 65) {
                    if (ssDCval < 48) {
                        if (diffMVdiffVal < 123) {
                            if (diffRsCsdiff < 131) {
                                if (gchDC < 15) {
                                    if (RsDiff < 4) {
                                        if (TSCindex < 7) {
                                            return 0;
                                        }
                                        else if (TSCindex >= 7) {
                                            if (Cs < 16) {
                                                return 0;
                                            }
                                            else if (Cs >= 16) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 4) {
                                        if (gchDC < 4) {
                                            return 0;
                                        }
                                        else if (gchDC >= 4) {
                                            if (AFD < 23) {
                                                return 1;
                                            }
                                            else if (AFD >= 23) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 15) {
                                    if (Cs < 256) {
                                        return 0;
                                    }
                                    else if (Cs >= 256) {
                                        if (RsDiff < 4) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 131) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 123) {
                            if (ssDCval < 16) {
                                if (TSC < 6) {
                                    return 0;
                                }
                                else if (TSC >= 6) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 16) {
                                if (diffMVdiffVal < 919) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 919) {
                                    if (Cs < 1098) {
                                        return 1;
                                    }
                                    else if (Cs >= 1098) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 48) {
                        if (diffTSC < 6) {
                            if (diffTSC < 4) {
                                if (MVDiff < 2511) {
                                    return 0;
                                }
                                else if (MVDiff >= 2511) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 4) {
                                if (RsCsDiff < 75) {
                                    if (TSC < 8) {
                                        return 1;
                                    }
                                    else if (TSC >= 8) {
                                        if (refDCval < 57) {
                                            return 1;
                                        }
                                        else if (refDCval >= 57) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 75) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 6) {
                            if (AFD < 20) {
                                if (gchDC < 29) {
                                    if (gchDC < 12) {
                                        if (TSC < 14) {
                                            if (Cs < 90) {
                                                return 1;
                                            }
                                            else if (Cs >= 90) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 14) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 12) {
                                        if (TSC < 8) {
                                            return 1;
                                        }
                                        else if (TSC >= 8) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 29) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 20) {
                                if (negBalance < 24) {
                                    if (diffRsCsdiff < 23) {
                                        if (AFD < 22) {
                                            if (Cs < 75) {
                                                return 1;
                                            }
                                            else if (Cs >= 75) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 23) {
                                        if (diffMVdiffVal < 117) {
                                            if (RsDiff < 10) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 10) {
                                                if (refDCval < 42) {
                                                    if (Rs < 1974) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1974) {
                                                        return 0;
                                                    }
                                                }
                                                else if (refDCval >= 42) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 117) {
                                            if (Cs < 251) {
                                                return 1;
                                            }
                                            else if (Cs >= 251) {
                                                if (TSC < 15) {
                                                    return 0;
                                                }
                                                else if (TSC >= 15) {
                                                    if (RsDiff < 6) {
                                                        return 1;
                                                    }
                                                    else if (RsDiff >= 6) {
                                                        if (RsCsDiff < 89) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 89) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 24) {
                                    if (diffAFD < 15) {
                                        if (posBalance < 63) {
                                            return 0;
                                        }
                                        else if (posBalance >= 63) {
                                            if (diffMVdiffVal < 95) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 95) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 15) {
                                        if (diffMVdiffVal < 79) {
                                            if (MVDiff < 91) {
                                                if (AFD < 24) {
                                                    return 0;
                                                }
                                                else if (AFD >= 24) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 91) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 79) {
                                            if (negBalance < 99) {
                                                return 1;
                                            }
                                            else if (negBalance >= 99) {
                                                if (AFD < 28) {
                                                    return 0;
                                                }
                                                else if (AFD >= 28) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (AFD >= 30) {
            if (CsDiff < 7) {
                if (diffAFD < 23) {
                    if (diffTSC < 9) {
                        if (SC < 41) {
                            if (CsDiff < 4) {
                                if (diffRsCsdiff < -85) {
                                    return 1;
                                }
                                else if (diffRsCsdiff >= -85) {
                                    if (diffMVdiffVal < 1561) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 1561) {
                                        return 1;
                                    }
                                }
                            }
                            else if (CsDiff >= 4) {
                                return 1;
                            }
                        }
                        else if (SC >= 41) {
                            if (posBalance < 66) {
                                if (diffAFD < 13) {
                                    if (negBalance < 41) {
                                        return 1;
                                    }
                                    else if (negBalance >= 41) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            if (diffMVdiffVal < 65) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 65) {
                                                if (posBalance < 25) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 25) {
                                                    if (RsDiff < 4) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        if (Cs < 253) {
                                                            if (SC < 1545) {
                                                                return 1;
                                                            }
                                                            else if (SC >= 1545) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (Cs >= 253) {
                                                            if (refDCval < 148) {
                                                                return 0;
                                                            }
                                                            else if (refDCval >= 148) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 13) {
                                    if (gchDC < 19) {
                                        if (refDCval < 64) {
                                            if (Cs < 203) {
                                                return 1;
                                            }
                                            else if (Cs >= 203) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 64) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 19) {
                                        if (Scindex < 2) {
                                            if (Cs < 53) {
                                                return 1;
                                            }
                                            else if (Cs >= 53) {
                                                return 0;
                                            }
                                        }
                                        else if (Scindex >= 2) {
                                            if (SC < 823) {
                                                return 1;
                                            }
                                            else if (SC >= 823) {
                                                if (CsDiff < 3) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 3) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 66) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 9) {
                        if (RsCsDiff < 108) {
                            if (MVDiff < 1224) {
                                if (diffMVdiffVal < 304) {
                                    if (posBalance < 33) {
                                        return 0;
                                    }
                                    else if (posBalance >= 33) {
                                        if (ssDCval < 62) {
                                            if (diffTSC < 11) {
                                                return 1;
                                            }
                                            else if (diffTSC >= 11) {
                                                if (diffRsCsdiff < -10) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= -10) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 62) {
                                            if (Rs < 10) {
                                                return 1;
                                            }
                                            else if (Rs >= 10) {
                                                if (diffTSC < 17) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 17) {
                                                    if (refDCval < 61) {
                                                        return 0;
                                                    }
                                                    else if (refDCval >= 61) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 304) {
                                    if (refDCval < 44) {
                                        return 0;
                                    }
                                    else if (refDCval >= 44) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 1224) {
                                return 1;
                            }
                        }
                        else if (RsCsDiff >= 108) {
                            if (Cs < 548) {
                                return 1;
                            }
                            else if (Cs >= 548) {
                                if (Rs < 1810) {
                                    return 1;
                                }
                                else if (Rs >= 1810) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 23) {
                    if (negBalance < 41) {
                        if (negBalance < 2) {
                            return 0;
                        }
                        else if (negBalance >= 2) {
                            if (diffAFD < 26) {
                                return 0;
                            }
                            else if (diffAFD >= 26) {
                                if (gchDC < 77) {
                                    if (TSC < 29) {
                                        if (diffMVdiffVal < 73) {
                                            if (SC < 264) {
                                                return 0;
                                            }
                                            else if (SC >= 264) {
                                                if (Cs < 228) {
                                                    return 0;
                                                }
                                                else if (Cs >= 228) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 73) {
                                            if (posBalance < 116) {
                                                if (ssDCval < 139) {
                                                    if (ssDCval < 70) {
                                                        if (AFD < 39) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 39) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (ssDCval >= 70) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 139) {
                                                    return 0;
                                                }
                                            }
                                            else if (posBalance >= 116) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 29) {
                                        if (TSC < 52) {
                                            if (gchDC < 36) {
                                                if (RsDiff < 16) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 16) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 36) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 52) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 77) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 41) {
                        if (refDCval < 163) {
                            if (AFD < 83) {
                                if (diffMVdiffVal < 81) {
                                    if (RsCsDiff < 400) {
                                        if (negBalance < 66) {
                                            return 0;
                                        }
                                        else if (negBalance >= 66) {
                                            if (TSC < 27) {
                                                return 1;
                                            }
                                            else if (TSC >= 27) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 400) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 81) {
                                    if (CsDiff < 1) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 1) {
                                        if (gchDC < 39) {
                                            return 1;
                                        }
                                        else if (gchDC >= 39) {
                                            if (RsCsDiff < 201) {
                                                if (negBalance < 73) {
                                                    if (CsDiff < 3) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (negBalance >= 73) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 201) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 83) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 163) {
                            if (SC < 544) {
                                return 1;
                            }
                            else if (SC >= 544) {
                                if (diffAFD < 48) {
                                    return 0;
                                }
                                else if (diffAFD >= 48) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (CsDiff >= 7) {
                if (diffAFD < 18) {
                    if (gchDC < 23) {
                        if (diffMVdiffVal < 153) {
                            if (diffAFD < 15) {
                                if (SC < 1481) {
                                    if (SC < 1439) {
                                        return 0;
                                    }
                                    else if (SC >= 1439) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1481) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 15) {
                                if (CsDiff < 8) {
                                    return 1;
                                }
                                else if (CsDiff >= 8) {
                                    if (TSC < 33) {
                                        return 0;
                                    }
                                    else if (TSC >= 33) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 153) {
                            if (gchDC < 1) {
                                return 1;
                            }
                            else if (gchDC >= 1) {
                                if (diffAFD < 8) {
                                    return 0;
                                }
                                else if (diffAFD >= 8) {
                                    if (refDCval < 80) {
                                        return 1;
                                    }
                                    else if (refDCval >= 80) {
                                        if (SC < 1977) {
                                            return 1;
                                        }
                                        else if (SC >= 1977) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 23) {
                        if (ssDCval < 41) {
                            return 1;
                        }
                        else if (ssDCval >= 41) {
                            if (RsCsDiff < 607) {
                                if (Cs < 73) {
                                    return 1;
                                }
                                else if (Cs >= 73) {
                                    if (diffRsCsdiff < 98) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 98) {
                                        if (posBalance < 73) {
                                            if (posBalance < 18) {
                                                if (SC < 2902) {
                                                    return 0;
                                                }
                                                else if (SC >= 2902) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 18) {
                                                if (RsDiff < 18) {
                                                    return 1;
                                                }
                                                else if (RsDiff >= 18) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 73) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 607) {
                                if (Rs < 2189) {
                                    if (SC < 3703) {
                                        if (diffRsCsdiff < 247) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= 247) {
                                            if (SC < 752) {
                                                return 0;
                                            }
                                            else if (SC >= 752) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 3703) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 2189) {
                                    if (diffAFD < 15) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 15) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 18) {
                    if (diffAFD < 27) {
                        if (posBalance < 102) {
                            if (diffTSC < 6) {
                                if (Rs < 418) {
                                    return 1;
                                }
                                else if (Rs >= 418) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 6) {
                                if (AFD < 50) {
                                    if (diffMVdiffVal < 73) {
                                        if (Cs < 768) {
                                            return 1;
                                        }
                                        else if (Cs >= 768) {
                                            if (refDCval < 62) {
                                                return 0;
                                            }
                                            else if (refDCval >= 62) {
                                                if (gchDC < 29) {
                                                    if (posBalance < 53) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 53) {
                                                        return 1;
                                                    }
                                                }
                                                else if (gchDC >= 29) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 73) {
                                        if (refDCval < 71) {
                                            return 1;
                                        }
                                        else if (refDCval >= 71) {
                                            if (Cs < 819) {
                                                return 1;
                                            }
                                            else if (Cs >= 819) {
                                                if (ssDCval < 59) {
                                                    if (RsCsDiff < 281) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 281) {
                                                        return 0;
                                                    }
                                                }
                                                else if (ssDCval >= 59) {
                                                    if (Rs < 4514) {
                                                        if (AFD < 41) {
                                                            if (Rs < 1351) {
                                                                return 0;
                                                            }
                                                            else if (Rs >= 1351) {
                                                                if (diffRsCsdiff < 643) {
                                                                    return 0;
                                                                }
                                                                else if (diffRsCsdiff >= 643) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (AFD >= 41) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 4514) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 50) {
                                    if (SC < 2762) {
                                        return 0;
                                    }
                                    else if (SC >= 2762) {
                                        if (Rs < 1576) {
                                            return 1;
                                        }
                                        else if (Rs >= 1576) {
                                            if (Cs < 930) {
                                                return 1;
                                            }
                                            else if (Cs >= 930) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 102) {
                            if (ssDCval < 139) {
                                return 0;
                            }
                            else if (ssDCval >= 139) {
                                if (negBalance < 6) {
                                    if (Rs < 144) {
                                        return 1;
                                    }
                                    else if (Rs >= 144) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 6) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 27) {
                        if (diffMVdiffVal < 61) {
                            if (MVDiff < 112) {
                                return 0;
                            }
                            else if (MVDiff >= 112) {
                                if (ssDCval < 115) {
                                    return 1;
                                }
                                else if (ssDCval >= 115) {
                                    if (gchDC < 48) {
                                        if (ssDCval < 162) {
                                            if (MVDiff < 143) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 143) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 162) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 48) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 61) {
                            if (diffAFD < 33) {
                                if (gchDC < 21) {
                                    return 1;
                                }
                                else if (gchDC >= 21) {
                                    if (RsCsDiff < 417) {
                                        if (negBalance < 15) {
                                            return 0;
                                        }
                                        else if (negBalance >= 15) {
                                            if (TSC < 29) {
                                                return 1;
                                            }
                                            else if (TSC >= 29) {
                                                if (diffMVdiffVal < 326) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 326) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 417) {
                                        if (TSC < 36) {
                                            if (diffRsCsdiff < 871) {
                                                return 1;
                                            }
                                            else if (diffRsCsdiff >= 871) {
                                                if (diffRsCsdiff < 916) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 916) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (TSC >= 36) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 33) {
                                if (posBalance < 105) {
                                    if (refDCval < 132) {
                                        if (TSC < 40) {
                                            return 1;
                                        }
                                        else if (TSC >= 40) {
                                            if (diffTSC < 25) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 25) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 132) {
                                        if (Rs < 9511) {
                                            if (gchDC < 84) {
                                                if (MVDiff < 1235) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 1235) {
                                                    if (Rs < 819) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 819) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 84) {
                                                if (AFD < 41) {
                                                    return 0;
                                                }
                                                else if (AFD >= 41) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 9511) {
                                            if (Rs < 10400) {
                                                return 0;
                                            }
                                            else if (Rs >= 10400) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 105) {
                                    if (ssDCval < 168) {
                                        if (MVDiff < 362) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 362) {
                                            if (SC < 2014) {
                                                return 1;
                                            }
                                            else if (SC >= 2014) {
                                                if (Rs < 6578) {
                                                    return 0;
                                                }
                                                else if (Rs >= 6578) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 168) {
                                        if (CsDiff < 10) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 10) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect10(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 6) {
        if (diffMVdiffVal < 53) {
            if (gchDC < 6) {
                if (negBalance < 9) {
                    if (ssDCval < 21) {
                        if (RsCsDiff < 7) {
                            return 0;
                        }
                        else if (RsCsDiff >= 7) {
                            if (SC < 1346) {
                                return 1;
                            }
                            else if (SC >= 1346) {
                                if (RsDiff < 9) {
                                    if (TSC < 1) {
                                        return 0;
                                    }
                                    else if (TSC >= 1) {
                                        if (Cs < 1014) {
                                            return 1;
                                        }
                                        else if (Cs >= 1014) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 9) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 21) {
                        if (SC < 9) {
                            if (Cs < 4) {
                                return 0;
                            }
                            else if (Cs >= 4) {
                                if (MVDiff < 841) {
                                    return 0;
                                }
                                else if (MVDiff >= 841) {
                                    if (refDCval < 23) {
                                        return 0;
                                    }
                                    else if (refDCval >= 23) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (SC >= 9) {
                            return 0;
                        }
                    }
                }
                else if (negBalance >= 9) {
                    if (SC < 87) {
                        if (RsCsDiff < 2) {
                            return 0;
                        }
                        else if (RsCsDiff >= 2) {
                            if (diffRsCsdiff < 1) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 1) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 87) {
                        if (diffMVdiffVal < -123) {
                            if (diffMVdiffVal < -124) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= -124) {
                                if (MVDiff < 126) {
                                    return 0;
                                }
                                else if (MVDiff >= 126) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= -123) {
                            return 0;
                        }
                    }
                }
            }
            else if (gchDC >= 6) {
                if (refDCval < 229) {
                    if (gchDC < 154) {
                        if (AFD < 53) {
                            if (MVDiff < 224) {
                                if (TSC < 24) {
                                    if (diffAFD < 13) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 13) {
                                        if (posBalance < 21) {
                                            return 1;
                                        }
                                        else if (posBalance >= 21) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (TSC >= 24) {
                                    if (RsDiff < 32) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 32) {
                                        if (CsDiff < 30) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 30) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 224) {
                                if (diffTSC < -6) {
                                    return 0;
                                }
                                else if (diffTSC >= -6) {
                                    if (diffRsCsdiff < -96) {
                                        if (MVDiff < 228) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 228) {
                                            if (SC < 133) {
                                                return 1;
                                            }
                                            else if (SC >= 133) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= -96) {
                                        if (diffRsCsdiff < 5) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 5) {
                                            if (diffMVdiffVal < -210) {
                                                if (TSC < 8) {
                                                    return 1;
                                                }
                                                else if (TSC >= 8) {
                                                    if (TSC < 15) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= -210) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 53) {
                            if (negBalance < 74) {
                                if (Cs < 1585) {
                                    return 0;
                                }
                                else if (Cs >= 1585) {
                                    if (Cs < 1912) {
                                        return 1;
                                    }
                                    else if (Cs >= 1912) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 74) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 154) {
                        return 1;
                    }
                }
                else if (refDCval >= 229) {
                    return 1;
                }
            }
        }
        else if (diffMVdiffVal >= 53) {
            if (diffAFD < 8) {
                if (AFD < 47) {
                    if (ssDCval < 31) {
                        if (RsCsDiff < 2) {
                            if (TSCindex < 4) {
                                return 0;
                            }
                            else if (TSCindex >= 4) {
                                if (TSC < 5) {
                                    if (diffMVdiffVal < 219) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 219) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 5) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsCsDiff >= 2) {
                            if (diffRsCsdiff < 12) {
                                if (TSC < 3) {
                                    return 0;
                                }
                                else if (TSC >= 3) {
                                    if (MVDiff < 167) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 167) {
                                        if (gchDC < 10) {
                                            return 1;
                                        }
                                        else if (gchDC >= 10) {
                                            if (diffTSC < -9) {
                                                return 1;
                                            }
                                            else if (diffTSC >= -9) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 12) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 31) {
                        if (diffRsCsdiff < 2302) {
                            if (gchDC < 112) {
                                if (Cs < 378) {
                                    if (ssDCval < 41) {
                                        if (SC < 718) {
                                            if (diffTSC < 1) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 1) {
                                                if (AFD < 9) {
                                                    return 0;
                                                }
                                                else if (AFD >= 9) {
                                                    if (Cs < 149) {
                                                        if (Rs < 7) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 7) {
                                                            if (CsDiff < 1) {
                                                                return 0;
                                                            }
                                                            else if (CsDiff >= 1) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (Cs >= 149) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (SC >= 718) {
                                            if (Rs < 625) {
                                                return 1;
                                            }
                                            else if (Rs >= 625) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 41) {
                                        if (TSC < 15) {
                                            return 0;
                                        }
                                        else if (TSC >= 15) {
                                            if (TSC < 17) {
                                                if (negBalance < 44) {
                                                    if (refDCval < 112) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 112) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 44) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 17) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 378) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 112) {
                                if (TSC < 14) {
                                    return 1;
                                }
                                else if (TSC >= 14) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 2302) {
                            return 1;
                        }
                    }
                }
                else if (AFD >= 47) {
                    if (RsCsDiff < 110) {
                        return 0;
                    }
                    else if (RsCsDiff >= 110) {
                        if (TSC < 34) {
                            if (SC < 1565) {
                                if (posBalance < 58) {
                                    return 1;
                                }
                                else if (posBalance >= 58) {
                                    return 0;
                                }
                            }
                            else if (SC >= 1565) {
                                if (posBalance < 20) {
                                    return 1;
                                }
                                else if (posBalance >= 20) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 34) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (AFD < 15) {
                    if (refDCval < 48) {
                        return 1;
                    }
                    else if (refDCval >= 48) {
                        if (Cs < 27) {
                            return 1;
                        }
                        else if (Cs >= 27) {
                            return 0;
                        }
                    }
                }
                else if (AFD >= 15) {
                    if (negBalance < 42) {
                        if (diffMVdiffVal < 120) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 120) {
                            if (RsDiff < 8) {
                                return 0;
                            }
                            else if (RsDiff >= 8) {
                                if (TSC < 11) {
                                    return 1;
                                }
                                else if (TSC >= 11) {
                                    if (Cs < 152) {
                                        return 1;
                                    }
                                    else if (Cs >= 152) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 42) {
                        if (refDCval < 82) {
                            if (SC < 543) {
                                if (negBalance < 87) {
                                    if (TSC < 15) {
                                        if (gchDC < 6) {
                                            if (diffRsCsdiff < -2) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= -2) {
                                                if (Rs < 278) {
                                                    return 1;
                                                }
                                                else if (Rs >= 278) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 15) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 87) {
                                    if (posBalance < 1) {
                                        if (diffMVdiffVal < 209) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 209) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 1) {
                                        if (CsDiff < 8) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 543) {
                                if (RsCsDiff < 47) {
                                    if (AFD < 30) {
                                        return 0;
                                    }
                                    else if (AFD >= 30) {
                                        if (ssDCval < 51) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 51) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 47) {
                                    if (AFD < 25) {
                                        if (Rs < 452) {
                                            return 0;
                                        }
                                        else if (Rs >= 452) {
                                            if (SC < 2247) {
                                                return 1;
                                            }
                                            else if (SC >= 2247) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 25) {
                                        if (RsCsDiff < 467) {
                                            if (AFD < 39) {
                                                if (TSC < 11) {
                                                    return 0;
                                                }
                                                else if (TSC >= 11) {
                                                    if (RsCsDiff < 105) {
                                                        return 0;
                                                    }
                                                    else if (RsCsDiff >= 105) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 39) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 467) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 82) {
                            if (SC < 4) {
                                return 1;
                            }
                            else if (SC >= 4) {
                                if (gchDC < 14) {
                                    return 0;
                                }
                                else if (gchDC >= 14) {
                                    if (refDCval < 165) {
                                        if (AFD < 47) {
                                            return 0;
                                        }
                                        else if (AFD >= 47) {
                                            if (RsCsDiff < 147) {
                                                if (Scindex < 1) {
                                                    return 1;
                                                }
                                                else if (Scindex >= 1) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsCsDiff >= 147) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 165) {
                                        if (ssDCval < 109) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 109) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 6) {
        if (diffAFD < 21) {
            if (ssDCval < 46) {
                if (diffAFD < 10) {
                    if (RsDiff < 4) {
                        if (negBalance < 64) {
                            return 0;
                        }
                        else if (negBalance >= 64) {
                            if (RsCsDiff < 23) {
                                return 1;
                            }
                            else if (RsCsDiff >= 23) {
                                return 0;
                            }
                        }
                    }
                    else if (RsDiff >= 4) {
                        if (negBalance < 87) {
                            return 1;
                        }
                        else if (negBalance >= 87) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 10) {
                    if (diffMVdiffVal < 61) {
                        if (diffRsCsdiff < 813) {
                            if (Rs < 884) {
                                if (CsDiff < 7) {
                                    if (gchDC < 19) {
                                        return 0;
                                    }
                                    else if (gchDC >= 19) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 7) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 884) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 813) {
                            return 1;
                        }
                    }
                    else if (diffMVdiffVal >= 61) {
                        if (MVDiff < 210) {
                            return 1;
                        }
                        else if (MVDiff >= 210) {
                            if (negBalance < 58) {
                                if (diffMVdiffVal < 191) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 191) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 58) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 46) {
                if (diffMVdiffVal < 74) {
                    if (MVDiff < 14) {
                        return 0;
                    }
                    else if (MVDiff >= 14) {
                        if (diffTSC < 24) {
                            if (RsDiff < 24) {
                                if (MVDiff < 185) {
                                    if (diffMVdiffVal < 25) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 25) {
                                        if (diffRsCsdiff < 124) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 124) {
                                            if (Cs < 1271) {
                                                if (negBalance < 42) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 42) {
                                                    if (diffTSC < 9) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Cs >= 1271) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 185) {
                                    if (ssDCval < 117) {
                                        if (diffTSC < 12) {
                                            if (refDCval < 37) {
                                                if (diffTSC < 9) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 9) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 37) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 12) {
                                            if (posBalance < 104) {
                                                if (AFD < 24) {
                                                    return 1;
                                                }
                                                else if (AFD >= 24) {
                                                    return 0;
                                                }
                                            }
                                            else if (posBalance >= 104) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 117) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsDiff >= 24) {
                                if (SC < 3594) {
                                    return 1;
                                }
                                else if (SC >= 3594) {
                                    if (diffMVdiffVal < 19) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 19) {
                                        if (diffAFD < 12) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 12) {
                                            if (Cs < 3439) {
                                                if (RsCsDiff < 1268) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 1268) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 3439) {
                                                if (diffTSC < 11) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 11) {
                                                    if (ssDCval < 96) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 96) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 24) {
                            if (refDCval < 110) {
                                return 1;
                            }
                            else if (refDCval >= 110) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 74) {
                    if (ssDCval < 125) {
                        if (diffRsCsdiff < 24) {
                            if (TSCindex < 9) {
                                return 1;
                            }
                            else if (TSCindex >= 9) {
                                if (SC < 332) {
                                    if (TSC < 11) {
                                        return 1;
                                    }
                                    else if (TSC >= 11) {
                                        if (diffTSC < 14) {
                                            if (Scindex < 1) {
                                                if (Rs < 52) {
                                                    return 0;
                                                }
                                                else if (Rs >= 52) {
                                                    return 1;
                                                }
                                            }
                                            else if (Scindex >= 1) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 14) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 332) {
                                    if (diffAFD < 17) {
                                        if (AFD < 59) {
                                            return 0;
                                        }
                                        else if (AFD >= 59) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 17) {
                                        if (TSC < 15) {
                                            if (negBalance < 30) {
                                                return 0;
                                            }
                                            else if (negBalance >= 30) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 15) {
                                            if (negBalance < 42) {
                                                return 0;
                                            }
                                            else if (negBalance >= 42) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 24) {
                            if (SC < 1069) {
                                if (posBalance < 3) {
                                    return 0;
                                }
                                else if (posBalance >= 3) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1069) {
                                if (negBalance < 44) {
                                    if (posBalance < 68) {
                                        return 1;
                                    }
                                    else if (posBalance >= 68) {
                                        if (ssDCval < 68) {
                                            if (SC < 2100) {
                                                if (RsCsDiff < 61) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 61) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2100) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 68) {
                                            if (Rs < 4166) {
                                                return 0;
                                            }
                                            else if (Rs >= 4166) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 44) {
                                    if (posBalance < 7) {
                                        return 0;
                                    }
                                    else if (posBalance >= 7) {
                                        if (SC < 4363) {
                                            return 1;
                                        }
                                        else if (SC >= 4363) {
                                            if (diffRsCsdiff < 953) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 953) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 125) {
                        if (SC < 9) {
                            return 1;
                        }
                        else if (SC >= 9) {
                            if (ssDCval < 150) {
                                if (CsDiff < 9) {
                                    return 0;
                                }
                                else if (CsDiff >= 9) {
                                    if (Cs < 1937) {
                                        return 1;
                                    }
                                    else if (Cs >= 1937) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 150) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 21) {
            if (diffMVdiffVal < 63) {
                if (TSC < 16) {
                    if (gchDC < 10) {
                        return 0;
                    }
                    else if (gchDC >= 10) {
                        if (TSC < 14) {
                            return 0;
                        }
                        else if (TSC >= 14) {
                            if (CsDiff < 30) {
                                return 1;
                            }
                            else if (CsDiff >= 30) {
                                return 0;
                            }
                        }
                    }
                }
                else if (TSC >= 16) {
                    if (diffTSC < 40) {
                        if (ssDCval < 66) {
                            if (gchDC < 20) {
                                return 1;
                            }
                            else if (gchDC >= 20) {
                                if (refDCval < 32) {
                                    return 1;
                                }
                                else if (refDCval >= 32) {
                                    if (Cs < 70) {
                                        return 1;
                                    }
                                    else if (Cs >= 70) {
                                        if (diffAFD < 31) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 31) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 66) {
                            if (posBalance < 91) {
                                if (posBalance < 40) {
                                    if (diffMVdiffVal < 56) {
                                        if (RsDiff < 85) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 85) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 56) {
                                        if (Cs < 3397) {
                                            return 1;
                                        }
                                        else if (Cs >= 3397) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (posBalance >= 40) {
                                    if (diffMVdiffVal < -16) {
                                        if (negBalance < 23) {
                                            return 0;
                                        }
                                        else if (negBalance >= 23) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -16) {
                                        if (diffAFD < 26) {
                                            if (AFD < 26) {
                                                return 1;
                                            }
                                            else if (AFD >= 26) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 26) {
                                            if (Cs < 2501) {
                                                return 1;
                                            }
                                            else if (Cs >= 2501) {
                                                if (negBalance < 53) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 53) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 91) {
                                if (RsCsDiff < 267) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 267) {
                                    if (Cs < 571) {
                                        return 1;
                                    }
                                    else if (Cs >= 571) {
                                        if (posBalance < 126) {
                                            return 0;
                                        }
                                        else if (posBalance >= 126) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 40) {
                        if (RsDiff < 9) {
                            return 0;
                        }
                        else if (RsDiff >= 9) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 63) {
                if (negBalance < 12) {
                    if (TSC < 46) {
                        if (refDCval < 54) {
                            if (diffRsCsdiff < 5) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= 5) {
                                if (RsCsDiff < 77) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 77) {
                                    if (Cs < 754) {
                                        return 1;
                                    }
                                    else if (Cs >= 754) {
                                        if (diffMVdiffVal < 96) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 96) {
                                            if (ssDCval < 173) {
                                                if (diffAFD < 53) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 53) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 173) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 54) {
                            if (diffRsCsdiff < 3) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 3) {
                                if (diffTSC < 9) {
                                    return 0;
                                }
                                else if (diffTSC >= 9) {
                                    if (gchDC < 67) {
                                        if (gchDC < 54) {
                                            return 1;
                                        }
                                        else if (gchDC >= 54) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 67) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 46) {
                        return 1;
                    }
                }
                else if (negBalance >= 12) {
                    if (gchDC < 18) {
                        if (diffRsCsdiff < -2) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= -2) {
                            if (diffMVdiffVal < 73) {
                                if (MVDiff < 71) {
                                    return 1;
                                }
                                else if (MVDiff >= 71) {
                                    if (diffTSC < 11) {
                                        if (Rs < 650) {
                                            if (SC < 647) {
                                                return 1;
                                            }
                                            else if (SC >= 647) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 650) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 11) {
                                        if (Rs < 9024) {
                                            return 1;
                                        }
                                        else if (Rs >= 9024) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 73) {
                                if (SC < 1348) {
                                    if (Rs < 1061) {
                                        if (AFD < 43) {
                                            return 1;
                                        }
                                        else if (AFD >= 43) {
                                            if (negBalance < 45) {
                                                if (Cs < 461) {
                                                    return 0;
                                                }
                                                else if (Cs >= 461) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 45) {
                                                if (refDCval < 151) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 151) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 1061) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 1348) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 18) {
                        if (diffAFD < 28) {
                            if (RsDiff < 35) {
                                if (refDCval < 140) {
                                    if (posBalance < 43) {
                                        if (AFD < 34) {
                                            if (SC < 1676) {
                                                return 1;
                                            }
                                            else if (SC >= 1676) {
                                                if (CsDiff < 12) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 12) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 34) {
                                            if (RsDiff < 22) {
                                                if (SC < 101) {
                                                    return 0;
                                                }
                                                else if (SC >= 101) {
                                                    if (negBalance < 84) {
                                                        if (Cs < 1548) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 1548) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (negBalance >= 84) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsDiff >= 22) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 43) {
                                        if (ssDCval < 63) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 63) {
                                            if (diffTSC < 12) {
                                                if (AFD < 43) {
                                                    return 1;
                                                }
                                                else if (AFD >= 43) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 12) {
                                                if (ssDCval < 105) {
                                                    if (RsDiff < 16) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 16) {
                                                        if (negBalance < 39) {
                                                            return 0;
                                                        }
                                                        else if (negBalance >= 39) {
                                                            if (CsDiff < 14) {
                                                                return 0;
                                                            }
                                                            else if (CsDiff >= 14) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 105) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 140) {
                                    if (ssDCval < 136) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 136) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsDiff >= 35) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 28) {
                            if (RsCsDiff < 263) {
                                if (refDCval < 57) {
                                    if (Rs < 1037) {
                                        if (diffRsCsdiff < 70) {
                                            if (Rs < 640) {
                                                if (AFD < 44) {
                                                    return 1;
                                                }
                                                else if (AFD >= 44) {
                                                    if (posBalance < 92) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 92) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 640) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 70) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1037) {
                                        if (TSC < 30) {
                                            if (TSC < 29) {
                                                return 0;
                                            }
                                            else if (TSC >= 29) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 30) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 57) {
                                    if (diffRsCsdiff < -10) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= -10) {
                                        if (Cs < 680) {
                                            if (diffTSC < 40) {
                                                if (ssDCval < 37) {
                                                    if (RsDiff < 4) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 4) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 37) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffTSC >= 40) {
                                                if (RsCsDiff < 26) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 26) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 680) {
                                            if (RsDiff < 14) {
                                                if (RsDiff < 7) {
                                                    if (refDCval < 90) {
                                                        if (negBalance < 44) {
                                                            return 1;
                                                        }
                                                        else if (negBalance >= 44) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (refDCval >= 90) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 14) {
                                                if (TSC < 27) {
                                                    return 1;
                                                }
                                                else if (TSC >= 27) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 263) {
                                if (TSC < 39) {
                                    if (diffMVdiffVal < 455) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= 455) {
                                        if (negBalance < 37) {
                                            if (Rs < 1686) {
                                                return 1;
                                            }
                                            else if (Rs >= 1686) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 37) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 39) {
                                    if (diffMVdiffVal < 171) {
                                        if (diffTSC < 20) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 20) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 171) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect11(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (gchDC < 5) {
        if (diffRsCsdiff < 70) {
            if (diffTSC < 6) {
                if (diffMVdiffVal < 54) {
                    if (negBalance < 1) {
                        if (diffRsCsdiff < 7) {
                            if (TSCindex < 2) {
                                return 0;
                            }
                            else if (TSCindex >= 2) {
                                if (AFD < 2) {
                                    if (posBalance < 68) {
                                        return 1;
                                    }
                                    else if (posBalance >= 68) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 7) {
                            if (ssDCval < 23) {
                                return 1;
                            }
                            else if (ssDCval >= 23) {
                                return 0;
                            }
                        }
                    }
                    else if (negBalance >= 1) {
                        if (diffMVdiffVal < -9) {
                            if (TSC < 11) {
                                return 0;
                            }
                            else if (TSC >= 11) {
                                if (SC < 568) {
                                    if (posBalance < 60) {
                                        return 0;
                                    }
                                    else if (posBalance >= 60) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 568) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= -9) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 54) {
                    if (Cs < 251) {
                        if (TSCindex < 5) {
                            if (diffMVdiffVal < 272) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 272) {
                                if (MVDiff < 356) {
                                    return 1;
                                }
                                else if (MVDiff >= 356) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSCindex >= 5) {
                            if (refDCval < 32) {
                                return 1;
                            }
                            else if (refDCval >= 32) {
                                if (diffAFD < 7) {
                                    if (negBalance < 41) {
                                        if (AFD < 25) {
                                            return 0;
                                        }
                                        else if (AFD >= 25) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 41) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 7) {
                                    if (refDCval < 50) {
                                        if (diffRsCsdiff < -4) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= -4) {
                                            if (Cs < 182) {
                                                return 1;
                                            }
                                            else if (Cs >= 182) {
                                                if (SC < 435) {
                                                    return 0;
                                                }
                                                else if (SC >= 435) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 50) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 251) {
                        return 0;
                    }
                }
            }
            else if (diffTSC >= 6) {
                if (diffAFD < 11) {
                    return 0;
                }
                else if (diffAFD >= 11) {
                    if (posBalance < 62) {
                        if (diffMVdiffVal < 73) {
                            if (AFD < 25) {
                                return 0;
                            }
                            else if (AFD >= 25) {
                                if (Rs < 160) {
                                    return 1;
                                }
                                else if (Rs >= 160) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 73) {
                            if (Rs < 28) {
                                return 0;
                            }
                            else if (Rs >= 28) {
                                return 1;
                            }
                        }
                    }
                    else if (posBalance >= 62) {
                        if (Rs < 1241) {
                            return 1;
                        }
                        else if (Rs >= 1241) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffRsCsdiff >= 70) {
            if (diffAFD < 21) {
                if (MVDiff < 54) {
                    if (ssDCval < 20) {
                        return 1;
                    }
                    else if (ssDCval >= 20) {
                        if (Cs < 535) {
                            if (diffTSC < 11) {
                                return 0;
                            }
                            else if (diffTSC >= 11) {
                                return 1;
                            }
                        }
                        else if (Cs >= 535) {
                            if (refDCval < 44) {
                                if (RsCsDiff < 813) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 813) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 44) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 54) {
                    if (diffMVdiffVal < 58) {
                        if (AFD < 10) {
                            if (diffRsCsdiff < 764) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 764) {
                                return 1;
                            }
                        }
                        else if (AFD >= 10) {
                            if (refDCval < 46) {
                                if (diffMVdiffVal < -64) {
                                    if (RsCsDiff < 110) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 110) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= -64) {
                                    if (CsDiff < 12) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 12) {
                                        if (RsDiff < 12) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 12) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 46) {
                                return 0;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 58) {
                        if (ssDCval < 78) {
                            if (negBalance < 43) {
                                if (ssDCval < 34) {
                                    return 1;
                                }
                                else if (ssDCval >= 34) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 43) {
                                if (RsDiff < 12) {
                                    return 1;
                                }
                                else if (RsDiff >= 12) {
                                    if (SC < 2533) {
                                        return 1;
                                    }
                                    else if (SC >= 2533) {
                                        if (SC < 3200) {
                                            return 0;
                                        }
                                        else if (SC >= 3200) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 78) {
                            if (diffAFD < 17) {
                                return 0;
                            }
                            else if (diffAFD >= 17) {
                                if (diffTSC < 12) {
                                    if (Rs < 1485) {
                                        return 1;
                                    }
                                    else if (Rs >= 1485) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 21) {
                if (MVDiff < 47) {
                    return 0;
                }
                else if (MVDiff >= 47) {
                    if (diffMVdiffVal < 73) {
                        if (refDCval < 115) {
                            if (diffTSC < 11) {
                                if (SC < 1039) {
                                    return 1;
                                }
                                else if (SC >= 1039) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (TSC < 14) {
                                    return 0;
                                }
                                else if (TSC >= 14) {
                                    if (diffAFD < 39) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 39) {
                                        if (SC < 9756) {
                                            return 1;
                                        }
                                        else if (SC >= 9756) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 115) {
                            if (diffTSC < 13) {
                                return 0;
                            }
                            else if (diffTSC >= 13) {
                                if (TSC < 24) {
                                    return 1;
                                }
                                else if (TSC >= 24) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 73) {
                        if (CsDiff < 8) {
                            if (ssDCval < 105) {
                                return 1;
                            }
                            else if (ssDCval >= 105) {
                                if (RsCsDiff < 233) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 233) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 8) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (gchDC >= 5) {
        if (diffMVdiffVal < 65) {
            if (refDCval < 25) {
                if (diffMVdiffVal < 3) {
                    if (ssDCval < 45) {
                        return 0;
                    }
                    else if (ssDCval >= 45) {
                        if (Rs < 399) {
                            return 1;
                        }
                        else if (Rs >= 399) {
                            if (SC < 3093) {
                                return 0;
                            }
                            else if (SC >= 3093) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 3) {
                    return 1;
                }
            }
            else if (refDCval >= 25) {
                if (diffAFD < 13) {
                    if (AFD < 47) {
                        if (TSC < 24) {
                            if (SC < 111) {
                                if (negBalance < 82) {
                                    return 0;
                                }
                                else if (negBalance >= 82) {
                                    if (Rs < 11) {
                                        if (MVDiff < 662) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 662) {
                                            if (ssDCval < 124) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 124) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 11) {
                                        if (MVDiff < 20) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 20) {
                                            if (RsCsDiff < 4) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 4) {
                                                if (Rs < 51) {
                                                    return 1;
                                                }
                                                else if (Rs >= 51) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 111) {
                                if (MVDiff < 224) {
                                    return 0;
                                }
                                else if (MVDiff >= 224) {
                                    if (Cs < 376) {
                                        return 0;
                                    }
                                    else if (Cs >= 376) {
                                        if (Scindex < 4) {
                                            if (ssDCval < 63) {
                                                if (diffTSC < -9) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= -9) {
                                                    if (CsDiff < 5) {
                                                        if (Cs < 378) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 378) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (CsDiff >= 5) {
                                                        if (SC < 1351) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1351) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 63) {
                                                return 0;
                                            }
                                        }
                                        else if (Scindex >= 4) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 24) {
                            if (RsCsDiff < 1303) {
                                if (ssDCval < 64) {
                                    if (Cs < 502) {
                                        return 0;
                                    }
                                    else if (Cs >= 502) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 64) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 1303) {
                                if (SC < 8521) {
                                    return 1;
                                }
                                else if (SC >= 8521) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (AFD >= 47) {
                        if (diffTSC < 3) {
                            if (RsDiff < 9) {
                                if (SC < 40) {
                                    return 1;
                                }
                                else if (SC >= 40) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 9) {
                                if (MVDiff < 258) {
                                    return 0;
                                }
                                else if (MVDiff >= 258) {
                                    if (SC < 3672) {
                                        if (MVDiff < 582) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 582) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 3672) {
                                        if (RsCsDiff < 1022) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 1022) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 3) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 13) {
                    if (RsDiff < 9) {
                        if (posBalance < 95) {
                            if (MVDiff < 107) {
                                return 0;
                            }
                            else if (MVDiff >= 107) {
                                if (TSC < 26) {
                                    if (diffMVdiffVal < -206) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= -206) {
                                        if (negBalance < 68) {
                                            if (posBalance < 89) {
                                                return 0;
                                            }
                                            else if (posBalance >= 89) {
                                                if (CsDiff < 2) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 2) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 68) {
                                            if (Rs < 50) {
                                                return 0;
                                            }
                                            else if (Rs >= 50) {
                                                if (Cs < 827) {
                                                    return 1;
                                                }
                                                else if (Cs >= 827) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 26) {
                                    return 0;
                                }
                            }
                        }
                        else if (posBalance >= 95) {
                            return 0;
                        }
                    }
                    else if (RsDiff >= 9) {
                        if (ssDCval < 77) {
                            if (diffRsCsdiff < 283) {
                                if (negBalance < 30) {
                                    return 0;
                                }
                                else if (negBalance >= 30) {
                                    if (posBalance < 29) {
                                        if (gchDC < 101) {
                                            return 0;
                                        }
                                        else if (gchDC >= 101) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 29) {
                                        if (TSC < 10) {
                                            return 0;
                                        }
                                        else if (TSC >= 10) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 283) {
                                if (CsDiff < 33) {
                                    return 1;
                                }
                                else if (CsDiff >= 33) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 77) {
                            if (MVDiff < 222) {
                                if (RsDiff < 108) {
                                    if (diffAFD < 37) {
                                        if (Cs < 4333) {
                                            return 0;
                                        }
                                        else if (Cs >= 4333) {
                                            if (Rs < 3473) {
                                                return 1;
                                            }
                                            else if (Rs >= 3473) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 37) {
                                        if (RsDiff < 21) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 21) {
                                            if (Rs < 4829) {
                                                return 1;
                                            }
                                            else if (Rs >= 4829) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 108) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 222) {
                                if (MVDiff < 289) {
                                    return 1;
                                }
                                else if (MVDiff >= 289) {
                                    if (MVDiff < 401) {
                                        if (ssDCval < 99) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 99) {
                                            if (RsCsDiff < 273) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 273) {
                                                if (posBalance < 100) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 100) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 401) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffTSC < 7) {
                if (diffAFD < 8) {
                    if (refDCval < 44) {
                        if (diffRsCsdiff < 11) {
                            if (gchDC < 16) {
                                return 0;
                            }
                            else if (gchDC >= 16) {
                                if (AFD < 19) {
                                    if (refDCval < 42) {
                                        if (AFD < 13) {
                                            return 1;
                                        }
                                        else if (AFD >= 13) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 42) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 19) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 11) {
                            if (diffMVdiffVal < 100) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= 100) {
                                if (posBalance < 50) {
                                    return 1;
                                }
                                else if (posBalance >= 50) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 44) {
                        if (AFD < 58) {
                            if (gchDC < 123) {
                                if (diffRsCsdiff < -38) {
                                    if (AFD < 40) {
                                        if (diffAFD < 7) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 40) {
                                        if (Rs < 926) {
                                            if (CsDiff < 4) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (TSC < 15) {
                                                    return 0;
                                                }
                                                else if (TSC >= 15) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Rs >= 926) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= -38) {
                                    if (diffTSC < -8) {
                                        if (diffAFD < 3) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= -8) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 123) {
                                if (ssDCval < 24) {
                                    return 1;
                                }
                                else if (ssDCval >= 24) {
                                    if (CsDiff < 7) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 7) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 58) {
                            if (Rs < 1544) {
                                return 0;
                            }
                            else if (Rs >= 1544) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffAFD >= 8) {
                    if (Cs < 211) {
                        if (Rs < 157) {
                            if (ssDCval < 34) {
                                if (diffRsCsdiff < 7) {
                                    return 1;
                                }
                                else if (diffRsCsdiff >= 7) {
                                    if (ssDCval < 19) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 19) {
                                        if (TSC < 11) {
                                            return 1;
                                        }
                                        else if (TSC >= 11) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 34) {
                                if (CsDiff < 4) {
                                    if (gchDC < 14) {
                                        if (Cs < 26) {
                                            return 0;
                                        }
                                        else if (Cs >= 26) {
                                            if (TSC < 5) {
                                                return 0;
                                            }
                                            else if (TSC >= 5) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 14) {
                                        if (MVDiff < 390) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 390) {
                                            if (RsDiff < 4) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 4) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    if (diffTSC < -5) {
                                        return 0;
                                    }
                                    else if (diffTSC >= -5) {
                                        if (MVDiff < 672) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 672) {
                                            if (AFD < 37) {
                                                return 1;
                                            }
                                            else if (AFD >= 37) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 157) {
                            if (refDCval < 49) {
                                if (RsCsDiff < 27) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 27) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 49) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 211) {
                        if (diffMVdiffVal < 94) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 94) {
                            if (MVDiff < 135) {
                                return 1;
                            }
                            else if (MVDiff >= 135) {
                                if (ssDCval < 62) {
                                    if (RsCsDiff < 150) {
                                        if (gchDC < 9) {
                                            return 1;
                                        }
                                        else if (gchDC >= 9) {
                                            if (Cs < 243) {
                                                if (AFD < 38) {
                                                    return 1;
                                                }
                                                else if (AFD >= 38) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 243) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 150) {
                                        if (Cs < 477) {
                                            if (RsDiff < 11) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 11) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 477) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 62) {
                                    if (RsDiff < 40) {
                                        if (MVDiff < 458) {
                                            if (gchDC < 7) {
                                                if (diffTSC < 4) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 4) {
                                                    if (Rs < 1703) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1703) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 7) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 458) {
                                            if (AFD < 59) {
                                                if (diffMVdiffVal < 517) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 517) {
                                                    if (MVDiff < 653) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 653) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 59) {
                                                if (Rs < 1750) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1750) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 40) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 7) {
                if (diffAFD < 20) {
                    if (ssDCval < 47) {
                        if (SC < 104) {
                            if (diffAFD < 9) {
                                return 0;
                            }
                            else if (diffAFD >= 9) {
                                return 1;
                            }
                        }
                        else if (SC >= 104) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 47) {
                        if (refDCval < 129) {
                            if (diffAFD < 14) {
                                if (gchDC < 22) {
                                    return 0;
                                }
                                else if (gchDC >= 22) {
                                    if (Cs < 741) {
                                        if (refDCval < 56) {
                                            return 0;
                                        }
                                        else if (refDCval >= 56) {
                                            if (RsCsDiff < 37) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 37) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 741) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 14) {
                                if (negBalance < 42) {
                                    if (gchDC < 37) {
                                        if (TSC < 11) {
                                            return 1;
                                        }
                                        else if (TSC >= 11) {
                                            if (negBalance < 17) {
                                                return 0;
                                            }
                                            else if (negBalance >= 17) {
                                                if (refDCval < 54) {
                                                    if (diffMVdiffVal < 128) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 128) {
                                                        return 1;
                                                    }
                                                }
                                                else if (refDCval >= 54) {
                                                    if (diffMVdiffVal < 104) {
                                                        if (diffAFD < 17) {
                                                            return 0;
                                                        }
                                                        else if (diffAFD >= 17) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 104) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 37) {
                                        if (Cs < 406) {
                                            if (diffRsCsdiff < 0) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 0) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 406) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (negBalance >= 42) {
                                    if (Rs < 3044) {
                                        if (Cs < 557) {
                                            if (AFD < 37) {
                                                return 1;
                                            }
                                            else if (AFD >= 37) {
                                                if (TSC < 16) {
                                                    return 0;
                                                }
                                                else if (TSC >= 16) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 557) {
                                            if (MVDiff < 138) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 138) {
                                                if (ssDCval < 79) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 79) {
                                                    if (Cs < 1781) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1781) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 3044) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 129) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 20) {
                    if (RsCsDiff < 460) {
                        if (negBalance < 30) {
                            if (ssDCval < 94) {
                                if (MVDiff < 191) {
                                    if (diffTSC < 20) {
                                        if (diffMVdiffVal < 78) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 78) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 20) {
                                        if (diffAFD < 41) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 41) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 191) {
                                    if (diffMVdiffVal < 199) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 199) {
                                        if (negBalance < 21) {
                                            if (TSC < 18) {
                                                return 1;
                                            }
                                            else if (TSC >= 18) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 21) {
                                            if (MVDiff < 1504) {
                                                if (Rs < 908) {
                                                    return 1;
                                                }
                                                else if (Rs >= 908) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 1504) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 94) {
                                if (diffRsCsdiff < 29) {
                                    if (SC < 671) {
                                        if (ssDCval < 142) {
                                            if (RsCsDiff < 13) {
                                                if (AFD < 36) {
                                                    return 1;
                                                }
                                                else if (AFD >= 36) {
                                                    if (Rs < 33) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 33) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 13) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 142) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 671) {
                                        if (refDCval < 78) {
                                            return 0;
                                        }
                                        else if (refDCval >= 78) {
                                            if (Rs < 1232) {
                                                return 1;
                                            }
                                            else if (Rs >= 1232) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 29) {
                                    if (diffMVdiffVal < 102) {
                                        if (diffAFD < 37) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 37) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 102) {
                                        if (SC < 1702) {
                                            return 1;
                                        }
                                        else if (SC >= 1702) {
                                            if (RsDiff < 7) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 7) {
                                                if (SC < 2274) {
                                                    return 1;
                                                }
                                                else if (SC >= 2274) {
                                                    if (gchDC < 76) {
                                                        if (diffAFD < 41) {
                                                            return 0;
                                                        }
                                                        else if (diffAFD >= 41) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 76) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 30) {
                            if (AFD < 46) {
                                if (Cs < 3430) {
                                    if (MVDiff < 76) {
                                        if (SC < 865) {
                                            return 0;
                                        }
                                        else if (SC >= 865) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 76) {
                                        if (CsDiff < 6) {
                                            if (Rs < 1705) {
                                                if (diffMVdiffVal < 129) {
                                                    if (Cs < 91) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 91) {
                                                        if (refDCval < 41) {
                                                            return 0;
                                                        }
                                                        else if (refDCval >= 41) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 129) {
                                                    if (posBalance < 13) {
                                                        if (RsCsDiff < 24) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 24) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (posBalance >= 13) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1705) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (MVDiff < 97) {
                                                if (Cs < 1059) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1059) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 97) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 3430) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 46) {
                                if (CsDiff < 7) {
                                    if (diffAFD < 42) {
                                        if (MVDiff < 166) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 166) {
                                            if (diffMVdiffVal < 159) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 159) {
                                                if (diffAFD < 26) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 26) {
                                                    if (negBalance < 50) {
                                                        if (Cs < 519) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 519) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (negBalance >= 50) {
                                                        if (Rs < 137) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 137) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 42) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 7) {
                                    if (diffMVdiffVal < 125) {
                                        if (RsCsDiff < 156) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 156) {
                                            if (diffAFD < 31) {
                                                if (TSC < 29) {
                                                    if (RsDiff < 11) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 11) {
                                                        return 1;
                                                    }
                                                }
                                                else if (TSC >= 29) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 31) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 125) {
                                        if (diffAFD < 29) {
                                            if (Cs < 897) {
                                                if (Rs < 168) {
                                                    return 0;
                                                }
                                                else if (Rs >= 168) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 897) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 29) {
                                            if (RsDiff < 5) {
                                                if (Rs < 191) {
                                                    return 0;
                                                }
                                                else if (Rs >= 191) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 5) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 460) {
                        if (AFD < 42) {
                            if (diffAFD < 24) {
                                if (RsDiff < 24) {
                                    if (refDCval < 46) {
                                        return 0;
                                    }
                                    else if (refDCval >= 46) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 24) {
                                    if (Cs < 3157) {
                                        return 0;
                                    }
                                    else if (Cs >= 3157) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffAFD >= 24) {
                                if (RsCsDiff < 3210) {
                                    if (diffMVdiffVal < 229) {
                                        if (AFD < 41) {
                                            return 1;
                                        }
                                        else if (AFD >= 41) {
                                            if (negBalance < 36) {
                                                return 0;
                                            }
                                            else if (negBalance >= 36) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 229) {
                                        if (diffMVdiffVal < 233) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 233) {
                                            if (diffAFD < 28) {
                                                if (AFD < 38) {
                                                    return 1;
                                                }
                                                else if (AFD >= 38) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 28) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 3210) {
                                    if (gchDC < 54) {
                                        return 1;
                                    }
                                    else if (gchDC >= 54) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 42) {
                            if (diffMVdiffVal < 458) {
                                if (posBalance < 17) {
                                    if (Rs < 1571) {
                                        return 0;
                                    }
                                    else if (Rs >= 1571) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 17) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 458) {
                                if (negBalance < 39) {
                                    return 0;
                                }
                                else if (negBalance >= 39) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect12(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 14) {
        if (diffMVdiffVal < 52) {
            if (RsCsDiff < 837) {
                if (MVDiff < 200) {
                    if (Rs < 1148) {
                        if (refDCval < 24) {
                            if (diffRsCsdiff < 5) {
                                if (MVDiff < 6) {
                                    if (gchDC < 109) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 109) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 6) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 5) {
                                if (SC < 1207) {
                                    if (RsDiff < 10) {
                                        if (gchDC < 23) {
                                            return 0;
                                        }
                                        else if (gchDC >= 23) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 10) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1207) {
                                    if (RsCsDiff < 17) {
                                        if (ssDCval < 17) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 17) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 17) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 24) {
                            if (gchDC < 31) {
                                if (TSCindex < 9) {
                                    return 0;
                                }
                                else if (TSCindex >= 9) {
                                    if (Rs < 1) {
                                        return 1;
                                    }
                                    else if (Rs >= 1) {
                                        if (diffRsCsdiff < 48) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 48) {
                                            if (SC < 705) {
                                                return 1;
                                            }
                                            else if (SC >= 705) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 31) {
                                if (Cs < 52) {
                                    if (Cs < 30) {
                                        if (RsCsDiff < 325) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 325) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 30) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 52) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (Rs >= 1148) {
                        return 0;
                    }
                }
                else if (MVDiff >= 200) {
                    if (gchDC < 1) {
                        return 0;
                    }
                    else if (gchDC >= 1) {
                        if (diffRsCsdiff < 41) {
                            if (negBalance < 1) {
                                if (refDCval < 24) {
                                    if (diffAFD < -25) {
                                        return 0;
                                    }
                                    else if (diffAFD >= -25) {
                                        if (Cs < 2) {
                                            if (AFD < 3) {
                                                return 0;
                                            }
                                            else if (AFD >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 2) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 24) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 1) {
                                if (Rs < 2) {
                                    if (Cs < 53) {
                                        if (MVDiff < 662) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 662) {
                                            if (TSC < 4) {
                                                return 0;
                                            }
                                            else if (TSC >= 4) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 53) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 2) {
                                    if (TSC < 16) {
                                        return 0;
                                    }
                                    else if (TSC >= 16) {
                                        if (posBalance < 29) {
                                            return 0;
                                        }
                                        else if (posBalance >= 29) {
                                            if (diffMVdiffVal < -205) {
                                                if (posBalance < 76) {
                                                    if (RsDiff < 3) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 76) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -205) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 41) {
                            if (refDCval < 51) {
                                if (Cs < 855) {
                                    if (SC < 1438) {
                                        return 1;
                                    }
                                    else if (SC >= 1438) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 855) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 51) {
                                if (AFD < 57) {
                                    return 0;
                                }
                                else if (AFD >= 57) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (RsCsDiff >= 837) {
                if (ssDCval < 47) {
                    if (Rs < 1198) {
                        return 0;
                    }
                    else if (Rs >= 1198) {
                        return 1;
                    }
                }
                else if (ssDCval >= 47) {
                    if (MVDiff < 140) {
                        if (TSC < 24) {
                            return 0;
                        }
                        else if (TSC >= 24) {
                            if (Cs < 2454) {
                                return 1;
                            }
                            else if (Cs >= 2454) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 140) {
                        if (TSC < 24) {
                            return 0;
                        }
                        else if (TSC >= 24) {
                            if (Rs < 2620) {
                                if (Cs < 1912) {
                                    return 1;
                                }
                                else if (Cs >= 1912) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 2620) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 52) {
            if (diffAFD < 7) {
                if (diffRsCsdiff < 5) {
                    if (gchDC < 51) {
                        if (diffAFD < 6) {
                            if (RsCsDiff < 101) {
                                if (negBalance < 81) {
                                    return 0;
                                }
                                else if (negBalance >= 81) {
                                    if (ssDCval < 24) {
                                        if (Cs < 36) {
                                            return 0;
                                        }
                                        else if (Cs >= 36) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 24) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 101) {
                                if (Rs < 156) {
                                    return 1;
                                }
                                else if (Rs >= 156) {
                                    if (diffRsCsdiff < -4) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= -4) {
                                        if (AFD < 34) {
                                            return 0;
                                        }
                                        else if (AFD >= 34) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 6) {
                            if (refDCval < 27) {
                                return 1;
                            }
                            else if (refDCval >= 27) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 51) {
                        if (Rs < 1) {
                            return 1;
                        }
                        else if (Rs >= 1) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 5) {
                    if (diffMVdiffVal < 160) {
                        if (refDCval < 28) {
                            return 1;
                        }
                        else if (refDCval >= 28) {
                            if (RsCsDiff < 423) {
                                return 0;
                            }
                            else if (RsCsDiff >= 423) {
                                if (Cs < 763) {
                                    if (MVDiff < 143) {
                                        if (SC < 5269) {
                                            return 1;
                                        }
                                        else if (SC >= 5269) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 143) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 763) {
                                    if (MVDiff < 238) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 238) {
                                        if (ssDCval < 56) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 56) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 160) {
                        if (refDCval < 31) {
                            return 1;
                        }
                        else if (refDCval >= 31) {
                            if (gchDC < 33) {
                                if (negBalance < 72) {
                                    return 0;
                                }
                                else if (negBalance >= 72) {
                                    if (posBalance < 44) {
                                        return 0;
                                    }
                                    else if (posBalance >= 44) {
                                        if (gchDC < 8) {
                                            return 0;
                                        }
                                        else if (gchDC >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 33) {
                                if (AFD < 34) {
                                    if (TSC < 15) {
                                        return 1;
                                    }
                                    else if (TSC >= 15) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 34) {
                                    if (negBalance < 54) {
                                        if (Cs < 190) {
                                            return 1;
                                        }
                                        else if (Cs >= 190) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 7) {
                if (AFD < 15) {
                    if (gchDC < 7) {
                        if (diffMVdiffVal < 65) {
                            if (MVDiff < 85) {
                                return 1;
                            }
                            else if (MVDiff >= 85) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 65) {
                            if (SC < 490) {
                                return 1;
                            }
                            else if (SC >= 490) {
                                if (Cs < 324) {
                                    return 0;
                                }
                                else if (Cs >= 324) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 7) {
                        if (diffTSC < 3) {
                            return 0;
                        }
                        else if (diffTSC >= 3) {
                            if (Scindex < 2) {
                                if (gchDC < 15) {
                                    if (Cs < 55) {
                                        return 0;
                                    }
                                    else if (Cs >= 55) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 15) {
                                    return 1;
                                }
                            }
                            else if (Scindex >= 2) {
                                if (diffAFD < 11) {
                                    return 0;
                                }
                                else if (diffAFD >= 11) {
                                    if (Cs < 264) {
                                        return 1;
                                    }
                                    else if (Cs >= 264) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (AFD >= 15) {
                    if (Cs < 211) {
                        if (Rs < 102) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (TSC < 24) {
                                    if (diffRsCsdiff < 8) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            if (refDCval < 46) {
                                                if (SC < 159) {
                                                    return 1;
                                                }
                                                else if (SC >= 159) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 46) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 8) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 24) {
                                    return 0;
                                }
                            }
                        }
                        else if (Rs >= 102) {
                            if (CsDiff < 3) {
                                return 0;
                            }
                            else if (CsDiff >= 3) {
                                if (TSC < 11) {
                                    return 1;
                                }
                                else if (TSC >= 11) {
                                    if (posBalance < 54) {
                                        return 0;
                                    }
                                    else if (posBalance >= 54) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 211) {
                        if (ssDCval < 38) {
                            if (Rs < 447) {
                                if (RsDiff < 3) {
                                    return 1;
                                }
                                else if (RsDiff >= 3) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 447) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 38) {
                            if (MVDiff < 347) {
                                if (diffTSC < 12) {
                                    if (ssDCval < 69) {
                                        if (diffAFD < 12) {
                                            if (gchDC < 7) {
                                                if (ssDCval < 66) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 66) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 7) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 12) {
                                            if (negBalance < 36) {
                                                return 0;
                                            }
                                            else if (negBalance >= 36) {
                                                if (posBalance < 41) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 41) {
                                                    if (SC < 491) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 491) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 69) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    if (MVDiff < 170) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 170) {
                                        if (Rs < 3064) {
                                            if (RsCsDiff < 505) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 505) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 3064) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 347) {
                                if (SC < 1674) {
                                    if (posBalance < 77) {
                                        if (diffMVdiffVal < 86) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 86) {
                                            if (TSC < 14) {
                                                if (Cs < 231) {
                                                    return 0;
                                                }
                                                else if (Cs >= 231) {
                                                    if (refDCval < 87) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 87) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 14) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 77) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 1674) {
                                    if (negBalance < 60) {
                                        if (RsCsDiff < 124) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 124) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 60) {
                                        if (Rs < 1665) {
                                            return 1;
                                        }
                                        else if (Rs >= 1665) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 14) {
        if (AFD < 37) {
            if (diffMVdiffVal < 52) {
                if (TSC < 33) {
                    if (gchDC < 2) {
                        return 0;
                    }
                    else if (gchDC >= 2) {
                        if (RsDiff < 55) {
                            if (MVDiff < 132) {
                                if (refDCval < 31) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 31) {
                                    if (negBalance < 80) {
                                        if (Rs < 213) {
                                            if (RsDiff < 3) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 213) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 80) {
                                        if (AFD < 18) {
                                            if (gchDC < 12) {
                                                return 1;
                                            }
                                            else if (gchDC >= 12) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 132) {
                                if (diffAFD < 31) {
                                    if (gchDC < 5) {
                                        return 1;
                                    }
                                    else if (gchDC >= 5) {
                                        if (Scindex < 6) {
                                            if (diffMVdiffVal < -532) {
                                                if (AFD < 31) {
                                                    return 1;
                                                }
                                                else if (AFD >= 31) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -532) {
                                                if (diffRsCsdiff < 188) {
                                                    if (TSC < 8) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 8) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 188) {
                                                    if (TSC < 20) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 20) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Scindex >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 31) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 55) {
                            if (CsDiff < 37) {
                                return 1;
                            }
                            else if (CsDiff >= 37) {
                                if (CsDiff < 75) {
                                    return 0;
                                }
                                else if (CsDiff >= 75) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (TSC >= 33) {
                    return 1;
                }
            }
            else if (diffMVdiffVal >= 52) {
                if (ssDCval < 66) {
                    if (gchDC < 15) {
                        if (diffRsCsdiff < 86) {
                            if (TSC < 12) {
                                if (Rs < 235) {
                                    return 1;
                                }
                                else if (Rs >= 235) {
                                    if (MVDiff < 122) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 122) {
                                        if (Cs < 325) {
                                            return 0;
                                        }
                                        else if (Cs >= 325) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 12) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 86) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 15) {
                        if (negBalance < 22) {
                            if (refDCval < 34) {
                                if (CsDiff < 7) {
                                    if (RsCsDiff < 51) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 51) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 7) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 34) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 22) {
                            if (Rs < 1921) {
                                if (diffMVdiffVal < 63) {
                                    if (Cs < 797) {
                                        if (negBalance < 46) {
                                            return 0;
                                        }
                                        else if (negBalance >= 46) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 797) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 63) {
                                    if (diffAFD < 23) {
                                        if (RsCsDiff < 128) {
                                            if (RsDiff < 5) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 5) {
                                                if (ssDCval < 45) {
                                                    if (Cs < 21) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 21) {
                                                        if (Cs < 596) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 596) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 45) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 128) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 23) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 1921) {
                                if (SC < 4972) {
                                    return 0;
                                }
                                else if (SC >= 4972) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 66) {
                    if (diffAFD < 22) {
                        if (diffMVdiffVal < 94) {
                            if (negBalance < 42) {
                                return 0;
                            }
                            else if (negBalance >= 42) {
                                if (AFD < 31) {
                                    if (refDCval < 64) {
                                        return 0;
                                    }
                                    else if (refDCval >= 64) {
                                        if (Cs < 233) {
                                            return 0;
                                        }
                                        else if (Cs >= 233) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 31) {
                                    if (TSC < 20) {
                                        return 0;
                                    }
                                    else if (TSC >= 20) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 94) {
                            if (gchDC < 12) {
                                return 1;
                            }
                            else if (gchDC >= 12) {
                                if (SC < 775) {
                                    if (CsDiff < 4) {
                                        if (TSC < 11) {
                                            return 1;
                                        }
                                        else if (TSC >= 11) {
                                            if (negBalance < 61) {
                                                return 0;
                                            }
                                            else if (negBalance >= 61) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 4) {
                                        if (diffMVdiffVal < 98) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 98) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 775) {
                                    if (diffRsCsdiff < 183) {
                                        if (TSC < 19) {
                                            return 0;
                                        }
                                        else if (TSC >= 19) {
                                            if (AFD < 35) {
                                                if (RsDiff < 4) {
                                                    if (AFD < 32) {
                                                        return 1;
                                                    }
                                                    else if (AFD >= 32) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 35) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 183) {
                                        if (Rs < 2473) {
                                            return 1;
                                        }
                                        else if (Rs >= 2473) {
                                            if (AFD < 29) {
                                                return 1;
                                            }
                                            else if (AFD >= 29) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 22) {
                        if (refDCval < 50) {
                            if (negBalance < 3) {
                                return 0;
                            }
                            else if (negBalance >= 3) {
                                if (SC < 1394) {
                                    return 1;
                                }
                                else if (SC >= 1394) {
                                    if (Cs < 2716) {
                                        return 0;
                                    }
                                    else if (Cs >= 2716) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 50) {
                            if (diffMVdiffVal < 135) {
                                if (ssDCval < 143) {
                                    if (Rs < 237) {
                                        if (Rs < 124) {
                                            return 1;
                                        }
                                        else if (Rs >= 124) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 237) {
                                        if (negBalance < 27) {
                                            if (MVDiff < 67) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 67) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 27) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 143) {
                                    if (posBalance < 90) {
                                        if (MVDiff < 75) {
                                            if (ssDCval < 167) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 167) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 75) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 90) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 135) {
                                if (Rs < 4773) {
                                    return 1;
                                }
                                else if (Rs >= 4773) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (AFD >= 37) {
            if (diffAFD < 29) {
                if (RsCsDiff < 148) {
                    if (gchDC < 26) {
                        if (diffRsCsdiff < 68) {
                            if (refDCval < 91) {
                                if (diffTSC < 16) {
                                    if (refDCval < 87) {
                                        return 0;
                                    }
                                    else if (refDCval >= 87) {
                                        if (Cs < 53) {
                                            return 1;
                                        }
                                        else if (Cs >= 53) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 16) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 91) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 68) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 26) {
                        return 0;
                    }
                }
                else if (RsCsDiff >= 148) {
                    if (posBalance < 78) {
                        if (diffTSC < 8) {
                            if (ssDCval < 67) {
                                return 1;
                            }
                            else if (ssDCval >= 67) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 8) {
                            if (MVDiff < 301) {
                                if (CsDiff < 27) {
                                    if (AFD < 38) {
                                        if (SC < 2526) {
                                            return 1;
                                        }
                                        else if (SC >= 2526) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 38) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 27) {
                                    if (refDCval < 129) {
                                        if (SC < 4291) {
                                            if (diffMVdiffVal < 156) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 156) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 4291) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 129) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 301) {
                                if (MVDiff < 473) {
                                    if (posBalance < 49) {
                                        if (TSC < 50) {
                                            if (AFD < 45) {
                                                if (Cs < 824) {
                                                    return 1;
                                                }
                                                else if (Cs >= 824) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 45) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 50) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 49) {
                                        if (TSC < 33) {
                                            return 1;
                                        }
                                        else if (TSC >= 33) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 473) {
                                    if (posBalance < 68) {
                                        return 1;
                                    }
                                    else if (posBalance >= 68) {
                                        if (gchDC < 11) {
                                            return 1;
                                        }
                                        else if (gchDC >= 11) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 78) {
                        if (diffRsCsdiff < 354) {
                            if (ssDCval < 91) {
                                if (MVDiff < 240) {
                                    return 1;
                                }
                                else if (MVDiff >= 240) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 91) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 354) {
                            if (Rs < 2263) {
                                return 1;
                            }
                            else if (Rs >= 2263) {
                                if (Cs < 2172) {
                                    return 0;
                                }
                                else if (Cs >= 2172) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 29) {
                if (diffMVdiffVal < 49) {
                    if (RsDiff < 6) {
                        return 0;
                    }
                    else if (RsDiff >= 6) {
                        if (TSC < 23) {
                            if (SC < 2301) {
                                if (gchDC < 18) {
                                    return 1;
                                }
                                else if (gchDC >= 18) {
                                    return 0;
                                }
                            }
                            else if (SC >= 2301) {
                                return 0;
                            }
                        }
                        else if (TSC >= 23) {
                            if (diffTSC < 39) {
                                return 1;
                            }
                            else if (diffTSC >= 39) {
                                if (ssDCval < 122) {
                                    return 1;
                                }
                                else if (ssDCval >= 122) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 49) {
                    if (RsCsDiff < 37) {
                        if (TSC < 32) {
                            if (ssDCval < 55) {
                                return 0;
                            }
                            else if (ssDCval >= 55) {
                                if (CsDiff < 2) {
                                    return 0;
                                }
                                else if (CsDiff >= 2) {
                                    return 1;
                                }
                            }
                        }
                        else if (TSC >= 32) {
                            if (MVDiff < 974) {
                                if (posBalance < 59) {
                                    if (AFD < 64) {
                                        return 0;
                                    }
                                    else if (AFD >= 64) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 59) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 974) {
                                return 1;
                            }
                        }
                    }
                    else if (RsCsDiff >= 37) {
                        if (refDCval < 54) {
                            if (negBalance < 12) {
                                if (diffAFD < 55) {
                                    return 0;
                                }
                                else if (diffAFD >= 55) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 12) {
                                if (TSC < 60) {
                                    if (negBalance < 43) {
                                        if (negBalance < 42) {
                                            if (RsCsDiff < 255) {
                                                if (Rs < 1342) {
                                                    if (MVDiff < 757) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 757) {
                                                        if (SC < 1086) {
                                                            return 1;
                                                        }
                                                        else if (SC >= 1086) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 1342) {
                                                    if (AFD < 46) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 46) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (RsCsDiff >= 255) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 42) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 60) {
                                    return 0;
                                }
                            }
                        }
                        else if (refDCval >= 54) {
                            if (refDCval < 132) {
                                if (posBalance < 75) {
                                    if (diffRsCsdiff < 218) {
                                        if (diffMVdiffVal < 81) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 81) {
                                            if (RsDiff < 14) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 14) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 218) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 75) {
                                    if (diffAFD < 38) {
                                        if (SC < 1736) {
                                            return 1;
                                        }
                                        else if (SC >= 1736) {
                                            if (ssDCval < 117) {
                                                if (RsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 117) {
                                                if (CsDiff < 24) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 24) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 38) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 132) {
                                if (ssDCval < 156) {
                                    if (posBalance < 72) {
                                        if (diffRsCsdiff < 53) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 53) {
                                            if (SC < 6154) {
                                                return 1;
                                            }
                                            else if (SC >= 6154) {
                                                if (Cs < 6832) {
                                                    if (Rs < 3561) {
                                                        if (AFD < 55) {
                                                            return 0;
                                                        }
                                                        else if (AFD >= 55) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 3561) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 6832) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 72) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 156) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect13(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (diffMVdiffVal < 71) {
            if (refDCval < 24) {
                if (gchDC < 2) {
                    if (diffRsCsdiff < 204) {
                        if (diffRsCsdiff < 7) {
                            if (diffAFD < -21) {
                                if (ssDCval < 24) {
                                    return 0;
                                }
                                else if (ssDCval >= 24) {
                                    if (Cs < 10) {
                                        return 1;
                                    }
                                    else if (Cs >= 10) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= -21) {
                                if (CsDiff < 3) {
                                    return 0;
                                }
                                else if (CsDiff >= 3) {
                                    if (MVDiff < 6) {
                                        if (RsDiff < 2) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 2) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 6) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 7) {
                            if (ssDCval < 18) {
                                return 1;
                            }
                            else if (ssDCval >= 18) {
                                return 0;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 204) {
                        return 1;
                    }
                }
                else if (gchDC >= 2) {
                    if (SC < 1197) {
                        if (AFD < 1) {
                            return 1;
                        }
                        else if (AFD >= 1) {
                            return 0;
                        }
                    }
                    else if (SC >= 1197) {
                        return 1;
                    }
                }
            }
            else if (refDCval >= 24) {
                if (MVDiff < 55) {
                    if (gchDC < 9) {
                        return 0;
                    }
                    else if (gchDC >= 9) {
                        if (Cs < 48) {
                            if (diffMVdiffVal < -235) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= -235) {
                                if (ssDCval < 213) {
                                    if (gchDC < 39) {
                                        return 0;
                                    }
                                    else if (gchDC >= 39) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 213) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 48) {
                            return 0;
                        }
                    }
                }
                else if (MVDiff >= 55) {
                    if (gchDC < 15) {
                        if (diffAFD < 9) {
                            if (MVDiff < 175) {
                                if (RsDiff < 37) {
                                    if (AFD < 40) {
                                        return 0;
                                    }
                                    else if (AFD >= 40) {
                                        if (refDCval < 133) {
                                            return 0;
                                        }
                                        else if (refDCval >= 133) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsDiff >= 37) {
                                    if (Rs < 2335) {
                                        if (SC < 3109) {
                                            return 0;
                                        }
                                        else if (SC >= 3109) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2335) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 175) {
                                if (refDCval < 30) {
                                    if (refDCval < 28) {
                                        return 0;
                                    }
                                    else if (refDCval >= 28) {
                                        if (gchDC < 3) {
                                            return 0;
                                        }
                                        else if (gchDC >= 3) {
                                            if (AFD < 14) {
                                                return 1;
                                            }
                                            else if (AFD >= 14) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 30) {
                                    if (TSC < 14) {
                                        return 0;
                                    }
                                    else if (TSC >= 14) {
                                        if (RsCsDiff < 323) {
                                            if (diffMVdiffVal < -10) {
                                                if (diffRsCsdiff < 216) {
                                                    if (ssDCval < 59) {
                                                        if (Cs < 388) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 388) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 59) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 216) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= -10) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 323) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 9) {
                            if (ssDCval < 44) {
                                if (diffTSC < 3) {
                                    return 0;
                                }
                                else if (diffTSC >= 3) {
                                    if (negBalance < 41) {
                                        return 0;
                                    }
                                    else if (negBalance >= 41) {
                                        if (AFD < 28) {
                                            if (MVDiff < 128) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 128) {
                                                if (RsDiff < 6) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 6) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 28) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 44) {
                                if (Cs < 2369) {
                                    return 0;
                                }
                                else if (Cs >= 2369) {
                                    if (refDCval < 82) {
                                        if (SC < 6502) {
                                            return 0;
                                        }
                                        else if (SC >= 6502) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 82) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 15) {
                        if (CsDiff < 19) {
                            if (diffAFD < 13) {
                                if (ssDCval < 62) {
                                    if (TSC < 26) {
                                        if (AFD < 27) {
                                            return 0;
                                        }
                                        else if (AFD >= 27) {
                                            if (posBalance < 57) {
                                                return 0;
                                            }
                                            else if (posBalance >= 57) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 26) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 62) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 13) {
                                if (RsCsDiff < 307) {
                                    if (diffTSC < -1) {
                                        return 1;
                                    }
                                    else if (diffTSC >= -1) {
                                        if (ssDCval < 87) {
                                            if (refDCval < 47) {
                                                return 0;
                                            }
                                            else if (refDCval >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 87) {
                                            if (diffMVdiffVal < 40) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 40) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 307) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 19) {
                            if (ssDCval < 41) {
                                return 1;
                            }
                            else if (ssDCval >= 41) {
                                if (AFD < 43) {
                                    return 0;
                                }
                                else if (AFD >= 43) {
                                    if (Cs < 1894) {
                                        if (refDCval < 111) {
                                            return 1;
                                        }
                                        else if (refDCval >= 111) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 1894) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 71) {
            if (diffAFD < 8) {
                if (diffMVdiffVal < 203) {
                    if (AFD < 61) {
                        if (diffRsCsdiff < 10) {
                            if (RsDiff < 7) {
                                return 0;
                            }
                            else if (RsDiff >= 7) {
                                if (Scindex < 3) {
                                    if (RsCsDiff < 172) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 172) {
                                        return 1;
                                    }
                                }
                                else if (Scindex >= 3) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 10) {
                            if (TSC < 6) {
                                if (Rs < 735) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 735) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 6) {
                                if (AFD < 7) {
                                    return 1;
                                }
                                else if (AFD >= 7) {
                                    if (negBalance < 115) {
                                        if (diffAFD < -4) {
                                            if (diffAFD < -5) {
                                                return 0;
                                            }
                                            else if (diffAFD >= -5) {
                                                if (ssDCval < 117) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 117) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= -4) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 115) {
                                        if (diffTSC < 2) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 2) {
                                            if (AFD < 28) {
                                                return 1;
                                            }
                                            else if (AFD >= 28) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 61) {
                        return 1;
                    }
                }
                else if (diffMVdiffVal >= 203) {
                    if (SC < 1070) {
                        if (diffAFD < 3) {
                            if (Cs < 618) {
                                if (diffAFD < -44) {
                                    if (TSC < 10) {
                                        return 1;
                                    }
                                    else if (TSC >= 10) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= -44) {
                                    if (posBalance < 128) {
                                        if (MVDiff < 438) {
                                            if (MVDiff < 428) {
                                                if (Rs < 381) {
                                                    return 0;
                                                }
                                                else if (Rs >= 381) {
                                                    if (RsCsDiff < 15) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 15) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 428) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 438) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 128) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 618) {
                                if (RsDiff < 4) {
                                    return 0;
                                }
                                else if (RsDiff >= 4) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 3) {
                            if (MVDiff < 266) {
                                if (diffMVdiffVal < 229) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 229) {
                                    if (AFD < 9) {
                                        return 1;
                                    }
                                    else if (AFD >= 9) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 266) {
                                if (diffMVdiffVal < 343) {
                                    if (Rs < 191) {
                                        if (RsCsDiff < 4) {
                                            if (diffAFD < 6) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 6) {
                                                if (Rs < 69) {
                                                    return 1;
                                                }
                                                else if (Rs >= 69) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 4) {
                                            if (diffTSC < 1) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 1) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 191) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 343) {
                                    if (posBalance < 55) {
                                        if (gchDC < 29) {
                                            if (CsDiff < 3) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 29) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 55) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 1070) {
                        return 0;
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (TSC < 10) {
                    if (ssDCval < 51) {
                        if (diffRsCsdiff < 4) {
                            if (diffMVdiffVal < 214) {
                                if (negBalance < 70) {
                                    if (AFD < 15) {
                                        return 1;
                                    }
                                    else if (AFD >= 15) {
                                        if (Rs < 25) {
                                            return 1;
                                        }
                                        else if (Rs >= 25) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (negBalance >= 70) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 214) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 4) {
                            if (TSC < 5) {
                                return 0;
                            }
                            else if (TSC >= 5) {
                                if (gchDC < 18) {
                                    return 1;
                                }
                                else if (gchDC >= 18) {
                                    if (gchDC < 19) {
                                        if (AFD < 21) {
                                            return 0;
                                        }
                                        else if (AFD >= 21) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 19) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 51) {
                        if (MVDiff < 285) {
                            if (diffTSC < 6) {
                                return 0;
                            }
                            else if (diffTSC >= 6) {
                                if (ssDCval < 53) {
                                    return 1;
                                }
                                else if (ssDCval >= 53) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 285) {
                            if (RsCsDiff < 46) {
                                if (negBalance < 76) {
                                    return 0;
                                }
                                else if (negBalance >= 76) {
                                    if (AFD < 20) {
                                        return 0;
                                    }
                                    else if (AFD >= 20) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 46) {
                                return 1;
                            }
                        }
                    }
                }
                else if (TSC >= 10) {
                    if (Cs < 555) {
                        if (CsDiff < 6) {
                            if (diffMVdiffVal < 142) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 142) {
                                if (SC < 398) {
                                    if (RsCsDiff < 11) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 11) {
                                        if (refDCval < 69) {
                                            return 1;
                                        }
                                        else if (refDCval >= 69) {
                                            if (negBalance < 124) {
                                                return 0;
                                            }
                                            else if (negBalance >= 124) {
                                                if (RsDiff < 4) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (SC >= 398) {
                                    if (Cs < 249) {
                                        if (AFD < 47) {
                                            return 1;
                                        }
                                        else if (AFD >= 47) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 249) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 6) {
                            if (SC < 2281) {
                                if (ssDCval < 146) {
                                    return 1;
                                }
                                else if (ssDCval >= 146) {
                                    if (TSC < 19) {
                                        return 1;
                                    }
                                    else if (TSC >= 19) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 2281) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 555) {
                        if (SC < 2727) {
                            if (MVDiff < 267) {
                                return 0;
                            }
                            else if (MVDiff >= 267) {
                                if (SC < 2432) {
                                    if (AFD < 45) {
                                        if (diffAFD < 10) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 10) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 45) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 2432) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 2727) {
                            return 0;
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (diffAFD < 21) {
            if (Cs < 777) {
                if (ssDCval < 47) {
                    if (diffRsCsdiff < 2) {
                        if (diffMVdiffVal < 200) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 200) {
                            return 1;
                        }
                    }
                    else if (diffRsCsdiff >= 2) {
                        if (posBalance < 56) {
                            if (refDCval < 151) {
                                if (negBalance < 38) {
                                    return 0;
                                }
                                else if (negBalance >= 38) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 151) {
                                return 0;
                            }
                        }
                        else if (posBalance >= 56) {
                            if (SC < 251) {
                                return 1;
                            }
                            else if (SC >= 251) {
                                if (Cs < 560) {
                                    if (CsDiff < 3) {
                                        if (Cs < 83) {
                                            return 1;
                                        }
                                        else if (Cs >= 83) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        if (posBalance < 69) {
                                            if (Rs < 386) {
                                                if (MVDiff < 114) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 114) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 386) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 69) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 560) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 47) {
                    if (CsDiff < 6) {
                        if (refDCval < 129) {
                            if (TSC < 10) {
                                if (diffMVdiffVal < 84) {
                                    if (Cs < 97) {
                                        return 1;
                                    }
                                    else if (Cs >= 97) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 84) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 10) {
                                if (diffAFD < 15) {
                                    if (refDCval < 31) {
                                        if (Rs < 237) {
                                            return 1;
                                        }
                                        else if (Rs >= 237) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 31) {
                                        if (Cs < 206) {
                                            if (gchDC < 9) {
                                                if (RsDiff < 4) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 4) {
                                                    return 1;
                                                }
                                            }
                                            else if (gchDC >= 9) {
                                                if (TSC < 15) {
                                                    return 0;
                                                }
                                                else if (TSC >= 15) {
                                                    if (posBalance < 114) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 114) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 206) {
                                            if (TSC < 30) {
                                                if (diffTSC < 13) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 13) {
                                                    if (SC < 1090) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 1090) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 30) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 15) {
                                    if (diffMVdiffVal < 151) {
                                        if (negBalance < 28) {
                                            if (TSC < 12) {
                                                return 1;
                                            }
                                            else if (TSC >= 12) {
                                                if (diffTSC < 28) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 28) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 28) {
                                            if (AFD < 29) {
                                                return 1;
                                            }
                                            else if (AFD >= 29) {
                                                if (SC < 729) {
                                                    return 1;
                                                }
                                                else if (SC >= 729) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 151) {
                                        if (posBalance < 88) {
                                            return 1;
                                        }
                                        else if (posBalance >= 88) {
                                            if (Cs < 296) {
                                                if (AFD < 26) {
                                                    return 1;
                                                }
                                                else if (AFD >= 26) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 296) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 129) {
                            return 0;
                        }
                    }
                    else if (CsDiff >= 6) {
                        if (diffRsCsdiff < 46) {
                            if (Rs < 401) {
                                return 1;
                            }
                            else if (Rs >= 401) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 46) {
                            if (RsCsDiff < 173) {
                                return 1;
                            }
                            else if (RsCsDiff >= 173) {
                                if (ssDCval < 134) {
                                    if (diffRsCsdiff < 321) {
                                        if (SC < 663) {
                                            return 1;
                                        }
                                        else if (SC >= 663) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 321) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 134) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (Cs >= 777) {
                if (diffAFD < 14) {
                    if (negBalance < 87) {
                        if (posBalance < 55) {
                            if (diffRsCsdiff < 975) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 975) {
                                if (gchDC < 3) {
                                    return 0;
                                }
                                else if (gchDC >= 3) {
                                    if (refDCval < 88) {
                                        return 1;
                                    }
                                    else if (refDCval >= 88) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 55) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 87) {
                        if (gchDC < 38) {
                            return 0;
                        }
                        else if (gchDC >= 38) {
                            if (posBalance < 17) {
                                return 0;
                            }
                            else if (posBalance >= 17) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffAFD >= 14) {
                    if (MVDiff < 46) {
                        if (SC < 17265) {
                            return 0;
                        }
                        else if (SC >= 17265) {
                            if (Rs < 9867) {
                                return 1;
                            }
                            else if (Rs >= 9867) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 46) {
                        if (RsCsDiff < 141) {
                            if (MVDiff < 1273) {
                                return 0;
                            }
                            else if (MVDiff >= 1273) {
                                return 1;
                            }
                        }
                        else if (RsCsDiff >= 141) {
                            if (Rs < 702) {
                                return 1;
                            }
                            else if (Rs >= 702) {
                                if (refDCval < 122) {
                                    if (negBalance < 43) {
                                        if (RsCsDiff < 215) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 215) {
                                            if (AFD < 37) {
                                                if (diffTSC < 13) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 13) {
                                                    if (Cs < 983) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 983) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (AFD >= 37) {
                                                if (gchDC < 42) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 42) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        if (diffMVdiffVal < 53) {
                                            if (Rs < 13392) {
                                                if (negBalance < 46) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 46) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 13392) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 53) {
                                            if (Cs < 1524) {
                                                return 1;
                                            }
                                            else if (Cs >= 1524) {
                                                if (diffAFD < 15) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 15) {
                                                    if (CsDiff < 14) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 14) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 122) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 21) {
            if (diffTSC < 15) {
                if (MVDiff < 51) {
                    if (diffTSC < 13) {
                        return 0;
                    }
                    else if (diffTSC >= 13) {
                        if (gchDC < 10) {
                            return 0;
                        }
                        else if (gchDC >= 10) {
                            if (refDCval < 45) {
                                return 1;
                            }
                            else if (refDCval >= 45) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 51) {
                    if (Rs < 1869) {
                        if (diffMVdiffVal < 50) {
                            if (ssDCval < 58) {
                                if (MVDiff < 600) {
                                    return 0;
                                }
                                else if (MVDiff >= 600) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 58) {
                                if (refDCval < 95) {
                                    if (Cs < 830) {
                                        return 1;
                                    }
                                    else if (Cs >= 830) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 95) {
                                    if (refDCval < 130) {
                                        if (Cs < 306) {
                                            return 0;
                                        }
                                        else if (Cs >= 306) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 130) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 50) {
                            if (posBalance < 117) {
                                if (AFD < 69) {
                                    if (AFD < 37) {
                                        if (diffTSC < 12) {
                                            if (Rs < 278) {
                                                return 1;
                                            }
                                            else if (Rs >= 278) {
                                                if (negBalance < 58) {
                                                    if (diffMVdiffVal < 67) {
                                                        if (gchDC < 3) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 3) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 67) {
                                                        if (ssDCval < 48) {
                                                            if (SC < 821) {
                                                                return 0;
                                                            }
                                                            else if (SC >= 821) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (ssDCval >= 48) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (negBalance >= 58) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 37) {
                                        if (diffMVdiffVal < 91) {
                                            if (MVDiff < 166) {
                                                if (RsDiff < 9) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 166) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 91) {
                                            if (TSC < 27) {
                                                if (Cs < 2207) {
                                                    if (SC < 3045) {
                                                        if (RsCsDiff < 10) {
                                                            return 0;
                                                        }
                                                        else if (RsCsDiff >= 10) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 3045) {
                                                        if (Rs < 1575) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 1575) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 2207) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 27) {
                                                if (Rs < 979) {
                                                    return 0;
                                                }
                                                else if (Rs >= 979) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 69) {
                                    if (diffMVdiffVal < 341) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 341) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 117) {
                                if (SC < 1716) {
                                    return 0;
                                }
                                else if (SC >= 1716) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (Rs >= 1869) {
                        if (diffMVdiffVal < 61) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 61) {
                            if (AFD < 29) {
                                if (CsDiff < 12) {
                                    return 1;
                                }
                                else if (CsDiff >= 12) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 29) {
                                if (CsDiff < 15) {
                                    if (RsDiff < 21) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 21) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 15) {
                                    if (refDCval < 96) {
                                        if (gchDC < 18) {
                                            return 1;
                                        }
                                        else if (gchDC >= 18) {
                                            if (RsDiff < 47) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 47) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (refDCval >= 96) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 15) {
                if (posBalance < 91) {
                    if (diffMVdiffVal < 62) {
                        if (RsDiff < 13) {
                            if (diffAFD < 24) {
                                if (TSC < 21) {
                                    return 1;
                                }
                                else if (TSC >= 21) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 24) {
                                if (MVDiff < 17) {
                                    return 1;
                                }
                                else if (MVDiff >= 17) {
                                    if (negBalance < 54) {
                                        if (TSC < 21) {
                                            return 1;
                                        }
                                        else if (TSC >= 21) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 54) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 13) {
                            if (Cs < 2537) {
                                if (negBalance < 82) {
                                    return 1;
                                }
                                else if (negBalance >= 82) {
                                    if (gchDC < 29) {
                                        return 0;
                                    }
                                    else if (gchDC >= 29) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 2537) {
                                if (gchDC < 11) {
                                    return 0;
                                }
                                else if (gchDC >= 11) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 62) {
                        if (CsDiff < 4) {
                            if (diffRsCsdiff < 23) {
                                if (diffTSC < 18) {
                                    return 1;
                                }
                                else if (diffTSC >= 18) {
                                    if (negBalance < 31) {
                                        return 0;
                                    }
                                    else if (negBalance >= 31) {
                                        if (gchDC < 36) {
                                            return 1;
                                        }
                                        else if (gchDC >= 36) {
                                            if (TSC < 23) {
                                                return 0;
                                            }
                                            else if (TSC >= 23) {
                                                if (posBalance < 17) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 17) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 23) {
                                if (refDCval < 47) {
                                    return 0;
                                }
                                else if (refDCval >= 47) {
                                    if (AFD < 44) {
                                        return 1;
                                    }
                                    else if (AFD >= 44) {
                                        if (AFD < 51) {
                                            return 0;
                                        }
                                        else if (AFD >= 51) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 4) {
                            if (posBalance < 7) {
                                if (Cs < 2127) {
                                    if (RsDiff < 7) {
                                        if (TSC < 34) {
                                            return 1;
                                        }
                                        else if (TSC >= 34) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 7) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 2127) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 7) {
                                if (CsDiff < 11) {
                                    if (negBalance < 39) {
                                        if (gchDC < 24) {
                                            if (Rs < 266) {
                                                return 0;
                                            }
                                            else if (Rs >= 266) {
                                                if (MVDiff < 624) {
                                                    if (diffAFD < 22) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= 22) {
                                                        if (CsDiff < 5) {
                                                            return 0;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 624) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 24) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 39) {
                                        if (AFD < 46) {
                                            return 1;
                                        }
                                        else if (AFD >= 46) {
                                            if (diffMVdiffVal < 121) {
                                                if (Rs < 697) {
                                                    return 1;
                                                }
                                                else if (Rs >= 697) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 121) {
                                                if (MVDiff < 491) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 491) {
                                                    if (diffMVdiffVal < 464) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 464) {
                                                        if (SC < 780) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 780) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 11) {
                                    if (SC < 1452) {
                                        if (diffAFD < 32) {
                                            if (Cs < 734) {
                                                if (gchDC < 25) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 25) {
                                                    return 0;
                                                }
                                            }
                                            else if (Cs >= 734) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 32) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 1452) {
                                        if (diffRsCsdiff < 10398) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= 10398) {
                                            if (MVDiff < 75) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 75) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (posBalance >= 91) {
                    if (diffAFD < 26) {
                        if (SC < 4270) {
                            if (Cs < 34) {
                                return 1;
                            }
                            else if (Cs >= 34) {
                                if (negBalance < 2) {
                                    if (MVDiff < 117) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 117) {
                                        if (Cs < 949) {
                                            return 0;
                                        }
                                        else if (Cs >= 949) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (SC >= 4270) {
                            if (Cs < 5284) {
                                return 1;
                            }
                            else if (Cs >= 5284) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 26) {
                        if (diffRsCsdiff < 163) {
                            if (Cs < 709) {
                                if (TSC < 28) {
                                    if (CsDiff < 5) {
                                        if (gchDC < 25) {
                                            if (Rs < 246) {
                                                return 1;
                                            }
                                            else if (Rs >= 246) {
                                                return 0;
                                            }
                                        }
                                        else if (gchDC >= 25) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 28) {
                                    if (negBalance < 7) {
                                        return 1;
                                    }
                                    else if (negBalance >= 7) {
                                        if (refDCval < 62) {
                                            return 0;
                                        }
                                        else if (refDCval >= 62) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 709) {
                                if (diffMVdiffVal < 354) {
                                    if (Cs < 1718) {
                                        if (diffTSC < 21) {
                                            if (Rs < 742) {
                                                if (AFD < 40) {
                                                    return 0;
                                                }
                                                else if (AFD >= 40) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 742) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 21) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 1718) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 354) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 163) {
                            if (SC < 1650) {
                                return 1;
                            }
                            else if (SC >= 1650) {
                                if (diffAFD < 40) {
                                    if (MVDiff < 150) {
                                        if (ssDCval < 87) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 87) {
                                            if (negBalance < 12) {
                                                return 0;
                                            }
                                            else if (negBalance >= 12) {
                                                if (MVDiff < 115) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 115) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 150) {
                                        if (diffAFD < 33) {
                                            if (Cs < 2511) {
                                                if (diffMVdiffVal < 138) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 138) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 2511) {
                                                return 1;
                                            }
                                        }
                                        else if (diffAFD >= 33) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 40) {
                                    if (diffTSC < 40) {
                                        return 1;
                                    }
                                    else if (diffTSC >= 40) {
                                        if (gchDC < 101) {
                                            return 1;
                                        }
                                        else if (gchDC >= 101) {
                                            if (Cs < 1861) {
                                                return 0;
                                            }
                                            else if (Cs >= 1861) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect14(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffRsCsdiff < 69) {
        if (diffAFD < 8) {
            if (negBalance < 66) {
                if (diffAFD < 3) {
                    if (TSC < 30) {
                        if (refDCval < 24) {
                            if (diffRsCsdiff < 5) {
                                if (ssDCval < 35) {
                                    return 0;
                                }
                                else if (ssDCval >= 35) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 5) {
                                if (posBalance < 2) {
                                    if (diffMVdiffVal < 2) {
                                        if (RsDiff < 2) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 2) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 2) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (refDCval >= 24) {
                            if (gchDC < 4) {
                                return 0;
                            }
                            else if (gchDC >= 4) {
                                if (MVDiff < 223) {
                                    return 0;
                                }
                                else if (MVDiff >= 223) {
                                    if (MVDiff < 225) {
                                        if (negBalance < 47) {
                                            return 0;
                                        }
                                        else if (negBalance >= 47) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 225) {
                                        if (ssDCval < 129) {
                                            if (ssDCval < 128) {
                                                if (CsDiff < 4) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 4) {
                                                    if (refDCval < 47) {
                                                        if (TSCindex < 9) {
                                                            return 0;
                                                        }
                                                        else if (TSCindex >= 9) {
                                                            if (Cs < 863) {
                                                                if (Rs < 732) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 732) {
                                                                    return 0;
                                                                }
                                                            }
                                                            else if (Cs >= 863) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                    else if (refDCval >= 47) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 128) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 129) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 30) {
                        if (Rs < 1239) {
                            return 0;
                        }
                        else if (Rs >= 1239) {
                            if (MVDiff < 272) {
                                return 1;
                            }
                            else if (MVDiff >= 272) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffAFD >= 3) {
                    if (gchDC < 63) {
                        if (Cs < 71) {
                            if (MVDiff < 252) {
                                return 0;
                            }
                            else if (MVDiff >= 252) {
                                if (ssDCval < 33) {
                                    return 1;
                                }
                                else if (ssDCval >= 33) {
                                    return 0;
                                }
                            }
                        }
                        else if (Cs >= 71) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                if (diffRsCsdiff < 48) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 48) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 63) {
                        if (posBalance < 128) {
                            return 0;
                        }
                        else if (posBalance >= 128) {
                            return 1;
                        }
                    }
                }
            }
            else if (negBalance >= 66) {
                if (SC < 974) {
                    if (diffRsCsdiff < 38) {
                        if (ssDCval < 26) {
                            if (RsCsDiff < 3) {
                                if (gchDC < 84) {
                                    return 0;
                                }
                                else if (gchDC >= 84) {
                                    if (Rs < 6) {
                                        return 1;
                                    }
                                    else if (Rs >= 6) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 3) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 26) {
                            if (posBalance < 59) {
                                if (diffRsCsdiff < -1668) {
                                    return 1;
                                }
                                else if (diffRsCsdiff >= -1668) {
                                    if (diffMVdiffVal < 99) {
                                        if (posBalance < 1) {
                                            if (refDCval < 226) {
                                                return 0;
                                            }
                                            else if (refDCval >= 226) {
                                                if (diffAFD < -13) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= -13) {
                                                    if (ssDCval < 226) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 226) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (posBalance >= 1) {
                                            if (ssDCval < 36) {
                                                if (AFD < 17) {
                                                    return 0;
                                                }
                                                else if (AFD >= 17) {
                                                    if (posBalance < 36) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 36) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 36) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 99) {
                                        if (diffAFD < 7) {
                                            if (gchDC < 163) {
                                                if (posBalance < 53) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 53) {
                                                    if (SC < 572) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 572) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 163) {
                                                return 1;
                                            }
                                        }
                                        else if (diffAFD >= 7) {
                                            if (AFD < 15) {
                                                return 1;
                                            }
                                            else if (AFD >= 15) {
                                                if (Rs < 429) {
                                                    return 0;
                                                }
                                                else if (Rs >= 429) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 59) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 38) {
                        if (Rs < 415) {
                            return 1;
                        }
                        else if (Rs >= 415) {
                            if (ssDCval < 38) {
                                return 1;
                            }
                            else if (ssDCval >= 38) {
                                return 0;
                            }
                        }
                    }
                }
                else if (SC >= 974) {
                    if (diffTSC < 4) {
                        if (SC < 1443) {
                            if (RsCsDiff < 197) {
                                if (ssDCval < 28) {
                                    if (TSC < 11) {
                                        return 1;
                                    }
                                    else if (TSC >= 11) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 28) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 197) {
                                if (SC < 1428) {
                                    return 0;
                                }
                                else if (SC >= 1428) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 1443) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= 4) {
                        if (AFD < 58) {
                            return 0;
                        }
                        else if (AFD >= 58) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 8) {
            if (MVDiff < 84) {
                if (diffTSC < 6) {
                    return 0;
                }
                else if (diffTSC >= 6) {
                    if (Cs < 58) {
                        if (RsCsDiff < 2) {
                            if (AFD < 25) {
                                return 0;
                            }
                            else if (AFD >= 25) {
                                return 1;
                            }
                        }
                        else if (RsCsDiff >= 2) {
                            return 1;
                        }
                    }
                    else if (Cs >= 58) {
                        if (CsDiff < 16) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                if (refDCval < 24) {
                                    return 1;
                                }
                                else if (refDCval >= 24) {
                                    if (diffMVdiffVal < 73) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 73) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 16) {
                            return 1;
                        }
                    }
                }
            }
            else if (MVDiff >= 84) {
                if (Cs < 741) {
                    if (diffMVdiffVal < 86) {
                        if (negBalance < 62) {
                            if (MVDiff < 118) {
                                if (gchDC < 18) {
                                    return 0;
                                }
                                else if (gchDC >= 18) {
                                    if (SC < 573) {
                                        if (SC < 411) {
                                            return 0;
                                        }
                                        else if (SC >= 411) {
                                            if (Rs < 388) {
                                                return 1;
                                            }
                                            else if (Rs >= 388) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (SC >= 573) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 118) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 62) {
                            if (gchDC < 46) {
                                if (gchDC < 6) {
                                    if (Cs < 141) {
                                        if (AFD < 21) {
                                            return 1;
                                        }
                                        else if (AFD >= 21) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 141) {
                                        if (MVDiff < 227) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 227) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 6) {
                                    if (ssDCval < 100) {
                                        if (MVDiff < 173) {
                                            if (refDCval < 53) {
                                                return 1;
                                            }
                                            else if (refDCval >= 53) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 173) {
                                            if (diffTSC < -1) {
                                                return 1;
                                            }
                                            else if (diffTSC >= -1) {
                                                if (SC < 925) {
                                                    if (Rs < 705) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 705) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 925) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 100) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 46) {
                                if (negBalance < 113) {
                                    return 0;
                                }
                                else if (negBalance >= 113) {
                                    if (diffTSC < 7) {
                                        return 1;
                                    }
                                    else if (diffTSC >= 7) {
                                        if (RsCsDiff < 179) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 179) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 86) {
                        if (ssDCval < 44) {
                            if (gchDC < 15) {
                                if (posBalance < 97) {
                                    if (Rs < 276) {
                                        if (diffMVdiffVal < 96) {
                                            if (RsCsDiff < 5) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 5) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 96) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 276) {
                                        if (Rs < 353) {
                                            if (TSC < 6) {
                                                return 1;
                                            }
                                            else if (TSC >= 6) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 353) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 97) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 15) {
                                if (gchDC < 18) {
                                    if (diffMVdiffVal < 273) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 273) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 18) {
                                    if (diffMVdiffVal < 287) {
                                        if (diffTSC < 4) {
                                            if (TSCindex < 9) {
                                                return 1;
                                            }
                                            else if (TSCindex >= 9) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 287) {
                                        if (refDCval < 106) {
                                            if (diffRsCsdiff < -5) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= -5) {
                                                if (negBalance < 70) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 70) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 106) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 44) {
                            if (diffTSC < 5) {
                                if (diffRsCsdiff < 37) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 37) {
                                    if (ssDCval < 90) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 90) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 5) {
                                if (gchDC < 7) {
                                    if (negBalance < 43) {
                                        if (diffAFD < 32) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 32) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 43) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 7) {
                                    if (AFD < 70) {
                                        if (diffRsCsdiff < 16) {
                                            if (Scindex < 3) {
                                                if (AFD < 49) {
                                                    if (negBalance < 6) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 6) {
                                                        if (diffRsCsdiff < 4) {
                                                            if (Rs < 119) {
                                                                if (RsCsDiff < 1) {
                                                                    return 0;
                                                                }
                                                                else if (RsCsDiff >= 1) {
                                                                    if (diffTSC < 9) {
                                                                        if (AFD < 15) {
                                                                            return 1;
                                                                        }
                                                                        else if (AFD >= 15) {
                                                                            if (RsCsDiff < 22) {
                                                                                return 0;
                                                                            }
                                                                            else if (RsCsDiff >= 22) {
                                                                                return 1;
                                                                            }
                                                                        }
                                                                    }
                                                                    else if (diffTSC >= 9) {
                                                                        return 1;
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 119) {
                                                                if (SC < 211) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 211) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 4) {
                                                            if (Rs < 260) {
                                                                if (CsDiff < 3) {
                                                                    return 1;
                                                                }
                                                                else if (CsDiff >= 3) {
                                                                    if (MVDiff < 266) {
                                                                        return 1;
                                                                    }
                                                                    else if (MVDiff >= 266) {
                                                                        if (diffAFD < 24) {
                                                                            return 0;
                                                                        }
                                                                        else if (diffAFD >= 24) {
                                                                            return 1;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                            else if (Rs >= 260) {
                                                                if (AFD < 18) {
                                                                    return 0;
                                                                }
                                                                else if (AFD >= 18) {
                                                                    if (SC < 459) {
                                                                        return 1;
                                                                    }
                                                                    else if (SC >= 459) {
                                                                        return 0;
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                                else if (AFD >= 49) {
                                                    return 0;
                                                }
                                            }
                                            else if (Scindex >= 3) {
                                                if (Scindex < 4) {
                                                    if (TSC < 13) {
                                                        if (MVDiff < 624) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 624) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 13) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Scindex >= 4) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 16) {
                                            if (Scindex < 3) {
                                                if (diffMVdiffVal < 134) {
                                                    if (diffAFD < 22) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 22) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 134) {
                                                    return 1;
                                                }
                                            }
                                            else if (Scindex >= 3) {
                                                if (SC < 934) {
                                                    if (gchDC < 107) {
                                                        if (diffTSC < 8) {
                                                            return 0;
                                                        }
                                                        else if (diffTSC >= 8) {
                                                            if (diffAFD < 32) {
                                                                if (refDCval < 48) {
                                                                    if (ssDCval < 67) {
                                                                        return 0;
                                                                    }
                                                                    else if (ssDCval >= 67) {
                                                                        if (TSC < 16) {
                                                                            return 1;
                                                                        }
                                                                        else if (TSC >= 16) {
                                                                            if (SC < 753) {
                                                                                return 1;
                                                                            }
                                                                            else if (SC >= 753) {
                                                                                return 0;
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                                else if (refDCval >= 48) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffAFD >= 32) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (gchDC >= 107) {
                                                        return 0;
                                                    }
                                                }
                                                else if (SC >= 934) {
                                                    if (TSC < 26) {
                                                        if (diffRsCsdiff < 49) {
                                                            if (posBalance < 90) {
                                                                if (diffAFD < 21) {
                                                                    return 0;
                                                                }
                                                                else if (diffAFD >= 21) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (posBalance >= 90) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 49) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (TSC >= 26) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 70) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 741) {
                    if (diffTSC < 20) {
                        return 0;
                    }
                    else if (diffTSC >= 20) {
                        if (gchDC < 24) {
                            return 1;
                        }
                        else if (gchDC >= 24) {
                            if (Cs < 3004) {
                                return 0;
                            }
                            else if (Cs >= 3004) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffRsCsdiff >= 69) {
        if (diffAFD < 21) {
            if (ssDCval < 60) {
                if (MVDiff < 45) {
                    if (diffRsCsdiff < 638) {
                        if (Rs < 1155) {
                            if (Rs < 934) {
                                if (CsDiff < 14) {
                                    return 0;
                                }
                                else if (CsDiff >= 14) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 934) {
                                return 1;
                            }
                        }
                        else if (Rs >= 1155) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 638) {
                        if (TSC < 6) {
                            return 1;
                        }
                        else if (TSC >= 6) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                return 1;
                            }
                        }
                    }
                }
                else if (MVDiff >= 45) {
                    if (diffRsCsdiff < 516) {
                        if (diffAFD < 10) {
                            if (SC < 229) {
                                return 1;
                            }
                            else if (SC >= 229) {
                                if (ssDCval < 37) {
                                    if (AFD < 28) {
                                        return 0;
                                    }
                                    else if (AFD >= 28) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 37) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 10) {
                            if (CsDiff < 14) {
                                if (Rs < 458) {
                                    return 1;
                                }
                                else if (Rs >= 458) {
                                    if (diffMVdiffVal < 43) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 43) {
                                        if (gchDC < 18) {
                                            if (RsDiff < 6) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 6) {
                                                if (posBalance < 65) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 65) {
                                                    if (Rs < 764) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 764) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 14) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 516) {
                        if (SC < 7396) {
                            return 1;
                        }
                        else if (SC >= 7396) {
                            if (gchDC < 17) {
                                return 1;
                            }
                            else if (gchDC >= 17) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 60) {
                if (diffTSC < 9) {
                    if (SC < 1833) {
                        if (AFD < 25) {
                            return 0;
                        }
                        else if (AFD >= 25) {
                            if (gchDC < 4) {
                                if (Cs < 541) {
                                    if (RsCsDiff < 254) {
                                        if (Cs < 171) {
                                            return 0;
                                        }
                                        else if (Cs >= 171) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 254) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 541) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 4) {
                                if (MVDiff < 960) {
                                    if (ssDCval < 68) {
                                        if (AFD < 28) {
                                            return 0;
                                        }
                                        else if (AFD >= 28) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 68) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 960) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (SC >= 1833) {
                        if (MVDiff < 55) {
                            return 0;
                        }
                        else if (MVDiff >= 55) {
                            if (diffRsCsdiff < 1109) {
                                if (Rs < 1260) {
                                    if (AFD < 35) {
                                        return 0;
                                    }
                                    else if (AFD >= 35) {
                                        if (diffRsCsdiff < 202) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 202) {
                                            if (RsCsDiff < 735) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 735) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 1260) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 1109) {
                                if (Cs < 3730) {
                                    if (posBalance < 68) {
                                        if (SC < 4691) {
                                            return 1;
                                        }
                                        else if (SC >= 4691) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 68) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 3730) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 9) {
                    if (gchDC < 2) {
                        if (SC < 2786) {
                            return 1;
                        }
                        else if (SC >= 2786) {
                            return 0;
                        }
                    }
                    else if (gchDC >= 2) {
                        if (SC < 1542) {
                            return 1;
                        }
                        else if (SC >= 1542) {
                            if (diffMVdiffVal < 75) {
                                if (AFD < 29) {
                                    if (CsDiff < 4) {
                                        if (posBalance < 47) {
                                            return 1;
                                        }
                                        else if (posBalance >= 47) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 4) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 29) {
                                    if (refDCval < 57) {
                                        return 0;
                                    }
                                    else if (refDCval >= 57) {
                                        if (ssDCval < 118) {
                                            if (negBalance < 39) {
                                                if (refDCval < 59) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 59) {
                                                    return 0;
                                                }
                                            }
                                            else if (negBalance >= 39) {
                                                if (TSC < 23) {
                                                    if (SC < 5898) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 5898) {
                                                        return 0;
                                                    }
                                                }
                                                else if (TSC >= 23) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 118) {
                                            if (CsDiff < 35) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 35) {
                                                if (Cs < 5177) {
                                                    return 1;
                                                }
                                                else if (Cs >= 5177) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 75) {
                                if (diffRsCsdiff < 784) {
                                    if (RsCsDiff < 769) {
                                        if (ssDCval < 87) {
                                            if (refDCval < 52) {
                                                if (SC < 2181) {
                                                    return 1;
                                                }
                                                else if (SC >= 2181) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 52) {
                                                if (Cs < 1518) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1518) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 87) {
                                            if (RsCsDiff < 424) {
                                                if (diffMVdiffVal < 85) {
                                                    return 1;
                                                }
                                                else if (diffMVdiffVal >= 85) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsCsDiff >= 424) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 769) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 784) {
                                    if (diffMVdiffVal < 79) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 79) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 21) {
            if (ssDCval < 78) {
                if (AFD < 32) {
                    if (TSC < 23) {
                        if (diffMVdiffVal < 62) {
                            if (RsDiff < 16) {
                                if (Cs < 104) {
                                    return 1;
                                }
                                else if (Cs >= 104) {
                                    if (diffTSC < 15) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 15) {
                                        if (Cs < 1148) {
                                            return 0;
                                        }
                                        else if (Cs >= 1148) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 16) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 62) {
                            if (AFD < 22) {
                                return 0;
                            }
                            else if (AFD >= 22) {
                                if (CsDiff < 2) {
                                    return 0;
                                }
                                else if (CsDiff >= 2) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (TSC >= 23) {
                        return 0;
                    }
                }
                else if (AFD >= 32) {
                    if (ssDCval < 48) {
                        if (RsDiff < 31) {
                            if (AFD < 40) {
                                return 1;
                            }
                            else if (AFD >= 40) {
                                if (gchDC < 23) {
                                    return 0;
                                }
                                else if (gchDC >= 23) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 31) {
                            if (Rs < 12617) {
                                return 0;
                            }
                            else if (Rs >= 12617) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 48) {
                        if (diffMVdiffVal < 10) {
                            if (Rs < 1594) {
                                return 1;
                            }
                            else if (Rs >= 1594) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 10) {
                            if (diffAFD < 27) {
                                if (refDCval < 87) {
                                    return 1;
                                }
                                else if (refDCval >= 87) {
                                    if (AFD < 37) {
                                        return 1;
                                    }
                                    else if (AFD >= 37) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 27) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 78) {
                if (diffTSC < 13) {
                    if (MVDiff < 107) {
                        return 0;
                    }
                    else if (MVDiff >= 107) {
                        if (TSC < 29) {
                            if (diffMVdiffVal < 107) {
                                if (AFD < 30) {
                                    return 1;
                                }
                                else if (AFD >= 30) {
                                    if (diffAFD < 30) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 30) {
                                        if (ssDCval < 115) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 115) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 107) {
                                if (Cs < 214) {
                                    if (posBalance < 64) {
                                        return 0;
                                    }
                                    else if (posBalance >= 64) {
                                        if (Rs < 8) {
                                            return 0;
                                        }
                                        else if (Rs >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 214) {
                                    if (Rs < 3089) {
                                        return 1;
                                    }
                                    else if (Rs >= 3089) {
                                        if (Cs < 3240) {
                                            return 0;
                                        }
                                        else if (Cs >= 3240) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 29) {
                            return 0;
                        }
                    }
                }
                else if (diffTSC >= 13) {
                    if (diffMVdiffVal < 57) {
                        if (AFD < 56) {
                            if (ssDCval < 104) {
                                if (diffRsCsdiff < 275) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 275) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 104) {
                                if (RsCsDiff < 6215) {
                                    if (gchDC < 8) {
                                        if (negBalance < 48) {
                                            if (refDCval < 126) {
                                                return 0;
                                            }
                                            else if (refDCval >= 126) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 48) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 8) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 6215) {
                                    if (CsDiff < 56) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 56) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 56) {
                            if (posBalance < 96) {
                                return 1;
                            }
                            else if (posBalance >= 96) {
                                if (Rs < 984) {
                                    return 1;
                                }
                                else if (Rs >= 984) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 57) {
                        if (diffRsCsdiff < 351) {
                            if (negBalance < 4) {
                                if (AFD < 53) {
                                    return 0;
                                }
                                else if (AFD >= 53) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 4) {
                                if (posBalance < 95) {
                                    if (SC < 3173) {
                                        if (diffMVdiffVal < 413) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 413) {
                                            if (negBalance < 53) {
                                                if (posBalance < 77) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 77) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 53) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 3173) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 95) {
                                    if (posBalance < 97) {
                                        return 0;
                                    }
                                    else if (posBalance >= 97) {
                                        if (diffRsCsdiff < 158) {
                                            if (SC < 1534) {
                                                return 1;
                                            }
                                            else if (SC >= 1534) {
                                                if (RsCsDiff < 213) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 213) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 158) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 351) {
                            if (diffTSC < 17) {
                                if (negBalance < 39) {
                                    if (Rs < 2166) {
                                        return 1;
                                    }
                                    else if (Rs >= 2166) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 39) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 17) {
                                if (posBalance < 125) {
                                    if (diffRsCsdiff < 10398) {
                                        return 1;
                                    }
                                    else if (diffRsCsdiff >= 10398) {
                                        if (CsDiff < 69) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 69) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 125) {
                                    if (Cs < 844) {
                                        return 1;
                                    }
                                    else if (Cs >= 844) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect15(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffAFD, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffRsCsdiff, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (Cs < 1116) {
        if (diffMVdiffVal < 65) {
            if (diffRsCsdiff < 64) {
                if (diffTSC < 5) {
                    if (posBalance < 2) {
                        if (diffRsCsdiff < 5) {
                            if (RsCsDiff < 11) {
                                if (refDCval < 229) {
                                    return 0;
                                }
                                else if (refDCval >= 229) {
                                    if (TSC < 3) {
                                        return 0;
                                    }
                                    else if (TSC >= 3) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 11) {
                                if (MVDiff < 11) {
                                    if (MVDiff < 3) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 3) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 11) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (ssDCval < 20) {
                                if (Rs < 687) {
                                    return 1;
                                }
                                else if (Rs >= 687) {
                                    if (SC < 1728) {
                                        return 0;
                                    }
                                    else if (SC >= 1728) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 20) {
                                return 0;
                            }
                        }
                    }
                    else if (posBalance >= 2) {
                        if (gchDC < 18) {
                            if (MVDiff < 145) {
                                return 0;
                            }
                            else if (MVDiff >= 145) {
                                if (refDCval < 51) {
                                    if (diffAFD < 8) {
                                        if (TSC < 14) {
                                            if (MVDiff < 1538) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 1538) {
                                                if (Cs < 137) {
                                                    return 1;
                                                }
                                                else if (Cs >= 137) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (TSC >= 14) {
                                            if (negBalance < 53) {
                                                return 0;
                                            }
                                            else if (negBalance >= 53) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 8) {
                                        if (Cs < 79) {
                                            return 1;
                                        }
                                        else if (Cs >= 79) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 51) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 18) {
                            if (Scindex < 1) {
                                if (gchDC < 20) {
                                    return 1;
                                }
                                else if (gchDC >= 20) {
                                    if (AFD < 4) {
                                        return 1;
                                    }
                                    else if (AFD >= 4) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Scindex >= 1) {
                                if (Rs < 116) {
                                    if (Rs < 111) {
                                        return 0;
                                    }
                                    else if (Rs >= 111) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 116) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 5) {
                    if (SC < 993) {
                        if (ssDCval < 47) {
                            if (SC < 826) {
                                if (diffMVdiffVal < 5) {
                                    if (RsCsDiff < 63) {
                                        if (diffAFD < -6) {
                                            return 1;
                                        }
                                        else if (diffAFD >= -6) {
                                            if (negBalance < 46) {
                                                return 0;
                                            }
                                            else if (negBalance >= 46) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsCsDiff >= 63) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 5) {
                                    return 0;
                                }
                            }
                            else if (SC >= 826) {
                                if (diffTSC < 6) {
                                    return 0;
                                }
                                else if (diffTSC >= 6) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 47) {
                            if (refDCval < 25) {
                                if (AFD < 42) {
                                    return 1;
                                }
                                else if (AFD >= 42) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 25) {
                                if (RsCsDiff < 18) {
                                    if (gchDC < 46) {
                                        if (diffRsCsdiff < 8) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 8) {
                                            if (diffMVdiffVal < 40) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 40) {
                                                if (AFD < 31) {
                                                    return 1;
                                                }
                                                else if (AFD >= 31) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 46) {
                                        if (SC < 239) {
                                            if (Cs < 32) {
                                                return 0;
                                            }
                                            else if (Cs >= 32) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 239) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 18) {
                                    if (AFD < 58) {
                                        if (diffMVdiffVal < 54) {
                                            if (diffAFD < 18) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 18) {
                                                if (refDCval < 60) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 60) {
                                                    if (TSC < 11) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 11) {
                                                        if (SC < 921) {
                                                            if (Cs < 244) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 244) {
                                                                if (diffMVdiffVal < -1) {
                                                                    return 0;
                                                                }
                                                                else if (diffMVdiffVal >= -1) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (SC >= 921) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 54) {
                                            if (Rs < 155) {
                                                return 0;
                                            }
                                            else if (Rs >= 155) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (AFD >= 58) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 993) {
                        if (RsCsDiff < 108) {
                            return 0;
                        }
                        else if (RsCsDiff >= 108) {
                            if (SC < 1355) {
                                if (SC < 1268) {
                                    return 0;
                                }
                                else if (SC >= 1268) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1355) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 64) {
                if (SC < 1424) {
                    if (AFD < 31) {
                        if (CsDiff < 12) {
                            return 0;
                        }
                        else if (CsDiff >= 12) {
                            if (MVDiff < 43) {
                                return 1;
                            }
                            else if (MVDiff >= 43) {
                                if (gchDC < 4) {
                                    return 1;
                                }
                                else if (gchDC >= 4) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (AFD >= 31) {
                        if (ssDCval < 93) {
                            if (negBalance < 89) {
                                return 1;
                            }
                            else if (negBalance >= 89) {
                                if (Rs < 156) {
                                    return 1;
                                }
                                else if (Rs >= 156) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 93) {
                            return 0;
                        }
                    }
                }
                else if (SC >= 1424) {
                    if (diffAFD < 13) {
                        if (Rs < 1024) {
                            if (diffTSC < 6) {
                                return 0;
                            }
                            else if (diffTSC >= 6) {
                                if (CsDiff < 6) {
                                    return 0;
                                }
                                else if (CsDiff >= 6) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 1024) {
                            return 0;
                        }
                    }
                    else if (diffAFD >= 13) {
                        if (diffTSC < 9) {
                            if (CsDiff < 3) {
                                return 1;
                            }
                            else if (CsDiff >= 3) {
                                if (refDCval < 40) {
                                    return 1;
                                }
                                else if (refDCval >= 40) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 9) {
                            if (Rs < 1069) {
                                return 0;
                            }
                            else if (Rs >= 1069) {
                                if (gchDC < 17) {
                                    if (RsCsDiff < 164) {
                                        if (Cs < 692) {
                                            return 1;
                                        }
                                        else if (Cs >= 692) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 164) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 17) {
                                    if (AFD < 48) {
                                        if (gchDC < 53) {
                                            return 0;
                                        }
                                        else if (gchDC >= 53) {
                                            if (Rs < 1398) {
                                                return 1;
                                            }
                                            else if (Rs >= 1398) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 48) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffAFD < 10) {
                if (ssDCval < 41) {
                    if (diffRsCsdiff < 2) {
                        if (diffMVdiffVal < 105) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 105) {
                            if (TSCindex < 4) {
                                return 0;
                            }
                            else if (TSCindex >= 4) {
                                if (RsCsDiff < 48) {
                                    if (Rs < 101) {
                                        if (gchDC < 62) {
                                            if (SC < 91) {
                                                if (diffAFD < 8) {
                                                    if (negBalance < 38) {
                                                        if (MVDiff < 405) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 405) {
                                                            if (AFD < 9) {
                                                                return 0;
                                                            }
                                                            else if (AFD >= 9) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (negBalance >= 38) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 8) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 91) {
                                                if (SC < 193) {
                                                    return 1;
                                                }
                                                else if (SC >= 193) {
                                                    if (TSC < 9) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (gchDC >= 62) {
                                            if (Cs < 9) {
                                                return 1;
                                            }
                                            else if (Cs >= 9) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 101) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 48) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 2) {
                        if (negBalance < 77) {
                            if (diffAFD < 7) {
                                if (Rs < 59) {
                                    return 0;
                                }
                                else if (Rs >= 59) {
                                    if (CsDiff < 6) {
                                        if (ssDCval < 38) {
                                            if (posBalance < 39) {
                                                if (Rs < 371) {
                                                    return 0;
                                                }
                                                else if (Rs >= 371) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 39) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 38) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 6) {
                                        if (diffAFD < 5) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 5) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 7) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 77) {
                            if (diffTSC < 3) {
                                if (RsDiff < 9) {
                                    return 0;
                                }
                                else if (RsDiff >= 9) {
                                    if (TSC < 18) {
                                        return 0;
                                    }
                                    else if (TSC >= 18) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 3) {
                                if (ssDCval < 30) {
                                    return 1;
                                }
                                else if (ssDCval >= 30) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 41) {
                    if (TSC < 11) {
                        if (gchDC < 132) {
                            if (negBalance < 30) {
                                if (TSC < 9) {
                                    return 0;
                                }
                                else if (TSC >= 9) {
                                    if (diffMVdiffVal < 104) {
                                        if (gchDC < 8) {
                                            return 1;
                                        }
                                        else if (gchDC >= 8) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 104) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 30) {
                                return 0;
                            }
                        }
                        else if (gchDC >= 132) {
                            return 1;
                        }
                    }
                    else if (TSC >= 11) {
                        if (diffRsCsdiff < 57) {
                            if (refDCval < 42) {
                                if (CsDiff < 2) {
                                    return 0;
                                }
                                else if (CsDiff >= 2) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 42) {
                                if (Rs < 4) {
                                    if (diffAFD < -2) {
                                        return 0;
                                    }
                                    else if (diffAFD >= -2) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 4) {
                                    if (Cs < 526) {
                                        return 0;
                                    }
                                    else if (Cs >= 526) {
                                        if (TSC < 19) {
                                            return 0;
                                        }
                                        else if (TSC >= 19) {
                                            if (Cs < 555) {
                                                return 1;
                                            }
                                            else if (Cs >= 555) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 57) {
                            if (diffRsCsdiff < 106) {
                                if (diffAFD < 6) {
                                    return 0;
                                }
                                else if (diffAFD >= 6) {
                                    if (ssDCval < 63) {
                                        if (Rs < 575) {
                                            return 1;
                                        }
                                        else if (Rs >= 575) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 63) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 106) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 10) {
                if (CsDiff < 7) {
                    if (gchDC < 15) {
                        if (diffTSC < 5) {
                            if (ssDCval < 41) {
                                return 1;
                            }
                            else if (ssDCval >= 41) {
                                if (posBalance < 75) {
                                    return 0;
                                }
                                else if (posBalance >= 75) {
                                    if (Cs < 418) {
                                        return 0;
                                    }
                                    else if (Cs >= 418) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 5) {
                            if (Rs < 230) {
                                if (diffRsCsdiff < -1) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= -1) {
                                    if (posBalance < 101) {
                                        if (diffRsCsdiff < 3) {
                                            if (Scindex < 2) {
                                                return 1;
                                            }
                                            else if (Scindex >= 2) {
                                                if (gchDC < 10) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 10) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 101) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Rs >= 230) {
                                if (diffMVdiffVal < 73) {
                                    if (RsCsDiff < 24) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 24) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 73) {
                                    if (gchDC < 5) {
                                        return 1;
                                    }
                                    else if (gchDC >= 5) {
                                        if (diffRsCsdiff < 34) {
                                            if (SC < 608) {
                                                return 0;
                                            }
                                            else if (SC >= 608) {
                                                if (diffTSC < 7) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 7) {
                                                    if (Cs < 463) {
                                                        if (AFD < 17) {
                                                            return 0;
                                                        }
                                                        else if (AFD >= 17) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 463) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 34) {
                                            if (diffTSC < 6) {
                                                if (Cs < 193) {
                                                    return 1;
                                                }
                                                else if (Cs >= 193) {
                                                    if (Rs < 366) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 366) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= 6) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 15) {
                        if (negBalance < 56) {
                            if (ssDCval < 86) {
                                if (diffAFD < 24) {
                                    if (SC < 149) {
                                        return 1;
                                    }
                                    else if (SC >= 149) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            if (posBalance < 93) {
                                                return 0;
                                            }
                                            else if (posBalance >= 93) {
                                                if (refDCval < 31) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 31) {
                                                    if (AFD < 23) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 23) {
                                                        if (CsDiff < 5) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            if (Cs < 335) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 335) {
                                                                return 0;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 24) {
                                    if (TSC < 33) {
                                        if (gchDC < 18) {
                                            return 1;
                                        }
                                        else if (gchDC >= 18) {
                                            if (CsDiff < 2) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 2) {
                                                if (diffAFD < 28) {
                                                    if (Cs < 216) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 216) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 28) {
                                                    if (refDCval < 32) {
                                                        return 0;
                                                    }
                                                    else if (refDCval >= 32) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 33) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 86) {
                                if (diffRsCsdiff < 21) {
                                    if (diffTSC < 9) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 9) {
                                        if (RsCsDiff < 25) {
                                            if (negBalance < 19) {
                                                if (RsCsDiff < 5) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 5) {
                                                    if (ssDCval < 144) {
                                                        if (Cs < 293) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 293) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (ssDCval >= 144) {
                                                        if (diffTSC < 10) {
                                                            return 1;
                                                        }
                                                        else if (diffTSC >= 10) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 19) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 25) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 21) {
                                    if (posBalance < 121) {
                                        if (RsCsDiff < 217) {
                                            if (SC < 1010) {
                                                return 1;
                                            }
                                            else if (SC >= 1010) {
                                                if (RsCsDiff < 53) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 53) {
                                                    if (diffAFD < 26) {
                                                        return 0;
                                                    }
                                                    else if (diffAFD >= 26) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 217) {
                                            if (CsDiff < 5) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 5) {
                                                if (negBalance < 29) {
                                                    if (RsCsDiff < 361) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 361) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 29) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 121) {
                                        if (posBalance < 128) {
                                            if (AFD < 59) {
                                                return 0;
                                            }
                                            else if (AFD >= 59) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 128) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 56) {
                            if (diffRsCsdiff < 4) {
                                if (refDCval < 203) {
                                    if (Cs < 1) {
                                        return 1;
                                    }
                                    else if (Cs >= 1) {
                                        if (diffAFD < 16) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 16) {
                                            if (TSC < 24) {
                                                if (SC < 33) {
                                                    return 0;
                                                }
                                                else if (SC >= 33) {
                                                    if (Cs < 208) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 208) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 24) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 203) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 4) {
                                if (diffAFD < 17) {
                                    if (TSC < 24) {
                                        if (negBalance < 98) {
                                            if (diffAFD < 11) {
                                                return 1;
                                            }
                                            else if (diffAFD >= 11) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 98) {
                                            if (gchDC < 20) {
                                                return 0;
                                            }
                                            else if (gchDC >= 20) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 24) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 17) {
                                    if (ssDCval < 84) {
                                        if (CsDiff < 4) {
                                            if (AFD < 41) {
                                                return 1;
                                            }
                                            else if (AFD >= 41) {
                                                if (ssDCval < 63) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 63) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 84) {
                                        if (diffAFD < 26) {
                                            if (Cs < 232) {
                                                return 1;
                                            }
                                            else if (Cs >= 232) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 26) {
                                            if (diffMVdiffVal < 121) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 121) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (CsDiff >= 7) {
                    if (diffAFD < 20) {
                        if (ssDCval < 63) {
                            if (negBalance < 47) {
                                if (diffMVdiffVal < 100) {
                                    if (CsDiff < 11) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 11) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 100) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 47) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 63) {
                            if (diffMVdiffVal < 95) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 95) {
                                if (ssDCval < 148) {
                                    if (diffMVdiffVal < 106) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= 106) {
                                        if (refDCval < 74) {
                                            if (gchDC < 17) {
                                                return 1;
                                            }
                                            else if (gchDC >= 17) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 74) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 148) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 20) {
                        if (negBalance < 5) {
                            if (AFD < 63) {
                                if (gchDC < 139) {
                                    return 0;
                                }
                                else if (gchDC >= 139) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 63) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 5) {
                            if (diffAFD < 29) {
                                if (TSC < 23) {
                                    if (CsDiff < 8) {
                                        if (diffMVdiffVal < 80) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 80) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 8) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 23) {
                                    if (diffRsCsdiff < 112) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 112) {
                                        if (posBalance < 29) {
                                            return 0;
                                        }
                                        else if (posBalance >= 29) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 29) {
                                if (MVDiff < 646) {
                                    return 1;
                                }
                                else if (MVDiff >= 646) {
                                    if (posBalance < 69) {
                                        return 1;
                                    }
                                    else if (posBalance >= 69) {
                                        if (refDCval < 133) {
                                            if (RsDiff < 5) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 5) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 133) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (Cs >= 1116) {
        if (diffTSC < 12) {
            if (MVDiff < 69) {
                if (refDCval < 44) {
                    if (RsCsDiff < 649) {
                        if (diffAFD < 3) {
                            return 0;
                        }
                        else if (diffAFD >= 3) {
                            if (ssDCval < 25) {
                                return 1;
                            }
                            else if (ssDCval >= 25) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 649) {
                        if (TSC < 12) {
                            return 1;
                        }
                        else if (TSC >= 12) {
                            return 0;
                        }
                    }
                }
                else if (refDCval >= 44) {
                    if (RsDiff < 37) {
                        return 0;
                    }
                    else if (RsDiff >= 37) {
                        if (SC < 3914) {
                            if (AFD < 9) {
                                return 1;
                            }
                            else if (AFD >= 9) {
                                if (SC < 3568) {
                                    return 0;
                                }
                                else if (SC >= 3568) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 3914) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 69) {
                if (diffRsCsdiff < 253) {
                    if (TSC < 31) {
                        if (diffTSC < 9) {
                            return 0;
                        }
                        else if (diffTSC >= 9) {
                            if (negBalance < 52) {
                                return 0;
                            }
                            else if (negBalance >= 52) {
                                if (RsCsDiff < 203) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 203) {
                                    if (SC < 2364) {
                                        return 1;
                                    }
                                    else if (SC >= 2364) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (TSC >= 31) {
                        if (refDCval < 111) {
                            return 1;
                        }
                        else if (refDCval >= 111) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 253) {
                    if (ssDCval < 49) {
                        if (Cs < 4226) {
                            if (TSCindex < 9) {
                                if (Cs < 2441) {
                                    return 0;
                                }
                                else if (Cs >= 2441) {
                                    return 1;
                                }
                            }
                            else if (TSCindex >= 9) {
                                return 1;
                            }
                        }
                        else if (Cs >= 4226) {
                            return 0;
                        }
                    }
                    else if (ssDCval >= 49) {
                        if (MVDiff < 424) {
                            if (RsDiff < 37) {
                                if (TSC < 21) {
                                    if (ssDCval < 51) {
                                        if (Rs < 1511) {
                                            return 1;
                                        }
                                        else if (Rs >= 1511) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 51) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 21) {
                                    if (diffRsCsdiff < 812) {
                                        if (ssDCval < 79) {
                                            if (SC < 4402) {
                                                return 1;
                                            }
                                            else if (SC >= 4402) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 79) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 812) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsDiff >= 37) {
                                if (diffMVdiffVal < 53) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 53) {
                                    if (AFD < 49) {
                                        if (Scindex < 7) {
                                            if (diffAFD < 14) {
                                                if (SC < 7881) {
                                                    return 1;
                                                }
                                                else if (SC >= 7881) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 14) {
                                                return 1;
                                            }
                                        }
                                        else if (Scindex >= 7) {
                                            if (RsDiff < 64) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 64) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 49) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 424) {
                            if (diffAFD < 5) {
                                return 0;
                            }
                            else if (diffAFD >= 5) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 12) {
            if (diffAFD < 25) {
                if (diffMVdiffVal < 65) {
                    if (RsDiff < 61) {
                        if (AFD < 61) {
                            if (diffTSC < 25) {
                                if (diffRsCsdiff < 408) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 408) {
                                    if (MVDiff < 46) {
                                        if (Rs < 2319) {
                                            return 1;
                                        }
                                        else if (Rs >= 2319) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 46) {
                                        if (MVDiff < 75) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 75) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 25) {
                                return 1;
                            }
                        }
                        else if (AFD >= 61) {
                            return 1;
                        }
                    }
                    else if (RsDiff >= 61) {
                        return 1;
                    }
                }
                else if (diffMVdiffVal >= 65) {
                    if (SC < 9051) {
                        if (gchDC < 14) {
                            return 1;
                        }
                        else if (gchDC >= 14) {
                            if (diffMVdiffVal < 223) {
                                if (diffTSC < 16) {
                                    return 0;
                                }
                                else if (diffTSC >= 16) {
                                    if (posBalance < 81) {
                                        return 1;
                                    }
                                    else if (posBalance >= 81) {
                                        if (Cs < 2675) {
                                            return 0;
                                        }
                                        else if (Cs >= 2675) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 223) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 9051) {
                        return 0;
                    }
                }
            }
            else if (diffAFD >= 25) {
                if (negBalance < 27) {
                    if (diffTSC < 27) {
                        if (negBalance < 8) {
                            return 0;
                        }
                        else if (negBalance >= 8) {
                            if (Rs < 2012) {
                                if (diffRsCsdiff < 195) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 195) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 2012) {
                                if (ssDCval < 77) {
                                    return 1;
                                }
                                else if (ssDCval >= 77) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 27) {
                        if (AFD < 69) {
                            if (negBalance < 1) {
                                if (diffAFD < 35) {
                                    return 1;
                                }
                                else if (diffAFD >= 35) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 1) {
                                if (diffTSC < 33) {
                                    if (Cs < 2054) {
                                        return 1;
                                    }
                                    else if (Cs >= 2054) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 33) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 69) {
                            return 0;
                        }
                    }
                }
                else if (negBalance >= 27) {
                    if (diffMVdiffVal < 55) {
                        if (MVDiff < 47) {
                            return 0;
                        }
                        else if (MVDiff >= 47) {
                            if (negBalance < 57) {
                                return 1;
                            }
                            else if (negBalance >= 57) {
                                if (Rs < 1308) {
                                    return 1;
                                }
                                else if (Rs >= 1308) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 55) {
                        if (RsCsDiff < 131) {
                            return 0;
                        }
                        else if (RsCsDiff >= 131) {
                            if (SC < 6645) {
                                return 1;
                            }
                            else if (SC >= 6645) {
                                if (negBalance < 121) {
                                    if (MVDiff < 75) {
                                        if (RsCsDiff < 8537) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 8537) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 75) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 121) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}

bool SCDetectRF(
    mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 MVDiff, mfxU32 Rs, mfxU32 AFD,
    mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 TSC, mfxU32 gchDC, mfxI32 diffRsCsdiff,
    mfxU32 posBalance, mfxU32 SC, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 Cs,
    mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxU32 refDCval, mfxU32 RsDiff,
    mfxU8 control) {
    mfxU8 sum = 0;
    sum += SCDetect1(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect2(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect3(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect4(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect5(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect6(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect7(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect8(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect9(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectA(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectB(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectC(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectD(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectE(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectF(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect10(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect11(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect12(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect13(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect14(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect15(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffAFD, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffRsCsdiff, negBalance, ssDCval, diffMVdiffVal);
    return(sum > RF_DECISION_LEVEL + control);
}
